/*
 * This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel


   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ProgressSemaphore.hh"

#include <Base/Debug/Error.hh>

using namespace BIAS;

ProgressSemaphore::ProgressSemaphore()
{
#ifdef BIAS_HAVE_PTHREADS
	pthread_mutex_init(&Lock_, NULL);
#endif
	Reset();
}


ProgressSemaphore::~ProgressSemaphore()
{}
 

float ProgressSemaphore::GetProgress()
{
  float ratio;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  if (NumSteps_>0) 
    ratio = float(CurrentStep_)/float(NumSteps_);
  else
    ratio=0.0f;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
  BIASASSERT(ratio>=0. && ratio<=1.);
  return ratio;
}
 

void ProgressSemaphore::SetNumSteps(unsigned int steps)
{
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  NumSteps_ = steps;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
}
 

void ProgressSemaphore::SetCurrentStep(unsigned int currentstep)
{
  BIASASSERT(currentstep<=NumSteps_);
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  CurrentStep_ = currentstep;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
}

void ProgressSemaphore::IncrementStep()
{
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  CurrentStep_++;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
  BIASASSERT(CurrentStep_<=NumSteps_);
}


void ProgressSemaphore::SetTerminateFlag()
{
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  Terminate_ = true;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
}

void ProgressSemaphore::ResetTerminateFlag()
{
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  Terminate_ =false;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
}

bool ProgressSemaphore::IsTerminationRequested()
{
  bool flag;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  flag = Terminate_;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
  return flag;
}

void ProgressSemaphore::Reset()
{
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&Lock_);
#endif
  NumSteps_ = 100;
  CurrentStep_ = 0;
  Terminate_ = false;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&Lock_);
#endif
}
