/*
 * This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel


   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ProgressSemaphore_hh__
#define __ProgressSemaphore_hh__

// includes
#include <bias_config.h>
#ifdef BIAS_HAVE_PTHREADS
#include <pthread.h>
#endif

namespace BIAS {

  /** @class ProgressSemaphore
      @brief this class represents a progress counter which multiple processes
      can use. It is implemented using locked access to the progress variables
      using pthread mutexes
			An example is that one thread starts a progress and has to wait till
			this progress is done and wants to display a progress bar. In the meanwhile
			the working progress can update the progress values in the ProgressSemaphore
			to inform the waiting progress about its progress.
      @author evers 03/2008 */
  class BIASCommon_EXPORT ProgressSemaphore
  {
  public:
    ProgressSemaphore();
    /// destructor
    ~ProgressSemaphore();
    /** \brief returns current progress state in standard intervall */
    float GetProgress();
    /** \brief Set the number of steps to progress will need to complete */
    void SetNumSteps(unsigned int steps);
    /** \brief Set the current step */
    void SetCurrentStep(unsigned int CurrentStep);
    /** \brief Increment the current step */
    void IncrementStep();
    /** \brief Sets a termination flag to inform e.g. another process that
     * an abort is requested */
    void SetTerminateFlag();
    /** \brief Chancels the termination flag */
    void ResetTerminateFlag();
    /** \brief Tells whether the termination flag is set.*/
    bool IsTerminationRequested();
    /** Resets all values */
    void Reset();

  protected:
#ifdef BIAS_HAVE_PTHREADS
    pthread_mutex_t Lock_;
#endif
	unsigned int NumSteps_;
    unsigned int CurrentStep_;
    bool Terminate_;
    
  }; // class ProgressSemaphore

} // namespace

#endif // __ProgressSemaphore_hh__
