/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIASTIMESTAMP_HH__
#define __BIASTIMESTAMP_HH__

#ifdef WIN32
#include <winsock.h>
#else //WIN32
#include <sys/time.h>
#endif //WIN32

#include <Base/Debug/Debug.hh>
//for compatibility with Win32 (i.e. for gettimeofday)
#include <Base/Common/W32Compat.hh>

namespace BIAS {

  /**
     @class TimeStamp 
     @brief holds system time in milliseconds
  */
  //typedef long long TimeStamp;
  class BIASCommon_EXPORT TimeStamp {
  public:
    /**
       @brief Ctors
     */
    TimeStamp() {msec_ = 0;}
    TimeStamp(long long ts) {msec_ = ts;}
    ~TimeStamp(){}

    /**
       @brief Operators for comparsion, duplication and basic arithmetics
     */
    bool operator>(const TimeStamp &ts) const;
    bool operator>(const long long &ts) const;
    bool operator<(const TimeStamp &ts) const;
    bool operator<(const long long &ts) const;
    bool operator>=(const TimeStamp &ts) const;
    bool operator>=(const long long &ts) const;
    bool operator<=(const TimeStamp &ts) const;
    bool operator<=(const long long &ts) const;
    bool operator==(const TimeStamp &ts) const;
    bool operator==(const long long &ts) const;
    bool operator!=(const TimeStamp &ts) const;
    bool operator!=(const long long &ts) const;

    TimeStamp& operator=(const TimeStamp &ts);
    TimeStamp& operator=(const long long &ts);

    TimeStamp operator+(const TimeStamp &ts) const;
    TimeStamp operator+(const long long &ts) const;
    TimeStamp operator-(const TimeStamp &ts) const;
    TimeStamp operator-(const long long &ts) const;
    TimeStamp operator*(const TimeStamp &ts) const;
    TimeStamp operator*(const long long &ts) const;
    TimeStamp operator/(const TimeStamp &ts) const;
    TimeStamp operator/(const long long &ts) const;
    TimeStamp operator%(const TimeStamp &ts) const;
    TimeStamp operator%(const long long &ts) const;

    /**
       @brief Get the timestamp as long long int
     */
    inline long long GetMSec() const {return msec_;}

    /**
       @brief Returns the absolute value timestamp
    */
    inline TimeStamp Abs() {if (msec_<0) return TimeStamp(-msec_); 
                            return *this;}

  private:
    long long msec_;
  };

  /**
     @class TimeStampGen
     @author 23.1.07 apetersen
     @ingroup Sensors
     @brief Generates absolute timestamps with a certain offset from 
            given base-time
   */
  class TimeStampGen : public BIAS::Debug {

  public:
    /**
       @brief Standard Ctor and Dtor
    */
    TimeStampGen() {
      baseTime_.tv_sec = 0;
      baseTime_.tv_usec = 0;
      msecBaseTime_ = 0;
    }
    ~TimeStampGen(){}

    /**
       @brief Generate a timestamp with offset 'msec' from given base-time.
              For msec==0 just returns base-time in ms.

       @param msec Offset from  base-time for timestamp

       @return Timestamp with value 'base-time + msec'
     */
    inline TimeStamp GenerateTimeStamp(const long long &msec)
    { return msecBaseTime_ + msec; }
    inline TimeStamp GenerateTimeStamp(const TimeStamp &ts)
    { return msecBaseTime_ + ts; }

    /**
       @brief Generate a timestamp from system time

       @return actual system time in ms
    */
    inline TimeStamp GenerateTimeStamp()
    { timeval tmp;
      gettimeofday(&tmp,NULL);
      return TimeStamp(((long long) tmp.tv_sec)*1000 + (tmp.tv_usec / 1000));
    }

    /**
       @brief Sets system time as base-time for stamp generation.
              'msecOffset' will be added to base-time as timing correction.
              Timestamps are conform with (unix) gettimeofday; baseTime_ holds
              the 'raw' timeval struct and msecBaseTime_ the (corresponding) 
              milliseconds since beginning of epoch (unix: 1.1.1970 00:00).

       @param msecOffset Offset from actual time to base-time
    */
    inline void SetBaseTime(const long long msecOffset=0)
    { //get actual system time and add offset
      gettimeofday(&baseTime_, NULL);
      baseTime_.tv_sec = baseTime_.tv_sec + long(msecOffset/1000LL);
      baseTime_.tv_usec = baseTime_.tv_usec + long(msecOffset%1000)*1000;
      //combine timeval-struct to fit in one long long
      msecBaseTime_ = ((long long)baseTime_.tv_sec) * 1000 
                      + (baseTime_.tv_usec / 1000);
    }

  protected:
    timeval baseTime_;
    TimeStamp msecBaseTime_;
  };

  /**
     @brief Standard output operator for TimeStamps
  */
  BIASCommon_EXPORT std::ostream& operator<<(std::ostream& os, 
                                             const BIAS::TimeStamp &ts);
  
  /**
     @brief Standard input operator for TimeStamps
  */
  BIASCommon_EXPORT std::istream& operator>>(std::istream& is, 
                                             BIAS::TimeStamp &ts);
  
} //namespace BIAS

#endif //__BIASTIMESTAMP_HH__
