/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


template <class QUAT_TYPE> inline
DualQuaternion<QUAT_TYPE>::DualQuaternion()
{
  //SetIdentity();
}

template <class QUAT_TYPE> inline
DualQuaternion<QUAT_TYPE>::DualQuaternion(const DualQuaternion<QUAT_TYPE> &d)
{ 
  real = d.real;
  dual = d.dual;
} 

template <class QUAT_TYPE> inline
DualQuaternion<QUAT_TYPE>::DualQuaternion(const Quaternion<QUAT_TYPE> &q,
                                          const Quaternion<QUAT_TYPE> &p)
{
  real = q;
  dual = p;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Invert()
{
  real.Invert();
  dual.Invert();
}
 
template <class QUAT_TYPE> inline DualQuaternion<QUAT_TYPE>
DualQuaternion<QUAT_TYPE>::Inverse() const
{
  DualQuaternion<QUAT_TYPE> res(*this);
  res.Invert();
  return res;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::MakeUnique()
{  
  // @todo consider all cases from Quaternion::MakeUnique()!
  if (real[3] < 0) {
    real *= 1.0;
    dual *= 1.0;
  }
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Normalize()
{
  // enforce unit length constraint for real part
  real.Normalize();
  // enforce orthogonality of real and dual part
  dual -= dual.ScalarProduct(real) * real;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::GetDualNorm(QUAT_TYPE &r, QUAT_TYPE &d) const
{
  // real part of result is norm of real quaternion part
  r = (QUAT_TYPE)real.NormL2();
  // dual part of result is inner product of real and dual quaternion parts
  d = (QUAT_TYPE)(real.ScalarProduct(dual) + dual.ScalarProduct(real));
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Add(const DualQuaternion<QUAT_TYPE>& r)
{
  real += r.real;
  dual += r.dual;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Add(const DualQuaternion<QUAT_TYPE>& r,
                               DualQuaternion<QUAT_TYPE>& res) const
{
  res = *this;
  res.Add(r);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Sub(const DualQuaternion<QUAT_TYPE>& r)
{
  real -= r.real;
  dual -= r.dual;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Sub(const DualQuaternion<QUAT_TYPE>& r,
                               DualQuaternion<QUAT_TYPE>& res) const
{
  res = *this;
  res.Sub(r);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Mult(const QUAT_TYPE &scalar)
{
  real *= scalar;
  dual *= scalar;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Mult(const QUAT_TYPE &scalar,
                                DualQuaternion<QUAT_TYPE> &res) const
{
  res = *this;
  res.Mult(scalar);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Div(const QUAT_TYPE &scalar)
{
  real /= scalar;
  dual /= scalar;  
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Div(const QUAT_TYPE &scalar,
                               DualQuaternion<QUAT_TYPE> &res) const
{
  res = *this;
  res.Div(scalar);
}

// ---------------------------------------------

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Mult(const DualQuaternion<QUAT_TYPE> &r)
{
  dual = (real * r.dual) + (dual * r.real);
  real.Mult(r.real);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Mult(const DualQuaternion<QUAT_TYPE> &r,
                                DualQuaternion<QUAT_TYPE> &res) const
{
  res = *this;
  res.Mult(r);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::MultLeft(const DualQuaternion<QUAT_TYPE> &l)
{
  dual = (l.real * dual) + (l.dual * real);
  real.MultLeft(l.real);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::MultLeft(const DualQuaternion<QUAT_TYPE> &l,
                                    DualQuaternion<QUAT_TYPE> &res) const
{
  res = *this;
  res.MultLeft(l);
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::MultVec(const Vector3<QUAT_TYPE> &vec,
                                   Vector3<QUAT_TYPE> &res) const
{
  // @todo
}

template <class QUAT_TYPE> inline Vector3<QUAT_TYPE>
DualQuaternion<QUAT_TYPE>::MultVec(const Vector3<QUAT_TYPE> &vec) const
{ 
  Vector3<QUAT_TYPE> res;
  MultVec(vec, res);
  return res;
}

// ---------------------------------------------

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::SetIdentity()
{
  real.SetIdentity();
  dual.SetZero();
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::Set(const Quaternion<QUAT_TYPE> &q,
                               const Quaternion<QUAT_TYPE> &p)
{
  real = q;
  dual = p;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::SetRealPart(const Quaternion<QUAT_TYPE> &q)
{
  real = q;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::SetDualPart(const Quaternion<QUAT_TYPE> &p)
{
  dual = p;
}

template <class QUAT_TYPE> inline Quaternion<QUAT_TYPE>
DualQuaternion<QUAT_TYPE>::GetRealPart() const
{
  return real;
}

template <class QUAT_TYPE> inline Quaternion<QUAT_TYPE>
DualQuaternion<QUAT_TYPE>::GetDualPart() const
{
  return dual;
}

template <class QUAT_TYPE> inline Vector<QUAT_TYPE>
DualQuaternion<QUAT_TYPE>::GetVector() const
{
  Vector<QUAT_TYPE> vec(8);
  for (register int i = 0; i < 4; i++) {
    vec[i] = real[i];
    vec[i+4] = dual[i];
  }
  return vec;
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::
GetRigidMotion(RMatrixBase &rotation,
               Vector3<QUAT_TYPE> &translation) const
{
  Quaternion<QUAT_TYPE> quat;
  GetRigidMotion(quat, translation);
  rotation.SetFromQuaternion(Quaternion<ROTATION_MATRIX_TYPE>((ROTATION_MATRIX_TYPE)quat[0],
                                                              (ROTATION_MATRIX_TYPE)quat[1],
                                                              (ROTATION_MATRIX_TYPE)quat[2],
                                                              (ROTATION_MATRIX_TYPE)quat[3]));
}

template <class QUAT_TYPE> inline void
DualQuaternion<QUAT_TYPE>::
SetFromRigidMotion(const RMatrixBase &rotation,
                   const Vector3<QUAT_TYPE> &translation)
{
  Quaternion<ROTATION_MATRIX_TYPE> quat;
  rotation.GetQuaternion(quat);
  SetFromRigidMotion(Quaternion<QUAT_TYPE>((QUAT_TYPE)quat[0],
                                           (QUAT_TYPE)quat[1],
                                           (QUAT_TYPE)quat[2],
                                           (QUAT_TYPE)quat[3]), translation);
}
