/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
#include "RMatrixBase.hh"
#include <Base/Common/W32Compat.hh>
#include <Base/Debug/Exception.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Geometry/Quaternion.hh>


#ifdef WIN32
#  define _USE_MATH_DEFINES
#endif //WIN32

#include <math.h>

// accuracy for constraint check in Check() function calls
#define R_CONSTRAINT_ACCURACY 1e-10

// if rotation angle is below this value, assume zero rotation
#define MINIMUM_ROTATION 1e-6

// if real part of quaternion = cos(alpha) is larger than this + 1,
// GetQuaternion() will produce an error and return -1
#define MAX_QUATERNION_COS_ERROR 1e-12

// show warnings when orthonormality constraints are enforced
//#define R_CONSTRAINTS_WARNING

#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

RMatrixBase::
RMatrixBase()
  : Matrix3x3<ROTATION_MATRIX_TYPE>(MatrixIdentity) 
{}


RMatrixBase::
RMatrixBase(const RMatrixBase& r)
  : Matrix3x3<ROTATION_MATRIX_TYPE>(r)
{
  if (!Check(R_CONSTRAINT_ACCURACY)) {
#ifdef R_CONSTRAINTS_WARNING    
    BIASWARN("Rotation matrix is invalid: "
             << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
             << "\nEnforcing orthonormality and positive determinant!\n");
#endif
    EnforceConstraints();
  }
}


RMatrixBase::
RMatrixBase(const MatrixInitType& i) : 
  Matrix3x3<ROTATION_MATRIX_TYPE>(i) 
{
  if (i != MatrixIdentity) {
    //SetIdentity();
    BEXCEPTION("Can not initialize a zero rotation matrix!");
  }
}


RMatrixBase::
RMatrixBase(const TNT::Matrix<ROTATION_MATRIX_TYPE>& mat)
  : Matrix3x3<ROTATION_MATRIX_TYPE>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BEXCEPTION("Wrong rotation matrix size " << mat.num_rows() << "x"
               << mat.num_cols());
  }
  if (!Check(R_CONSTRAINT_ACCURACY)) {
#ifdef R_CONSTRAINTS_WARNING    
     BIASWARN("Rotation matrix is invalid: "
              << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
              << "\nEnforcing orthonormality and positive determinant!\n");
#endif
    EnforceConstraints();
  }
}


RMatrixBase::
RMatrixBase(const Matrix3x3<ROTATION_MATRIX_TYPE>& mat)
  : Matrix3x3<ROTATION_MATRIX_TYPE>(mat) 
{
  if (!Check(R_CONSTRAINT_ACCURACY)) {
#ifdef R_CONSTRAINTS_WARNING    
    BIASWARN("Rotation matrix is invalid: "
             << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
             << "\nEnforcing orthonormality and positive determinant!\n");
#endif
    EnforceConstraints();
  }
}


RMatrixBase::
RMatrixBase(Vector3<ROTATION_MATRIX_TYPE> w, ROTATION_MATRIX_TYPE phi)
{
  Set(w, phi);
}


RMatrixBase::
~RMatrixBase() 
{}


void RMatrixBase::
SetXYZ(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Rx*Ry*Rz;
}


void RMatrixBase::
SetZYX(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Rz*Ry*Rx;
}


void RMatrixBase::
SetYXZ(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Ry*Rx*Rz;
}


void RMatrixBase::
SetZXY(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Rz*Rx*Ry;
}

void RMatrixBase::
SetYZX(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Ry*Rz*Rx;
}


void RMatrixBase::
SetXZY(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
       ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx(MatrixIdentity);
  RMatrixBase Ry(MatrixIdentity);
  RMatrixBase Rz(MatrixIdentity);
  
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; // = sin(PhiX);
  Rx[2][2] = Rx[1][1];  // = cos(PhiX);

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  Ry[2][0] = -Ry[0][2]; // = -sin(PhiY);
  Ry[2][2] = Ry[0][0];  // = cos(PhiY);
  
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1]; // = sin(PhiZ);
  Rz[1][1] = Rz[0][0];  // = cos(PhiZ);

  (*this) = Rx*Rz*Ry;
}

void RMatrixBase::
Set(const Vector3<ROTATION_MATRIX_TYPE> &wa, const ROTATION_MATRIX_TYPE phi)
{
  // zero rotation  results in identity matrix
  if (Equal(fabs(phi), 0.0)){
    SetIdentity();
    return;
  }
  double wnorm = wa.NormL2();
  if (wnorm<1e-6){
    BEXCEPTION("Vector "<<wa<<" is close to zero (norm = "<<wnorm
               <<"), solution is instable!");
  }
  
  Vector3<ROTATION_MATRIX_TYPE> w(wa);
  w /= wnorm;

  Matrix3x3<ROTATION_MATRIX_TYPE> Omega, I, A;
  I.SetIdentity();

  Omega[0][0] = 0.0;
  Omega[0][1] = -w[2];
  Omega[0][2] = w[1];
  Omega[1][0] = w[2];
  Omega[1][1] = 0.0;
  Omega[1][2] = -w[0];
  Omega[2][0] = -w[1] ;
  Omega[2][1] = w[0];
  Omega[2][2] = 0.0;

  Matrix3x3<ROTATION_MATRIX_TYPE> Sin_O, O_O, Cos_O_O;

  Sin_O = Omega *  sin(phi);

  Cos_O_O = (Omega * Omega) * (1-cos(phi));
  
  A = Matrix3x3<ROTATION_MATRIX_TYPE>(I + Sin_O + Cos_O_O);
  (*this) = RMatrixBase(I + Sin_O + Cos_O_O);
}


void RMatrixBase::
SetFromAxisAngle(Vector3<ROTATION_MATRIX_TYPE> w)
{
  double dAngle = w.NormL2();
  if (fabs(dAngle) < MINIMUM_ROTATION) {
    SetIdentity();
  } else {
    w.Normalize();
    Set(w, dAngle);
  }
}


int RMatrixBase::
GetQuaternion(Quaternion<ROTATION_MATRIX_TYPE>& quat) const
{
#ifdef BIAS_DEBUG
  // This function only works for true rotation matrices!
  if (!Check(R_CONSTRAINT_ACCURACY)) {
//#ifdef R_CONSTRAINTS_WARNING
    //BIASWARN("Rotation matrix is invalid: "
    //         << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
    //         << "\nCall EnforceConstraints() to enforce orthonormality "
    //         << "and positive determinant before quaternion conversion!\n");
//#endif
    BEXCEPTION("Rotation matrix is invalid: "
               << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
               << "\nCall EnforceConstraints() to enforce orthonormality "
               << "and positive determinant before quaternion conversion!\n");
    return -1;
  }
#endif

  // Conversion method from OpenSG
  // @attention Removed old conversion in Linux build which had numerical
  //            issues for rotations close to 180 degree! (esquivel 03/2011)

  double tr = (*this)[0][0] + (*this)[1][1] + (*this)[2][2];

  if (tr > 0.0)
  {
    // Use default formula if trace is large enough
    double s = sqrt(tr + 1.0);

    quat[3] = s * 0.5;

    s *= 2.0;
    quat[0] = ((*this)[2][1] - (*this)[1][2]) / s;
    quat[1] = ((*this)[0][2] - (*this)[2][0]) / s;
    quat[2] = ((*this)[1][0] - (*this)[0][1]) / s;

  }
  else
  {
    // Use alternate formula using largest trace element
    double s;
    unsigned int i;
    unsigned int j;
    unsigned int k;

    // Find largest trace element i and its successors j, k
    if((*this)[1][1] > (*this)[0][0])
      i = 1;
    else
      i = 0;
    if((*this)[2][2] > (*this)[i][i])
      i = 2;

    j = (i + 1) % 3;
    k = (j + 1) % 3;

    // Compute discriminator using largest trace element
    s = sqrt((*this)[i][i] - ((*this)[j][j] + (*this)[k][k]) + 1.0);

    // Compute quaternion entries
    quat[i] = s * 0.5;
    s *= 2.0;

    quat[j] = ((*this)[i][j] + (*this)[j][i]) / s;
    quat[k] = ((*this)[i][k] + (*this)[k][i]) / s;

    quat[3] = ((*this)[k][j] - (*this)[j][k] ) / s;
  }

/*
  // Conversion method using derivation depending on max. denominator
  // @see http://en.wikipedia.org/wiki/Rotation_representation_%28mathematics%29#Rotation_matrix_.E2.86.94_quaternion
  // @deprecated

  // Determine max. denominator
  double d[4] = { 1 + (*this)[0][0] - (*this)[1][1] - (*this)[2][2],
                  1 - (*this)[0][0] + (*this)[1][1] - (*this)[2][2],
                  1 - (*this)[0][0] - (*this)[1][1] + (*this)[2][2],
                  1 + (*this)[0][0] + (*this)[1][1] + (*this)[2][2] };
  int maxidx = 0;
  for (int i = 1; i < 4; i++) {
    if (d[maxidx] < d[i]) maxidx = i;
  }

  // Compute quaternion starting with max. denominator
  if (maxidx == 0) {
    quat[0]=0.5*sqrt(1 + (*this)[0][0] - (*this)[1][1] - (*this)[2][2]);
    quat[3]=0.25*((*this)[2][1]-(*this)[1][2])/quat[0];
    quat[1]=0.25*((*this)[0][1]+(*this)[1][0])/quat[0];
    quat[2]=0.25*((*this)[0][2]+(*this)[2][0])/quat[0];
  } else if (maxidx == 1) {
    quat[1]=0.5*sqrt(1 - (*this)[0][0] + (*this)[1][1] - (*this)[2][2]);
    quat[3]=0.25*((*this)[0][2]-(*this)[2][0])/quat[1];
    quat[2]=0.25*((*this)[1][2]+(*this)[2][1])/quat[1];
    quat[0]=0.25*((*this)[1][0]+(*this)[0][1])/quat[1];
  } else if (maxidx == 2) {
    quat[2]=0.5*sqrt(1 - (*this)[0][0] - (*this)[1][1] + (*this)[2][2]);
    quat[3]=0.25*((*this)[1][0]-(*this)[0][1])/quat[2];
    quat[0]=0.25*((*this)[2][0]+(*this)[0][2])/quat[2];
    quat[1]=0.25*((*this)[2][1]+(*this)[1][2])/quat[2];
  } else {
    quat[3]=0.5*sqrt(1 + (*this)[0][0] + (*this)[1][1] + (*this)[2][2]);
    quat[0]=0.25*((*this)[2][1]-(*this)[1][2])/quat[3];
    quat[1]=0.25*((*this)[0][2]-(*this)[2][0])/quat[3];
    quat[2]=0.25*((*this)[1][0]-(*this)[0][1])/quat[3];
  }
*/

/*
  // More compact implementation of the conversion method from OpenSG
  // @deprecated

  double trace = (*this)[0][0] + (*this)[1][1] + (*this)[2][2];
  if ( trace > 0 ) {
    double s = 2.0 * sqrt(trace + 1.0);
    quat[3] = 0.25 * s;
    quat[0] = ( (*this)[2][1] - (*this)[1][2] ) / s;
    quat[1] = ( (*this)[0][2] - (*this)[2][0] ) / s;
    quat[2] = ( (*this)[1][0] - (*this)[0][1] ) / s;
  } else {
    if ( (*this)[0][0] > (*this)[1][1] && (*this)[0][0] > (*this)[2][2] ) {
      double s = 2.0 * sqrt( 1.0 + (*this)[0][0] - (*this)[1][1] - (*this)[2][2]);
      quat[3] = ((*this)[2][1] - (*this)[1][2] ) / s;
      quat[0] = 0.25 * s;
      quat[1] = ((*this)[0][1] + (*this)[1][0] ) / s;
      quat[2] = ((*this)[0][2] + (*this)[2][0] ) / s;
    } else if ((*this)[1][1] > (*this)[2][2]) {
      double s = 2.0 * sqrt( 1.0 + (*this)[1][1] - (*this)[0][0] - (*this)[2][2]);
      quat[3] = ((*this)[0][2] - (*this)[2][0] ) / s;
      quat[0] = ((*this)[0][1] + (*this)[1][0] ) / s;
      quat[1] = 0.25 * s;
      quat[2] = ((*this)[1][2] + (*this)[2][1] ) / s;
    } else {
      double s = 2.0 * sqrt( 1.0 + (*this)[2][2] - (*this)[0][0] - (*this)[1][1] );
      quat[3] = ((*this)[1][0] - (*this)[0][1] ) / s;
      quat[0] = ((*this)[0][2] + (*this)[2][0] ) / s;
      quat[1] = ((*this)[1][2] + (*this)[2][1] ) / s;
      quat[2] = 0.25 * s;
    }
  }
*/

/*
  // Old conversion method
  // @see http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
  // @deprecated
  // @attention Computes wrong quaternion for rotation close to 180 degree!

  quat[3] = sqrt( max( 0.0, 1.0 + (*this)[0][0] + (*this)[1][1] + (*this)[2][2] ) ) / 2.0; 
  quat[0] = sqrt( max( 0.0, 1.0 + (*this)[0][0] - (*this)[1][1] - (*this)[2][2] ) ) / 2.0; 
  quat[1] = sqrt( max( 0.0, 1.0 - (*this)[0][0] + (*this)[1][1] - (*this)[2][2] ) ) / 2.0; 
  quat[2] = sqrt( max( 0.0, 1.0 - (*this)[0][0] - (*this)[1][1] + (*this)[2][2] ) ) / 2.0; 
  quat[0] = copysign( quat[0], (*this)[2][1] - (*this)[1][2] ) ;
  quat[1] = copysign( quat[1], (*this)[0][2] - (*this)[2][0] ) ;
  quat[2] = copysign( quat[2], (*this)[1][0] - (*this)[0][1] ) ;
*/

  // Normalize resulting quaternion
  quat /= quat.NormL2();

  // Correct quaternion so that scalar part is non-negative
  if (quat[3] < 0) quat.MultiplyIP(-1.0);

  return 0;
}


int RMatrixBase::
GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& Axis, 
                     ROTATION_MATRIX_TYPE& Angle) const
{
  // nicked from OpenSG
  Quaternion<ROTATION_MATRIX_TYPE> quat;
  if (GetQuaternion(quat)!=0)
    return -1;
  return quat.GetAxisAngle(Axis, Angle);
}


Vector3<ROTATION_MATRIX_TYPE> RMatrixBase::
GetRotationAxis() const
{
  Vector3<ROTATION_MATRIX_TYPE> Axis;
  ROTATION_MATRIX_TYPE Angle;

#ifdef BIASASSERT_ISACTIVE
  int res = GetRotationAxisAngle(Axis, Angle);
  BIASASSERT(res==0);
#else //avoid warning in release builds
  GetRotationAxisAngle(Axis, Angle);
#endif

  return Axis;
}


ROTATION_MATRIX_TYPE RMatrixBase::
GetRotationAngle() const
{
  Vector3<ROTATION_MATRIX_TYPE> Axis;
  ROTATION_MATRIX_TYPE Angle;

#ifdef BIASASSERT_ISACTIVE
  int res = GetRotationAxisAngle(Axis, Angle);
  BIASASSERT(res==0);
#else //avoid warning in release builds
  GetRotationAxisAngle(Axis, Angle);
#endif

  return Angle;
}


int RMatrixBase::
GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& w) const
{
  double angle;
  // nicked from above
  Quaternion<ROTATION_MATRIX_TYPE> quat;
  if (GetQuaternion(quat)!=0)     return -1;
  quat.GetAxisAngle(w , angle);
  w.Normalize();
  //get a three-parameter representation by encoding the angle into the axis:
  w *= angle;
  return 0;
}


int RMatrixBase::
GetRotationAnglesXYZ(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[0][2]) == 1.0) {
    BIASWARN("Rotation for Y-axis is +-PI/2, can not decompose X-/Z-component!");
    return -1;
  }
  PhiY = asin((*this)[0][2]);
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
  PhiX = atan2(-(*this)[1][2],(*this)[2][2]);
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiZ = atan2(-(*this)[0][1],(*this)[0][0]);
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
/*
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetXYZ(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


int RMatrixBase::
GetRotationAnglesZYX(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[2][0]) == 1.0) {
    BIASWARN("Rotation for Y-axis is +-PI/2, can not decompose X-/Z-component!");
    return -1;
  }
  PhiY = asin(-(*this)[2][0]); 
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
  PhiX = atan2((*this)[2][1],(*this)[2][2]);
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiZ = atan2((*this)[1][0],(*this)[0][0]);
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
/*
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetZYX(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


int RMatrixBase::
GetRotationAnglesZXY(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[2][1]) == 1.0) {
    BIASWARN("Rotation for X-axis is +-PI/2, can not decompose Y-/Z-component!");
    return -1;
  }
  PhiX = asin((*this)[2][1]); 
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiY = atan2(-(*this)[2][0],(*this)[2][2]); 
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
  PhiZ = atan2(-(*this)[0][1],(*this)[1][1]);
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
/*   
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetZXY(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


int RMatrixBase::
GetRotationAnglesYXZ(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[1][2]) == 1.0) {
    BIASWARN("Rotation for X-axis is +-PI/2, can not decompose Y-/Z-component!");
    return -1;
  }
  PhiX = asin(-(*this)[1][2]); 
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiY = atan2((*this)[0][2],(*this)[2][2]);
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
  PhiZ = atan2((*this)[1][0],(*this)[1][1]); 
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
/*
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetYXZ(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


int RMatrixBase::
GetRotationAnglesXZY(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[0][1]) == 1.0) {
    BIASWARN("Rotation for Z-axis is +-PI/2, can not decompose X-/Y-component!");
    return -1;
  }
  PhiZ = asin(-(*this)[0][1]); 
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
  PhiX = atan2((*this)[2][1],(*this)[1][1]);
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiY = atan2((*this)[0][2],(*this)[0][0]);
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
/*
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetXZY(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


int RMatrixBase::
GetRotationAnglesYZX(double& PhiX, double& PhiY, double& PhiZ) const
{
  if (fabs((*this)[1][0]) == 1.0) {
    BIASWARN("Rotation for Z-axis is +-PI/2, can not decompose X-/Y-component!");
    return -1;
  }
  PhiZ = asin((*this)[1][0]); 
  if (BIAS_ISNAN(PhiZ) || BIAS_ISINF(PhiZ)) return -3;
  PhiX = atan2(-(*this)[1][2],(*this)[1][1]);
  if (BIAS_ISNAN(PhiX) || BIAS_ISINF(PhiX)) return -1;
  PhiY = atan2(-(*this)[2][0],(*this)[0][0]);
  if (BIAS_ISNAN(PhiY) || BIAS_ISINF(PhiY)) return -2;
/*
#ifdef BIAS_DEBUG
  RMatrixBase R;
  R.SetYZX(PhiX, PhiY, PhiZ);
  if (R == (*this))
    res = 0;
  else
    res = -1;
#endif
*/
  return 0;
}


bool BIAS::operator==(const RMatrixBase& left, const RMatrixBase& right)
{
  bool res=true;
  for (register int i=0; i<9; i++){
    if (fabs(left.GetData()[i]-right.GetData()[i])>
        numeric_limits<double>::epsilon()){
      res=false;
      break;
    }
  }
  return res;
}


int RMatrixBase::
SetFromQuaternion(const Quaternion<ROTATION_MATRIX_TYPE> &qu) 
{
  // fast version, see below for old code:
  const double norm = 1.0/qu.NormL2();
  const double q0 = qu[0]*norm;
  const double q1 = qu[1]*norm;
  const double q2 = qu[2]*norm;
  const double q3 = qu[3]*norm;

  double* pData = GetData();
  
  *pData++ = 1.0 - 2.0 * (q1 * q1 + q2 * q2);
  *pData++ =       2.0 * (q0 * q1 - q2 * q3);
  *pData++ =       2.0 * (q2 * q0 + q1 * q3);

  *pData++ =       2.0 * (q0 * q1 + q2 * q3);
  *pData++ = 1.0 - 2.0 * (q2 * q2 + q0 * q0);
  *pData++ =       2.0 * (q1 * q2 - q0 * q3);

  *pData++ =       2.0 * (q2 * q0 - q1 * q3);
  *pData++ =       2.0 * (q1 * q2 + q0 * q3);
  *pData++ = 1.0 - 2.0 * (q1 * q1 + q0 * q0);


#ifdef WANT_TO_COMPARE_WITH_OLD_CODE
  
  Quaternion<ROTATION_MATRIX_TYPE> q = qu; // de-const
  q.Normalize();

  // nicked from OpenSG
  RMatrixBase RCompare(MatrixZero);
  RCompare[0][0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
  RCompare[1][0] =       2.0 * (q[0] * q[1] + q[2] * q[3]);
  RCompare[2][0] =       2.0 * (q[2] * q[0] - q[1] * q[3]);

  RCompare[0][1] =       2.0 * (q[0] * q[1] - q[2] * q[3]);
  RCompare[1][1] = 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
  RCompare[2][1] =       2.0 * (q[1] * q[2] + q[0] * q[3]);

  RCompare[0][2] =       2.0 * (q[2] * q[0] + q[1] * q[3]);
  RCompare[1][2] =       2.0 * (q[1] * q[2] - q[0] * q[3]);
  RCompare[2][2] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);

  for (unsigned int i=0; i<3; i++) {
    for (unsigned int j=0; j<3; j++) {
      if (!Equal((*this)[i][j], RCompare[i][j], 1e-12)) {
        cout<<"Matrices are not same !"<<*this<<" "<<RCompare<<endl;
        cout<<" at "<<i<<" "<<j<<" "<< ((*this)[i][j])-(RCompare[i][j])<<endl;
        BIASABORT;
      }
    }
  }
#endif

  return 0;
}


void  RMatrixBase::
SetFromOrthogonalBasis(const BIAS::Vector3<ROTATION_MATRIX_TYPE> &ex,
                       const BIAS::Vector3<ROTATION_MATRIX_TYPE> &ey,
                       const BIAS::Vector3<ROTATION_MATRIX_TYPE> &ez)
{
  SetFromRowVectors(ex.GetNormalized(), 
                    ey.GetNormalized(), 
                    ez.GetNormalized());
  if (!Check(R_CONSTRAINT_ACCURACY)) {
#ifdef R_CONSTRAINTS_WARNING    
    BIASWARN("Rotation matrix is invalid: "
             << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
             << "\nEnforcing orthonormality and positive determinant!\n");
#endif
    EnforceConstraints();
  }
}


void RMatrixBase::
SetFromHV(const BIAS::Vector3<ROTATION_MATRIX_TYPE> &xh,
          const BIAS::Vector3<ROTATION_MATRIX_TYPE> &vy)
{
  // normalize
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> x0(xh.GetNormalized());
  
  // compute base z in RHS
  const BIAS::Vector3<ROTATION_MATRIX_TYPE>
    z0(xh.CrossProduct(vy).GetNormalized());
  
  // compute y base orthogonal in RHS
  const BIAS::Vector3<ROTATION_MATRIX_TYPE>
    y0(z0.CrossProduct(x0).GetNormalized());
  
  (*this).SetFromColumnVectors(x0, y0, z0);
  if (!Check(R_CONSTRAINT_ACCURACY)) {
#ifdef R_CONSTRAINTS_WARNING    
    BIASWARN("Rotation matrix is invalid: "
             << GetCheckFailureReason(R_CONSTRAINT_ACCURACY)
             << "\nEnforcing orthonormality and positive determinant!\n");
#endif
    EnforceConstraints();
  }
}


int RMatrixBase::
SetFromOriUpGL(BIAS::Vector3<ROTATION_MATRIX_TYPE> ori,
               BIAS::Vector3<ROTATION_MATRIX_TYPE> up)
{
  // calculate orthogonal vectors
  BIAS::Vector3<double> right;
  ori.Normalize();
  ori.CrossProduct(up, right);
  right.Normalize();
  right.CrossProduct(ori, up);
  up.Normalize();
  ori *= -1.0; // we need the right handed system
  SetFromColumnVectors(right, up, ori);

  return 0;
}


int RMatrixBase::
SetFromOriUp(BIAS::Vector3<ROTATION_MATRIX_TYPE> ori,
             BIAS::Vector3<ROTATION_MATRIX_TYPE> up)
{
  // calculate orthogonal vectors
  BIAS::Vector3<double> right;
  ori.Normalize();
  ori.CrossProduct(up, right);
  right.Normalize();
  ori.CrossProduct(right, up);
  up.Normalize();
  SetFromColumnVectors(right, up, ori);

  return 0;
}


bool RMatrixBase::
Check(const double eps, const bool verbose) const
{
  // check determinant
  bool ok = Equal(GetDeterminant(), 1.0, eps);
  // check unit column vector length
  Vector3<double> c[3];
  if (ok) {
    for (register unsigned int i = 0; i < 3 && ok; i++) {
      GetColumn(i, c[i]); 
      ok = Equal(c[i].NormL2(), 1.0, eps);
    }
  }
  // check orthogonality
  if (ok)
    ok = Equal(c[0].ScalarProduct(c[1]), 0.0, eps);
  if (ok)
    ok = Equal(c[0].ScalarProduct(c[2]), 0.0, eps);
  if (ok)
    ok = Equal(c[1].ScalarProduct(c[2]), 0.0, eps);
  // print result in case of failure  
  if (!ok && verbose) {
    BIASWARN("Rotation matrix is invalid: " << GetCheckFailureReason(eps));
  }
  return ok;
}


std::string RMatrixBase::
GetCheckFailureReason(const double eps) const
{
  ostringstream res(" ");

  double det = GetDeterminant();
  if (!Equal(det, 1.0, eps)) {
    res << "det != 1 (det=" << setprecision(30) << det << ") ";
  }
  Vector3<double> c[3];
  for (register unsigned int i = 0; i < 3; i++) {
    GetColumn(i, c[i]); 
    if (!Equal(c[i].NormL2(), 1.0, eps)){
      res << "col[" << i << "].NormL2() != 1) (= " << c[i].NormL2() << ") ";
    }
  }
  if (!Equal(c[0].ScalarProduct(c[1]), 0.0, eps)){
    res << "col[0].ScalarProduct(c[1]) != 0 (= "
        << c[0].ScalarProduct(c[1]) << ") ";
  }
  if (!Equal(c[0].ScalarProduct(c[2]), 0.0, eps)){
    res << "col[0].ScalarProduct(c[2]) != 0 (= "
        << c[0].ScalarProduct(c[2]) << ") ";
  }
  if (!Equal(c[1].ScalarProduct(c[2]), 0.0, eps)){
    res << "col[1].ScalarProduct(c[2]) != 0 (= "
        << c[1].ScalarProduct(c[2]) << ") ";
  }
  return res.str();
}


void RMatrixBase::EnforceConstraints()
{
  // @todo Enforcing constraints is not working right now!
#ifdef BIAS_DEBUG
  Matrix3x3<double> oldR = *this;
#endif
  const double det = GetDeterminant();
  Vector3<double> c[3];
  for (register unsigned int i = 0; i < 3; i++) {
    GetColumn(i, c[i]);
    if (Equal(c[i].NormL2(), 0.0)) 
      BEXCEPTION("Rotation matrix has not full rank. Can not enforce "
                 "unit length of zero column vector!");
    c[i] /= c[i].NormL2();
  }
  // enforce orthonormality of c[0] and c[1];
  c[1] -= c[0].ScalarProduct(c[1])*c[0];
  c[1] /= c[1].NormL2();
  // set c[2] perpendicular to c[0] and c[1]
  c[0].CrossProduct(c[1], c[2]);
  c[2] /= c[2].NormL2();
  // flip c[0] and c[1] if determinant is negative
  if (det < 0.0) {
    c[0] *= -1.0;
    c[1] *= -1.0;
  }
  // set columns
  for (register unsigned int i = 0; i < 3; i++) {
    SetColumn(i, c[i]);
  }
  BIASASSERT(GetDeterminant() > 0.0);

/*
  // flip c[2] if cross product inverted direction
  // @todo check if this happens only when determinant is negative!!
  int maxRow2 = 0;
  if (fabs(c[2][1]) > fabs(c[2][maxRow2])) maxRow2 = 1;
  if (fabs(c[2][2]) > fabs(c[2][maxRow2])) maxRow2 = 2;
  if ((c[2][maxRow2] < 0 && oldR[maxRow2][2] > 0) ||
      (c[2][maxRow2] > 0 && oldR[maxRow2][2] < 0)) {
    c[2] *= -1.0;
#ifdef BIAS_DEBUG
    // show warning when 3rd column is flipped (happens for input
    // rotation matrix with determinant -1)
    Matrix3x3<double> tmpR;
    for (register unsigned int i = 0; i < 3; i++) tmpR.SetColumn(i, c[i]);
    BIASWARN("\n [ATTENTION] Flipped 3rd column in R:\n- old R : "
             << oldR << "- new R : " << tmpR << "- old determinant : "
             << oldR.GetDeterminant() << "\n- new determinant : "
             << tmpR.GetDeterminant() << endl);
    BIASASSERT(oldR.GetDeterminant() < 0);
#endif
  } else {
    BIASASSERT(oldR.GetDeterminant() > 0);
  }
  // set columns
  for (register unsigned int i = 0; i < 3; i++) {
    SetColumn(i, c[i]);
  }
  if (GetDeterminant() < 0.0) *this *= -1.0;
*/

/*
  if (!Check()) {
#ifdef R_CONSTRAINTS_WARNING    
    BIASERR("- R beforce constraint enforcement: " << oldR
            << "\n- R after constraint enforcement: "<< *this
            << "\n-> orthonormality enforcement failed with reason "
            << GetCheckFailureReason());
#endif
  }
*/
}
