#ifndef __WrapBias2Ipl__hh__
#define __WrapBias2Ipl__hh__

#include <bias_config.h>


#ifndef BIAS_HAVE_OPENCV
#  error BIAS_HAVE_OPENCV not defined. Please recompile BIAS with USE_OPENCV to use WrapBias2Ipl.
#endif

#ifdef BIAS_HAVE_OPENCV

#include <Base/Image/ImageBase.hh>

// forwrad declaration for cxtypes.h to avoid cv.h inclusion:
typedef struct _IplImage IplImage;

#define ALLOW_ALPHA_WIN_DEFAULT true

namespace BIAS {

  /** @class WrapBias2Ipl
      @brief wrapper around a BIAS image to be used as an OpenCv IPlimage
      with shared data area. 
      @author Jan Woetzel 07/2005 */
  class BIASImageBase_EXPORT WrapBias2Ipl {
  public:
    ~WrapBias2Ipl();
    WrapBias2Ipl();
    //WrapBias2Ipl(BIAS::ImageBase * PtrImgBias);
    WrapBias2Ipl(const BIAS::ImageBase * PtrImgBias);

    /** create the internal IPL image which shares the data area with the src Bias image. 
    And save the pointer to this source BIAS img. 
    @return 0 for succes, other=error
    @author Jan Woetzel 07/2005 */
    int Bind(const BIAS::ImageBase * p_src);

    /** helper function to create a slave ipl image which shares the data area with the master ref. BIAS image
    Ipl image is slave of BIAS  image.
    @return 0 for succes, other=error
    @author Jan Woetzel 07/2005 */
    static int CreateIplImageShared(const BIAS::ImageBase * p_src, IplImage * &p_dest);

    /** helper function to create a copy (!) BIAS image from a master Ipl image 
    Data area is NOT shared but copied from ipl to bias. JW */
    static int CreateBiasImageCopy(const IplImage * p_src, BIAS::ImageBase & dest);

    /** @return false if the wrapped images are inconsistent JW */
    bool Consistent() const;

    /** wrapper for cvWaitKey 
    @return pressed key (if any) 
    JW 2005 */
    static int WaitKey(const unsigned int & delay=0);

    /** Display img as simple popup window. 
    @param waitForKey if true delaMsec is waited for keypres. 0=forever.
    @param delayMsec if waitForKey is true wait for key with 0=forever, see cvWaitkey.
    @return pressed key (if any)
    @author JW 2005 
    TODO: add interface this in BIAS/Gui as GuiCV with static functions including sliders, mouse motion etc. (jw)
    */
    static int Display(
      const IplImage* img,
      const std::string & DestWin, 
      const bool & autoresize,
      const bool & moveToTopLeft,
      const bool & waitForKey,
      const unsigned int & delayMsec,
      const float & scale=DEFAULT_32to8_scale,
      const bool & allowAlphaWindow=ALLOW_ALPHA_WIN_DEFAULT
      );

    /** Display img as simple popup window. 
    If you experience color channel Seq. problems, use highgui with by RGB/BGR fix(jw)
    @param delayMsec: 
    @return pressed key (if any)
    @author JW */
    int Display(const bool & waitForKey=false, 
      const unsigned int & delayMsec=0,
      const float & scale=DEFAULT_32to8_scale
      ) const;

    int Display(const char* DestWin, 
      const bool & autoresize=true,
      const bool & moveToTopLeft=true,
      const bool & waitForKey=false,
      const unsigned int & delayMsec=0,
      const float & scale=DEFAULT_32to8_scale,
      const bool & allowAlphaWindow=ALLOW_ALPHA_WIN_DEFAULT
      ) const;

    /** Display img as simple popup HighGUI window. 
    @param actually scale is used only for float to 8 Byte range conversion 
    @return pressed key (if any) 
    @author Jan Woetzel 2005 */
    int Display(
      const std::string & DestWin, 
      const bool & autoresize=true,
      const bool & moveToTopLeft=true,
      const bool & waitForKey=false,
      const unsigned int & delayMsec=0,
      const float & scale=DEFAULT_32to8_scale,
      const bool & allowAlphaWindow=ALLOW_ALPHA_WIN_DEFAULT
      ) const;

    /** 32 bit float to displayable 8 bit conversion
    Both images should be pre-initilaized to equal size and correct storage type.
    @param scale mapping function, e.g. 255 for 0..1 --> 0..255
    @author Jan Woetzel */
    static void ConvertIPL32to8(const IplImage* src32, 
      IplImage* dest8,
      const float & scale=DEFAULT_32to8_scale
      );

    /** wrapper for CV destroiing all HighGUI windows JW */
    static void DestroyAllWindows();

    /* wrapper for CV destroiing one HighGUI window JW */
    static void DestroyWindow( const char* name );
    static void DestroyWindow( const std::string & name );

    /* wrapper for OpenCV,changes size of HighGUI window JW */
    static void ResizeWindow( const char* name, 
      const unsigned int width, 
      const unsigned int height );
    static void ResizeWindow( const std::string & name, 
      const unsigned int width, 
      const unsigned int height );

    /* wrapper for OpenCV, moves HighGUI window position JW */
    static void MoveWindow( const char* name, const int newX, const int newY );
    static void MoveWindow( const std::string & name, const int newX, const int newY );

    /* wrapper for OPenCV, creates a HighGUI window JW */
    static void NamedWindow(const char* name, const int flags=1);
    static void NamedWindow(const std::string& name, const int flags=1);

  public:
    const BIAS::ImageBase* p_imgBias;
    //bool p_imgBiasDestroyable;

    IplImage * p_imgIpl;
    bool p_imgIplDestroyable;
  };
} // namespace BIAS
#endif // BIAS_HAVE_OPENCV

#endif //__WrapBias2Ipl__hh__
