/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "ImageCalc.hh"

using namespace std;
using namespace BIAS;

////////////////////////////////////////////////////////////////////////////////
// calculations between two images                                            //
////////////////////////////////////////////////////////////////////////////////

template<class StorageType>
void ImageCalc<StorageType>::Add(const Image<StorageType> &in1,
                                 const Image<StorageType> &in2,
                                 Image<StorageType> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount(),
           in1.GetStorageType());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] + in2Data[i];
  }
}

template<class StorageType>
void ImageCalc<StorageType>::Subtract(const Image<StorageType> &in1,
                                      const Image<StorageType> &in2,
                                      Image<StorageType> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] - in2Data[i];
  }
}

template<class StorageType>
void ImageCalc<StorageType>::Multiply(const Image<StorageType> &in1,
                                      const Image<StorageType> &in2,
                                      Image<StorageType> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = in1Data[i] * in2Data[i];
  }
}

template<class StorageType>
void ImageCalc<StorageType>::Divide(const Image<StorageType> &in1,
                                    const Image<StorageType> &in2,
                                    Image<StorageType> &out) {
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in2Data[i] != (StorageType)0) {
      outData[i] = in1Data[i] / in2Data[i];
    }
    else {
      BIASWARN("prevented division by zero")
      outData[i] = (StorageType)0;
    }
  }
}

template <class StorageType>
void ImageCalc<StorageType>::And(const Image<StorageType> &in1,
                                 const Image<StorageType> &in2,
                                 Image<StorageType> &out)
{
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in1Data[i] != 0 && in2Data[i] != 0) {
      outData[i] = 1;
    }
    else {
      outData[i] = 0;
    }
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Nand(const Image<StorageType> &in1,
                                  const Image<StorageType> &in2,
                                  Image<StorageType> &out)
{
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in1Data[i] == 0 || in2Data[i] == 0) { // = !(in1Data[i] != 0 && in2Data[i] != 0)
      outData[i] = 1;
    }
    else {
      outData[i] = 0;
    }
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Or(const Image<StorageType> &in1,
                                const Image<StorageType> &in2,
                                Image<StorageType> &out)
{
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in1Data[i] != 0 || in2Data[i] != 0) {
      outData[i] = 1;
    }
    else {
      outData[i] = 0;
    }
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Nor(const Image<StorageType> &in1,
                                 const Image<StorageType> &in2,
                                 Image<StorageType> &out)
{
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if (in1Data[i] == 0 && in2Data[i] == 0) { // = !(in1Data[i] != 0 || in2Data[i] != 0)
      outData[i] = 1;
    }
    else {
      outData[i] = 0;
    }
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Xor(const Image<StorageType> &in1,
                                 const Image<StorageType> &in2,
                                 Image<StorageType> &out)
{
  // assert same size
  BIASASSERT(   in1.GetWidth()        == in2.GetWidth()
             && in1.GetHeight()       == in2.GetHeight()
             && in1.GetChannelCount() == in2.GetChannelCount())

  // init output image
  out.Init(in1.GetWidth(), in1.GetHeight(), in1.GetChannelCount());

  // get pointers to data
  const StorageType *in1Data = in1.GetImageData();
  const StorageType *in2Data = in2.GetImageData();
  StorageType *outData       = out.GetImageData();

  const unsigned int size =   in1.GetWidth()
                            * in1.GetHeight()
                            * in1.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    if ((in1Data[i] != 0 && in2Data[i] == 0)
        ||
        (in1Data[i] == 0 && in2Data[i] != 0)) {
      outData[i] = 1;
    }
    else {
      outData[i] = 0;
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
//calculations between one image and a scalar                                 //
////////////////////////////////////////////////////////////////////////////////

template <class StorageType>
void ImageCalc<StorageType>::AddScalar(const Image<StorageType> &in,
                                       StorageType scalar,
                                       Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] + scalar;
  }
}

template <class StorageType>
void ImageCalc<StorageType>::SubtractScalar(const Image<StorageType> &in,
                                            StorageType scalar,
                                            Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] - scalar;
  }
}

template <class StorageType>
void ImageCalc<StorageType>::MultiplyScalar(const Image<StorageType> &in,
                                            StorageType scalar,
                                            Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] * scalar;
  }
}

template <class StorageType>
void ImageCalc<StorageType>::DivideScalar(const Image<StorageType> &in,
                                          StorageType scalar,
                                          Image<StorageType> &out) {
  BIASASSERT(scalar != (StorageType)0)

  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size =   in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = inData[i] / scalar;
  }
}

////////////////////////////////////////////////////////////////////////////////
// calculations on a single image                                             //
////////////////////////////////////////////////////////////////////////////////

template <class StorageType>
void ImageCalc<StorageType>::Log(const Image<StorageType> &in,
                                 Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size = in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = StorageType(log(double(inData[i])));
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Log10(const Image<StorageType> &in,
                                   Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size = in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = StorageType(log10(double(inData[i])));
  }
}

template <class StorageType>
void ImageCalc<StorageType>::Abs(const Image<StorageType> &in,
                                 Image<StorageType> &out) {
  // init output image
  out.Init(in.GetWidth(), in.GetHeight(), in.GetChannelCount());

  // get pointers to data
  const StorageType *inData = in.GetImageData();
  StorageType *outData      = out.GetImageData();

  const unsigned int size = in.GetWidth()
                            * in.GetHeight()
                            * in.GetChannelCount();

  // perform operation
  for (unsigned int i = 0; i < size; i++) {
    outData[i] = (StorageType)abs(inData[i]);
  }
}

// explicit instantiation
#define INSTANCE_ImageCalc(type)\
template class BIASImageUtilsBase_EXPORT ImageCalc<type>;
namespace BIAS{
INSTANCE_ImageCalc(unsigned char)
INSTANCE_ImageCalc(float)
#ifdef BUILD_IMAGE_INT
  INSTANCE_ImageCalc(int)
#endif
#ifdef BUILD_IMAGE_CHAR
  INSTANCE_ImageCalc(char)
#endif
#ifdef BUILD_IMAGE_SHORT
  INSTANCE_ImageCalc(short)
#endif
#ifdef BUILD_IMAGE_USHORT
  INSTANCE_ImageCalc(unsigned short)
#endif
#ifdef BUILD_IMAGE_DOUBLE
  INSTANCE_ImageCalc(double)
#endif
#ifdef BUILD_IMAGE_UINT
  INSTANCE_ImageCalc(unsigned int)
#endif
}
