/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#ifndef __SymmetricMatrix3x3_hh__
#define __SymmetricMatrix3x3_hh__

#include "bias_config.h" 
#include <iostream>

namespace BIAS {

  // forward declarations
  template <class T> class Matrix3x3;
  template <class T> class Matrix;

  // forward declaration for stream operator
  template <class T> class BIASMathBase_EXPORT SymmetricMatrix3x3;

  // stream operators
  template <class T> BIASMathBase_EXPORT
  std::ostream& operator<<(std::ostream& os, const SymmetricMatrix3x3<T>& mat);

  template <class T> BIASMathBase_EXPORT
  std::istream& operator>>(std::istream& is, SymmetricMatrix3x3<T>& mat);

  /** @class SymmetricMatrix3x3
      @ingroup g_math
      @brief is a 'fixed size' symmetric quadratic matrix of dim. 3 x 3 which 
      is templated over the element-type. The class is memory efficient,
      storing only the necessary 6 independent elements of the symmetric matrix.
     
      @author woelk 11/2007 (c) www.vision-n.de */
  template <class T=double>
  class BIASMathBase_EXPORT SymmetricMatrix3x3 
  {
  public:
    SymmetricMatrix3x3() {};

    SymmetricMatrix3x3(const SymmetricMatrix3x3<T>& src);

    SymmetricMatrix3x3(const Matrix3x3<T>& src);

    SymmetricMatrix3x3(const Matrix<T>& src);

    ~SymmetricMatrix3x3();

    SymmetricMatrix3x3<T>& operator=(const SymmetricMatrix3x3<T>& src);

    /// element access
    T&       operator()(const unsigned row, const unsigned col);
    const T& operator()(const unsigned row, const unsigned col) const;

    /// conversion to full matrix
    void GetMatrix(Matrix3x3<T>& dst) const;
    Matrix3x3<T> GetMatrix() const;
    
    /// binary io
    int Write(std::ostream& os) const;
    int Read(std::istream& is);

    /// stream operators are declared as friends
    friend BIASMathBase_EXPORT std::ostream&
      operator<<<T>(std::ostream& os, const SymmetricMatrix3x3<T>& mat);
    friend BIASMathBase_EXPORT std::istream&
      operator>><T>(std::istream& is, SymmetricMatrix3x3<T>& mat);

    T* GetData() { return Data_; };
    const T* GetData() const { return Data_; }
  protected:
  	
    unsigned SymMatrix3x3ToIndex_(const unsigned row, const unsigned col) const;
    
    /// The matrix data is stored as follows
    /// | Data_[0] Data_[1] Data_[2] |
    /// | Data_[1] Data_[3] Data_[4] |
    /// | Data_[2] Data_[4] Data_[5] |
    T Data_[6];
    
  }; // class


} // namespace


#endif // __SymmetricMatrix3x3_hh__
