# - Find clcc executable and provides a macro to generate custom build rules
#
# The module defines the following variables:
#  CLCC_FOUND - true is clcc executable is found
#  CLCC_EXECUTABLE - the path to the clcc executable
#  CLCC_VERSION - the version of clcc
#
# If clcc is found on the system, the module provides the macro:
#  CLCC_TARGET(Name OpenCLInput [COMPILE_FLAGS <string>])
# which creates a custom command  compile the <OpenCLInput> file.
# If  COMPILE_FLAGS option is specified, the next
# parameter is added to the clcc  command line. Name is an alias used to
# get  details of  this custom  command.  Indeed the  macro defines  the
# following variables:
#  CLCC_${Name}_DEFINED - true is the macro ran successfully
#  CLCC_${Name}_INPUT - the opencl source file, an alias for ${OpenCLInput}
#
#  ====================================================================
#  Example:
#
#   find_package(CLCC)
#
#   CLCC_TARGET(MyOpenCLProgram kernels.cl)
#
#  ====================================================================

#=============================================================================
#
# This Module is based on the FindFLEX.cmake module.
# Copyright 2014 Stefan Reinhold
# 
# Copyright of the original module:
# 
# Copyright 2009 Kitware, Inc.
# Copyright 2006 Tristan Carel
#
# Distributed under the OSI-approved BSD License (the "License");
# see accompanying file Copyright.txt for details.
#
# This software is distributed WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the License for more information.
#=============================================================================
# (To distributed this file outside of CMake, substitute the full
#  License text for the above reference.)

FIND_PROGRAM(CLCC_EXECUTABLE clcc DOC "path to the clcc executable")
MARK_AS_ADVANCED(CLCC_EXECUTABLE)

IF(CLCC_EXECUTABLE)

  EXECUTE_PROCESS(COMMAND ${CLCC_EXECUTABLE} --version
    OUTPUT_VARIABLE CLCC_version_output
    ERROR_VARIABLE CLCC_version_error
    RESULT_VARIABLE CLCC_version_result
    OUTPUT_STRIP_TRAILING_WHITESPACE)
  IF(NOT ${CLCC_version_result} EQUAL 0)
    MESSAGE(SEND_ERROR "Command \"${CLCC_EXECUTABLE} --version\" failed with output:\n${CLCC_version_error}")
  ELSE()
    STRING(REGEX REPLACE "^clcc (.*)$" "\\1"
      CLCC_VERSION "${CLCC_version_output}")
  ENDIF()

  #============================================================
  # CLCC_TARGET (public macro)
  #============================================================
  #
  MACRO(CLCC_TARGET Name Input)
    SET(CLCC_TARGET_usage "CLCC_TARGET(<Name> <Input> [COMPILE_FLAGS <string>]")
#    IF(${ARGC} GREATER 2)
#      IF(${ARGC} EQUAL 4)
#        IF("${ARGV3}" STREQUAL "COMPILE_FLAGS")
#          SET(CLCC_EXECUTABLE_opts  "${ARGV4}")
#          SEPARATE_ARGUMENTS(CLCC_EXECUTABLE_opts)
#        ELSE()
#          MESSAGE(SEND_ERROR ${CLCC_TARGET_usage})
#        ENDIF()
#      ELSE()
#        MESSAGE(SEND_ERROR ${CLCC_TARGET_usage})
#      ENDIF()
#    ENDIF()

    ADD_CUSTOM_COMMAND(TARGET ${Name} PRE_LINK
      COMMAND ${CLCC_EXECUTABLE}
      ARGS ${CLCC_EXECUTABLE_opts} ${Input}
      COMMENT "[CLCC][${Name}] Building OpenCL program with clcc ${CLCC_VERSION}"
      VERBATIM)

    ADD_DEPENDENCIES(${Name} ${Input})

    SET(CLCC_${Name}_DEFINED TRUE)
    SET(CLCC_${Name}_INPUT ${Input})
    SET(CLCC_${Name}_COMPILE_FLAGS ${CLCC_EXECUTABLE_opts})
  ENDMACRO(CLCC_TARGET)
  #============================================================


ENDIF(CLCC_EXECUTABLE)

INCLUDE(FindPackageHandleStandardArgs)
FIND_PACKAGE_HANDLE_STANDARD_ARGS(CLCC DEFAULT_MSG CLCC_EXECUTABLE)

# FindCLCC.cmake ends here
