#ifndef __BLOBDETECTOR_LEVELSET_HH__
#define __BLOBDETECTOR_LEVELSET_HH__

// bias include
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include "BlobDetectorBase.hh"

namespace BIAS{

  /**
     @class BlobDetectorLevelSet 
     @brief Evaluates a segementation into fore/back-ground.        
     Usage: 
     Initialise class with an image by calling Init(),
     if the image roi has a valid mask it is used as evaluation area.
     Set an initial level set square (use InitLevelSetSquare).
     Set a evaluation square, if the image has no valid roi mask.
     Than call Evolute() round about 100 times (or what ever 
     it will needs).

     What you get:
     The segmentation of image is given by GetLevel(), but
     GetSegmentation() creats an image much better for visalisation.

     ---- 

     Please take a look at example ExampleSegmentationLevelSet, which
     visualises the Evolute() steps.
     @test tested with ExampleBlobDetectorLevelSet.cpp
     @author Dennis Herzog, 2005-07-28, edited ischiller 08/08
     @ingroup g_feature     
  */
  template <class StorageType>
  class BIASFeatureDetector_EXPORT BlobDetectorLevelSet :
    public BlobDetectorBase<StorageType>
  {

  public:
    BlobDetectorLevelSet();
    virtual ~BlobDetectorLevelSet();

    int Detect(Image<StorageType>& image,
               std::vector<BIAS::BIASBlob>& blobs);
  
  
    /** @brief initialise this instance with given image,
        note: image must not be modified until you delete 
        this instance or an other image is set, 
        the image's roi msk defines the evaluation area (if given)      
        @author Dennis Herzog, 2005-07-28 */
    void Init(const BIAS::Image<StorageType> &image, bool useROI = true);

    /** @brief sets image to operate on,
        note: image must not be modified until you delete 
        this instance or an other image is set, 
        the image's roi msk defines the evaluation area (if given)      
        @author Dennis Herzog, 2005-08-01 */
    void Set(const BIAS::Image<StorageType> &image, bool useROI = true);

    /** @brief look at class discription 
        @author Dennis Herzog, 2005-07-28 */
    void InitLevelSetSquare(const BIAS::Vector2<int>& p1,
                            const BIAS::Vector2<int>& p2);

    /** @brief can be used only, iff image's roi is not in use
        (look at class discription)
        @author Dennis Herzog, 2005-07-28 */
    void SetEvaluationSquare(const BIAS::Vector2<int>& p1,
                             const BIAS::Vector2<int>& p2);
    /// look at SetEvaluationSquare()
    void AddEvaluationSquare(const BIAS::Vector2<int>& p1,
                             const BIAS::Vector2<int>& p2);

    /** @brief  does the hole segmentation work by modifying the level set
        (look at class discription)
        @author Bodo Rosenhahn / ported and optimised by Dennis Herzog, 
        2005-07-28 */
    void Evolute();

  public:
    /** @brief returns reference of currently used image 
        @author Dennis Herzog, 2005-07-28 */
    const BIAS::Image<StorageType>& GetImage() const { return *image_; }

    /** @brief returns (roi) msk (the evaluation array) as an image
        @author Dennis Herzog, 2005-07-28 */
    const BIAS::Image<StorageType>& GetEval()  const { return eval_;   }

    /** @brief returns (roi) msk (the evaluation array) as an image
        (is allowed to be modified, iff image's roi is not in use)
        @author Dennis Herzog, 2005-07-28 */
    BIAS::Image<StorageType>& GetEval()  
    { BIASASSERT( eval_ownsdata_ == true ); return eval_; }
    
    /** @brief returns the actual level set, modified each time you call 
        Evolute()
        @author Dennis Herzog, 2005-07-28 */
    const BIAS::Image<StorageType>& GetLevel() const { return *level_; }

    /** @brief returns the actual level set, modified each time you call 
        Evolute(),
        (is allowed to be modified),
        (0 is background, != 0 is foreground)
        @author Dennis Herzog, 2005-07-28 */
    BIAS::Image<StorageType>& GetLevel() { return *level_; }

    /** @brief creates an image for visulalistion of the actual segmentation
        given by actual level set
        @author Dennis Herzog, 2005-07-28 */
    void GetSegmentation(BIAS::Image<StorageType>& im) const;

    /** @returns histogram of foreground (hist=1) and background (hist=0)
        @author Dennis Herzog, 2005-07-28 */
    std::vector<int> GetHistogram(int hist);

  protected:
    int _hist0[256];
    int _hist1[256];

    const BIAS::Image<StorageType>   *image_;   
    BIAS::Image<StorageType>         *level_;   // level/level_s are swapped
    BIAS::Image<StorageType>         *level_s_; // used temporarily by Evolute()
    BIAS::Image<StorageType>         eval_;    // encaps _image's msk   
    bool  eval_ownsdata_;                       // does _eval owns the data it uses

    static void Resize(BIAS::Image<StorageType> &image,
                       const int w, const int h);
    static void ClearHistogram(int hist[]);
    void ClearHistograms();
    void RecalcHistograms();
    void AddSquare(BIAS::Image<StorageType>& image, 
                   const BIAS::Vector2<int>& p1, 
                   const BIAS::Vector2<int>& p2);

  };//class



}//namespace



#endif // __BLOBDETECTOR_LEVELSET_HH__

