
#include <Base/ImageUtils/ImageDraw.hh>
#include <FeatureDetector/BlobDetectorCCA.hh>

#include "CheckerboardDetectorCrossFilter.hh"

#include <Base/Common/W32Compat.hh>


using namespace std;
using namespace BIAS;


// Instance pointer if an instance was created. This hack is needed to stay compatible
// to the static function calls of the other checkerboard detectors
CheckerboardDetectorCrossFilter* CheckerboardDetectorCrossFilter::instancePointer_ = NULL;


CheckerboardDetectorCrossFilter::CheckerboardDetectorCrossFilter(){
  // Set default values

  // The default filter size
  defaultHalfWinSize_ = 20;

  // The default number of roations of the filter
  defaultAngleSteps_ = 16;

  // The initial threshold for the filter response. This threshold is lowered
  // if to few corners are detected
  initialFilterThreshold_ = 70000;


  initialLineThreshold_ = 8000;

  // Set the static instance pointer so the static function can
  // use the current instance
  instancePointer_ = this;
}

CheckerboardDetectorCrossFilter::~CheckerboardDetectorCrossFilter(){
  // Set the static instance pointer to NULL so the static function knows
  // it must generate her own instance
  instancePointer_ = NULL;
}


int CheckerboardDetectorCrossFilter::Compute(Image<unsigned char> &img,
                                             const unsigned int &xCorners,
                                             const unsigned int & yCorners,
                                             t_CoordVecf &coords,
                                             const bool &saveDebugImage){

  coords.clear();

  // Create an instance of the checkerboard detector if none hase been created
  // already, call GetCheckerboard and delete the instance afterwards, if created.
  bool removeInstance = false;
  if (instancePointer_ == NULL){
    instancePointer_ = new CheckerboardDetectorCrossFilter();
    removeInstance = true;
  }
  vector<HomgPoint2D> retVal =
                         instancePointer_->GetCheckerboard(img,xCorners,yCorners,true);

  if (removeInstance) delete instancePointer_;

  if (retVal.size() != xCorners * yCorners) return -1;


  for (unsigned int i = 0; i < retVal.size(); i++)
    coords.push_back(std::pair<float,float>((float)retVal[i][0],(float)retVal[i][1]));

  return 0;
}



void CheckerboardDetectorCrossFilter::GetAvgBlackWhiteValue(Image<unsigned char> &img,
                                                            vector<HomgPoint2D> points,
                                                            double &avgBlack,
                                                            double &avgWhite, int angleSteps){

  // Calculate the average black and white values of the cross filter for a given
  // set of points

  avgBlack = 0.0;
  avgWhite = 0.0;

  unsigned char** ida = img.GetImageDataArray();

  double filterAngleStepSize = M_PI / 2.0 / (double)angleSteps;
  int numChannels = img.GetChannelCount();

  for (int i = 0; i < (int)points.size(); i++)
    for (int angleStep = 0; angleStep < defaultAngleSteps_ ; angleStep++){
      double angle = filterAngleStepSize * (double)angleStep;
      double sine = sin(angle);
      double cosine = cos(angle);
      double dirA = 0.0;
      double dirB = 0.0;

      // Calculate the filter response
      for (int c = 0; c < numChannels; c++){
          for (int walkX = -defaultHalfWinSize_; walkX <= defaultHalfWinSize_; walkX++){
            dirA += ida[(int)points[i][1]+(int)rint((double)walkX*sine)]
                       [((int)points[i][0]+(int)rint((double)walkX*cosine))*numChannels+c];
          }
          for (int walkY = -defaultHalfWinSize_; walkY <= defaultHalfWinSize_; walkY++){
            dirB += ida[(int)points[i][1]+(int)rint((double)walkY*cosine)]
                       [((int)points[i][0]-(int)rint((double)walkY*sine))*numChannels+c];
          }
        }

      // Assign the results to the black / white sum
      if (dirA > dirB){
        avgWhite += dirA;
        avgBlack += dirB;
      } else {
        avgBlack += dirA;
        avgWhite += dirB;
      }
    }

  // Divide everything by the number of summands
  avgBlack /= (double)points.size() * (double)angleSteps *
                     (defaultHalfWinSize_ * 2+1)* (double)numChannels;
  avgWhite /= (double)points.size() * (double)angleSteps *
                     (defaultHalfWinSize_ * 2+1)* (double)numChannels;
}



void CheckerboardDetectorCrossFilter::GetFilterResponseSum(Image<unsigned char> &img,
                                                           vector<HomgPoint2D> &response,
                                                           float threshold, int halfWinSize, bool fastMode){

  // Calculate filter responses for the full image and return the set of coordinates
  // above the threshold

  int width = (int)img.GetWidth() -halfWinSize - 1;
  int height = (int)img.GetHeight() -halfWinSize - 1;

  // Cache the sine and cosine results to save some performance
  double angleInc = (M_PI/2.0) / (double)defaultAngleSteps_ ;
  vector<double> angleSin;
  for (int i = 0; i < defaultAngleSteps_ ; i++)
    angleSin.push_back(sin(angleInc * (double)i));
  vector<double> angleCos;
  for (int i = 0; i < defaultAngleSteps_ ; i++)
    angleCos.push_back(cos(angleInc * (double)i));

  unsigned char** ida = img.GetImageDataArray();
  int channels = img.GetChannelCount();

  int angleIncrement = 1;
  int from = halfWinSize+1;
  int increment = 3;
  int toX = width-2;
  int toY = height-2;

  int offFrom = -1;
  int offTo = 2;

  if (fastMode){
    angleIncrement = 2;

    from = halfWinSize+2;
    increment = 5;
    toX = width-3;
    toY = height-3;

    offFrom = -2;
    offTo = 3;
  }

  double lowThreshold = 2000.0 / (double)angleIncrement;


  // Calculate every third pixel in every third line to speed
  // things up. If the filter result is high enough, also
  // examine the neighboring pixels
#ifdef BIAS_HAVE_OPENMP
  #pragma omp parallel for
#endif
  for (int x = from; x < toX; x+=increment){
    for (int y = from; y < toY; y+=increment){
      double absDiff = 0.0;
      for (int i = 0; i < defaultAngleSteps_ ; i+=angleIncrement)
         absDiff += filterFast(ida, x, y, channels, halfWinSize, angleSin[i], angleCos[i]);

      // If the filter result is above 500, calculate the filter results for all neighbors
      if (absDiff > lowThreshold){
        for (int xOff = offFrom; xOff < offTo; xOff++)
          for (int yOff = offFrom; yOff < offTo; yOff++){
             double diffTest = 0.0;
             for (int i = 0; i < defaultAngleSteps_ ; i+=angleIncrement)
                diffTest += filterFast(ida, x+xOff, y+yOff, channels,
                              halfWinSize, angleSin[i], angleCos[i]);
             if (diffTest > threshold / (3.0*angleIncrement)){
               double diff = 0.0;
               for (int i = 0; i < defaultAngleSteps_ ; i++)
                 diff += filter(ida, x+xOff, y+yOff, channels,
                                halfWinSize, angleSin[i], angleCos[i]);

               if (diff > threshold){
               #ifdef BIAS_HAVE_OPENMP
                 #pragma omp critical
               #endif
               {
                 response.push_back(HomgPoint2D(x+xOff,y+yOff,diff));
               }}
             }
        }
      }
    }
  }
}


void CheckerboardDetectorCrossFilter::StretchImageValues(Image<unsigned char> &img,
                                                         unsigned int from, unsigned int to){

  // Stretch the image values such that from becomes 0 and to becomes 255
  unsigned char** ida = img.GetImageDataArray();

  double whiteColor = to;
  double blackColor = from;

  for (int x = 0; x < (int)img.GetWidth(); x++)
    for (int y = 0; y < (int)img.GetHeight(); y++)
      for (int c = 0; c < 3; c++){
        double newVal = (double)(ida[y][x*3+c]- blackColor) * (255.0 / (whiteColor  - blackColor)) ;
        if (newVal < 0.0) newVal = 0.0;
        if (newVal > 255.0) newVal = 255.0;
        ida[y][x*3+c] = (unsigned char)rint(newVal);
      }
}

void CheckerboardDetectorCrossFilter::GetCCACenter(const vector<HomgPoint2D> pixels,
                                                   vector<HomgPoint2D> &centers,
                                                   int width, int height){

  // Make a connected components analysis by drawing every pixel into an image
  // and then perform a blob detecting CCA on it

  Image<unsigned char> foundCornerImg = Image<unsigned char>(width,height,1);
  foundCornerImg.SetZero();
  unsigned char** fciIda = foundCornerImg.GetImageDataArray();
  int inputSize = (int)pixels.size();
  for (int i = 0; i < inputSize; i++){
    fciIda[(int)pixels[i][1]][(int)pixels[i][0]] = 255;
  }

  // Do the blob detection
  BlobDetectorCCA<unsigned char> cca;
  vector<BIASBlob> ccaBlobs;
  cca.Detect(foundCornerImg,ccaBlobs);

  // Get the result and reformat it
  size_t outputSize = ccaBlobs.size();
  centers.resize(outputSize-1);
  for (size_t i = 1; i < outputSize; i++){
    centers[i-1] = HomgPoint2D(ccaBlobs[i].centerofmass[0],
                               ccaBlobs[i].centerofmass[1],1.0);

  }
}

double CheckerboardDetectorCrossFilter::NonNormalizedLength(const HomgPoint2D &a,
                                                            const HomgPoint2D &b){
  // Calculate the euclidean distance between two points neglecting
  // the 3rd component
  return sqrt((a[0]-b[0])*(a[0]-b[0])+(a[1]-b[1])*(a[1]-b[1]));
}

void CheckerboardDetectorCrossFilter::GenerateRectangularTopology(Image<unsigned char> &img,
                                                        vector<HomgPoint2D> &knotPositions,
                                                        vector<vector<int> > &adjacence,
                                                        deque<deque< int > > &rectTopolog){


  //Search for connection that with the same origin that go into the same direction
  // and remove the longer of these two connections

  for (int t = 0; t < (int)knotPositions.size(); t++){
    if (adjacence[t].size() > 2){
      vector<int> markForRemoval;
      for (int nA = 0; nA < (int)adjacence[t].size(); nA++)
         for (int nB = nA+1; nB < (int)adjacence[t].size(); nB++){
           if (GetOpeningAngle(knotPositions[t],knotPositions[adjacence[t][nA]],
                               knotPositions[adjacence[t][nB]]) < M_PI / 8.0){
             double lengthA = NonNormalizedSquaredLength(knotPositions[t],
                                                  knotPositions[adjacence[t][nA]]);
             double lengthB = NonNormalizedSquaredLength(knotPositions[t],
                                                  knotPositions[adjacence[t][nB]]);
             if (lengthA > lengthB) markForRemoval.push_back(nA);
             else markForRemoval.push_back(nB);
           }

      }
      //Create a new vector without the marked entries and replace the adjancences with it
      if (markForRemoval.size() > 0){
        vector<int> newAdjacences;
        for (int i = 0; i < (int)adjacence[t].size(); i++)
          if (find(markForRemoval.begin(), markForRemoval.end(), i)==markForRemoval.end())
            newAdjacences.push_back(adjacence[t][i]);
        adjacence[t] = newAdjacences;
      }
    }
  }

  //Find the first knot with 4 adjacences to start the rectTopolog structure with...
  // Find a start knot
  int firstKnot = -1;
  for (int i = 0; i< (int)knotPositions.size();i++){
    if (adjacence[i].size() == 4){
      firstKnot = i;
      break;
    }
  }
  if (firstKnot == -1) return;

  // The vector containing the elements already in the rectTopolog structure:
  vector<int> isInrectTopolog;

  // rectTopolog element convention:
  // -1 no object available (tried to find one but didn't succeed)
  // -2 didn't try to add and object
  rectTopolog.clear();
  // Add the first element
  deque<int> firstLine; firstLine.push_back(firstKnot);
  rectTopolog.push_back(firstLine);
  isInrectTopolog.push_back(firstKnot);

  /*cout << "1) checking if topology structure is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/


  // Add the first adjacence on the same line (simply take the first entry):
  rectTopolog[0].push_back(adjacence[firstKnot][0]);
  isInrectTopolog.push_back(adjacence[firstKnot][0]);

  /*cout << "2) checking if topology structure is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/

  // Find the knot in the oposite direction
  int op = 0;
  for (int i = 1; i < 4; i++){
    if (fabs(GetOpeningAngle(knotPositions[firstKnot],
                             knotPositions[adjacence[firstKnot][0]],
                             knotPositions[adjacence[firstKnot][i]])) > M_PI * 0.8)
      op = i;
  }

  // Return if no knot was found on the other side
  if (op == 0) return;


  // If a knot was found, add it
  rectTopolog[0].push_front(adjacence[firstKnot][op]);

  // Add the other entries as column entries:
  deque<int> upperLine; upperLine.push_back(-2);
  if (op == 1) upperLine.push_back(adjacence[firstKnot][2]);
  else upperLine.push_back(adjacence[firstKnot][1]);
  upperLine.push_back(-2);
  rectTopolog.push_front(upperLine);

  /*cout << "3) checking if topology structure is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/

  deque<int> lowerLine; lowerLine.push_back(-2);
  if (op == 3) lowerLine.push_back(adjacence[firstKnot][2]);
  else lowerLine.push_back(adjacence[firstKnot][3]);
  lowerLine.push_back(-2);
  rectTopolog.push_back(lowerLine);

  /*cout << "4) checking if topology structure is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/

  isInrectTopolog.push_back(adjacence[firstKnot][1]);
  isInrectTopolog.push_back(adjacence[firstKnot][2]);
  isInrectTopolog.push_back(adjacence[firstKnot][3]);

 /* cout << "Entering the extension procedure, checking if topology structure is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/



  // Generate the rectTopolog:
  int noNewEntriesFound = 0;
  int searchDirection = 0;
  while (noNewEntriesFound < 5){

    // Step 1 complete rectangle:
    // Find the -2 entries and try to associate them with a knot:
    bool changedEntry = true;
    while(changedEntry){
      changedEntry = false;
      for (int row = 0; row < (int)rectTopolog.size(); row++)
        for (int col = 0; col < (int)rectTopolog[row].size(); col++)
          if (rectTopolog[row][col] == -2){
            // Analyse the neighbours that are not unknown to find the missing knot in the topology:
            int neighbourRow = 0; int neighbourCol = 0;
            if (row < (int)rectTopolog.size()- 1 && rectTopolog[row+1][col] >= 0) neighbourRow = 1;
            if (row > 0 && rectTopolog[row-1][col] >= 0) neighbourRow = -1;
            if (col < (int)rectTopolog[row].size() - 1 && rectTopolog[row][col+1] >= 0) neighbourCol = 1;
            if (col > 0 && rectTopolog[row][col-1] >= 0) neighbourCol = -1;

            if (neighbourCol != 0 && neighbourRow != 0){
              int newKnot = -1;
              for (int adjRow = 0; adjRow < (int)adjacence[rectTopolog[row+neighbourRow][col]].size(); adjRow++)
                for (int adjCol = 0; adjCol < (int)adjacence[rectTopolog[row][col+neighbourCol]].size(); adjCol++)
                  if (adjacence[rectTopolog[row+neighbourRow][col]][adjRow] == adjacence[rectTopolog[row][col+neighbourCol]][adjCol]
                      && find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row+neighbourRow][col]][adjRow]) == isInrectTopolog.end())
                     newKnot = adjacence[rectTopolog[row+neighbourRow][col]][adjRow];
               rectTopolog[row][col] = newKnot;
              if (newKnot != -1) isInrectTopolog.push_back(newKnot);
              changedEntry = true;
            }
         }
    }

    // Step 2 search the borders for a new knot to extend the rectTopolog:
    bool addedNewEntry = false;
    if (searchDirection < 2){ // Search in column direction
      int colDir = (searchDirection == 0 ? -1 : 1);
      int colBorder = (searchDirection == 0 ? 0 : rectTopolog[0].size() - 1);

      for (int row = 0; row < (int)rectTopolog.size(); row++){
        if (rectTopolog[row][colBorder] >= 0)
        for (int candidate = 0; candidate < (int)adjacence[rectTopolog[row][colBorder]].size(); candidate++){

          if (!addedNewEntry
              && find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row][colBorder]][candidate]) == isInrectTopolog.end()
              && rectTopolog[row][colBorder] >= 0 && rectTopolog[row][colBorder-colDir] >= 0){

              if (CouldBeNeighbour(knotPositions[rectTopolog[row][colBorder]],
                                      knotPositions[rectTopolog[row][colBorder-colDir]],
                                      knotPositions[adjacence[rectTopolog[row][colBorder]][candidate]])){

            int newKnot = adjacence[rectTopolog[row][colBorder]][candidate];
            for (int i = 0; i < (int)rectTopolog.size(); i++){
              if (searchDirection == 0) rectTopolog[i].push_front(-2);
              else rectTopolog[i].push_back(-2);
            }
            if (searchDirection == 0) rectTopolog[row][0] = newKnot;
            if (searchDirection == 1) rectTopolog[row][rectTopolog[row].size()-1] = newKnot;
            noNewEntriesFound = -1;
            addedNewEntry = true;
            isInrectTopolog.push_back(newKnot);

            /*cout << "Search direction is " << searchDirection << " Checking if topology structure is still valid:  ";
            for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
              cout << rectTopolog[qwe].size() << " ";
            cout << endl;*/

          }}
        }
      }
    }
    if (searchDirection >=2 && searchDirection <= 3){ // Search in row direction
      int rowDir = (searchDirection == 2 ? -1 : 1);
      int rowBorder = (searchDirection == 2 ? 0 : rectTopolog.size() - 1);

      for (int col = 0; col < (int)rectTopolog[0].size(); col++){
        if (rectTopolog[rowBorder][col] >= 0)
        for (int candidate = 0; candidate < (int)adjacence[rectTopolog[rowBorder][col]].size(); candidate++){

          if (!addedNewEntry
              && find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[rowBorder][col]][candidate]) == isInrectTopolog.end()
              && rectTopolog[rowBorder][col] >= 0 && rectTopolog[rowBorder-rowDir][col] >= 0
              && CouldBeNeighbour(knotPositions[rectTopolog[rowBorder][col]],
                                      knotPositions[rectTopolog[rowBorder-rowDir][col]],
                                      knotPositions[adjacence[rectTopolog[rowBorder][col]][candidate]])){

            int newKnot = adjacence[rectTopolog[rowBorder][col]][candidate];
            // Generate the new row:
            deque<int> newLine;
            newLine.assign(rectTopolog[0].size(),-2);

            if (searchDirection == 2) rectTopolog.push_front(newLine);
            else rectTopolog.push_back(newLine);

            if (searchDirection == 2) rectTopolog[0][col] = newKnot;
            if (searchDirection == 3) rectTopolog[rectTopolog.size()-1][col] = newKnot;
            noNewEntriesFound = -1;
            addedNewEntry = true;
            isInrectTopolog.push_back(newKnot);

            /*cout << "Search direction is " << searchDirection << " Checking if topology structure is still valid:  ";
            for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
              cout << rectTopolog[qwe].size() << " ";
            cout << endl;*/

          }
        }
      }
    }
    if (searchDirection == 4){ // try to associate the -1 labeled entries

      for (int row = 0; row < (int)rectTopolog.size(); row++)
        for (int col = 0; col < (int)rectTopolog[row].size(); col++){

        if (rectTopolog[row][col] == -1){
          vector<int> candidates;
          // Start with neighbours:
          if (row > 0 && rectTopolog[row-1][col] >= 0)
            for (int n = 0; n < (int)adjacence[rectTopolog[row-1][col]].size(); n++)
              if (find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row-1][col]][n]) == isInrectTopolog.end())
                candidates.push_back(adjacence[rectTopolog[row-1][col]][n]);
          if (col > 0 && rectTopolog[row][col-1] >= 0)
            for (int n = 0; n < (int)adjacence[rectTopolog[row][col-1]].size(); n++)
              if (find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row][col-1]][n]) == isInrectTopolog.end())
                candidates.push_back(adjacence[rectTopolog[row][col-1]][n]);
          if (row < (int)rectTopolog.size()-1 && rectTopolog[row+1][col] >= 0)
            for (int n = 0; n < (int)adjacence[rectTopolog[row+1][col]].size(); n++)
              if (find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row+1][col]][n]) == isInrectTopolog.end())
                candidates.push_back(adjacence[rectTopolog[row+1][col]][n]);
          if (col < (int)rectTopolog[0].size()-1 && rectTopolog[row][col+1] >= 0)
            for (int n = 0; n < (int)adjacence[rectTopolog[row][col+1]].size(); n++)
              if (find(isInrectTopolog.begin(),isInrectTopolog.end(), adjacence[rectTopolog[row][col+1]][n]) == isInrectTopolog.end())
                candidates.push_back(adjacence[rectTopolog[row][col+1]][n]);

          // Look if we can find an entry thats more than one time available (that would be the first choice)
          int newKnot = -1;

          for (int i = 0; i < (int)candidates.size();i++)
            for (int j = i+1; j < (int)candidates.size();j++){
              if (candidates[i] == candidates[j])
                newKnot = candidates[i];
            }

          if (newKnot != -1){
            rectTopolog[row][col] = newKnot;
            continue;
          }

          int numberOfDirectionAngles = 0;
          if (row > 1 && rectTopolog[row-1][col] >= 0 && rectTopolog[row-2][col] >= 0)
                           numberOfDirectionAngles++;
          if (col > 1 && rectTopolog[row][col-1] >= 0 && rectTopolog[row][col-2] >= 0)
                           numberOfDirectionAngles++;
          if (row < (int)rectTopolog.size() -2 && rectTopolog[row+1][col] >= 0 && rectTopolog[row+2][col] >= 0)
                           numberOfDirectionAngles++;
          if (col < (int)rectTopolog[0].size() -2 && rectTopolog[row][col+1] >= 0 && rectTopolog[row][col+1] >= 0)
                           numberOfDirectionAngles++;

          // See if we find a suitable point in the candidates in respect to the direction (angle)
          if (numberOfDirectionAngles > 0){
            for (int i = 0; i < (int)candidates.size();i++){
              int hits = 0;

              if (row > 1 && rectTopolog[row-1][col] >= 0 && rectTopolog[row-2][col] >= 0 &&
                  CouldBeNeighbour(knotPositions[rectTopolog[row-1][col]],
                                   knotPositions[rectTopolog[row-2][col]],
                                   knotPositions[candidates[i]]))
                   hits++;

              if (col > 1 && rectTopolog[row][col-1] >= 0 && rectTopolog[row][col-2] >= 0 &&
                  CouldBeNeighbour(knotPositions[rectTopolog[row][col-1]],
                                   knotPositions[rectTopolog[row][col-2]],
                                   knotPositions[candidates[i]]))
                   hits++;

              if (row < (int)rectTopolog.size() -2 && rectTopolog[row+1][col] >= 0 && rectTopolog[row+2][col] >= 0 &&
                  CouldBeNeighbour(knotPositions[rectTopolog[row+1][col]],
                                   knotPositions[rectTopolog[row+2][col]],
                                   knotPositions[candidates[i]]))
                   hits++;

              if (col < (int)rectTopolog[0].size() -2 && rectTopolog[row][col+1] >= 0 && rectTopolog[row][col+2] >= 0 &&
                  CouldBeNeighbour(knotPositions[rectTopolog[row][col+1]],
                                   knotPositions[rectTopolog[row][col+2]],
                                   knotPositions[candidates[i]]))
                   hits++;

              if (hits == numberOfDirectionAngles){
                rectTopolog[row][col] = candidates[i];
                isInrectTopolog.push_back(candidates[i]);
              }
            }
          }
        }
      }
    }
    searchDirection = (searchDirection + 1) % 5;
    noNewEntriesFound ++;
  }
}

void CheckerboardDetectorCrossFilter::SearchForMissingTopologyEntries(Image<unsigned char> &img,
                                                                      vector<HomgPoint2D> &knotPositions,
                                                                      deque< deque< int > > &rectTopolog){
  bool foundOne = true;
  while(foundOne){
  foundOne = false;
  for (int row = 0; row < (int)rectTopolog.size(); row++)
    for (int col = 0; col < (int)rectTopolog[0].size(); col++)
      if (rectTopolog[row][col] < 0){

        // Analyse the neighbours that are not unknown to find the missing knot in the topology:
        int neighbourRow = 0; int neighbourCol = 0;
        if (row < (int)rectTopolog.size()- 1 && rectTopolog[row+1][col] >= 0) neighbourRow = 1;
        if (row > 0 && rectTopolog[row-1][col] >= 0) neighbourRow = -1;
        if (col < (int)rectTopolog[row].size() - 1 && rectTopolog[row][col+1] >= 0) neighbourCol = 1;
        if (col > 0 && rectTopolog[row][col-1] >= 0) neighbourCol = -1;
        if (neighbourCol != 0 && neighbourRow != 0 && rectTopolog[row][col+neighbourCol] >= 0 &&
            rectTopolog[row+neighbourRow][col] >= 0 && 
            rectTopolog[row+neighbourRow][col+neighbourCol] >= 0)
        {

        double searchRadius;
        HomgPoint2D searchCenter;
        GetSearchRegionForNeighbor(knotPositions[rectTopolog[row+neighbourRow][col]],
                                   knotPositions[rectTopolog[row][col+neighbourCol]],
                                   knotPositions[rectTopolog[row+neighbourRow][col+neighbourCol]],
                                   searchCenter,searchRadius);

        HomgPoint2D a = FindNewEdgePosition(img,knotPositions[rectTopolog[row+neighbourRow][col]],
                                                knotPositions[rectTopolog[row][col+neighbourCol]],
                                                knotPositions[rectTopolog[row+neighbourRow][col+neighbourCol]],
                                                searchCenter,searchRadius);

        knotPositions.push_back(a);
        rectTopolog[row][col] = (int)knotPositions.size() - 1;
        foundOne = true;
      }
    }
  }
}

void CheckerboardDetectorCrossFilter::ExtendTopology(Image<unsigned char> &img,
                                                     vector<HomgPoint2D> &knotPositions,
                                                     deque< deque< int > > &rectTopolog){

  // Search all 4 directions for possible extension:
  // 0 <=> row negative, 1 <=> row positive, 2 <=> column negative , 3 <=> column positive

 //cout << "Extending Topology with " << knotPositions.size() << " and rect = "
  //       << rectTopolog.size() << "x" << rectTopolog[0].size() << endl;
  double bestFitness = 0.0;
  int bestSearchDir = 0;
  vector<HomgPoint2D> bestNewPoints;

  for (int searchDir = 0; searchDir < 4; searchDir++){
    double currentFitness  = 0.0;
    vector<int> attachPoints;
    vector<int> supportPointsNear;
    vector<int> supportPointsFar;

    switch(searchDir){
    case 0:
      for (int i = 0; i < (int)rectTopolog.size(); i++){ // 0 <=> along the rows, negative
        attachPoints.push_back(rectTopolog[i][0]);
        supportPointsNear.push_back(rectTopolog[i][1]);
        supportPointsFar.push_back(rectTopolog[i][2]);
      }
      break;
    case 1:
      for (int i = 0; i < (int)rectTopolog.size(); i++){ // 1 <=> along the row positive
        attachPoints.push_back(rectTopolog[i][rectTopolog[0].size()-1]);
        supportPointsNear.push_back(rectTopolog[i][rectTopolog[0].size()-2]);
        supportPointsFar.push_back(rectTopolog[i][rectTopolog[0].size()-3]);
      }
      break;
    case 2:
      for (int i = 0; i < (int)rectTopolog[0].size(); i++){ // 2 <=> along the column negative
        attachPoints.push_back(rectTopolog[0][i]);
        supportPointsNear.push_back(rectTopolog[1][i]);
        supportPointsFar.push_back(rectTopolog[2][i]);
      }
      break;
    case 3:
      for (int i = 0; i < (int)rectTopolog[0].size(); i++){ // 3 <=> along the column positive
        attachPoints.push_back(rectTopolog[rectTopolog.size() -1][i]);
        supportPointsNear.push_back(rectTopolog[rectTopolog.size() -2][i]);
        supportPointsFar.push_back(rectTopolog[rectTopolog.size() -3][i]);
      }
      break;
    }

    vector<HomgPoint2D> expectedPoints(attachPoints.size());

    // Generate expected point positions:
    #ifdef BIAS_HAVE_OPENMP
      #pragma omp parallel for
    #endif
    for (int i = 0; i < (int)attachPoints.size(); i++){
      double farAngle = atan2(knotPositions[supportPointsNear[i]][0]
                                 -knotPositions[supportPointsFar[i]][0],
                              knotPositions[supportPointsNear[i]][1]
                                 -knotPositions[supportPointsFar[i]][1]);

      double nearAngle = atan2(knotPositions[attachPoints[i]][0]
                                 -knotPositions[supportPointsNear[i]][0],
                               knotPositions[attachPoints[i]][1]
                                 -knotPositions[supportPointsNear[i]][1]);

      // prevent the 2pi ambiguity to destroy the calculation
      if ((nearAngle - farAngle) > M_PI) farAngle += 2.0 * M_PI;
      if ((nearAngle - farAngle) < -M_PI) farAngle -= 2.0 * M_PI;

      double expectedAngle = nearAngle + (nearAngle -farAngle);

      double farLength = NonNormalizedLength(knotPositions[supportPointsFar[i]],
                                             knotPositions[supportPointsNear[i]]);
      double nearLength = NonNormalizedLength(knotPositions[supportPointsNear[i]],
                                              knotPositions[attachPoints[i]]);

      double expectedLength = nearLength + (nearLength - farLength);

      HomgPoint2D expectedPoint;
      expectedPoint[0] = knotPositions[attachPoints[i]][0]
                         + sin(expectedAngle) * expectedLength;
      expectedPoint[1] = knotPositions[attachPoints[i]][1]
                         + cos(expectedAngle) * expectedLength;

      HomgPoint2D avoidPointA = HomgPoint2D(0.0,0.0,0.0);
      HomgPoint2D avoidPointB = HomgPoint2D(0.0,0.0,0.0);
      if (i > 0) avoidPointA = knotPositions[attachPoints[i-1]];
      if (i < (int)attachPoints.size()-1) avoidPointB = knotPositions[attachPoints[i+1]];
      expectedPoint = FindNewExtensionPosition(img,knotPositions[attachPoints[i]],
                                               avoidPointA,avoidPointB,
                                               expectedPoint,expectedLength/2.0);

      expectedPoints[i] = expectedPoint;

      // Add the filterresponse to the fitness
      #ifdef BIAS_HAVE_OPENMP
        #pragma omp critical
      #endif
      {
      currentFitness += GetSingleFilterResponse(img,expectedPoints[i],20) * 10.0
                        / (double)attachPoints.size();

      // Add the quality of line from the attach point to the new point to the fitness
      currentFitness += GetCheckerLineFactor(img,knotPositions[attachPoints[i]],
                                             expectedPoints[i]) / (double)(attachPoints.size() + attachPoints.size() - 1);

      }

      // Take a look if the checker board continues a little further..
      farAngle = atan2(knotPositions[attachPoints[i]][0]-knotPositions[supportPointsNear[i]][0],
                               knotPositions[attachPoints[i]][1]-knotPositions[supportPointsNear[i]][1]);

      nearAngle = atan2(expectedPoint[0]-knotPositions[attachPoints[i]][0],
                               expectedPoint[1]-knotPositions[attachPoints[i]][1]);

      expectedAngle = nearAngle + (nearAngle -farAngle);

      farLength = NonNormalizedLength(knotPositions[supportPointsNear[i]],
                                      knotPositions[attachPoints[i]]);
      nearLength = NonNormalizedLength(knotPositions[attachPoints[i]],
                                       expectedPoint);
      expectedLength = nearLength + (nearLength - farLength);

      HomgPoint2D outsidePoint;
      outsidePoint[0] = knotPositions[attachPoints[i]][0]
                            + sin(expectedAngle) * expectedLength;
      outsidePoint[1] = knotPositions[attachPoints[i]][1]
                            + cos(expectedAngle) * expectedLength;

      // Add the quality of the line from the new point in the extrapolated direction
      // to the fitness (see if the checkerboard continues)
      #ifdef BIAS_HAVE_OPENMP
        #pragma omp critical
      #endif
      {
      currentFitness += GetCheckerLineFactor(img,expectedPoint,outsidePoint)
                            / (double)attachPoints.size();
      }
    }

    // Add the quality of the line connecting the new points to the fitness
    for (int i = 1; i < (int)attachPoints.size(); i++){
      currentFitness += GetCheckerLineFactor(img,expectedPoints[i-1],expectedPoints[i])
                                         / (double)(expectedPoints.size()-1);

    }

    // Update the best fitness, if this iteration was better
    if (currentFitness > bestFitness || searchDir == 0){
      bestFitness = currentFitness;
      bestSearchDir = searchDir;
      bestNewPoints = expectedPoints;
    }
  }

 /* cout << "Creating new structure, checking if topology is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/

  switch(bestSearchDir){
  case 0:
    for (int i = 0; i < (int)rectTopolog.size(); i++){ // 0 <=> row negative
      rectTopolog[i].push_front((int)knotPositions.size());
      knotPositions.push_back(bestNewPoints[i]);
    }
    break;
  case 1:
    for (int i = 0; i < (int)rectTopolog.size(); i++){ // 1 <=> row positive
      rectTopolog[i].push_back((int)knotPositions.size());
      knotPositions.push_back(bestNewPoints[i]);
    }
    break;
  case 2:
    {deque<int> newRow;
    for (int i = 0; i < (int)rectTopolog[0].size(); i++){ // 2 <=> col negative
      newRow.push_back((int)knotPositions.size());
      knotPositions.push_back(bestNewPoints[i]);
    }
    rectTopolog.push_front(newRow);}
    break;
  case 3:
    {deque<int> newRow;
    for (int i = 0; i < (int)rectTopolog[0].size(); i++){ // 3 <=> col positive
      newRow.push_back((int)knotPositions.size());
      knotPositions.push_back(bestNewPoints[i]);
    }
    rectTopolog.push_back(newRow);}
    break;

  }
  /*cout << "Created new structure, checking if topology is still valid:  ";
  for (int qwe  = 0; qwe < rectTopolog.size(); qwe++)
    cout << rectTopolog[qwe].size() << " ";
  cout << endl;*/
}


vector<HomgPoint2D>
    CheckerboardDetectorCrossFilter::GetFinalCheckerboard(Image<unsigned char> &img,
                                                    vector<HomgPoint2D> &knotPositions,
                                                    deque< deque< int > > &rectTopolog,
                                                    int numberOfRows, int numberOfCols,
                                                    bool upperLeftIsBlack){

  // Flip if rows and columns are swaped:
  if ((int)rectTopolog.size() != numberOfRows){
    deque< deque <int> > newrectTopolog;
    newrectTopolog.resize(rectTopolog[0].size());
    for (int i = 0; i < (int)rectTopolog[0].size(); i++)
      newrectTopolog[i].resize(rectTopolog.size());
    for (int i = 0; i < (int)rectTopolog.size(); i++)
      for (int j = 0; j < (int)rectTopolog[0].size(); j++)
        newrectTopolog[j][i] = rectTopolog[i][j];
    rectTopolog = newrectTopolog;
  }

  // Make the upper left black
  HomgPoint2D pos;
  unsigned char** ida =  img.GetImageDataArray();

  pos = knotPositions[rectTopolog[0][0]] + knotPositions[rectTopolog[0][1]]
             + knotPositions[rectTopolog[1][0]] + knotPositions[rectTopolog[1][1]];
  double upperLeftCol = ida[(int)rint(pos[1]/4.0)][(int)rint(pos[0]/4.0)*img.GetChannelCount()];

  pos = knotPositions[rectTopolog[0][1]] + knotPositions[rectTopolog[0][2]]
             + knotPositions[rectTopolog[1][1]] + knotPositions[rectTopolog[1][2]];
  double otherCol = ida[(int)rint(pos[1]/4.0)][(int)rint(pos[0]/4.0)*img.GetChannelCount()];

  pos = knotPositions[rectTopolog[1][0]] + knotPositions[rectTopolog[1][1]]
             + knotPositions[rectTopolog[2][0]] + knotPositions[rectTopolog[2][1]];
  otherCol += ida[(int)rint(pos[1]/4.0)][(int)rint(pos[0]/4.0)*img.GetChannelCount()];


  // Try to change this by reversing the side that has an uneven number of points:
  if (2.0 * upperLeftCol > otherCol){

    if  (numberOfCols %2 == 1){
      for (int i = 0; i < (int)rectTopolog.size(); i++)
        for (int j = 0; j <  (int)rectTopolog[0].size()/2; j++){
          int tmp = rectTopolog[i][j];
          rectTopolog[i][j] = rectTopolog[i][rectTopolog[0].size() - 1 - j];
          rectTopolog[i][rectTopolog[0].size() - 1 - j] = tmp;
         }
      }
    } else {

      for (int i = 0; i < (int)rectTopolog.size()/2; i++)
        for (int j = 0; j < (int)rectTopolog[0].size(); j++){
          int tmp = rectTopolog[i][j];
          rectTopolog[i][j] = rectTopolog[rectTopolog.size() - 1 - i][j];
          rectTopolog[rectTopolog.size() - 1 - i][j] = tmp;
         }
     }

    // Check if we are looking from the right side to the checkerboard:
    Vector3<double> vecA = Vector3<double>(knotPositions[rectTopolog[0][1]][0]-knotPositions[rectTopolog[0][0]][0],
                                   knotPositions[rectTopolog[0][1]][1]-knotPositions[rectTopolog[0][0]][1], 0.0);
    Vector3<double> vecB = Vector3<double>(knotPositions[rectTopolog[1][0]][0]-knotPositions[rectTopolog[0][0]][0],
                                   knotPositions[rectTopolog[1][0]][1]-knotPositions[rectTopolog[0][0]][1], 0.0);
    Vector3<double> crossProduct = vecA.CrossProduct(vecB);

    if (crossProduct[2] > 0.0){
;
      if  (numberOfCols %2 == 0){
        for (int i = 0; i < (int)rectTopolog.size(); i++)
          for (int j = 0; j <  (int)rectTopolog[0].size()/2; j++){
            int tmp = rectTopolog[i][j];
            rectTopolog[i][j] = rectTopolog[i][rectTopolog[0].size() - 1 - j];
            rectTopolog[i][rectTopolog[0].size() - 1 - j] = tmp;
           }
        }
      } else {

        for (int i = 0; i < (int)rectTopolog.size()/2; i++)
          for (int j = 0; j < (int)rectTopolog[0].size(); j++){
            int tmp = rectTopolog[i][j];
            rectTopolog[i][j] = rectTopolog[rectTopolog.size() - 1 - i][j];
            rectTopolog[rectTopolog.size() - 1 - i][j] = tmp;
           }
       }

      // Push everything on the return vector:
      vector<HomgPoint2D> retVec;
      for (unsigned int r=0;r<rectTopolog.size();r++)
        for (unsigned int c = 0; c < rectTopolog[r].size(); c++)
          retVec.push_back(knotPositions[rectTopolog[r][c]]);

      return retVec;

  }





  #ifdef CHECKER_DETECT_DEBUG_MODE
  vector<HomgPoint2D> CheckerboardDetectorCrossFilter::GetCheckerboard(Image<unsigned char> &img,
                                                      int numberOfCols, int numberOfRows,
                                                      bool upperLeftIsBlack){
  #else
  vector<HomgPoint2D> CheckerboardDetectorCrossFilter::GetCheckerboard(Image<unsigned char> img,
                                                        int numberOfCols, int numberOfRows,
                                                        bool upperLeftIsBlack){
  #endif
  vector<HomgPoint2D> errorResult;
  // Detect the checkerboard corners in an image

  int filterThreshold = initialFilterThreshold_;
  double whiteColor, blackColor;
  unsigned char** ida = img.GetImageDataArray();

  //int width = (int)img.GetWidth() -defSize;
  //int height = (int)img.GetHeight() -winSize;
  int channels = (int)img.GetChannelCount();

  vector<HomgPoint2D> knotPositions;

  // Try to find enough filter response clusters to start the detection
  bool foundEnoughPoints = false;
  int filterCycles = 0;

  //cout << "Prefiltering the image..." << endl;

  while (!foundEnoughPoints){
    filterCycles++;
    // Filter the image. Skip the fast search if this the cycle counter is > 3.
    vector<HomgPoint2D> edges;
    GetFilterResponseSum(img, edges,(float)filterThreshold, defaultHalfWinSize_, filterCycles > 2);

    // Perform a CCA
    GetCCACenter(edges,knotPositions,img.GetWidth(),img.GetHeight());
#ifdef CHECKER_DETECT_DEBUG_MODE
    cout << "Edges : " << edges.size() << "              Knots : " << knotPositions.size() << endl;
#endif

    // Lower the threshold if it was to high
    if ((int)knotPositions.size() < (numberOfCols * numberOfRows * 2) / 3) filterThreshold -= 10000;
    else foundEnoughPoints = true;
    if (filterThreshold < 10000.0) {
       #ifdef CHECKER_DETECT_DEBUG_MODE
         deque<deque< int> > tmp2;
         vector<pair<int,int> > tmp;
         DrawDebugImage(img,knotPositions,tmp,tmp2);
      #endif
      return errorResult;
    }
  }

   //cout << "done." << endl;

  // Write the filter results in the homogenous component of the position vector
  for (int i = 0; i < (int)knotPositions.size(); i++)
     knotPositions[i][2] = sumFilter(ida,(int)knotPositions[i][0],
                                     (int)knotPositions[i][1],
                                     channels,defaultHalfWinSize_,M_PI * 0.05);


  // Normalize the image values so the checkboard becomes black - white
  GetAvgBlackWhiteValue(img,knotPositions,blackColor,whiteColor,defaultAngleSteps_ );
  StretchImageValues(img,(int)rint(blackColor),(int)rint(whiteColor));


  // Remove all cluster centers that appear to be double
  vector<HomgPoint2D> cleanknotPositions;
  vector<int> markForRemoval;
  for (int i = 0; i < (int)knotPositions.size(); i++)
    if (find(markForRemoval.begin(),markForRemoval.end(),i)==markForRemoval.end()){
      for (int j = i+1; j < (int)knotPositions.size(); j++){
        if (NonNormalizedSquaredLength(knotPositions[i],knotPositions[j]) < 100.0){
          markForRemoval.push_back(j);
          knotPositions[i][0] = (knotPositions[i][0] + knotPositions[j][0])/2.0;
          knotPositions[i][1] = (knotPositions[i][1] + knotPositions[j][1])/2.0;
        }
      }
      cleanknotPositions.push_back(knotPositions[i]);
    }
  knotPositions = cleanknotPositions;
#ifdef CHECKER_DETECT_DEBUG_MODE
  cout << "Knots after removing doubles: " << knotPositions.size() << endl;
#endif
  // Build a topology on the knots:
  bool foundEnoughFullConnectedKnots = false;
  double lineThreshold = initialLineThreshold_;
  vector<pair<int,int> > topology;
  vector<vector<int> > adjacence;
  deque<deque<int> > rectTopolog;
  while (!foundEnoughFullConnectedKnots){

    topology.clear();
    for (int i = 0; i < (int)knotPositions.size(); i++)
      for (int j = i+1; j < (int)knotPositions.size(); j++)
        if (i != j && fabs(knotPositions[i][0] - knotPositions[j][0]) < 300
                   && fabs(knotPositions[i][1] - knotPositions[j][1]) < 300
                   && GetCheckerLineFactor(img,knotPositions[i],knotPositions[j]) > lineThreshold)
          topology.push_back(pair<int,int>(i,j));
#ifdef CHECKER_DETECT_DEBUG_MODE
    cout << "Topology: " << topology.size() << endl;
#endif

    // Build adjacence vector. This vector contains the set of neighbors for each entry
    adjacence.clear();
    adjacence.resize(knotPositions.size());
    for (int i = 0; i < (int)topology.size(); i++){
      adjacence[topology[i].first].push_back(topology[i].second);
      adjacence[topology[i].second].push_back(topology[i].first);
    }
    int numberOfFullyConnectedKnots= 0;
    for (int i = 0; i < (int)knotPositions.size(); i++)
      if (adjacence[i].size() > 3) numberOfFullyConnectedKnots++;



    // Generate a rectangular topology out of the adjacences
    GenerateRectangularTopology(img,knotPositions,adjacence,rectTopolog);
#ifdef CHECKER_DETECT_DEBUG_MODE
    cout << "RectTopolg: " << rectTopolog.size() << endl;
#endif
    if (rectTopolog.size() >= 3 && rectTopolog[0].size() >= 3) foundEnoughFullConnectedKnots = true;
    else lineThreshold -= 1000.0;
    if (lineThreshold < 1000.0) break;
  }

  SearchForMissingTopologyEntries(img,knotPositions,rectTopolog);

  if (rectTopolog.size() < 3 || rectTopolog[0].size() < 3){
    #ifdef CHECKER_DETECT_DEBUG_MODE
       DrawDebugImage(img,knotPositions,topology,rectTopolog);
    #endif
    return errorResult;
  }


  // Extend the rectangular topology while it has not the same size as the checkerboard
  // is expected to be
  int maxSize = (numberOfRows > numberOfCols ? numberOfRows : numberOfCols);
  while(!(((int)rectTopolog.size() == numberOfRows && (int)rectTopolog[0].size() == numberOfCols)
        || ((int)rectTopolog.size() == numberOfCols && (int)rectTopolog[0].size() == numberOfRows)
        || (int)rectTopolog.size() > maxSize  || (int)rectTopolog[0].size() > maxSize)){
    ExtendTopology(img,knotPositions,rectTopolog);
    if (((int)rectTopolog.size() > numberOfRows && (int)rectTopolog[0].size() > numberOfCols) ||
        ((int)rectTopolog.size() > numberOfCols && (int)rectTopolog[0].size() > numberOfRows))
      break;
  }

  if ((int)rectTopolog.size() * (int)rectTopolog[0].size() != numberOfCols * numberOfRows) return errorResult;
  
  // Check if all knot positions are inside the image
  int width = img.GetWidth();
  int height = img.GetHeight();
  for (int kpos = 0; kpos < (int)knotPositions.size(); kpos++)
    if (knotPositions[kpos][0] < 0.0 || knotPositions[kpos][1] < 0.0 ||
        knotPositions[kpos][0] >= width - 1|| knotPositions[kpos][1] >= height - 1)
      return errorResult;
      

  // Flip and rotate the checkerboard into its unique orientation
  vector<HomgPoint2D> retVal =
      GetFinalCheckerboard(img, knotPositions, rectTopolog, numberOfRows, numberOfCols, upperLeftIsBlack);

  //TODO: Perform a final check on the corners to avoid false positives

  #ifdef CHECKER_DETECT_DEBUG_MODE
  DrawDebugImage(img,knotPositions,topology,rectTopolog);
  #endif


  return retVal;

}

  void CheckerboardDetectorCrossFilter::DrawDebugImage(Image<unsigned char> &img,
                                                       std::vector<HomgPoint2D> &knotPositions,
                                                       vector<pair<int,int> > topology,
                                                       std::deque< std::deque< int > > &rectTopolog){


      // Draw the initial topology
      unsigned char lineCol[]  = {0,0,255};
      for (int i = 0; i < (int)topology.size(); i++)
        ImageDraw<unsigned char>::InterpolatedLine( img,knotPositions[topology[i].first],
                                                   knotPositions[topology[i].second],lineCol,10.0f,0.15f);



      //Paint the rows of the rectTopolog:
      unsigned char rowCol[]  = {255,0,0};
      for (int c = 0; c < (int)rectTopolog.size(); c++)
        for (int r = 0; r < (int)rectTopolog[c].size()-1;r++)
          if (rectTopolog[c][r] > -1 && rectTopolog[c][r+1] > -1)
            ImageDraw<unsigned char>::InterpolatedLine(img,knotPositions[rectTopolog[c][r]],knotPositions[rectTopolog[c][r+1]],rowCol,1.7f,1.0f);

      //Paint the cols of the rectTopolog:
      unsigned char colCol[]  = {0,255,0};
      if (rectTopolog.size() > 0)
      for (int r = 0; r < (int)rectTopolog[0].size(); r++)
        for (int c = 0; c < (int)rectTopolog.size()-1;c++)
          if (rectTopolog[c][r] > -1 && rectTopolog[c+1][r] > -1)
            ImageDraw<unsigned char>::InterpolatedLine(img,knotPositions[rectTopolog[c][r]],knotPositions[rectTopolog[c+1][r]],colCol,1.8f,1.0f);

      unsigned char colSum[]  = {0,255,0};

      for (int i = 0; i < (int)rectTopolog.size(); i++)
        for (int j = 0; j < (int)rectTopolog[0].size();j++){
          ImageDraw<unsigned char>::InterpolatedCircleCenter(img,knotPositions[rectTopolog[i][j]],7,colSum,1.0f,1.0f);
        ImgText(debugImage_,i*rectTopolog[0].size()+j,(int)knotPositions[rectTopolog[i][j]][0],(int)knotPositions[rectTopolog[i][j]][1]);
      }

      debugImage_ = img;


  }


bool CheckerboardDetectorCrossFilter::doLinesIntersect(HomgPoint2D &a, HomgPoint2D &b,HomgPoint2D &c, HomgPoint2D &d){
  double ua = ((d[0]-c[0])*(a[1]-c[1]) - (d[1]-c[1])*(a[0]-c[0]));
  ua /= ((d[1]-c[1])*(b[0]-a[0]) - (d[0]-c[0])*(b[1]-a[1]));

  double ub = ((b[0]-a[0])*(a[1]-c[1])-(b[1]-a[1])*(a[0]-c[0]));
  ub /= ((d[1]-c[1])*(b[0]-a[0])-(d[0]-c[0])*(b[1]-a[1]));
  if (ua > 0.0 && ua < 1.0 && ub > 0.0 && ub < 1.0) return true;
  return false;
}

HomgPoint2D CheckerboardDetectorCrossFilter::lineIntersection(HomgPoint2D &a, HomgPoint2D &b,HomgPoint2D &c, HomgPoint2D &d){
  double ua = ((d[0]-c[0])*(a[1]-c[1]) - (d[1]-c[1])*(a[0]-c[0]));
  ua /= ((d[1]-c[1])*(b[0]-a[0]) - (d[0]-c[0])*(b[1]-a[1]));
  return a + ((b*ua)-(a* ua)) ;
}


double CheckerboardDetectorCrossFilter::getAverage(BIAS::Image<unsigned char> &img, BIAS::HomgPoint2D center, int windowSize){
  int minX = (int)center[0] - windowSize;
  int maxX = (int)center[0] + windowSize;
  int minY = (int)center[1] - windowSize;
  int maxY = (int)center[1] + windowSize;
  if (minX < 0) minX = 0;
  if (minY < 0) minY = 0;
  if (maxX > (int)img.GetWidth()) maxX = (int)img.GetWidth();
  if (maxY > (int)img.GetHeight()) maxY = (int)img.GetHeight();
  double numPx = (double)(maxX - minX) * (double)(maxY - minY) * 3.0;
  double pxSum = 0.0;
  for (int x = minX; x < maxX; x++)
    for (int y = minY; y < maxY; y++)
      for (int c = 0; c < 3; c++)
      pxSum += (double)img.GetImageDataArray()[y][x*3+c];

  return pxSum / numPx;

}

double CheckerboardDetectorCrossFilter::GetCheckerLineFactor(Image<unsigned char> img, HomgPoint2D a, HomgPoint2D b){
  int samples = 20;
  double sumLeft = 0.0;
  double sumRight = 0.0;
  double retVal = 0.0;
  int height = img.GetHeight();
  int width = img.GetWidth();
  unsigned char** ida = img.GetImageDataArray();

  for (int sample = 0; sample < samples; sample++){
    double ratio = (double)(sample + 1) / (double)(samples+1);
    HomgPoint2D p;
    p[0] = (1.0 - ratio) * a[0] + ratio * (b[0]);
    p[1] = (1.0 - ratio) * a[1] + ratio * (b[1]);
    if (ratio > 0.5) ratio = 1.0 - ratio;

    int leftX = (int)rint(p[0] + 0.2 * ratio * (b[1]-a[1]));
    int leftY = (int)rint(p[1] - 0.2 * ratio * (b[0]-a[0]));

    int rightX = (int)rint(p[0] - 0.2 * ratio * (b[1]-a[1]));
    int rightY = (int)rint(p[1] + 0.2 * ratio * (b[0]-a[0]));


    if (leftY >= 0 && leftY < height && leftX >= 0 && leftX < width &&
        rightY >= 0 && rightY < height && rightX >= 0 && rightX < width){
      sumLeft += ida[leftY][leftX*3];
      sumRight += ida[rightY][rightX*3];
      sumLeft += ida[leftY][(leftX*3)+1];
      sumRight += ida[rightY][(rightX*3)+1];
      sumLeft += ida[leftY][(leftX*3)+2];
      sumRight += ida[rightY][(rightX*3)+2];
    }


  }

  retVal += fabs(sumLeft - sumRight);
  return retVal * 0.7; // 0.7 is the balance factor between the filter reponse and this line factor (normalization)
}


double CheckerboardDetectorCrossFilter::GetOpeningAngle(HomgPoint2D from, HomgPoint2D toA, HomgPoint2D toB){
   double angleA = atan2(from[0]-toA[0],from[1]-toA[1]);
   double angleB = atan2(from[0]-toB[0],from[1]-toB[1]);
   double angleDiff = angleA - angleB;
   if (angleDiff < -M_PI) angleDiff += 2.0*M_PI;
   if (angleDiff > M_PI) angleDiff -= 2.0*M_PI;
   return fabs(angleDiff);
}


void CheckerboardDetectorCrossFilter::ImgText(Image<unsigned char> &img, int label, int posX, int posY){
  ostringstream temp;
  temp << label;
  string s = temp.str();

  ImageDraw< unsigned char >::Text(  	img, s, posX-1, posY-1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX-1, posY+1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX+1, posY-1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX+1, posY+1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX, posY-1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX, posY+1, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX-1, posY, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX+1, posY, ColourRGB< unsigned char >(0,0,0));
  ImageDraw< unsigned char >::Text(  	img, s, posX, posY, ColourRGB< unsigned char >(255,255,255));
}

bool CheckerboardDetectorCrossFilter::CouldBeNeighbour(const HomgPoint2D &edgePoint,
                                                       const HomgPoint2D &innerPoint,
                                                       const HomgPoint2D &candidate){

  if (fabs(GetOpeningAngle(edgePoint,innerPoint,candidate)) <= M_PI * 0.9) return false;
  double innerLength = NonNormalizedLength(edgePoint ,innerPoint);
  double candidateLength = NonNormalizedLength(edgePoint, candidate);
  if (innerLength * 1.9< candidateLength ) return false;
  if (innerLength / 1.9 > candidateLength ) return false;
  return true;

}

void CheckerboardDetectorCrossFilter::GetSearchRegionForNeighbor(const HomgPoint2D &edgeA, const HomgPoint2D &edgeB,
                                                                 const HomgPoint2D &innerEdge, HomgPoint2D &searchCenter,
                                                                 double &searchRadius){
  HomgPoint2D mirrorPoint = HomgPoint2D(0.0,0.0,0.0);
  mirrorPoint[0] = 0.5 * edgeA[0] + 0.5 * edgeB[0];
  mirrorPoint[1] = 0.5 * edgeA[1] + 0.5 * edgeB[1];
  mirrorPoint = innerEdge - mirrorPoint;
  searchCenter = innerEdge - mirrorPoint - mirrorPoint;
  mirrorPoint[0] = fabs(mirrorPoint[0]);
  mirrorPoint[1] = fabs(mirrorPoint[1]);
  searchRadius = sqrt(mirrorPoint[0]*mirrorPoint[0] + mirrorPoint[1]*mirrorPoint[1]);
  /*cout << "Calculated search radius " << searchRadius << endl;
  cout << "   Calculated from " << edgeA << endl;
  cout << "                   " << edgeB << endl;
  cout << "                   " << innerEdge << endl;
  cout << "                   " << searchCenter << endl;*/
}

double CheckerboardDetectorCrossFilter::GetEdgeFitness(Image<unsigned char> &img,const HomgPoint2D &edgeA,
                                                       const HomgPoint2D &edgeB, const HomgPoint2D &newEdge,
                                                       int halfWinSize){

  double fitness = GetCheckerLineFactor(img, edgeA, newEdge);
  fitness += GetCheckerLineFactor(img,edgeB,newEdge);
  fitness += GetSingleFilterResponse(img,newEdge,halfWinSize) * 3.0;
  return fitness;
}

double CheckerboardDetectorCrossFilter::GetExtensionFitness(Image<unsigned char> &img, const HomgPoint2D &edge,
                                                            const HomgPoint2D &newEdge, int halfWinSize){

  double fitness = GetCheckerLineFactor(img, edge, newEdge) * 2.0;
  fitness += GetSingleFilterResponse(img,newEdge,halfWinSize) * (30.0 / (double)halfWinSize);
  return fitness;

}

double CheckerboardDetectorCrossFilter::GetSingleFilterResponse(Image<unsigned char> &img,
                                                                const HomgPoint2D &edge, int halfWinSize){

  int edgeX = (int)rint(edge[0]);
  int edgeY = (int)rint(edge[1]);
  if (edgeX < halfWinSize){
    if (edgeX < 0) return -DBL_MAX;
    halfWinSize = edgeX;
  }
  if (edgeY < halfWinSize){
    if (edgeY < 0.0) return -DBL_MAX;
    halfWinSize = (int)edgeY;
  }
  if (edgeX >= (int)img.GetWidth() - halfWinSize){
    if (edgeX >= (int)img.GetWidth()) return -DBL_MAX;
    halfWinSize = (int)img.GetWidth() - edgeX - 1;
  }
  if (edgeY >= (int)img.GetHeight() - halfWinSize){
    if (edgeY >= (int)img.GetHeight()) return -DBL_MAX;
    halfWinSize = (int)img.GetHeight() - edgeY - 1;
  }

  double fitness = 0.0;
  for (double angle = 0.0; angle < M_PI/2.0; angle += M_PI / defaultAngleSteps_ )
    fitness += filter(img.GetImageDataArray(), edgeX, edgeY,
                      img.GetChannelCount(), halfWinSize, sin(angle), cos(angle))/32.0;

  return fitness;

}


HomgPoint2D CheckerboardDetectorCrossFilter::FindNewEdgePosition(Image<unsigned char> &img, const HomgPoint2D &edgeA,
                                                                 const HomgPoint2D &edgeB, const HomgPoint2D avoidPoint,
                                                                 HomgPoint2D searchCenter, double searchRadius){

  double forbiddenRadius = searchRadius / 2.0;
  searchRadius /= 4.0;
  int halfWinSize = defaultHalfWinSize_;
  // Make the filter smaller if the checkerboard fields seem to be tiny
  if (searchRadius < 1.5 * (double)defaultHalfWinSize_) halfWinSize = defaultHalfWinSize_ / 2;

  HomgPoint2D bestResult = searchCenter;
  if (NonNormalizedSquaredLength(bestResult,edgeA) > forbiddenRadius*forbiddenRadius
      && NonNormalizedSquaredLength(bestResult,edgeB) > forbiddenRadius*forbiddenRadius)
            bestResult[2] = GetEdgeFitness(img,edgeA,edgeB,bestResult,halfWinSize);
  else bestResult[2] = -DBL_MAX;

  //bool moved = false;

  while(searchRadius > 1.0){

    bool moved = false;
    for (double searchAngle = 0.0; searchAngle < 2.0*M_PI; searchAngle += M_PI/3.0){
       HomgPoint2D searchPoint = HomgPoint2D(searchCenter[0]
                                 + sin(searchAngle)*searchRadius, searchCenter[1]
                                 + cos(searchAngle)*searchRadius, -DBL_MAX);

       if (NonNormalizedSquaredLength(searchPoint,edgeA) > forbiddenRadius*forbiddenRadius
           && NonNormalizedSquaredLength(searchPoint,edgeB) > forbiddenRadius*forbiddenRadius
           && NonNormalizedSquaredLength(avoidPoint,searchPoint) > forbiddenRadius*forbiddenRadius)
                 searchPoint[2] = GetEdgeFitness(img,edgeA,edgeB,searchPoint,halfWinSize);

       //if (searchPoint[2] > bestResult[2]){
       // searchPoint[2] = bestResult[2] * (1.0 - ((NonNormalizedLength(searchPoint,searchCenter)/searchRadius * 4.0)*(NonNormalizedLength(searchPoint,searchCenter)/searchRadius * 4.0)));
         if (searchPoint[2] > bestResult[2]){
           bestResult = searchPoint;
           moved = true;
        // }
       }
    }
    searchCenter = bestResult;
    //searchRadius *= 0.7;
    if (moved) searchRadius *= 0.7;
    else searchRadius *= 0.5;
  }


  return searchCenter;

}




HomgPoint2D CheckerboardDetectorCrossFilter::FindNewExtensionPosition(Image<unsigned char> &img, const HomgPoint2D &edge,
                                                               const HomgPoint2D avoidPointA, const HomgPoint2D avoidPointB,
                                                               HomgPoint2D searchCenter, double searchRadius){

  double forbiddenRadius = searchRadius / 2.0;
  searchRadius /= 4.0;
  int halfWinSize = defaultHalfWinSize_;
  // Make the filter smaller if the checkerboard fields seem to be tiny
  if (searchRadius < 1.5 * (double)defaultHalfWinSize_) halfWinSize = defaultHalfWinSize_ / 2;



  HomgPoint2D bestResult = searchCenter;
  if (NonNormalizedSquaredLength(bestResult,edge) >  forbiddenRadius*forbiddenRadius
      && NonNormalizedSquaredLength(bestResult,avoidPointA) >  forbiddenRadius*forbiddenRadius
      && NonNormalizedSquaredLength(avoidPointB,edge) >  forbiddenRadius*forbiddenRadius)
            bestResult[2] = GetExtensionFitness(img,edge,bestResult,halfWinSize);
  else bestResult[2] = -DBL_MAX;
  bool moved = false;
  //cout << "Entering Extension search with radius " << searchRadius << endl;
  while(searchRadius > 1.0){
    for (double searchAngle = 0.0; searchAngle < 2.0*M_PI; searchAngle += M_PI/2.0){
       HomgPoint2D searchPoint = HomgPoint2D(searchCenter[0]
                                 + sin(searchAngle)*searchRadius, searchCenter[1]
                                 + cos(searchAngle)*searchRadius, -DBL_MAX);

       if (NonNormalizedSquaredLength(searchPoint,edge) >  forbiddenRadius*forbiddenRadius
           && NonNormalizedSquaredLength(searchPoint,avoidPointA) >  forbiddenRadius*forbiddenRadius
           && NonNormalizedSquaredLength(avoidPointB,searchPoint) >  forbiddenRadius*forbiddenRadius)
                 searchPoint[2] = GetExtensionFitness(img,edge,searchPoint,halfWinSize);

       //if (searchPoint[2] > bestResult[2]){

       //  searchPoint[2] = bestResult[2] * (1.0 - ((NonNormalizedLength(searchPoint,searchCenter)/searchRadius * 4.0)*(NonNormalizedLength(searchPoint,searchCenter)/searchRadius * 4.0)));
         if (searchPoint[2] > bestResult[2]){
           bestResult = searchPoint;
           moved = true;
        // }
       }
    }
   // cout << "Exiting Extension search " << endl;
    searchCenter = bestResult;
    //searchRadius *= 0.7;
    if (moved) searchRadius *= 0.7;
    else searchRadius *= 0.5;

  }

  return searchCenter;

}
