#ifndef CHECKERBOARDDETECT_HH
#define CHECKERBOARDDETECT_HH

#include <vector>
#include <deque>
#include <algorithm>

#include <Base/Math/Vector2.hh>
#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgLine2D.hh>

// / Overwrites the work image with the debug image
//#define CHECKER_DETECT_DEBUG_MODE


namespace BIAS{

#ifndef __CheckerboardDetector_hh__
/// Compatibility definition to provide the same interface as CheckerboardDetector
typedef std::vector< std::pair<float,float> > t_CoordVecf;
#endif

/**
  * @class CheckerboardDetectorCrossFilter
  * @brief Detects a checkerboard pattern (typically 7x4 inner
  *        corners, asymmetric layout recommended) in an image using a cross shaped
  *        image filter
  * @note This checker board detecter uses a rotating cross filter that sums up the absolute differences
  *       between the two lines of that cross along with a symmetry penalty term. In addition, a line fitness
  *       based on differences of the line vicinity and homogeneity is used together with a lot of heuristics.
  * @author Andreas Jordt
  * @date 2010
  * @ingroup g_feature
*/

class BIASFeatureDetector_EXPORT CheckerboardDetectorCrossFilter {



public:

  /// Standard constructor
  CheckerboardDetectorCrossFilter();

  virtual ~CheckerboardDetectorCrossFilter();

  /** Find a checkerboard of size numberOfCols x numberOfRows in the given image
    * @param img Input image
    * @param numberOfCols The number of corners on the checkerboard
    *                     in horizontal direction
    * @param numberOfRows The number of corners on the checkerboard in vertical direction
    * @param upperLeftIsBlack Set true if the field between corners (0,0), (0,1),
    *        (1,0) and (1,1) is black
    */
    #ifdef CHECKER_DETECT_DEBUG_MODE
    std::vector<HomgPoint2D> GetCheckerboard(Image<unsigned char> &img, int numberOfCols = 7,
                                             int numberOfRows = 4, bool upperLeftIsBlack = true);
    #else
    std::vector<HomgPoint2D> GetCheckerboard(Image<unsigned char> img, int numberOfCols = 7,
                                             int numberOfRows = 4, bool upperLeftIsBlack = true);
    #endif


    /** Detects a (recommended asymmetric) checkerboard pattern.
      * @param img Input image
      * @param xCorners Number of inner checkerboard corners in x direction
      * @param yCorners Number of inner checkerboard corners in y direction
      * @param coords The resulting set of corner coordinates
      * @param saveDebugImage ignored (just for compatibility)
      * @return 0 for success, <0 for error (e.g. could not detect all inner corners).
      */
    static int Compute(BIAS::Image<unsigned char> &img,
                       const unsigned int &xCorners,
                       const unsigned int & yCorners,
                       t_CoordVecf &coords,
                       const bool &saveDebugImage = false);

protected:

  /** Calculate a connected components analysis for a given set of pixel coordinates.
    * @param pixels The input pixel coordinates
    * @param centers The centers of weight of the resulting cluster
    * @param width The maximum x component of the input pixels
    * @param height The maximum y component of the input pixels
    */
  void GetCCACenter(const std::vector<HomgPoint2D> pixels, std::vector<HomgPoint2D> &centers,
                    int width, int height);

  /** Calculate the filter result for the whole image and return the pixel coordinates above
    *  the given threshold
    * @param img Input image
    * @param response The set of pixels above the threshold
    * @param threshold Filter threshold
    * @param halfWinSize The size of the cross filter
    * @param fastMode speed up the filter search but risk missing some points
    */
  void GetFilterResponseSum(Image<unsigned char> &img, std::vector<HomgPoint2D> &response,
                            float threshold = 25000.0, int halfWinSize = 20, bool fastMode = true);

  /** Stretch the image values such that from => 0 and to => 255
    * @param img Input image
    * @param from The value that is pulled to 0
    * @param to The value that is pushed to 255
    */
  void StretchImageValues(Image<unsigned char> &img, unsigned int from, unsigned int to);

  /** Check if the finite lines a - b and c - d intersect
    */
  bool doLinesIntersect(HomgPoint2D &a, HomgPoint2D &b,HomgPoint2D &c, HomgPoint2D &d);

  /** Calculate the intersection point of the infinite lines a - b and c - d
    */
  HomgPoint2D lineIntersection(HomgPoint2D &a, HomgPoint2D &b,HomgPoint2D &c, HomgPoint2D &d);

  /** Get the average image grey value (also for color images) in a window
    * @param img Input image
    * @param center The center of the square window
    * @param halfWinSize The half size of the square the value is calculated from
    */
  double getAverage(Image<unsigned char> &img, HomgPoint2D center, int halfWinSize);

  /** Calculate the average black and white values of the cross filter at a given set of points
    * @param img Input image
    * @param points The set of positions the filters are evaluated
    * @param avgBlack The resulting average black value
    * @param avgWhite The resulting average white value
    * @param angleStep The number of rotations of the cross filter
    */
  void GetAvgBlackWhiteValue(Image<unsigned char> &img,std::vector<HomgPoint2D> points,
                             double &avgBlack, double &avgWhite, int angleSteps);

  /** Calculate a fitness value for the line between a and b to be a checkerboard line
    */
  double GetCheckerLineFactor(Image<unsigned char> img, HomgPoint2D a, HomgPoint2D b);

  /** Generate a rectangular topology out of the adjcence lists.
    * @param img Input image
    * @param knotPositions The image coordinates of the knots
    * @param adjacence The adjacence list for each knot
    * @param rectTopolog The resulting rectangular output structure
    */
  void GenerateRectangularTopology(Image<unsigned char> &img,
                                   std::vector<HomgPoint2D> &knotPositions,
                                   std::vector<std::vector<int> > &adjacence,
                                   std::deque<std::deque< int > > &rectTopolog);

  /** Fill the topology entries that are unknown (negative indices) by generating an
    * appropriate knot
    * @param img Input image
    * @param knotPositions The image coordinates of the knots
    * @param rectTopolog The rectangular topology
    */
  void SearchForMissingTopologyEntries(Image<unsigned char> &img,
                                       std::vector<HomgPoint2D> &knotPositions,
                                       std::deque< std::deque< int > > &rectTopolog);

  /** Find the side of the rectangular topology which is most likely not on the edge
    * of the checkerboard and extend the checkerboard to it.
    * @param img Input image
    * @param knotPositions The image coordinates of the knots
    * @param rectTopolog The rectangular topology of the knots
    */
  void ExtendTopology(Image<unsigned char> &img, std::vector<HomgPoint2D> &knotPositions,
                      std::deque< std::deque< int > > &rectTopolog);


  /** Analyse the rectangular topology and rotate and flip the checkerboard to its
    * unique representation
    * @param img Input image
    * @param knotPositions The image coordinates of the knots
    * @param rectTopolog The rectangular topology
    * @param numberOfRows The expected number of rows of the checkerboard
    * @param numberOfCols The expected number of columns of the checkerboard
    * @param upperLeftIsBlack Set true if the field between corners (0,0), (0,1),
    *        (1,0) and (1,1) is black
    */
  std::vector<HomgPoint2D>
      GetFinalCheckerboard(Image<unsigned char> &img, std::vector<HomgPoint2D> &knotPositions,
                                           std::deque< std::deque< int > > &rectTopolog, int numberOfRows,
                                           int numberOfCols, bool upperLeftIsBlack);

  /// The default half window size of the filter window
  int defaultHalfWinSize_;

  /// The default number of rotation stpes performed at each filter evaluation
  int defaultAngleSteps_;

  /// The initial value for the filter response threshold for a pixel to be evaluated further
  int initialFilterThreshold_;

  /// The initial value for the line threshold in the topology generation
  int initialLineThreshold_;

  /// Instance pointer if an instance was created. This hack is needed to stay compatible
  /// to the static function calls of the other checkerboard detectors
  static CheckerboardDetectorCrossFilter* instancePointer_;

private:

  /// Draw debug information onto the image
  void DrawDebugImage(Image<unsigned char> &originalImage,
                      std::vector<HomgPoint2D> &knotPositions,
                      std::vector<std::pair<int,int> > topology,
                      std::deque< std::deque< int > > &rectTopolog);

  /// Calculate the fitness value for a point to be the extension to a checkerboard
  double GetExtensionFitness(Image<unsigned char> &img,const HomgPoint2D &edge,
                             const HomgPoint2D &newEdge, int halfWinSize);

  /// Calculate the fitness value for a point to be the missing edge on the checkerboard
  double GetEdgeFitness(Image<unsigned char> &img,const HomgPoint2D &edgeA,
                        const HomgPoint2D &edgeB, const HomgPoint2D &newEdge, int halfWinSize);

  /// Calculate the filter response for a given coordinate
  double GetSingleFilterResponse(Image<unsigned char> &img,const HomgPoint2D &edge, int halfWinSize);

  /// Find the image coordinate that is most likely to be the missing edge
  HomgPoint2D FindNewEdgePosition(Image<unsigned char> &img, const HomgPoint2D &edgeA,
                                  const HomgPoint2D &edgeB, const HomgPoint2D avoidPoint,
                                  HomgPoint2D searchCenter, double searchRadius);

  /// Find the image coordinate that is most likely the extension of the
  /// checkerboard not yet in the topology
  HomgPoint2D FindNewExtensionPosition(Image<unsigned char> &img, const HomgPoint2D &edge,
                                       const HomgPoint2D avoidPointA, const HomgPoint2D avoidPointB,
                                       HomgPoint2D searchCenter, double searchRadius);

  /// Approximate the neighbor corner ny extrapolating the length and the angle of the
  /// given point connection
  void GetSearchRegionForNeighbor(const HomgPoint2D &edgeA, const HomgPoint2D &edgeB,
                                  const HomgPoint2D &innerEdge,
                                   HomgPoint2D &searchCenter, double &searchRadius);

  /// Calculate the euclidean distance between two homogenous point neglecting the 3rd component
  double NonNormalizedLength(const HomgPoint2D &a,const HomgPoint2D &b);


  inline double NonNormalizedSquaredLength(const HomgPoint2D &a,const HomgPoint2D &b){
    return a[0]*a[0]+b[0]*b[0];
  }

  /// Check length and angle of a given candidate to check if this is a possible neighbor
  bool CouldBeNeighbour(const HomgPoint2D &edgePoint, const HomgPoint2D &innerPoint,
                        const HomgPoint2D &candidate);

  /// Write text onto the image
  void ImgText(Image<unsigned char> &img,int label, int posX, int posY);

  /// Calculate the angle between two lines with one common point
  double GetOpeningAngle(HomgPoint2D from, HomgPoint2D toA, HomgPoint2D toB);

  /// Inline implementation of the image filter for one angle
  inline double filter(unsigned char** ida, int x, int y, int channels,
                       int halfWinSize, double sinAngle, double cosAngle){

      double valA = 0.0;
      double valB = 0.0;
      double asym= 0.0;
      for (int c = 0; c < channels; c++){
        for (int walkX = 0; walkX <= halfWinSize; walkX++){
          double pos =
            (double)ida[y+(int)rint((double)walkX*sinAngle)]
                       [(x+(int)rint((double)walkX*cosAngle))*channels+c];
          double neg =
            (double)ida[y+(int)rint((double)(-walkX*sinAngle))]
                       [(x+(int)rint((double)-walkX*cosAngle))*channels+c];
          valA += pos + neg;
          asym += fabs(pos - neg);
        }

        for (int walkY = 0; walkY <= halfWinSize; walkY++){
          double pos =
            (double)ida[y+(int)rint((double)walkY*cosAngle)]
                       [(x-(int)rint((double)walkY*sinAngle))*channels+c];
          double neg =
            (double)ida[y+(int)rint((double)-walkY*cosAngle)]
                       [(x-(int)rint((double)-walkY*sinAngle))*channels+c];
          valB += pos + neg;
          asym += fabs(pos - neg);
        }

      }
      return (fabs(valA - valB) - asym) * 0.5;

  }



  /// Inline implementation of the image filter sum over all orientations
  inline double sumFilter(unsigned char** ida, int x, int y, int channels,
                          int halfWinSize, double angleInc){
    double absDiff = 0.0;
    for (double angle = 0.0; angle < M_PI/2.0; angle += angleInc){
      double sine = sin(angle);
      double cosine = cos(angle);
      absDiff += filter(ida,x,y,channels,halfWinSize,sine,cosine);
    }
    return absDiff;
  }

  /// A sloppy filter implementation to check if a pixel is worth of beeing labeled
  inline double filterFast(unsigned char** ida, int x, int y, int channels,
                       int halfWinSize, double sinAngle, double cosAngle){

      double valA = 0.0;
      double valB = 0.0;
      double asym= 0.0;
      for (int c = 0; c < channels; c++){
        for (int walkX = 1; walkX <= halfWinSize; walkX+=2){
          double pos =
            (double)ida[y+(int)rint((double)walkX*sinAngle)]
                       [(x+(int)rint((double)walkX*cosAngle))*channels+c];
          double neg =
            (double)ida[y+(int)rint((double)(-walkX*sinAngle))]
                       [(x+(int)rint((double)-walkX*cosAngle))*channels+c];
          valA += pos + neg;
          asym += fabs(pos - neg);
        }

        for (int walkY = 1; walkY <= halfWinSize; walkY+=2){
          double pos =
            (double)ida[y+(int)rint((double)walkY*cosAngle)]
                       [(x-(int)rint((double)walkY*sinAngle))*channels+c];
          double neg =
            (double)ida[y+(int)rint((double)-walkY*cosAngle)]
                       [(x-(int)rint((double)-walkY*sinAngle))*channels+c];
          valB += pos + neg;
          asym += fabs(pos - neg);
        }

      }
      return (fabs(valA - valB) - asym) * 0.5;

  }

  Image<unsigned char> debugImage_;

};

}

#endif // CHECKERBOARDDETECT_HH
