/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ConstantRegionDetector_hh__
#define __ConstantRegionDetector_hh__

#include <Base/Image/Image.hh>
#include <Base/Geometry/HomgPoint2D.hh>

namespace BIAS {

/**
     @class ColFeat
     @brief  internal class for feature  (with color) passing only
     @author koeser 09/2004
  */
template<class StorageType>
  class BIASFeatureDetector_EXPORT ColFeat {
  public:
    int x, y;
    double val;
    BIAS::Vector3<StorageType> col;

    bool operator<(const ColFeat<StorageType>& r) const
    { return val>r.val; }

    ColFeat<StorageType>& operator=(const ColFeat<StorageType>& f)
    { x=f.x; y=f.y; val=f.val; col = f.col; return *this; }

//    std::ostream& operator<<(std::ostream & os, const ColFeat<StorageType>& f)
//      { os << "("<<f.x<<", "<<f.y<<") = "<<f.val << " color: " << f.col; return os; }

  }; // end of class Feat




/**
 @class ConstantRegionDetector
 @brief detects regions with close to zero gradients in images, works on color images only
 @author sedlazeck 12/2008
 */

template<class StorageType, class CalculationType>
class BIASFeatureDetector_EXPORT ConstantRegionDetector : public Debug {

public:
  ConstantRegionDetector();
  virtual ~ConstantRegionDetector();

  void SetThreshold(CalculationType threshold){
    threshold_ = threshold;
  };
  CalculationType GetThreshold(){
    return threshold_;
  };

  void SetHalfWinSize(int halfWinSize){
    halfWinSize_ = halfWinSize;
  };
  int GetHalfWinSize(){
    return halfWinSize_;
  };

  void SetMaskImage(BIAS::Image<unsigned char>& maskImg){
    maskImg_ = maskImg;
  };
  
  void SetMinDistance(int minDistance){
    minDistance_ = minDistance;
  };
  
  /**
   * restrict maximum number of features -1 is default and all features are returned.
   */
  void SetMaxNumFeatures(int maxNumFeatures){
    maxNumFeatures_ = maxNumFeatures;
  };

  int GetMaxNumFeatures(){
    return maxNumFeatures_;
  };

  /** @brief Detects constant regions in image src

   Constant regions around pixels in images are detected by a histogram-based method, that is applied to windows around
   the pixel in question. The differences in color are compared.
   threshold is the allowed average deviation from the color of the middle pixel.
   halfWinSize the size of the window used for detection.
   if useMedian is set to true, the input image will be filtered by a median filter first.
   The result is written into a vector with homogenuous 2D points and a segmentation image.
   @author sedlazeck 01/2009
   @todo optimize for ROI usage */
  virtual int Detect(Image<StorageType>& src,
                     std::vector<HomgPoint2D>& points2d, std::vector<Vector3<
                         StorageType> >& colors, bool useMedian = false);


protected:
  CalculationType threshold_;
  int halfWinSize_;
  std::vector<ColFeat<StorageType> > featureList_;
  /// restrict maximum number of features -1 is default and all features are returned.
  int maxNumFeatures_;
  /// required min distance between two constance regions
  int minDistance_;

  /// mask image usefull for water regions in images which usually have very constant regions compared to the rest of the image (set to one if water/ to to be used)
  BIAS::Image<unsigned char> maskImg_;

}; // class

} // namespace


#endif // __ConstantRegionDetector_hh__
