/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __CornerDetectorHarris_hh__
#define __CornerDetectorHarris_hh__

#include "CornerDetectorGradient.hh"

namespace BIAS {

  /** 
      @class CornerDetectorHarris
      @ingroup g_feature
      @brief Harris corner detector, detects grey value corners in images.

      Computes Harris corners as described in:
      [1] C. Harris, M. Stephens: A combined corner and edge detector.
          Proceedings of the 4th Alvey Vision Conference, p.147-151, 1988.
      Corners are found by evaluating the Eigenvalues of the structure
      tensor matrix A. Harris and Stephens note that exact computation of
      the eigenvalues is computationally expensive, since it requires the
      computation of a square root, and instead suggest the following
      cornerness function Mc, where k is a tunable sensitivity parameter:
      (1) Mc = det(A) - k * trace(A)^2
      Values from 0.04 to 0.15 have been found as feasible for k.

      @bug CalculationType int does not work, use float instead.
      @author koeser 09/2004 
  */
  template <class StorageType, class CalculationType>
  class BIASFeatureDetector_EXPORT CornerDetectorHarris 
    : public CornerDetectorGradient<StorageType, CalculationType>
  {
  public:
    CornerDetectorHarris();
    ~CornerDetectorHarris();

    /** @brief Set sensitivity parameter for corner detection (should
               be within the range 0.04 to 0.15, see also [1]).*/
    void SetSensitivityParameter(double k) { _Scale = k; };

    /** @brief Return current value of the sensitivity parameter for
               corner detection (see [1]).*/
    double GetSensitivityParameter() { return _Scale; };
 
  protected:
    /// weight of trace in harris approximation
    double _Scale;

    /** @brief Computes the Harris approximation of the smaller eigenvalue of 
        the structure tensor A over the image and stores it in _Cornerness
        (Mc = det(A) - scale * trace(A)^2). Also fills up _FeatList.
        @author koeser 09/2004 
        @todo optimize for ROI usage */
    virtual int _ComputeCornerness(Image<CalculationType>& Cornerness);
  }; // class

} // namespace


#endif // __CornerDetectorHarris_hh__
