/*
  This file is part of the BIAS library (Basic ImageAlgorithmS).

  Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


  BIAS is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.

  BIAS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with BIAS; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __LEDDETECTOR_HH__
#define __LEDDETECTOR_HH__

#include <bias_config.h>

#ifndef BIAS_HAVE_OPENCV
#  error Please recompile BIAS with USE_OPENCV to use this code. 
#endif

#include <Base/Common/W32Compat.hh>
#include <Base/Image/Image.hh>
#include <Base/Math/Vector3.hh>
#include <cv.h>
#include <vector>

namespace BIAS
{
  /**
     @class LEDDetector
     @brief This class contains methods to detect (IR)-LED-Spots,
            with template matching (new search) or in "recycling" mode
            (search at old positions).
     @ingroup g_feature
     @author haase 3/2007, reworked by esquivel 11/2013
  */
  class BIASFeatureDetector_EXPORT LEDDetector
  {

  public:

    LEDDetector();

    ~LEDDetector();

    /**
      @brief Detect LED spots in image.
      @param[in] image RGB or gray image to detect LED spots in.
      @param[in/out] results Returns LED spot positions in current image.
        Parameters of 3d vectors are (x, y, val) where val is brightness.
        Can be also used as input parameter to specify LED spots positions
        in last image to use "recycling" mode.
      @return Returns <0 in case of errors, >= 0 otherwise.
    */
    int Compute(const BIAS::Image<unsigned char> &image,
                std::vector<BIAS::Vector3<double> > &results);

    /**
      @brief Return attributes of detected LED spots.
      @param[out] minX Returns min. x coordinate for each detected LED spot
      @param[out] maxX Returns max. x coordinate for each detected LED spot
      @param[out] maxValues Returns max. gray value for each detected LED spot
    */
    void GetAttributes(std::vector<double> &minX, std::vector<double> &maxX,
                       std::vector<double> &maxValues);

    /**
      @brief Return neighbours of detected LED spots.
      @param[out] neighbours Returns vector of neighbour positions for each
                             detected LED spot
    */
    void GetNeighbours(std::vector<std::vector<BIAS::Vector3<double> > > &neighbours);

  protected:

    int CheckNeighboursIfWhite_(int x, int y, int& a, int& b);

    int CheckNeighboursIfWhite_inter_(int x, int y);

    void Search_(int x, int y);

    int Recycle_();

    void TemplateMatching_();

    int CheckNeighbours_(int a, int b);

    IplImage *gray_threshold, *gray_orig;
    unsigned char *data;

    std::vector<BIAS::Vector3<double> > resultpixel;
    std::vector<BIAS::Vector3<double> > resultpixel_old;
    std::vector<BIAS::Vector3<double> > whitepixel;

    std::vector<double> minxcoords, maxxcoords, maxvalues;
    std::vector<std::vector<BIAS::Vector3<double> > > nextpixel;

    // lookup matrix (0 = black, 255 = white, 127 = undecided)
    unsigned char *lookup;
    int lookupsize;

    int step;
    int width, height;
    double mean;
    double maxvalue;
    double count;
    double pixelweightvaluex;
    double pixelweightvaluey;
    double recursivedepth;
    double blobsize;
    double valuesum;
    double minx, maxx;
    double blobs;

  };

}

#endif // __LEDDetector_hh__
