/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "CensusTransform.hh"

#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Common/BIASpragma.hh>

using namespace BIAS;
using namespace std;

namespace BIAS {

  //////////////////////////////////////////////////////////////////////////
  //                 implementation
  //////////////////////////////////////////////////////////////////////////

  template<class InputStorageType, class OutputStorageType>
  CensusTransform<InputStorageType, OutputStorageType>::CensusTransform()
    : FilterNToN<InputStorageType,OutputStorageType>()
  {
  }

  template<class InputStorageType, class OutputStorageType>
  CensusTransform<InputStorageType, OutputStorageType>::
  CensusTransform(const CensusTransform<InputStorageType, OutputStorageType>& other)
    : FilterNToN<InputStorageType, OutputStorageType>(other)
  {
  }

  template<class InputStorageType, class OutputStorageType>
  CensusTransform<InputStorageType, OutputStorageType>::~CensusTransform()
  {
  }

  template<class InputStorageType, class OutputStorageType>
  int CensusTransform<InputStorageType, OutputStorageType>::
  Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
  {
    if (src.GetColorModel() != ImageBase::CM_Grey) {
      BIASERR("Source Image is not grey scaled");
      BIASABORT;
      return 0;
    }

    if (src.GetChannelCount() != 1) {
      BIASERR("Source Image is not grey scaled");
      BIASABORT;
      return -1;
    }

    unsigned int width = src.GetWidth();
    unsigned int height = src.GetHeight();

    if (dst.GetWidth() != width || dst.GetHeight() != height ||
        dst.GetChannelCount() != 1)
    {
      BIASERR("Source and Destination Image dimension mismatch");
      BIASABORT;
    }

    const InputStorageType **in = src.GetImageDataArray();
    OutputStorageType **out = dst.GetImageDataArray();

    for (unsigned int y=1; y<height-1;y++) {
      for (unsigned int x=1; x<width-1;x++) {
        // walk 3x3 window
        out[y][x] =  (in[y-1][x-1] < in[y][x]) ? (1 << 7) : 0;
        out[y][x] += (in[y-1][x]   < in[y][x]) ? (1 << 6) : 0;
        out[y][x] += (in[y-1][x+1] < in[y][x]) ? (1 << 5) : 0;
        out[y][x] += (in[y][x-1]   < in[y][x]) ? (1 << 4) : 0;
        out[y][x] += (in[y][x+1]   < in[y][x]) ? (1 << 3) : 0;
        out[y][x] += (in[y+1][x-1] < in[y][x]) ? (1 << 2) : 0;
        out[y][x] += (in[y+1][x]   < in[y][x]) ? (1 << 1) : 0;
        out[y][x] += (in[y+1][x+1] < in[y][x]) ? 1 : 0;
      }
    }

    return 0;
  }

  template <class InputStorageType, class OutputStorageType> 
  int CensusTransform<InputStorageType, OutputStorageType>::
  FilterInt(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
  {
    BIASERR("FilterInt makes no sense here, using general Filter interface.");
    return Filter(src, dst);
  }

  template <class InputStorageType, class OutputStorageType> 
  int CensusTransform<InputStorageType, OutputStorageType>::
  FilterFloat(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) 
  {
    BIASERR("FilterFloat makes no sense here, using general Filter interface.");
    return Filter(src, dst);
  }

  template <class InputStorageType, class OutputStorageType>
  void CensusTransform<InputStorageType, OutputStorageType>::
  GetBordersValid_(int& border_x, int& border_y) const
  {
    BIASERR("No parameter support");
    BIASABORT;
  }


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
#define FILTER_INSTANTIATION_CLASS CensusTransform
#include "Filterinst.hh"

} // namespace BIAS
