/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __DeInterlace_hh__
#define __DeInterlace_hh__

#include "FilterNToN.hh"
#include <Base/Common/BIASpragmaStart.hh>

namespace BIAS 
{

	/**
	 * @enum BIAS_DEINTERLACE_TYPE
	 * @brief the type of deinterlacing:
	 * BIAS_DI_121 : calculate new pixel from old neighbouring pixels with  new = (old1+2*old2+old3)/4
	 * BIAS_DI_22  : calculate new pixel from old neighbouring pixels with  new = (old1+old2)/2
	 */
  enum BIAS_DEINTERLACE_TYPE{
    BIAS_DI_121,
    BIAS_DI_22
  };


  /** @class DeInterlace
	 * @ingroup g_filter
	 * @brief Deinterlacer filter for images. There are several options
	 * how deinterlacing is done. You can use even or odd lines or automatically
	 * downsample the image while deinterlacing. Use the Set...() functions
	 * to choose your options
	 * @author ischiller
	 * @date 01/2005
	 * */
  template <class InputStorageType, class OutputStorageType>
  class BIASFilter_EXPORT DeInterlace 
		: public FilterNToN<InputStorageType, OutputStorageType>
  {
  public:
    DeInterlace();
    DeInterlace(const DeInterlace<InputStorageType, OutputStorageType>& other);
    virtual ~DeInterlace();
    
    /** @brief averages over a region with constant weights
        @author woelk 09/2004 */
    virtual int Filter(const Image<InputStorageType>& src, 
                       Image<OutputStorageType>& dst);

    /** @brief only calls filter, for consistency of params */
    virtual int FilterInt(const Image<InputStorageType>& src, 
                          Image<OutputStorageType>& dst); 
    
    /** @brief only calls filter, for consistency of params */
    virtual int FilterFloat(const Image<InputStorageType>& src, 
                            Image<OutputStorageType>& dst);

    virtual FilterNToN<InputStorageType, OutputStorageType>* Clone() const
    { return new DeInterlace<InputStorageType, OutputStorageType>(); }

    /** @biref applies deinterlacing to each channel sperately */
    virtual int FilterColorImg(const Image<InputStorageType>& src,
                               Image<OutputStorageType>& dst);


    /** Set if the  even or odd lines in the images should be used  */
    void SetUseEvenLines(bool bEvenLines=true);

    /** Set the Type of deinterlacing e.g. mean of 2 pixel or weighting with 121... */
    void SetDeInterlaceType(BIAS_DEINTERLACE_TYPE type);

    /** Set the downsampling factor while filter */
    void SetDownSamplingByTwo(bool bDoDownSamplingByTwo = true );
    
  protected:
    virtual void GetBordersValid_(int& border_x, int& border_y) const {};
  protected:
    bool bUseEvenLines_;
    bool bDoDownSamplingByTwo_;
    BIAS_DEINTERLACE_TYPE eDeInterlaceType_;
  }; // class




} // namespace

#include <Base/Common/BIASpragmaEnd.hh>

#endif // __DeInterlace_hh__
