/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "DensityFilter.hh"

using namespace std;
using namespace BIAS;

template <class InputStorageType, class OutputStorageType>
int DensityFilter<InputStorageType, OutputStorageType>::Filter(const Image<InputStorageType>& srcImage,
                                                               Image<OutputStorageType>& dstImage)
{
  if (srcImage.GetChannelCount() != 1) {
    BIASERR("only single-channel images supported so far. sorry!")
    return -1;
  }

  // initialisation
  dstImage.Init(srcImage.GetWidth(), srcImage.GetHeight(),
                srcImage.GetChannelCount());
  dstImage.SetZero();

  const InputStorageType* srcImageData = srcImage.GetImageData();
  OutputStorageType* dstImageData = dstImage.GetImageData();

  float numWinPixels = (2.0f * halfWinSize_) + 1.0f; // normalisation factor
  numWinPixels = numWinPixels * numWinPixels;

  // if we multiply the normalisation factor with the required density, we get
  // an absolute threshold and can skip one division in the loop
  const float absRequiredDensity = numWinPixels * requiredDensity_;

  // compute density map
  for (unsigned int y = halfWinSize_;
       y < srcImage.GetHeight() - halfWinSize_;
       ++y) {
    for (unsigned int x = halfWinSize_;
         x < srcImage.GetWidth() - halfWinSize_;
         ++x) {
      const unsigned int offset = y * srcImage.GetWidth() + x;

      // iterate over window only if current pixel is defined
      if (srcImageData[offset] != undefinedPixel_) {
        float density = 0.0;

        for (unsigned int winY = y - halfWinSize_;
             winY <= y + halfWinSize_;
             ++winY) {
          for (unsigned int winX = x - halfWinSize_;
               winX <= x + halfWinSize_;
               ++winX) {
            // and count number of valid depth values
            const unsigned int winOffset = winY * srcImage.GetWidth() + winX;
            if (srcImageData[winOffset] != undefinedPixel_) {
              ++density;
            }
          }
        }

        // take the pixel if the density is high enough
        if (density >= absRequiredDensity) {
          dstImageData[offset] = (OutputStorageType)srcImageData[offset];
        }
      }
    }
  }

  return 0;
}

// instantiation
namespace BIAS{
#define FILTER_INSTANTIATION_CLASS DensityFilter
#include "Filterinst.hh"
}
