/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DENSITY_FILTER_HH__
#define __DENSITY_FILTER_HH__

// BIAS
#include <Filter/FilterBase.hh>

namespace BIAS
{
  /**
   * @brief Removes areas from an image that have a low density (fill rate).
   *
   * The primary focus for this class is filtering depth maps. For each pixel,
   * it is determined how many valid neighbours it has. If this number is below
   * a certain threshold, it is removed. Borders are completely ignored.
   *
   * @author rwulff
   * @date   10/2011
   */
  template <class InputStorageType, class OutputStorageType>
  class BIASFilter_EXPORT DensityFilter
      : public FilterBase<InputStorageType, OutputStorageType>
  {

  public:

    /**
     * @brief Constructor.
     *
     * @param halfWinSize
     *        determines size of window
     * @param requiredDensity
     *        areas with a density less or equal to this will be removed, must
     *        be in [0, 1]
     * @param undefinedPixel
     *        pixel value that means empty
     */
    inline
    DensityFilter(unsigned int halfWinSize = 10,
                  float requiredDensity = 0.5,
                  InputStorageType undefinedPixel = 0)
    {
      SetHalfWinSize(halfWinSize);
      SetRequiredDensity(requiredDensity);
      SetUndefinedPixelValue(undefinedPixel);
    }

    /**
     * @brief Performs the filter operation.
     *
     * @param src
     *        source image
     * @param dst
     *        destination image
     *
     * @return 0 on success or < 0 on error
     */
    virtual
    int Filter(const Image<InputStorageType>& srcImage,
               Image<OutputStorageType>& dstImage);

    /**
     * @brief Returns the size of the borders.
     *
     * This is essentially the half win size.
     *
     * @param[out] border_x
     *             border in x direction
     * @param[out] border_y
     *             border in y direction
     */
    virtual inline
    void GetBordersValid_(int& border_x, int& border_y) const
    {
      border_x = border_y = halfWinSize_;
    }

    /**
     * @brief Sets the half size of the filter window.
     *
     * The full size of the window will be 2 * halfWinSize + 1.
     *
     * @param[in] halfWinSize
     *            half size of the filter window
     */
    inline
    void SetHalfWinSize(unsigned int halfWinSize)
    {
      halfWinSize_ = halfWinSize;
    }

    /**
     * @brief Sets the required density for a pixel to be accepted.
     *
     * @param[in] requiredDensity
     *            areas with a density less or equal to this will be removed,
     *            must be in [0, 1]
     */
    inline
    void SetRequiredDensity(float requiredDensity)
    {
      BIASASSERT(requiredDensity >= 0.0 && requiredDensity <= 1.0)
      requiredDensity_ = requiredDensity;
    }

    /**
     * @brief Sets the value that represents empty pixels.
     *
     * @param[in] undefinedPixel
     *            the pixel value to interprete as empty
     */
    inline
    void SetUndefinedPixelValue(InputStorageType undefinedPixel)
    {
      undefinedPixel_ = undefinedPixel;
    }

  private:

    unsigned int halfWinSize_;
    float requiredDensity_;
    InputStorageType undefinedPixel_;

  };
}

#endif // __DENSITY_FILTER_HH__
