/* This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003-2009    (see file CONTACT for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel

   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.

   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA*/
#include "Median.hh"

#include "bias_config.h"

#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <MathAlgo/Median1D.hh>

#include <algorithm>

using namespace BIAS;
using namespace std;

template<class InputStorageType, class OutputStorageType>
Median<InputStorageType, OutputStorageType>::Median()
  : FilterNToN<InputStorageType,OutputStorageType>()
{
  _MedianSize = 2;
  _secondSize = 2;
  _MinValue = -1;
}

template<class InputStorageType, class OutputStorageType>
Median<InputStorageType, OutputStorageType>::Median(
  const Median<InputStorageType, OutputStorageType>& other) :
  FilterNToN<InputStorageType, OutputStorageType>(other),
  _MedianSize(other._MedianSize)
{
  _secondSize = _MedianSize;
}

template<class InputStorageType, class OutputStorageType>
Median<InputStorageType, OutputStorageType>::~Median()
{
}

int comp(const void *a, const void *b)
{
  if (*(float*) a < *(float*) b)
    return -1;
  else if (*(float*) a > *(float*) b)
    return 1;
  else
    return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
Filter(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  unsigned int halfsizeX = _MedianSize;
  unsigned int halfsizeY = _secondSize;

  if (src.GetChannelCount() != 1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }

  //cerr << "halfsizeX " << halfsizeX << endl;
  //cerr << "halfsizeY " << halfsizeY << endl;

  Image<OutputStorageType> tmp;
  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();
  bool identimg = false;
  int sizeX = 2 * halfsizeX + 1;
  int sizeY = 2 * halfsizeY + 1;

  if ((void*) src.GetImageData() != (void*) dst.GetImageData())
  {
    if (!src.SamePixelAndChannelCount(dst))
    {
      if (!dst.IsEmpty())
        dst.Release();
      dst.Init(src.GetWidth(), src.GetHeight(), 1);
      dstida = dst.GetImageDataArray();
    }
  }
  else
  {
    identimg = true;
    tmp.Init(src.GetWidth(), src.GetHeight(), 1);
    dstida = tmp.GetImageDataArray();
  }

  float *z, zvalue;
  int startx, starty, endx, endy;
  unsigned int found;

  z = new float[sizeX * sizeY];
  unsigned int ULx, ULy, LRx, LRy, width, height;
  width = src.GetWidth();
  height = src.GetHeight();
  src.GetROI()->GetCorners(ULx, ULy, LRx, LRy);
  for (unsigned int j = ULy; j < LRy; j++)
  {

    for (unsigned int i = ULx; i < LRx; i++)
    {
      found = 0;
      startx = int(i) - (sizeX - 1) / 2;
      endx = int(i) + (sizeX - 1) / 2;
      starty = int(j) - (sizeY - 1) / 2;
      endy = int(j) + (sizeY - 1) / 2;
      if (startx < 0)
        startx = 0;
      if (starty < 0)
        starty = 0;
      if (endx > int(width - 1))
        endx = width - 1;
      if (endy > int(height - 1))
        endy = height - 1;
      for (int y = starty; y <= endy; y++)
        for (int x = startx; x <= endx; x++)
        {
          zvalue = (float) srcida[y][x];
          if (zvalue > _MinValue)
          {
            z[found] = zvalue;
            found++;
          }
        }
      if (found > 0)
      {
        qsort(z, found, sizeof(float), comp);
        dstida[j][i] = (OutputStorageType) z[found / 2];
      }
      else
      {
        dstida[j][i] = (OutputStorageType) 0.0;
      }
    }
  }

  delete[] z;

  if (identimg)
    dst.StealImage(tmp);
  dst.GetROI()->SetCorners(ULx, ULy, LRx, LRy);
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterInt(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  BIASERR("FilterInt makes no sense here, using general Filter interface.");
  return Filter(src, dst);
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterFloat(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{
  BIASERR("FilterFloat makes no sense here, using general Filter interface.");
  return Filter(src, dst);
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterIgnoreZero5x5(const Image<InputStorageType>& src, Image<OutputStorageType>& dst) const
{
#ifdef BIAS_DEBUG
  if (src.GetChannelCount()!=1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }
#endif
  float z[25], zvalue;
  unsigned found, xm, x2m, xp, x2p, ym, y2m, yp, y2p, w, maxwidth, maxheight,
    h, x, y;

  if (!src.SamePixelAndChannelCount(dst))
  {
    if (!dst.IsEmpty())
      dst.Release();
    dst.Init(src.GetWidth(), src.GetHeight(), 1);
  }

  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();

  memset((void*) dstida[0], 0, sizeof(OutputStorageType)
         * dst.GetPixelCount());

  w = src.GetWidth();
  h = src.GetHeight();
  maxheight = h - 2;
  maxwidth = w - 2;
  for (y = 2; y < maxheight; y++)
  {
    for (x = 2; x < maxwidth; x++)
    {
      if ((zvalue = (float) srcida[y][x]) != 0.0)
      {
        z[0] = zvalue;
        found = 1;
        xm = x - 1;
        x2m = x - 2;
        xp = x + 1;
        x2p = x + 2;
        ym = y - 1;
        y2m = y - 2;
        yp = y + 1;
        y2p = y + 2;
        if ((zvalue = (float) srcida[y2m][x2m]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][xm]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][x]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][xp]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][x2p]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x2m]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xm]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xp]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x2p]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][x2m]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][xm]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }

        if ((zvalue = (float) srcida[y][xp]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][x2p]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x2m]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xm]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xp]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x2p]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x2m]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][xm]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][xp]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x2p]) > 0.0)
        {
          z[found] = zvalue;
          found++;
        }

        if (found > 4)
        {
          qsort(z, found, sizeof(float), comp);
          dstida[y][x] = (OutputStorageType) (0.33 * z[(found >> 1) - 1] + 0.34 * z[found >> 1] + 0.33
                                              * z[(found >> 1) + 1]);

        }
      }
    }
  }

  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterOnlyZeroIgnoreZero5x5(const Image<InputStorageType>& src,
                            Image<OutputStorageType>& dst,
                            const float& Threshold) const
{
#ifdef BIAS_DEBUG
  if (src.GetChannelCount()!=1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }
#endif
  float z[25], zvalue;
  unsigned found, xm, x2m, xp, x2p, ym, y2m, yp, y2p, w, maxwidth, maxheight,
    h, x, y;

  if (!src.SamePixelAndChannelCount(dst))
  {
    if (!dst.IsEmpty())
      dst.Release();
    dst.Init(src.GetWidth(), src.GetHeight(), 1);
  }

  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();

  memset((void*) dstida[0], 0, sizeof(OutputStorageType)
         * dst.GetPixelCount());

  w = src.GetWidth();
  h = src.GetHeight();
  maxheight = h - 2;
  maxwidth = w - 2;
  for (y = 2; y < maxheight; y++)
  {
    for (x = 2; x < maxwidth; x++)
    {
      if ((zvalue = (float) srcida[y][x]) <= Threshold)
      {
        found = 0;
        xm = x - 1;
        x2m = x - 2;
        xp = x + 1;
        x2p = x + 2;
        ym = y - 1;
        y2m = y - 2;
        yp = y + 1;
        y2p = y + 2;
        if ((zvalue = (float) srcida[y2m][x2m]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2m][x2p]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x2m]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x2p]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][x2m]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if ((zvalue = (float) srcida[y][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][x2p]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x2m]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x2p]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x2m]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y2p][x2p]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if (found > 4)
        {

          qsort(z, found, sizeof(float), comp);
          dstida[y][x] = (OutputStorageType) (0.33 * z[(found >> 1) - 1] + 0.34 * z[found >> 1] + 0.33
                                              * z[(found >> 1) + 1]);
        }
        else
        {
          dstida[y][x] = (OutputStorageType) Threshold;
        }
      }
      else
      {
        dstida[y][x] = (OutputStorageType) zvalue;
      }
    }
  }

  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterOnlyZeroIgnoreZero3x3(const Image<InputStorageType>& src,
                            Image<OutputStorageType>& dst, const float& Threshold) const
{
#ifdef BIAS_DEBUG
  if (src.GetChannelCount()!=1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }
#endif
  float z[9], zvalue;
  unsigned found, xm, xp, ym, yp, w, maxwidth, maxheight, h, x, y;

  if (!src.SamePixelAndChannelCount(dst))
  {
    if (!dst.IsEmpty())
      dst.Release();
    dst.Init(src.GetWidth(), src.GetHeight(), 1);
  }

  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();
  memset((void*) dstida[0], 0, sizeof(OutputStorageType)
         * dst.GetPixelCount());

  w = src.GetWidth();
  h = src.GetHeight();
  maxheight = h - 1;
  maxwidth = w - 1;
  for (y = 1; y < maxheight; y++)
  {
    for (x = 1; x < maxwidth; x++)
    {
      if ((zvalue = (float) srcida[y][x]) <= Threshold)
      {
        found = 0;
        xm = x - 1;
        xp = x + 1;
        ym = y - 1;
        yp = y + 1;
        if ((zvalue = (float) srcida[ym][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if ((zvalue = (float) srcida[y][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xm]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xp]) > Threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if (found >= 4)
        {
          qsort(z, found, sizeof(float), comp);
          // average median elements
          dstida[y][x] = (OutputStorageType) (0.33 * z[(found >> 1) - 1] + 0.34 * z[found >> 1] + 0.33
                                              * z[(found >> 1) + 1]);
        }
        else
        {
          //dstida[y][x] = (OutputStorageType)Threshold;
        }
      }
      else
      {
        dstida[y][x] = (OutputStorageType) zvalue;
      }
    }
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterOnlyBelowIgnoreBelow3x3(const Image<InputStorageType>& src,
                              Image<OutputStorageType>& dst,
                              const float &threshold, const unsigned &min_support) const
{
  if (src.GetChannelCount() != 1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }

  float z[9], zvalue;
  int found, xm, xp, ym, yp, x, y;

  if (!src.SamePixelAndChannelCount(dst))
  {
    dst.Init(src.GetWidth(), src.GetHeight(), 1);
  }

  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();

  const int w = src.GetWidth(), h = src.GetHeight();
  const int maxheight = h - 1, maxwidth = w - 1;
  // #ifdef _OPENMP
  // # pragma omp parallel for private(found, xm, xp, ym, yp, x, y, z, zvalue)
  // #endif
  for (y = 1; y < maxheight; y++)
  {
    for (x = 1; x < maxwidth; x++)
    {
      if ((zvalue = (float) srcida[y][x]) < threshold)
      {
        found = 0;
        xm = x - 1;
        xp = x + 1;
        ym = y - 1;
        yp = y + 1;
        if ((zvalue = (float) srcida[ym][xm]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][x]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[ym][xp]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[y][xm]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if ((zvalue = (float) srcida[y][xp]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xm]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][x]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }
        if ((zvalue = (float) srcida[yp][xp]) > threshold)
        {
          z[found] = zvalue;
          found++;
        }

        if (found >= (int) min_support)
        {
          qsort(z, found, sizeof(float), comp);
          // average median elements
          dstida[y][x] = (OutputStorageType) (z[(found >> 1)]);
        }
        else
        {
          dstida[y][x] = (OutputStorageType) zvalue;
        }
      }
      else
      {
        dstida[y][x] = (OutputStorageType) zvalue;
      }
    }
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterIgnore3x3(const Image<InputStorageType>& src,
                Image<OutputStorageType>& dst, const float& threshold) const
{
  if (src.GetChannelCount() != 1)
  {
    BIASERR("Median() not implemented for more than one channel");
    return -1;
  }

  float z[9], zvalue;
  int x, y, found, minx, miny, maxx, maxy, xx, yy;
  const int w = src.GetWidth(), h = src.GetHeight();

  if (!src.SamePixelAndChannelCount(dst))
  {
    if (!dst.IsEmpty())
      dst.Release();
    dst.Init(src.GetWidth(), src.GetHeight(), 1);
  }

  OutputStorageType **dstida = dst.GetImageDataArray();
  const InputStorageType **srcida = src.GetImageDataArray();

  // #ifdef _OPENMP
  // #pragma omp parallel for private(x, y, found, minx, miny, maxx, maxy, xx, yy, z, zvalue)
  // #endif

  for (y = 0; y < h; y++)
  {
    miny = (y - 1 < 0) ? (0) : (y - 1);
    maxy = (y + 1 < h) ? (y + 1) : (h - 1);
    for (x = 0; x < w; x++)
    {
      minx = (x - 1 < 0) ? (0) : (x - 1);
      maxx = (x + 1 < w) ? (x + 1) : (w - 1);

      found = 0;
      for (yy = miny; yy <= maxy; yy++)
      {
        for (xx = minx; xx <= maxx; xx++)
        {
          if ((zvalue = (float) srcida[yy][xx]) > threshold)
          {
            z[found] = zvalue;
            found++;
          }
        }
      }

      if (found == 0)
      {
        dstida[y][x] = (OutputStorageType) (srcida[y][x]);
      }
      else
      {
        qsort(z, found, sizeof(float), comp);
        dstida[y][x] = (OutputStorageType) (z[found >> 1]);
      }

    }
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterRemoveSaltAndPepper(const Image<InputStorageType>& src,
                          Image<OutputStorageType>& dest,
                          const float &Threshold) const
{
  unsigned int halfsizeX = _MedianSize;
  unsigned int halfsizeY = _secondSize;
  int channels = src.GetChannelCount();
  if (channels < 1)
  {
    BIASERR("No channels found");
    return -1;
  }
  if (channels > 2)
  {
    BIASERR("RemoveSaltAndPepper() not implemented for more than two channels");
    return -1;
  }
  if (channels > 1 && !src.IsPlanar())
  {
    BIASERR("Image is not planar, Median::FilterRemoveSaltAndPepper can only "
            <<"handle planar images with more than one channel\n");
    return -1;

  }

  //  cerr << "Image has " << channels << " channel(s)" << endl;
  //  cerr << "halfsizeX " << halfsizeX << "\t halfsizeY " << halfsizeY << endl;

  Image<OutputStorageType> tmp;
  OutputStorageType **dest_ida = dest.GetImageDataArray();
  const InputStorageType **src_ida = src.GetImageDataArray();
  bool identimg = false;
  int sizeX = 2 * halfsizeX + 1;
  int sizeY = 2 * halfsizeY + 1;
  int imageheight = src.GetHeight();
  //  cerr << "sizeX " << sizeX << "\t sizeY " << sizeY << endl;

  if ((void*) src.GetImageData() != (void*) dest.GetImageData())
  {
    if (!src.SamePixelAndChannelCount(dest))
    {
      if (!dest.IsEmpty())
        dest.Release();
      dest.Init(src.GetWidth(), imageheight, channels,
                src.GetStorageType(), false);
      dest_ida = dest.GetImageDataArray();
    }
    dest.SetColorModel(src.GetColorModel());
  }
  else
  {
    identimg = true;
    tmp.Init(src.GetWidth(), imageheight, channels, src.GetStorageType(),
             false);
    dest_ida = tmp.GetImageDataArray();
  }

  float zvalue_ch1, zvalue_ch2, poi;
  int startx, starty, endx, endy;
  unsigned int found, count, output;
  vector<float> foundvec1, foundvec2;
  output = 0;
  count = 0;
  //  cerr << "Width " << src.GetWidth() << "\t Height " << src.GetHeight() << endl;

  for (unsigned int j = 0; j < src.GetHeight(); j++)
  {
    for (unsigned int i = 0; i < src.GetWidth(); i++)
    {
      // set center-pixel as point of interest
      poi = (float) src_ida[j][i];
      if (poi > _MinValue)
      { // pixel is valid
        foundvec1.clear();
        foundvec2.clear();
        found = 0;

        startx = int(i) - (sizeX - 1) / 2;
        endx = int(i) + (sizeX - 1) / 2;
        starty = int(j) - (sizeY - 1) / 2;
        endy = int(j) + (sizeY - 1) / 2;
        if (startx < 0)
          startx = 0;
        if (starty < 0)
          starty = 0;
        if (endx > int(src.GetWidth() - 1))
          endx = src.GetWidth() - 1;
        if (endy > int(src.GetHeight() - 1))
          endy = src.GetHeight() - 1;

        for (int y = starty; y <= endy; y++)
        {
          for (int x = startx; x <= endx; x++)
          {
            zvalue_ch1 = (float) src_ida[y][x];
            if (channels == 2)
              zvalue_ch2 = (float) src_ida[y + imageheight][x];
            else
              zvalue_ch2 = 0.0;

            if (zvalue_ch1 > _MinValue)
            {
              foundvec1.push_back(zvalue_ch1);
              if (channels == 2)
                foundvec2.push_back(zvalue_ch2);
              found++;
              output++;
            }
          }
        }
        if (found > 0)
        {
          Median1D<float> med(foundvec1);
          float median_fp1 = med.GetMedian();
          float x84_fp1 = med.GetX84();
          float median_fp2 = 0.0f;
          if (channels == 2) {
            med.Compute(foundvec2);
            median_fp2 = med.GetMedian();
          }

          // ----  start debug output  ----

          // if (output%250 == 0){
          //   cerr << "\n\nCycle: " << output << " of " << src.GetHeight()*src.GetWidth() << endl;
          //   cerr << "poi(" << startx+2 << "," << starty+2 << "): " << poi << endl;
          //   cerr << "median: " << median_fp1 << ", " << median_fp2
          //        << "\t x84: " << x84_fp1 << ", " << x84_fp2  << endl;
          // }

          // ----  end debug output  ----

          if (poi < (median_fp1 - Threshold * x84_fp1) ||
              poi > (median_fp1 + Threshold * x84_fp1))
          {
            // set filtered value as output
            count++;
            dest_ida[starty + _secondSize][startx + _MedianSize]
              = (OutputStorageType) median_fp1;
            if (channels == 2)
              dest_ida[starty + imageheight + _secondSize][startx + _MedianSize]
                = (OutputStorageType) median_fp2;
          }
          else
          { // copy output
            dest_ida[starty + _secondSize][startx + _MedianSize]
              = (OutputStorageType) (src_ida[starty + _secondSize][startx + _MedianSize]);
            if (channels == 2)
              dest_ida[starty + imageheight + _secondSize][startx + _MedianSize]
                = (OutputStorageType) (src_ida[starty + imageheight + _secondSize][startx + _MedianSize]);
          }
        }
        else
        { // pixel is invalid, set output pixel also invalid
          dest_ida[j][i] = (OutputStorageType) src_ida[j][i];
          if (channels == 2)
            dest_ida[starty + imageheight + _secondSize][startx + _MedianSize]
              = (OutputStorageType) (src_ida[starty + imageheight + _secondSize][startx + _MedianSize]);
          //dest_ida[j+imageheight][i] =  src_ida[j+imageheight][i];  // Vision N version...
        }

      } // if (poi >=0)
      else
      { // region is invalid, set output pixel also invalid
        dest_ida[j][i] = (OutputStorageType) src_ida[j][i];
        if (channels == 2)
          dest_ida[j + imageheight][i] = (OutputStorageType) src_ida[j + imageheight][i];
      }
    } // for all x
  } // for all y

  if (identimg)
  {
    dest.StealImage(tmp);
    dest.SetColorModel(src.GetColorModel());
  }

  cerr << "FilterRemoveSaltAndPepper touched " << count << " of "
       << src.GetHeight() * src.GetWidth() << " pixels" << endl;

  return 0;
}

template<class InputStorageType, class OutputStorageType>
void Median<InputStorageType, OutputStorageType>::
GetBordersValid_(int &border_x, int &border_y) const
{
  if (_secondSize > _MedianSize)
  {
    border_x = border_y = _secondSize;
  }
  else
  {
    border_x = border_y = _MedianSize;
  }
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterColorImg(const Image<InputStorageType>& src, Image<OutputStorageType>& dst)
{

  Image<InputStorageType> srcCopy, temp;
  //  ImageConvert::ConvertST(src, srcCopy, dst.GetStorageType());

  int width = src.GetWidth();
  int height = src.GetHeight();

  /// use planar image in order to apply median filter seperately to each channel for now
  if (!((src.GetColorModel() == ImageBase::CM_RGB) && src.IsPlanar()))
  {
    temp = src;
    ImageConvert::Convert(temp, srcCopy, ImageBase::CM_RGB, true);
  }

  Image<InputStorageType> red(width, height, 1);
  Image<OutputStorageType> redG(width, height, 1);
  Image<InputStorageType> green(width, height, 1);
  Image<OutputStorageType> greenG(width, height, 1);
  Image<InputStorageType> blue(width, height, 1);
  Image<OutputStorageType> blueG(width, height, 1);

  red.GetChannel(srcCopy, 0);
  green.GetChannel(srcCopy, 1);
  blue.GetChannel(srcCopy, 2);

  Filter(red, redG);
  Filter(green, greenG);
  Filter(blue, blueG);

  dst = Image<OutputStorageType> (width, height, 3, true);
  for (int i = 0; i < width; i++)
  {
    for (int j = 0; j < height; j++)
    {
      dst.SetPixel(redG.PixelValue(i, j), i, j, 0);
      dst.SetPixel(greenG.PixelValue(i, j), i, j, 1);
      dst.SetPixel(blueG.PixelValue(i, j), i, j, 2);
    }
  }

  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterColorImgVec(const Image<InputStorageType>& src,
                  Image<OutputStorageType>& dst,
                  unsigned int px, unsigned int py,
                  const Image<unsigned char>* ignorePixels)
{
  unsigned int width = src.GetWidth();
  unsigned int height = src.GetHeight();
  unsigned int channels = src.GetChannelCount();

  if (dst.GetWidth() != width || dst.GetHeight() != height
      || dst.GetChannelCount() != channels)
  {
    dst.Init(width, height, channels);
    dst.Clear(0);
  }

  float minSum = FLT_MAX;

  int minX = px, minY = py;

  if (px >= 0 && px < width && py >= 0 && py < height)
  {

    int x1 = max(0, (int) px - _MedianSize);
    int y1 = max(0, (int) py - _secondSize);

    int x2 = min(width - 1, px + _MedianSize);
    int y2 = min(height - 1, py + _secondSize);

    for (int x = x1; x <= x2; x++)
    {
      for (int y = y1; y <= y2; y++)
      {
        unsigned char ignoreFlag = 0;
        if (ignorePixels != NULL)
        {
          ignoreFlag = ignorePixels->PixelValue(x, y);
        }

        if (ignoreFlag == 0)
        {
          float sum = 0;

          //---------------
          for (int tmp_x = x1; tmp_x <= x2; tmp_x++)
          {

            for (int tmp_y = y1; tmp_y <= y2; tmp_y++)
            {

              if (ignorePixels != NULL)
              {
                ignoreFlag = ignorePixels->PixelValue(tmp_x,
                                                      tmp_y);
              }

              if (ignoreFlag == 0)
              {

                float tmp_sum = 0;
                for (unsigned short ch = 0; ch < channels; ch++)
                {
                  float val1 = (float) src.PixelValue(x, y,
                                                      ch);
                  float val2 = (float) src.PixelValue(tmp_x,
                                                      tmp_y, ch);
                  float diff = pow(fabs(val1 - val2), 2);

                  tmp_sum += diff;
                }
                sum += sqrt(tmp_sum);
              }

            }
          }

          //--------------

          if (sum < minSum)
          {
            minSum = sum;
            minX = x;
            minY = y;
          }
        }
      }
    }

    for (unsigned short ch = 0; ch < channels; ch++)
    {
      OutputStorageType val = (OutputStorageType) src.PixelValue(minX,
                                                                 minY, ch);
      dst.SetPixel(val, px, py, ch);
    }
  }
  else
  {
    BIASERR("pixel coordinates "<< px << " , "<<py<< " exceed image range");
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterColorImgVec(const Image<InputStorageType>& src,
                  Image<OutputStorageType>& dst,
                  const Image<unsigned char>* ignorePixels)
{

  unsigned int width = src.GetWidth();
  unsigned int height = src.GetHeight();
  unsigned int channels = src.GetChannelCount();

  if (dst.GetWidth() != width || dst.GetHeight() != height
      || dst.GetChannelCount() != channels)
  {
    dst.Init(width, height, channels);
  }

  int x, y;
#pragma omp parallel for private(x,y)
  for (y = 0; y < (int) height; y++)
  {
    for (x = 0; x < (int) width; x++)
    {
      unsigned char ignoreValue = 0;
      if (ignorePixels != NULL)
      {
        ignoreValue = ignorePixels->PixelValue(x, y);
      }

      if (ignoreValue == 0)
      {
        FilterColorImgVec(src, dst, x, y, ignorePixels);
      }
    }
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
FilterParallel(Image<InputStorageType>& source,
               Image<OutputStorageType>& dest, unsigned int numberOfThreads)
{
  int width = source.GetWidth();
  int height = source.GetHeight();
  int size = width*height;

  if (dest.GetWidth() != (unsigned int) width || dest.GetHeight() != (unsigned int)height ||
      dest.GetChannelCount() != 1) {
    dest.Release();
    dest.Init(width,height,1);
  }

  InputStorageType* source_data = (InputStorageType*) source.GetImageData();
  OutputStorageType* dest_data = (OutputStorageType*) dest.GetImageData();

#ifdef _OPENMP
  vector<InputStorageType*> sort_vector(numberOfThreads);
  for(unsigned int i = 0; i < sort_vector.size(); i++) {
    sort_vector[i] = new InputStorageType[(2*_MedianSize+1)*(2*_secondSize+1)];
  }
#pragma omp parallel for num_threads(numberOfThreads)
#else
#define omp_get_thread_num() 0
  vector<InputStorageType*> sort_vector(numberOfThreads);
  sort_vector[0] = new InputStorageType[(2*_MedianSize+1)*(2*_secondSize+1)];
#endif
  for (int i = 0; i < size; i++) {

    int y = i / width;
    int x = i - y*width;

    int start_x = max((int)0,x - _MedianSize);
    int start_y = max((int)0,y - _secondSize);
    int end_x = min(width -1,x + _MedianSize);
    int end_y = min(height - 1,y + _secondSize);


    int array_width = end_x - start_x + 1;
    int array_height = end_y - start_y + 1;
    int array_size = array_width* array_height;

    unsigned int th_num = omp_get_thread_num();

    InputStorageType* sort_array = sort_vector[th_num];

    for (int tmp_x = start_x, array_x = 0; array_x < array_width; tmp_x++, array_x++) {
      for (int tmp_y = start_y, array_y =0; array_y < array_height; tmp_y++, array_y++) {
        int index = tmp_y * width + tmp_x;
        int array_index = array_y * array_width + array_x;
        sort_array[array_index] = source_data[index];
      }
    }

    sort(sort_array,sort_array+array_size);

    int med = min(array_size / 2+1, array_size - 1);

    dest_data[i] = (OutputStorageType) sort_array[med];

//		delete [] sort_array;
  }

  for(unsigned int i = 0; i < sort_vector.size(); i++) {
    delete sort_vector[i];
  }
  return 0;
}


template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
Filter3x3x3Grey(std::vector<Image<InputStorageType> >& srcs, Image<OutputStorageType>& dst)
{
  //
  if (srcs.size() != 3)
  {
    BIASERR("Number of input images not 3!");
    return -1;
  }

  unsigned int height = srcs[0].GetHeight();
  unsigned int width = srcs[0].GetWidth();

  for (int i = 0; i < 3; i++)
  {
    if (srcs[i].GetColorModel() != ImageBase::CM_Grey)
    {
      BIASERR("Images not grey!");
      return -1;
    }
    if (srcs[i].GetHeight() != height || srcs[i].GetWidth() != width)
    {
      BIASERR("Image size is not consistent!");
      return -1;
    }
  }

  dst = Image<OutputStorageType> (width, height, 1, true);
  //OutputStorageType** dstData = dst.GetImageDataArray();

  InputStorageType** imData[3];
  imData[0] = srcs[0].GetImageDataArray();
  imData[1] = srcs[1].GetImageDataArray();
  imData[2] = srcs[2].GetImageDataArray();
  //InputStorageType medValues[27];
  std::vector<InputStorageType> m;

  for (unsigned int w = 1; w < width - 1; w++)
  {
    for (unsigned int h = 1; h < height - 1; h++)
    {
      for (int i = 0; i < 3; i++)
      {
        m.push_back(imData[i][h - 1][w - 1]);
        m.push_back(imData[i][h - 1][w - 0]);
        m.push_back(imData[i][h - 1][w + 1]);
        m.push_back(imData[i][h - 0][w - 1]);
        m.push_back(imData[i][h - 0][w - 0]);
        m.push_back(imData[i][h - 0][w + 1]);
        m.push_back(imData[i][h + 1][w - 1]);
        m.push_back(imData[i][h + 1][w - 0]);
        m.push_back(imData[i][h + 1][w + 1]);
      }

      sort(m.begin(), m.end());
      //dstData[h][w] = (OutputStorageType)m[13];
      dst.SetPixel((OutputStorageType) m[13], w, h, 0);
      m.clear();
    }
  }
  return 0;
}

template<class InputStorageType, class OutputStorageType>
int Median<InputStorageType, OutputStorageType>::
Filter3x3x3Color(std::vector<Image<InputStorageType> >& srcs, Image<OutputStorageType>& dst)
{
  if (srcs.size() != 3)
  {
    BIASERR("Number of input images not 3!");
    return -1;
  }

  unsigned int height = srcs[0].GetHeight();
  unsigned int width = srcs[0].GetWidth();

  std::vector<Image<InputStorageType> > ImsRed, ImsGreen, ImsBlue;
  Image<InputStorageType> temp;

  for (int i = 0; i < 3; i++)
  {
    if (srcs[i].GetHeight() != height || srcs[i].GetWidth() != width)
    {
      BIASERR("Image size is not consistent!");
      return -1;
    }
    // use planar image in order to apply median filter seperately to each channel for now
    if (!((srcs[i].GetColorModel() == ImageBase::CM_RGB)
          && srcs[i].IsPlanar()))
    {
      ImageConvert::Convert(srcs[i], temp, ImageBase::CM_RGB, true);
    }
    else
    {
      temp = srcs[i];
    }
    Image<InputStorageType> red(width, height, 1);
    Image<InputStorageType> green(width, height, 1);
    Image<InputStorageType> blue(width, height, 1);

    red.GetChannel(temp, 0);
    green.GetChannel(temp, 1);
    blue.GetChannel(temp, 2);

    ImsRed.push_back(red);
    ImsGreen.push_back(green);
    ImsBlue.push_back(blue);
  }

  Image<OutputStorageType> redG(width, height, 1);
  Image<OutputStorageType> greenG(width, height, 1);
  Image<OutputStorageType> blueG(width, height, 1);

  Filter3x3x3Grey(ImsRed, redG);
  Filter3x3x3Grey(ImsGreen, greenG);
  Filter3x3x3Grey(ImsBlue, blueG);

  dst = Image<OutputStorageType> (width, height, 3, true);
  for (unsigned int i = 0; i < width; i++)
  {
    for (unsigned int j = 0; j < height; j++)
    {
      dst.SetPixel(redG.PixelValue(i, j), i, j, 0);
      dst.SetPixel(greenG.PixelValue(i, j), i, j, 1);
      dst.SetPixel(blueG.PixelValue(i, j), i, j, 2);
    }
  }

  return 0;

}

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

namespace BIAS{
#define FILTER_INSTANTIATION_CLASS Median
#include "Filterinst.hh"
}
