#ifndef __CONTROL_INTERFACE_HH__
#define __CONTROL_INTERFACE_HH__

#include <bias_config.h>

#ifndef BIAS_HAVE_OPENGL
#  error You need BIAS with USE_OpenGL Please recompile BIAS.
#endif // BUILD_GLviewer

#include <Base/Common/W32Compat.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Debug/Error.hh>

typedef enum {
    KEY_F1  = 0,
    KEY_F2,
    KEY_F3,
    KEY_F4,
    KEY_F5,
    KEY_F6,
    KEY_F7,
    KEY_F8,
    KEY_F9,
    KEY_F10,
    KEY_F11,
    KEY_F12,
    KEY_HOME,
    KEY_PGUP,
    KEY_PGDOWN,
    KEY_UP,
    KEY_DOWN,
    KEY_LEFT,
    KEY_RIGHT

} SpecialKey;

typedef enum {
  NONE = 0,
  ALT,
  CTRL,
  SHIFT
} modifiers;

namespace BIAS {

  class BIASGLviewer_EXPORT GLProjectionParametersInterface;

  /** \class ControlInterface
      \brief Defines the common interface used for communication between
      rendering components.
      All presented method return a boolean.
      Returning false means, that the called method has not
      reacted to the call.
      \author bartczak
      \ingroup g_glviewer
  */
  class BIASGLviewer_EXPORT ControlInterface : public BIAS::Debug
  {
  public:
  	/** \brief constructor */
    ControlInterface()
    {
    	listenerName_= "unnamed control listener";
    	controlledObject_=NULL;
    };
    /** \brief destructor */
    virtual ~ControlInterface(){}

    /** \brief Get the name of the controller */
    virtual std::string GetName()
    { return listenerName_;};

    /** \brief Set the name of the controller */
    virtual void SetName(std::string name)
    { listenerName_=name;};

    /** \brief react to mouse wheel usage
     *   overwrite in derived class of desired*/
    virtual bool MouseWheelUsed(double stepsize)
    {return false;};

    /** \brief react to left mouse double click
    *   overwrite in derived class of desired*/
    virtual bool LeftMouseDoubleClicked(int /*x*/, int /*y*/, int m)
    { return false; };

    /** \brief react to left mouse single click
    *   overwrite in derived class of desired*/
    virtual bool LeftMouseSingleClicked(int /*x*/, int /*y*/)
    { return false; };

    /** \brief react to mouse movement while left button held down
    *   overwrite in derived class of desired*/
    virtual bool LeftMouseMoved(int /*x1*/, int /*x2*/,int /*y1*/, int /*y2*/)
    { return false; };

    /** \brief react to left mouse button down event
    *   overwrite in derived class of desired*/
    virtual bool LeftMouseDown(int , int)
    { return false; };

    /** \brief react to left mouse button up event
    *   overwrite in derived class of desired*/
    virtual bool LeftMouseUp(int , int)
    { return false; };


    /** \brief react to right mouse single click event
    *   overwrite in derived class of desired*/
    virtual bool RightMouseSingleClicked(int /*x*/, int /*y*/)
    { return false; };

    /** \brief react to right mouse double click
    *   overwrite in derived class of desired*/
    virtual bool RightMouseDoubleClicked(int /*x*/, int /*y*/)
    { return false; };

    /** \brief react to mouse movement while right button held down
    *   overwrite in derived class of desired*/
    virtual bool RightMouseMoved(int /*x1*/, int /*x2*/, int /*y1*/, int /*y2*/)
    { return false; };

    /** \brief react to right mouse button down event
    *   overwrite in derived class of desired*/
    virtual bool RightMouseDown(int , int)
    { return false; };

    /** \brief react to right mouse button up event
    *   overwrite in derived class of desired*/
    virtual bool RightMouseUp(int , int)
    { return false; };

    /** \brief react to mouse movement while middle button held down
    *   overwrite in derived class of desired*/
    virtual bool MiddleMouseMoved(int /*x1*/, int /*x2*/,int /*y1*/, int /*y2*/)
    { return false; };

    /** \brief react to mouse movement while right and middle button held down
    *   overwrite in derived class of desired*/
    virtual bool LeftAndRightMouseMoved(int /*x1*/, int /*x2*/,int /*y1*/, int /*y2*/)
    { return false; };

    /** \brief react to middle mouse button down event
    *   overwrite in derived class of desired*/
    virtual bool MiddleMouseDown(int , int)
    { return false; };

    /** \brief react to middle mouse button up event
    *   overwrite in derived class of desired*/
    virtual bool MiddleMouseUp(int , int)
    { return false; };

    /** \brief react to mouse movement
    *   overwrite in derived class of desired*/
    virtual bool MouseMoved(int /*x*/, int /*y*/)
    {return false; };

    /** \brief react to mouse leaving the window
    *   overwrite in derived class of desired*/
    virtual bool MouseLeftWindow()
    { return false; };

    /** \brief react to press of special keys such as F1-F12, ...
    *   overwrite in derived class of desired*/
    virtual bool SpecialKeyPressed(int /*key*/)
    { return false; };

    /** \brief react to press of special keys such as F1-F12, ...
    *   overwrite in derived class of desired*/
    virtual bool SpecialKeyUp(int /*key*/)
    { return false; };

    /** \brief react to press of ascii-coded charakter plus pointer position */
    virtual bool StandardKeyPressed(unsigned char /*key*/, int, int)
    { return false; };

    /** \brief react to press of ascii-coded charakter plus pointer position */
    virtual bool StandardKeyUp(unsigned char /*key*/, int, int)
    { return false; };

    /** \brief this function is called by RenderContextBase when timer is expired
     * Overwrite and implement to do some timer dependend processing  */
    virtual bool TimerExpired()
    { return false; };

    /** \brief this function is called by RenderContextBase if the window is reshaped.
    * Overwrite and implement to do some processing upon reshape */
    virtual bool WindowReshape(int /*width*/, int /*height*/)
    { return false; };

    /** \brief this function is called by RenderContextBase when drawing is done.
      * Overwrite and implement to do some processing after drawing */
    virtual bool FinishedDraw()
    {return false;};

    /**\brief Set the camera projection parameters. Can be of type GLProjection or
     * any from  GLProjectionParametersBase derived class*/
    virtual void SetGLProjectionParametersInterface( BIAS::GLProjectionParametersInterface* interface_ ){
      if(interface_ != NULL)
        controlledObject_ =  interface_;
      else
        BIASWARN("Given interface is NULL");
    };
    /**\brief Get the camera projection parameters. Can be of type GLProjection or
     * any from  GLProjectionParametersBase derived class
     * \return Pointer to  GLProjectionParametersInterface */
    virtual BIAS::GLProjectionParametersInterface* GetGLProjectionParametersInterface()
    {	return controlledObject_; };

  protected:
    /** name of listener/controller */
    std::string listenerName_;
    /** the camera which is controlled by this controller*/
    BIAS::GLProjectionParametersInterface* controlledObject_;
  };

}//end of namespace

#endif //__CONTROL_INTERFACE_HH__
