/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


/**
   @example ExampleContextWx.cpp
   @relates ContextWX, GLProjectionParametersPerspective, SceneThreeDOutWrapper
   @brief Example for usage ContextWX, GLProjections and ThreeDOutscenes
   @ingroup g_examples
   @ingroup g_glviewer
   @author MIP, sedlazeck 04/2012
*/

#include <GLviewer/ContextWX.hh>
#include <Base/Common/BIASpragmaStart.hh>
#include <wx/wx.h>
#include <Base/Common/BIASpragmaEnd.hh>

#include <GLviewer/Scenes/SceneThreeDOutWrapper.hh>
#include <GLviewer/ThreeDOutOpenSceneGraph.hh>

#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>

#include <GLviewer/GLProjectionParametersPerspective.hh>
#include <GLviewer/Controller/TrackballControl.hh>



using namespace BIAS;
using namespace std;

/////////////////////////////////////
//////////   main window   //////////
/////////////////////////////////////

class CSimpleWxFrame : public wxFrame
{

public:

  CSimpleWxFrame();
  
protected:

  BIAS::TrackballControl* control_;
  ContextWX* context_;
  SceneThreeDOutWrapper sceneWrapper_;
  GLProjectionParametersPerspective*  proj_;
  BIAS::ThreeDOutOpenSceneGraph* tdo_;

};

CSimpleWxFrame::CSimpleWxFrame()
  : wxFrame((wxFrame *)NULL, -1, wxT("Wx"), wxPoint(0, 0), wxSize(800, 600))
{
  // init threeDOut and link to sceneWrapper
  tdo_ = new ThreeDOutOpenSceneGraph();
  
  // create a set of 3D points
  Vector3<double> p3d_1(-1.0, -1.0, 1.0);
  Vector3<double> p3d_2(1.0, -1.0, 1.0);
  Vector3<double> p3d_3(-1.0, 1.0, 1.0);
  Vector3<double> p3d_4(1.0, 1.0, 1.0);
   
  // set point drawing size and add points to threeDout
  tdo_->SetParamsPointSize(100);
  tdo_->AddPoint(p3d_1, RGBAuc(255, 0, 0, 128));
  tdo_->AddPoint(p3d_2, RGBAuc(0, 255, 0, 128));
  tdo_->AddPoint(p3d_3, RGBAuc(0, 0, 255, 128));
  tdo_->AddPoint(p3d_4, RGBAuc(255, 255, 0, 128));
  
  // add some lines between the points
  tdo_->AddLine(p3d_1, p3d_2, RGBAuc(255, 0, 0, 128));
  tdo_->AddLine(p3d_2, p3d_3, RGBAuc(255, 0, 0, 128));
  tdo_->AddLine(p3d_3, p3d_4, RGBAuc(255, 0, 0, 128));
  tdo_->AddLine(p3d_4, p3d_1, RGBAuc(255, 0, 0, 128));
   
  // SceneThreeDOutWrapper is derived from SceneBase and can 
  // therefore provide the link between scenes in a ThreeDOut(OpenSceneGraph)
  // object and the ContextWX
  sceneWrapper_.LinkWithThreeDOut(tdo_); 
  
  // init control
  control_ = new TrackballControl();
  control_->SetPointOfInterest(Vector3<double>(0,0,0));
  
  // init GL projection
  // note: the projection's extrinsics need to be set in order to be able
  // to see the scene upon starting the program. Otherwise, it can be
  // difficult to find the scene in 3D space.
  proj_ = new GLProjectionParametersPerspective();
  proj_->SetSimplePerspectiveCam(60.0, // field of view in degree
                                 400, // image width
                                 300); // image height
  proj_->SetAutoReshapeBehaviour(GLProjectionParametersBase::AutoRescaleParams);
  proj_->LookAt(Vector3<double>(0,0,-50), // eye - camera center
                Vector3<double>(0,0,1), // center point to look at
                Vector3<double>(0,1,0)); // up direction
  proj_->SubtractFromBackgroundImage(false);
  proj_->SetZClippingPlanes(0.1,50000);
  
  // initialize context
  context_ = new ContextWX(this, -1, wxDefaultPosition, wxSize(800,600));
  context_->SetGLProjectionParametersInterface(proj_);
  Vector4<float> clearColor(1,1,1,0);
  context_->SetClearColor(clearColor);
  context_->SetControl(control_);
  context_->SetAutoClipping(false);

  // add context to sizer within this wxFrame
  wxBoxSizer* mainSizer = new wxBoxSizer(wxVERTICAL);
  mainSizer->Add(context_, 1, wxEXPAND|wxALL|wxALIGN_CENTRE, 0);
  mainSizer->Fit(this);

  // show the whole thing
  Layout();
  Show(true);

  // append scene with threeDout stuff in context
  context_->AppendScene(&sceneWrapper_);

  // finish context initialization
  context_->InitializationComplete();
}

////////////////////////////////////////
//////////    main routine    //////////
////////////////////////////////////////

class CSimpleWxApp : public wxApp
{
public:
  bool OnInit();
};

bool CSimpleWxApp::OnInit()
{
  wxInitAllImageHandlers();
  CSimpleWxFrame* mainFrame = new CSimpleWxFrame();
  mainFrame->Show();
  return true;
}

IMPLEMENT_APP(CSimpleWxApp)
