/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __SCENETEXTUREDPLANE_HH__
#define __SCENETEXTUREDPLANE_HH__

#include <bias_config.h>

#include <GLviewer/Scenes/SceneBase.hh>
#include <Base/Image/Image.hh>
#include <Base/Math/Vector2.hh>
#include <Base/Math/Vector3.hh>

namespace BIAS
{
  /**
     @class SceneTexturedPlane
     @brief Display a textured plane. Lighting is computed appropriately.
     @author esquivel
     @ingroup g_glviewer
  */
  class BIASGLviewer_EXPORT SceneTexturedPlane : public SceneBase
  {
  public:

    /** @brief Create new plane with given position, size and orientation. */
    SceneTexturedPlane(const BIAS::Vector3<double> &topLeft,
                       const BIAS::Vector3<double> &width,
                       const BIAS::Vector3<double> &height,
                       const int widthSubsections = 1,
                       const int heightSubsections = 1,
                       const int minFilter = GL_NEAREST,
                       const int magFilter = GL_NEAREST,
                       const int wrapping = GL_REPEAT);

    virtual ~SceneTexturedPlane();

    /** @brief Reset position and size of plane. */
    void Resize(const BIAS::Vector3<double> &topLeft,
                const BIAS::Vector3<double> &width,
                const BIAS::Vector3<double> &height,
                const int widthSubsections = 1,
                const int heightSubsections = 1);

    /** @brief Set texture image and reset texture coordinates for plane.
        @attention Previously set texture will be deleted!
        @return Returns ID of created texture, or <0 in case of errors.
    */
    int SetImage(const BIAS::Image<unsigned char> &image);

    /** @brief Load texture image and reset texture coordinates for plane.
        @attention Previously set texture will be deleted!
        @return Returns ID of created texture, or <0 in case of errors.
    */
    int SetImage(const std::string &imageFilename);

    /** @brief Set texture image from existing texture ID and reset texture
               coordinates for plane.
        @attention Previously set texture will be deleted!
        @return Returns texture ID, or <0 in case of errors.
    */
    int SetImage(const int texID, const int width, const int height);

    /** @brief Set texture coordinates for plane texture in pixels. */
    void SetTextureCoordinates(const BIAS::Vector2<int> topLeft,
                               const BIAS::Vector2<int> bottomRight);

    /** @brief Set texture coordinates for plane texture in texels. */
    void SetTextureCoordinates(const BIAS::Vector2<float> topLeft,
                               const BIAS::Vector2<float> bottomRight);

    /** @brief Draw plane normal for debugging purposes? */
    void DrawNormal(const bool draw = true, const float length = 1.0f)
    { drawNormal_ = draw; if (draw) drawNormalLength_ = length; }

    /** @brief Implementation of the SceneBase GetBoundingBox method. */
    virtual void GetBoundingBox(BIAS::Vector3<double> &minVal,
                                BIAS::Vector3<double> &maxVal);

  protected:

    /** @brief Implementation of the SceneBase Draw method. */
    virtual void Draw();

    /** @brief Stores plane corners (top left/right, bottom right/left). */
    BIAS::Vector3<float> corners_[4];

    /** @brief Stores normal direction of plane. */
    BIAS::Vector3<float> normal_;

    /** @brief Stores top left and bottom right texture coordinates. */
    BIAS::Vector2<float> texCoords_[2];

    /** @brief Stores width and height of texture. */
    BIAS::Vector2<int> texSize_;

    /** @brief Initialize and bind texture.
        @return Returns ID of created texture, or <0 in case of errors.
    */
    int InitializeTexture_(BIAS::Image<unsigned char> &texture);

    /** @brief Specifies if texture of plane has been initialized. */
    bool isInitialized_;

    /** @brief Draw plane normal for debugging purposes? */
    bool drawNormal_;
    float drawNormalLength_;

    /** @brief Number of subsections in horizontal/vertical direction */
    int subsectionsX_, subsectionsY_;

    /** @brief Texture ID use to identify the texture binding */
    GLuint texID_;

    // Specifies OpenGL specific parameters for texture rendering. */
    GLint minFilter_, magFilter_;
    GLint wrapping_;

  };
}

#endif // __SCENETEXTUREDPLANE_HH__
