/** 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Geometry/ProjectionParametersOrthographic.hh"

using namespace std;
using namespace BIAS;
typedef ProjectionParametersOrthographic PPO;

bool PPO::Distort(HomgPoint2D& point2d) const{
  BIASERR("no distortion for orthographic projection implemented");
  BIASABORT;
  return false;
}

bool PPO::Undistort(HomgPoint2D& point2d) const{
  BIASERR("no undistortion for orthographic projection implemented");
  BIASABORT;
  return false;
}

HomgPoint3D PPO::UnProjectToImagePlane(const HomgPoint2D& pos, const double& depth, bool IgnoreDistortion) const{
  //  Distortion is not implemented for orthographic projections!
  BIASASSERT(IgnoreDistortion);
  Vector3<double> p(0, 0, 0);
  p[0] = (pos[0]-principalX_)*dX_ + rayDir_[0]*depth;
  p[1] = (pos[1]-principalY_)*dY_ + rayDir_[1]*depth;
  p[2] = rayDir_[2]*depth;
  HomgPoint3D local(p);
  return Pose_.LocalToGlobal(local);
}

HomgPoint2D PPO::ProjectLocal(const Vector3<double> &point, bool IgnoreDistortion) const{
  //  Distortion is not implemented for orthographic projections!
  BIASASSERT(IgnoreDistortion);
  HomgPoint2D res;
  res[2] = 1.0;
  res[1] = point[2]/rayDir_[2];
  res[0] = (point[0] - rayDir_[0]*res[1])/dX_+principalX_;
  res[1] = (point[1] - rayDir_[1]*res[1])/dY_+principalY_;
  return res;
}

const bool PPO::DoIntrinsicsDiffer(const PPB* p) const{
  if(PPB::DoIntrinsicsDiffer(p)){
    return true;
  }
  const PPO* ppo = dynamic_cast<const PPO*>(p);
  if(ppo == NULL){
    return true;
  }
  if(dX_ != ppo->dX_ || dY_ != ppo->dY_){
    return true;
  }
  if(rayDir_ != ppo->rayDir_){
    return true;
  }
  return false;
}

bool PPO::DoesPointProjectIntoImageLocal(const Vector3<double> &localP,
                               HomgPoint2D &p, bool IgnoreDistortion) const{
  //  Distortion is not implemented for orthographic projections!
  BIASASSERT(IgnoreDistortion);
  p = ProjectLocal(localP, IgnoreDistortion);
  p.Homogenize(); //neccessary ?
  return (p[0] >= 0.0 && p[1] >= 0.0 && 
          p[0] <= (double)(width_-1) && p[1] <= (double)(height_-1));
}


#ifdef BIAS_HAVE_XML2
int PPO::XMLGetClassName(string &TopLevelTag, double &Version) const{
  TopLevelTag = "ProjectionParametersOrthographic";
  Version = 0.1;
  return 0;
}

int PPO::XMLOut(const xmlNodePtr Node, XMLIO &XMLObject) const{
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  PPB::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  PPB::XMLOut(theNode, XMLObject);
  XMLObject.addAttribute(Node, "dX", dX_);
  XMLObject.addAttribute(Node, "dY", dY_);
  XMLObject.addAttribute(Node, "rayDirX", rayDir_[0]);
  XMLObject.addAttribute(Node, "rayDirY", rayDir_[1]);
  XMLObject.addAttribute(Node, "rayDirZ", rayDir_[2]);
  return 0; 
}

int PPO::XMLIn(const xmlNodePtr Node, XMLIO &XMLObject){
  string TopLevelName;
  double Version;
  PPB::XMLGetClassName(TopLevelName, Version);
  xmlNodePtr childNode = XMLObject.getChild(Node, TopLevelName); 
  if(!childNode){
    BIASERR("Error in XML, tag " << TopLevelName << " not found!");
    return -1;
  }
  if(PPB::XMLIn(childNode, XMLObject) != 0){
    return -1;
  }
  dX_ = XMLObject.getAttributeValueDouble(Node, "dX");
  dY_ = XMLObject.getAttributeValueDouble(Node, "dY");
  rayDir_[0] = XMLObject.getAttributeValueDouble(Node, "rayDirX");
  rayDir_[1] = XMLObject.getAttributeValueDouble(Node, "rayDirY");
  rayDir_[2] = XMLObject.getAttributeValueDouble(Node, "rayDirZ");
  return 0;
}
#endif // BIAS_HAVE_XML2
