#include "CameraViewFrustum.hh"
#include <Base/Common/W32Compat.hh>
#include <Base/Debug/Error.hh>


#include <iostream>
#include <math.h>

#ifdef BIAS_HAVE_OPENGL
#  include <Gui/biasgl.h>
#endif


using namespace std;
using namespace BIAS;


BIAS::CameraViewFrustum::CameraViewFrustum()
{
  InitMembers();
}


double BIAS::CameraViewFrustum::GetFovY() const {
  // unsymmetric frustum calculated separately
  double fovY = fabs(atan2(top, zNear)); // y,x ; y/x
  fovY += fabs(atan2(bottom, zNear));
  // convert from rad to degree
  fovY *= 180. / M_PI;
  return fovY;
}

double BIAS::CameraViewFrustum::GetFovX() const {
  double fovX = fabs(atan2(left,zNear)); // y,x
  fovX += fabs(atan2(right,zNear));
  fovX *= 180. / M_PI;
  return fovX;
}

double BIAS::CameraViewFrustum::GetAspect() const {
  const double w = right - left;
  const double h = top - bottom;
  BIASASSERT(h!=0);
  return w/h;
}

void BIAS::CameraViewFrustum::SetPerspective(const double & fovYdeg, 
                                             const double & fovAspect,
                                             const double & zNearX, 
                                             const double & zFarX)
{
  // set symmetric frustum like glulookat does
  this->zNear= zNearX;
  this->zFar = zFarX;
  top    = this->zNear * tan(fovYdeg/2. *M_PI/180.);
  bottom = -top;
  right  = fovAspect * top;
  left   = - right;
}

std::ostream & BIAS::CameraViewFrustum::Print(std::ostream & os) const
{
  os<<"CameraViewFrustum: "<<endl
    <<"left   "<<left<<endl
    <<"right  "<<right<<endl
    <<"bottom "<<bottom<<endl
    <<"top    "<<top<<endl
    <<"near   "<<zNear<<endl
    <<"far    "<<zFar<<endl
    ;
  return os;
}

void BIAS::CameraViewFrustum::AdaptWidth(const double & vpAspectRatioDesired  )
{
  BIASASSERT(vpAspectRatioDesired >0);
  BIASASSERT(this->GetAspect()>0);

  const double scale  = vpAspectRatioDesired  / this->GetAspect();
  // increase left and right arounf center 
  // to keep center in the middle
  const double center = left +0.5*(right-left);
  
  // shift to center
  this->left  -= center;
  this->right -= center;

  // scale
  this->left  *= scale;
  this->right *= scale;

  // shift back
  this->left  += center;
  this->right += center;
}


#ifdef BIAS_HAVE_OPENGL
void BIAS::CameraViewFrustum::DisplayGL() const
{
  BIASASSERT(left<right);
  BIASASSERT(bottom<top);
  BIASASSERT(zNear<zFar);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //gluPerspective( (GLdouble)GetFovY(), (GLdouble)GetAspect(), (GLdouble)zNear, (GLdouble)zFar);
  glFrustum(
    (GLdouble) left,
    (GLdouble) right,
    (GLdouble) bottom,
    (GLdouble) top,
    (GLdouble) zNear,
    (GLdouble) zFar
    );
  glMatrixMode(GL_MODELVIEW);
}
#endif // BIAS_HAVE_OPENGL

