/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Gui/FilterDialogRescale.hh>

using namespace BIAS;

template <class InputST, class OutputST>
FilterDialogRescale<InputST, OutputST>
::FilterDialogRescale(const Image<InputST>& origImage,
                      wxWindow* parent,
                      wxWindowID id,
                      const wxString& title,
                      const wxString& name)
: FilterDialogBase<InputST, OutputST>(origImage, parent, id, title, name),
  RescaleFilter(Rescale<InputST, OutputST>()),  
  RescaleParamSizer(new wxBoxSizer(wxVERTICAL)),
  LowPassFilter(NULL),
  MeanString(wxT("Mean")),
  GaussString(wxT("Gauss")),
  BinomString(wxT("Binomial")),
  GaussTDString(wxT("GaussThreshold")),
  Selection(BIAS::LPT_Mean)
{
	this->ParamPanel = new PanelForDoubleValueWx(this,                      // parent
																							 BIAS::ID_PANEL_DOUBLE_VAL, // id
																							 wxT("FACTOR:"),            // paramname
																							 0.1,                       // minlimit
																							 5.0,                       // maxlimit
																							 2.5);                      // initialval
  wxArrayString filterOps;
  filterOps.Add(this->MeanString);
  filterOps.Add(this->GaussString);
  filterOps.Add(this->BinomString);
  filterOps.Add(this->GaussTDString);

  this->LowPassFilter = new wxRadioBox(this,
                                       this->ID_RB_RADIOBOX,
                                       wxT("Low Pass Filter Types"),
                                       wxDefaultPosition,
                                       wxDefaultSize,
                                       filterOps,
                                       4,
                                       wxRA_SPECIFY_COLS);

  // we have to disable gaussthreshold because it is not implemented yet
  this->LowPassFilter->Enable(3, false);

  this->RescaleParamSizer->Add(this->ParamPanel,
                               0,
                               wxALIGN_CENTER_HORIZONTAL | wxALL,
                               0);
  this->RescaleParamSizer->Add(this->LowPassFilter,
                               0,
                               wxALIGN_LEFT | wxALL,
                               5);

  wxBoxSizer* paramSizer = this->GetParamSizer();
  paramSizer->Add(this->RescaleParamSizer, 1, wxALL, 0);

  this->Connect(BIAS::ID_BTN_FILTER_FDB,
                wxEVT_COMMAND_BUTTON_CLICKED,
                wxCommandEventHandler(FilterDialogRescale::OnFilterButton));

  this->Connect(this->ID_RB_RADIOBOX,
                wxEVT_COMMAND_RADIOBOX_SELECTED,
                wxCommandEventHandler(FilterDialogRescale::OnRadioBox));


  this->CreateStatusBar(1);
  this->Fit();
  this->Centre();
}


template <class InputST, class OutputST>
FilterDialogRescale<InputST, OutputST>::~FilterDialogRescale()
{
  if (!this->GetParent()->IsEnabled()) {
    this->GetParent()->Enable();
  }
}


/******************************************************************************/
/* ============================== protected ================================= */
/******************************************************************************/

template <class InputST, class OutputST>
void FilterDialogRescale<InputST, OutputST>
::OnFilterButton(wxCommandEvent& WXUNUSED(event))
{
  // disable some buttons to avoid interference, e.g. pressing preview during
  // filtering...
  this->EnableSaveButton(false);
  this->EnableFilterButton(false);
  this->EnablePreviewButton(false);
  this->EnableResetButton(false);
  this->EnableCancelButton(false);

  double factor = this->ParamPanel->GetParamValue();

  if ((factor == 1.0) && (this->GetImageIsFiltered())) {

    this->CopyOriginalImage();
    this->SetImageIsFiltered(false);
    this->SetStatusText(wxT("Reseted To Original Image Values"));

  } else if ((factor == 1.0) && !(this->GetImageIsFiltered())) {

    this->SetStatusText(wxT("Nothing Done To The Filtered Image"));

  } else {

    this->RescaleFilter.SetFactor(factor);
    this->RescaleFilter.SetLowPassType(this->Selection);
    Image<OutputST> dstImg;

    if (0 != this->RescaleFilter.Filter(this->GetOriginalImage(), dstImg)) {
      BIASERR("FilterDialogRescale::OnFilterButton: Filter Error");
      this->SetImageIsFiltered(false);
      this->SetStatusText(wxT("ERROR during FILTER"));
    } else {
      this->SetFilteredImage(dstImg);
      this->SetImageIsFiltered(true);
      wxString msg;
      msg.Printf(wxT("Image Successfully Rescaled With Factor: %f"), factor);
      this->SetStatusText(msg);
    }

  }

  // enable the buttons again
  this->EnableSaveButton();
  this->EnableFilterButton();
  this->EnablePreviewButton();
  this->EnableResetButton();
  this->EnableCancelButton();
}


template <class InputST, class OutputST>
void FilterDialogRescale<InputST, OutputST>
::OnResetButton(wxCommandEvent& event)
{
  // call the Base's class OnResetButton() method and pass the event
  this->FilterDialogBase<InputST, OutputST>::OnResetButton(event);
  this->SetImageIsFiltered(false);
  this->SetStatusText(wxT("Filtered Image Reseted"));
}


template <class InputST, class OutputST>
void FilterDialogRescale<InputST, OutputST>
::OnRadioBox(wxCommandEvent& WXUNUSED(event))
{
  wxString str = this->LowPassFilter->GetStringSelection();
  if (0 == str.Cmp(this->MeanString)) {
    this->Selection = BIAS::LPT_Mean;
  } else if (0 == str.Cmp(this->GaussString)) {
    this->Selection = BIAS::LPT_Gauss;
  } else if (0 == str.Cmp(this->BinomString)) {
    this->Selection = BIAS::LPT_Binomial;
  } else if (0 == str.Cmp(this->GaussTDString)) {
    this->Selection = BIAS::LPT_GaussThreshold;
  } else {
    // hmm...maybe someone has changed a label?!?...then be MEAN...grrr
    this->Selection = BIAS::LPT_Mean;
  }

}


/* =================== supported storagetype combinations =================== */
namespace BIAS{
template class FilterDialogRescale<unsigned char, unsigned char>;
template class FilterDialogRescale<unsigned char, float>;
//template class FilterDialogRescale<float, unsigned char>;
template class FilterDialogRescale<float, float>;
}
