#ifndef __GeometryGL_hh__
#define __GeometryGL_hh__

// ----------------------------------------------------------
// @author Jan Woetzel (c)  09/2003
// all rights reserved

#include <bias_config.h>

#ifndef BIAS_HAVE_OPENGL
#  error Please recompile BIAS with USE_OPENGL to use this lib. 
#endif // BIAS_HAVE_OPENGL


// STD
#include <vector>
#include <iostream>

#include <Base/Math/Vector3.hh>
#include <Geometry/PMatrix.hh>

#include <Gui/ImgObjGL.hh>
#include <Gui/RenderModeGL.hh>

#ifdef BIAS_HAVE_FONTGL
#  include <Utils/DrawTextGL.hh>
#endif // BIAS_HAVE_FONTGL


#ifndef VISIBLEALPHA
#  define VISIBLEALPHA 0.0
#endif

// used by OpenGLInfo: 
#define MAX_GL_DISPLIST_ID 1000
#define MAX_GL_TEXOBJ_ID   1000

namespace BIAS {

  // corners:
  enum CornerIndices {
    LOWER_LEFT=0
    ,LOWER_RIGHT
    ,UPPER_RIGHT
    ,UPPER_LEFT
    ,OPTICAL_CENTER
  };


  // jw ---------------------------------------------------------
  /** @brief Reusable Geometry an drawing functions for OpenGL 
  @author Jan Woetzel */
  class BIASGui_EXPORT GeometryGL {
  public:
    /// destructor
    virtual ~GeometryGL();

    /// ctor
    GeometryGL();

    /// utility functions

    /// deps to glut:
    //static void DisplayTeapotLit(const float & size=0.5f, const bool & solid=true);
    //static void DisplaySceneExampleLit(const bool & UseResetView=false);

    static void DisplayClear(const GLbitfield & mask=CLEAR_DEFAULT_MASK );

    static void DisplayRenderMode(const BIAS::RenderModeGL  & m);

    /** draw a coord cross as three lines in r,g,b 
    @param withAxes Label true to draw x/y/Z text as axes label JW */
    static void DisplayCoordCross(const float & size=1.0f,
      const float & linewidth=1.0f );

#ifdef BIAS_HAVE_FONTGL
    /** draw only the axes labels for a coord cross as text in r,g,b JW */
    static void DisplayCoordCrossAxesLabel(const float & size, 
      const BIAS::DrawTextGL & font,
      const std::string & LabelX=std::string("x"),
      const std::string & LabelY=std::string("y"),
      const std::string & LabelZ=std::string("z") );

    /** draw coord cross with labels in r,g,b JW */
    static void DisplayCoordCross(const float & size, 
      const float & linewidth,
      const BIAS::DrawTextGL & font,
      const std::string & LabelX=std::string("x"),
      const std::string & LabelY=std::string("y"),
      const std::string & LabelZ=std::string("z") );

#endif

    static void DisplayCameraInternalParams(
      const int & x0,
      const int & y0,
      const unsigned int & width,
      const unsigned int & height,
      const float & fovY,
      const float & zNear,
      const float & zFar );

    static void DisplayCameraExternalParams(BIAS::PMatrix & P);

    static void DisplayCameraExternalParams(const BIAS::Vector3<double> & C,
      const BIAS::Vector3<double> & up,
      const BIAS::Vector3<double> & a);

    /// print all GL state informatino to os including displaylist, texobj and matric info
    static void GetInfoGL(std::ostream & os);
    /// print individual information to os
    static void GetInfoGLenvironment(std::ostream & os);
    static void GetInfoGLTexID(const GLint id, int & foundTextures, std::ostream & os, const bool & noOutputIfInvalid=true);
    static void GetInfoGLTexObjects(std::ostream & os);
    static void GetInfoGLDisplayLists(std::ostream & os);
    static void GetInfoGLmatrices(std::ostream & os);
    static void GetInfoGLWriteMasks(std::ostream & os);

    /// return a name for an OpenGL flags nr.
    static std::string GetName(const GLint flag, 
      const bool withNumericDisplay=true,
      const bool numerixHexFormat=true);

    static BIAS::Vector3<double> GetImgCorner3d(BIAS::PMatrix & P
      ,const unsigned int & width
      ,const unsigned int & height
      ,const unsigned int & corner
      ,const double & scale );

    static BIAS::Vector2<double> GetImgCorner2d( const unsigned int & w
      ,const unsigned int & h
      ,const unsigned int & corner
      ,const int texMode
      ,const bool flipY );

    static void DisplayP(BIAS::PMatrix & P
      ,const unsigned int & w
      ,const unsigned int & h
      ,const double & scale
      ,const float & linewidth=1.0f );

    /// time dependant scene for samples
    static void DisplaySimpleAnimation(const double & time);

    // time independant scene for samples 
    static void DisplaySceneExampleSimple();

    /// draws a pkane with attched texture coordinates
    static void DisplayImgPlaneTextured(BIAS::PMatrix & P
      ,const unsigned int & w
      ,const unsigned int & h
      ,const int & texID
      ,const int & texTarget
      ,const double & scale
      ,const bool & flipY
      ,const BIAS::Vector4<float> & vertexColor=BIAS::Vector4<float>(1.0,0.0,1.0, 0.0)
      );

    /** compose an OpenGL 4x4 (projective) texture matrix 
    @param useRECTcoord true to use RECT 0..npixels
    instead of 0..1 normalized pw2 texture coordinate JW  */
    static BIAS::Matrix4x4<double>
      CreateTextureMatrix(BIAS::PMatrix & P
      ,const unsigned int w
      ,const unsigned int h
      ,const bool & useRECTcoord=false );

    static void DisplayQuadTexCoord2D(const BIAS::ImgObjGL & texobj, 
      const bool & additionalFlipY );

    static void DisplayQuadTexCoord2DFull(const BIAS::ImgObjGL & texobj,
      const bool & additionalFlipY);

    /// creates a sample scene as one dispaly list returned in dplistID once
    /// the arg is kept as it is if initialization was done before. 
    static void CreateSampleDisplayListScene(int & dplistID);

    // displays a simple displa ylist scene with animation 
    static void DisplayCallsSampleDisplayListScene(const double & curTime);

    /// fill
    static void FillCubemap(BIAS::ImgObjGL & im, const bool & mipmapColorChange=false);

  };
} // namespace BIAS
#endif
