#include "ColorSegmentation.hh"

#include <bias_config.h>
#ifdef BIAS_HAVE_PTHREADS
#  include <pthread.h>
#else
#  warning BIAS_HAVE_PTHREADS not defined. BIAS::ColorSegmentationThreaded will not work properly.
#endif

#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Debug/Exception.hh>
#include <cmath>

using namespace BIAS;
using namespace std;

namespace BIAS {

  /** @struct SegmentationData
      @brief Holds data for ColorSegmentation.
      @author africk */
  struct BIASImage_EXPORT SegmentationData
  {
    int tl[2], br[2];
    const unsigned char **imgHSV;
    unsigned char **imRes;
    int hue, hueIntervall, sat, satIntervall;
  };

}

ColorSegmentation::
ColorSegmentation()
{
  stdFarbRaum=BIAS::ImageBase::CM_HSL;
}


void ColorSegmentation::
SetColorSpace(BIAS::ImageBase::EColorModel farbraum)
{
  stdFarbRaum=farbraum;
}


void ColorSegmentation::
Segment(BIAS::Image<unsigned char> &src,
        BIAS::Image<unsigned char> &res,
        int hue, int hueIntervall, int sat, int satIntervall,
        int light, int lightIntervall)
{
  const unsigned int width = src.GetWidth();
  const unsigned int height = src.GetHeight();

  if (res.GetWidth() !=  src.GetWidth() ||
      res.GetHeight() != src.GetHeight() ||
      res.GetChannelCount() != 1){
    if (!res.IsEmpty()){
      res.Release();
    }
  }

  if(res.IsEmpty()){
    res = BIAS::Image<unsigned char>(width,height,1);
  }

  unsigned char **ida = src.GetImageDataArray();
  if (src.GetColorModel() != stdFarbRaum){
    BIAS::ImageConvert::Convert(src,image,stdFarbRaum);
    ida = image.GetImageDataArray();
  }

  unsigned char **rida = res.GetImageDataArray();
  int tl[] = {0, 0};
  int br[] = {(int)width, (int)height};
  Segment((const unsigned char **)ida, rida, tl, br, hue,
          hueIntervall, sat, satIntervall, light, lightIntervall);
}

void ColorSegmentation::
SegmentTreshold(BIAS::Image<unsigned char> &src, 
                BIAS::Image<unsigned char> &res, 
                int min,int max, int satTreshhold) {

  if (! (res.IsEmpty() && src.IsEmpty()) ) {

    int heightc = src.GetHeight();
    int widthc = src.GetWidth();

    int channelsc=src.GetChannelCount();
    int channelres = res.GetChannelCount();

    if ((channelsc == 3) && (channelres == 1)) {
      unsigned char ** datac = (unsigned char **)src.GetImageDataArray();
      unsigned char ** datar = (unsigned char **)res.GetImageDataArray();

      for(int i=0;i< (heightc);i++) {
        for(int j=0;j<(widthc);j++)
        {
          datar[i] [j*channelsc] = 0;
          if ((datac[i] [j*channelsc] >= min) && (datac[i] [j*channelsc] <= max)) {
            if ((datac[i] [j*channelsc + 1] >= satTreshhold)) {
              datar[i] [j*channelres] = 255;
            }
          }
        }
      }
    } else {
      BIASERR("Channelcount not ok!");
    }

  } else {
    BIASERR("Source or dest - picture is empty");
  }

}



void ColorSegmentation::
GetHueImage(BIAS::Image<unsigned char> &imgHue)
{
  unsigned int width=image.GetWidth();
  unsigned int height=image.GetHeight();
  if(imgHue.IsEmpty())
  {
    imgHue=BIAS::Image<unsigned char>(width,height,1);
  }
  for (unsigned int y=0; y<height; y++)
  {
    for (unsigned int x=0;x<width; x++)
    {
      imgHue.SetPixel(image.GetImageDataArray()[y][3*x],x,y,0);
    }
  }
}


void ColorSegmentation::
GetSatImage(BIAS::Image<unsigned char> &imgSat)
{

  unsigned int width=image.GetWidth();
  unsigned int height=image.GetHeight();
  if(imgSat.IsEmpty())
  {
    imgSat=BIAS::Image<unsigned char>(width,height,1);
  }
  for (unsigned int y=0; y<height; y++)
  {
    for (unsigned int x=0;x<width; x++)
    {
      imgSat.SetPixel(image.GetImageDataArray()[y][3*x+1],x,y,0);
    }
  }
}


void ColorSegmentation::
GetGreyImage(BIAS::Image<unsigned char> &imgGrey)
{
  unsigned int width=image.GetWidth();
  unsigned int height=image.GetHeight();
  if(imgGrey.IsEmpty())
  {
    imgGrey=BIAS::Image<unsigned char>(width,height,1);
  }
  for (unsigned int y=0; y<height; y++)
  {
    for (unsigned int x=0;x<width; x++)
    {
      imgGrey.SetPixel(image.GetImageDataArray()[y][3*x+2],x,y,0);
    }
  }
}


void ColorSegmentationThreaded::
Segment(BIAS::Image<unsigned char> &src,
        BIAS::Image<unsigned char> &res,
        int hue, int hueIntervall,int sat, int satIntervall)
{
  if (res.GetWidth() !=  src.GetWidth() ||
      res.GetHeight() != src.GetHeight() ||
      res.GetChannelCount() != 1){
    if (!res.IsEmpty()){
      res.Release();
    }
  }

  if(res.IsEmpty()){
    res = BIAS::Image<unsigned char>(src.GetWidth(), src.GetHeight(), 1);
  }

  Image<unsigned char> hsv;
  if (src.GetColorModel() != BIAS::ImageBase::CM_HSL){
    BIAS::ImageConvert::Convert(src, hsv, BIAS::ImageBase::CM_HSL);
  } else {
    hsv.StealImage(src);
  }

  SegmentationData d1, d2;
  const int width = (int)src.GetWidth();
  const int height = (int)src.GetHeight();
  d1.tl[0] = d1.tl[1] = 0;
  d1.br[0] = width; d1.br[1] = height/2;
  d2.tl[0] = 0;  d2.tl[1] = height/2-1;
  d2.br[0] = width; d2.br[1] = height;
  d1.imgHSV = d2.imgHSV = (const unsigned char **)hsv.GetImageDataArray();
  d1.imRes = d2.imRes = res.GetImageDataArray();
  d1.hue = d2.hue = hue;
  d1.hueIntervall = d2.hueIntervall = hueIntervall;
  d1.sat = d2.sat = sat;
  d1.satIntervall = d2.satIntervall = satIntervall;

#ifdef BIAS_HAVE_PTHREADS
  pthread_t thread1, thread2;
  if (pthread_create(&thread1, NULL,
                     &ColorSegmentationThreaded::Segment_, &d1)!=0){
    BEXCEPTION("error creating thread 1");
  }
  if (pthread_create(&thread2, NULL,
                     &ColorSegmentationThreaded::Segment_, &d2)!=0){
    BEXCEPTION("error creating thread 2");
  }
  if (pthread_join(thread1, NULL)!=0){
    BEXCEPTION("error joining thread 1");
  }
  if (pthread_join(thread2, NULL)!=0){
    BEXCEPTION("error joining thread 2");
  }
#else
  BIASWARNONCE("BIAS does not support pthreads! Using non-threaded segmentation as fallback.");
  ColorSegmentationThreaded::Segment_(&d1);
  ColorSegmentationThreaded::Segment_(&d2);
#endif
}


void *ColorSegmentationThreaded::
Segment_(void *data)
{
  SegmentationData *segdata = static_cast<SegmentationData *>(data);
  ColorSegmentation segmenter;
  segmenter.Segment(segdata->imgHSV, segdata->imRes, segdata->tl,
                    segdata->br, segdata->hue,
                    segdata->hueIntervall, segdata->sat, segdata->satIntervall);
  return (void *)(NULL);
}
