/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Tracker.hh"
#include "TrackerBaseSimple.hh"
#include "TrackerBaseAffine.hh"
#include "TrackerBaseAffine2.hh"
#include "TrackerBaseHomography.hh"
#include "TrackerBaseWeighted.hh"
#include "TrackerBaseWeighted1D.hh"
#include <Base/Image/ImageIO.hh>
#include <Base/Common/BIASpragma.hh>
#include <Base/Image/ImageConvert.hh>
#include <Filter/GradientSobel3x3.hh>
#include <Filter/Binomial.hh>

  //#define TIMING
#ifdef TIMING
#include <Base/Debug/TimeMeasure.hh>
#endif

using namespace BIAS;
using namespace std;

#ifdef BIAS_DEBUG
double totalitersum=0.0;
#endif

#define DFLT_PYSIZE 4


//////////////////////////////////////////////////////////////////////////
//                 implementation
//////////////////////////////////////////////////////////////////////////

template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
Tracker(int tracker_type) {
  this->tb_ = NULL;
  SetTrackerType(tracker_type);

  PyIndex_ = 0;

  MaxIter_.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) MaxIter_[i]=5;

  HalfWinSize_.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) HalfWinSize_[i]=3;

  MaxError_.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) MaxError_[i]=0.1;

  MaxResiduumMAD_.newsize(DFLT_PYSIZE);
  for (int i=0; i<DFLT_PYSIZE; i++) MaxResiduumMAD_[i]=10.0;

  UseCoarserPointIfLost_ = false;

  // filter masks
  LowpassFM_ = FilterMask(Vector3<FM_FLOAT>( 1, 2, 1),
                          Vector3<FM_FLOAT>( 1, 2, 1));
  GradXFM_ = FilterMask(Vector3<FM_FLOAT>(-1, 0, 1),
                        Vector3<FM_FLOAT>( 1, 2, 1));
  GradYFM_ = FilterMask(Vector3<FM_FLOAT>( 1, 2, 1),
                        Vector3<FM_FLOAT>(-1, 0, 1));

  LastImage_ = NULL;
  LastGradX_ = NULL;
  LastGradY_ = NULL;
}


template <class StorageType, class CalculationType>
Tracker<StorageType, CalculationType>::
~Tracker()
{
  if (tb_)
    delete tb_;
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ReplaceLastImages(PyramidImageInterface<CalculationType> *pim,
                  PyramidImageInterface<CalculationType> *gradx,
                  PyramidImageInterface<CalculationType> *grady)
{
  //   BIASASSERT((pim && gradx && grady) || (!pim && !gradx && !grady));
  //   BIASASSERT(!pim || (pim->Size() == gradx->Size() &&
  //                       pim->Size() == grady->Size()));

  if (LastImage_) { delete LastImage_; LastImage_ = NULL; }
  if (LastGradX_) { delete LastGradX_; LastGradX_ = NULL; }
  if (LastGradY_) { delete LastGradY_; LastGradY_ = NULL; }

  if (pim ){
    LastImage_ = pim->ShallowClone();
  }
  if(gradx && grady) {
    LastGradX_ = gradx->ShallowClone();
    LastGradY_ = grady->ShallowClone();
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ClearPoints()
{
  Points_.clear();
  PredictedPoints_.clear();
  NumIter_.clear();
  DisplInLastIter_.clear();
  ResiduiMAD_.clear();
  ResiduiMSD_.clear();
  res_.clear();                  // rabe, 11.11.2004
  Cov_.clear();
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ResizeInternals_(const int pysize)
{
  if (Points_.size() != (unsigned)pysize){
    Points_.resize(pysize);
    PredictedPoints_.resize(pysize);
    NumIter_.resize(pysize);
    DisplInLastIter_.resize(pysize);
    ResiduiMAD_.resize(pysize);
    ResiduiMSD_.resize(pysize);
    res_.resize(pysize);
    Cov_.resize(pysize);
    AffineResults_.resize(pysize);
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ResizeInternals_(const int pysize, const int num_points)
{
  ResizeInternals_(pysize);

  // dherzog: replaced '<numnew' by '!=numnew', because Track() asserts sizes
  if ((int)Points_[PyIndex_].size() != num_points){
    HomgPoint2D pz(42.0, 42.0, 1.0);
    p2tracked_.resize(num_points, pz);
    AffinePrediction_.resize(num_points);
    for (int p = PyIndex_; p < pysize; p++){
      res_[p].resize(num_points);
      Points_[p].resize(num_points);
      PredictedPoints_[p].resize(num_points);
      NumIter_[p].resize(num_points);
      DisplInLastIter_[p].resize(num_points);
      ResiduiMAD_[p].resize(num_points);
      ResiduiMSD_[p].resize(num_points);
      Cov_[p].resize(num_points);
      AffineResults_[p].resize(num_points);
    }
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
ReserveInternals_(const int pysize, const int newsize)
{
  int pyindex = PyIndex_;
  for (int p=pyindex; p<pysize; p++){
    res_[p].reserve(newsize);
    Points_[p].reserve(newsize);
    PredictedPoints_[p].reserve(newsize);
    NumIter_[p].reserve(newsize);
    DisplInLastIter_[p].reserve(newsize);
    ResiduiMAD_[p].reserve(newsize);
    ResiduiMSD_[p].reserve(newsize);
    Cov_[p].reserve(newsize);
    AffineResults_[p].resize(newsize);
  }
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
Reset()
{
  ClearPoints();
  ReplaceLastImages(NULL, NULL, NULL);
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
DetermineROI_(int &tlx, int &tly, int &brx, int &bry) const
{
  BIASASSERT(LastImage_ && LastGradX_ && LastGradY_);
  (*LastImage_)[PyIndex_]->GetROI()->GetCorners(tlx, tly, brx, bry);
  int gxtlx, gxtly, gxbrx, gxbry;
  (*LastGradX_)[PyIndex_]->GetROI()->GetCorners(gxtlx, gxtly, gxbrx, gxbry);
  int gytlx, gytly, gybrx, gybry;
  (*LastGradY_)[PyIndex_]->GetROI()->GetCorners(gytlx, gytly, gybrx, gybry);

  tlx = max(tlx, max(gxtlx, gytlx));
  tly = max(tly, max(gxtly, gytly));
  brx = min(brx, min(gxbrx, gybrx));
  bry = min(bry, min(gxbry, gybry));

  tlx += HalfWinSize_[PyIndex_];
  tly += HalfWinSize_[PyIndex_];
  brx -= HalfWinSize_[PyIndex_];
  bry -= HalfWinSize_[PyIndex_];
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
FillUpPoints(const vector<HomgPoint2D>& points,
             const vector<float>& qual)
{
  // compute MinBorderDist
  int lpx, lpy, gx, gy;
  LowpassFM_.GetSize(lpx, lpy);
  GradXFM_.GetSize(gx, gy);
  int MinBorderDist =
    max((lpx-1)/2, (lpy-1)/2) +
    max((gx-1)/2, (gy-1)/2) +
    HalfWinSize_[PyIndex_];

  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(Points_.size()>0);
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  int numnew=points.size();
  const int size=Points_[0].size();
  const int pyindex=PyIndex_;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();
  BIASCDOUT(D_TRACKER_INIT, "min distance to image border "
            <<MinBorderDist<<endl);
  int tlx, tly, brx, bry;
  (*LastImage_)[pyindex]->GetROI()->GetCorners(tlx, tly, brx, bry);
  const double minx=MinBorderDist+tlx;
  const double miny=MinBorderDist+tly;
  const double maxx=brx-MinBorderDist;
  const double maxy=bry-MinBorderDist;

#ifdef BIAS_DEBUG
  BIASCDOUT(D_TRACKER_FILL_UP, "minx" << minx << "miny" << miny <<
            "maxx" << maxx << "maxy" << maxy << endl);
#endif
  int n=0, num=0;
  while (n<numnew && (points[n].IsAtInfinity() ||
                      !(points[n][0]>minx && points[n][1]>miny &&
                        points[n][0]<maxx && points[n][1]<maxy)))
    n++;
#ifdef BIAS_DEBUG
  if (n<numnew)
    BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
  for (int i=0; i<size; i++){
    if (n<numnew){
      if (Points_[pyindex][i].IsAtInfinity()){
        p2d=points[n];
        p2d.Homogenize();
        BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<i<<" : replacing "
                  <<Points_[pyindex][i]<<" with "<<p2d<<endl);
        for (int p=pyindex; p<pysize; p++){
          Points_[p][i]=p2d;
          NumIter_[p][i]=-1;
          DisplInLastIter_[p][i]=-1.0;
          ResiduiMAD_[p][i]=-1.0;
          ResiduiMSD_[p][i]=-1.0;
          p2d[0]-=offset;
          p2d[1]-=offset;
          p2d[0]*=factor;
          p2d[1]*=factor;
          res_[p][i]=(int)rint(qual[n]);
          Cov_[p][i]=zero2x2;
        }
        n++; num++;
        while (n<numnew && ( points[n].IsAtInfinity() ||
                             !(points[n][0]>minx && points[n][1]>miny &&
                               points[n][0]<maxx && points[n][1]<maxy)))
          n++;
#ifdef BIAS_DEBUG
        if (n<numnew)
          BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
      } else {
#ifdef BIAS_DEBUG
        BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<i<<" : "<<Points_[pyindex][i]
                  <<" is still valid:  ");
        for (int p=pyindex+1; p<pysize; p++){
          BIASCDOUT(D_TRACKER_FILL_UP, p<<": "<<Points_[p][i]);
        }
        BIASCDOUT(D_TRACKER_FILL_UP, endl);
#endif
      }
    } else {
      num=-num;
      BIASCDOUT(D_TRACKER_FILL_UP, "not enought points in points "
                <<numnew<<endl);
      break;
    }
  }
  return num;
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
ReplacePoints(const vector<HomgPoint2D>& points,
              const vector<float>& qual)
{
  assert("deprecated, uese ReplaceAllPoints instead" && false);
  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  BIASASSERT(Points_.size()>0);
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  const int numnew=points.size();
#ifdef BIASASSERT_ISACTIVE
  const int size=Points_[0].size();
  BIASASSERT(size==numnew);
#endif
  const int pyindex=PyIndex_;
  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();

  int n=0, num=0;
  while (n<numnew){
    while (n<numnew && points[n].IsAtInfinity()) n++;
    if (n<numnew){
#ifdef BIAS_DEBUG
      BIASCDOUT(D_TRACKER_FILL_UP, "next new point is "<<points[n]<<endl);
#endif
      p2d=points[n];
      p2d.Homogenize();
      BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<n<<" : replacing "
                <<Points_[pyindex][n]<<" with "<<p2d<<endl);
      for (int p=pyindex; p<pysize; p++){
        Points_[p][n]=p2d;
        NumIter_[p][n]=-1;
        DisplInLastIter_[p][n]=-1.0;
        ResiduiMAD_[p][n]=-1.0;
        ResiduiMSD_[p][n]=-1.0;
        p2d[0]-=offset;
        p2d[1]-=offset;
        p2d[0]*=factor;
        p2d[1]*=factor;
        res_[p][n]=(int)rint(qual[n]);
        Cov_[p][n]=zero2x2;
      }
      n++; num++;
    }
  }
  BIASCDOUT(D_TRACKER_FILL_UP, "replaced "<<num<<" points"<<endl);
  return n;
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
ReplaceAllPoints(const vector<HomgPoint2D>& points,
                 const vector<float>& qual)
{
  //AddDebugLevel(D_TRACKER_FILL_UP);
  BIASASSERT(LastImage_!=NULL);
  const int numnew=points.size();
  const int pysize=(int)LastImage_->size();
  const double factor=1.0/LastImage_->GetRescaleFactor();
  const double offset=LastImage_->GetPositionOffset();
  const int pyindex=PyIndex_;
  // dherzog: replaced '<numnew' by '!=numnew', because Track() asserts sizes
  if ((int)Points_.size() != pysize || (int)Points_[pyindex].size()!=numnew) {
    ClearPoints();
    ResizeInternals_(pysize, numnew);
  }

  HomgPoint2D p2d;
  Matrix2x2<double> zero2x2; zero2x2.SetZero();

  for (int n=0; n<numnew; n++){
    p2d = points[n];
    p2d.Homogenize();
    BIASASSERT(!p2d.IsAtInfinity());
    BIASCDOUT(D_TRACKER_FILL_UP, setw(3)<<n<<" : replacing "
              <<Points_[pyindex][n]);
    for (int p=pyindex; p<pysize; p++){
      res_[p][n] = (int)rint(qual[n]);
      Points_[p][n] = p2d;
      PredictedPoints_[p][n] = p2d;
      BIASASSERT(!Points_[p][n].IsAtInfinity());
      BIASASSERT(!PredictedPoints_[p][n].IsAtInfinity());
      NumIter_[p][n] = -1;
      DisplInLastIter_[p][n] = -1.0;
      ResiduiMAD_[p][n] = -1.0;
      ResiduiMSD_[p][n] = -1.0;
      Cov_[p][n] = zero2x2;

      p2d[0] -= offset;
      p2d[1] -= offset;
      p2d[0] *= factor;
      p2d[1] *= factor;
    }
    BIASCDOUT(D_TRACKER_FILL_UP, " with "<<Points_[pyindex][n]<<endl);
  }
  BIASCDOUT(D_TRACKER_FILL_UP, "replaced "<<numnew<<" points"<<endl);

  PadPoints_(numnew);

  return numnew;
}


#ifdef BIAS_DEBUG
static int Tracker_GlobalImageCounter=0;
#endif

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
PreparePyramide(const Image<StorageType>& im,
                PyramidImageInterface<CalculationType>& pim,
                PyramidImageInterface<CalculationType>& gx,
                PyramidImageInterface<CalculationType>& gy,
                FilterNTo2N<CalculationType, CalculationType>& grad,
                FilterNToN<StorageType, CalculationType>& lowpass)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int fres=0, gres=0;
#ifdef BIAS_DEBUG
  BIASASSERT(im.GetChannelCount()==1);
  BIASASSERT(im.GetColorModel()==ImageBase::CM_Grey);
#endif
  //  cout << "pim before init:\n";
  //   pim.Dump();

  const unsigned py_size = pim.size();
  if (pim.size()==0 || !im.SamePixelAndChannelCount(*pim[0])){
    if (!pim.IsEmpty()) pim.Clear();   
    BIAS::Image<CalculationType> im2;
    ImageConvert::ConvertST(im,im2,im2.GetStorageType());
    pim.Init(im2,py_size);
    //pim.Init(im.GetWidth(), im.GetHeight(), 1, py_size); //this doesn'T create any image surfaces anymore...
    pim.SetZero();
  }
  if (gx.size()==0 || !im.SamePixelAndChannelCount(*gx[0])){
    if (!gx.IsEmpty()) gx.Clear();  
    Image<CalculationType> gr(im.GetWidth(),im.GetHeight()); //dummy image
      gx.Init(gr,py_size);
    //gx.Init(im.GetWidth(), im.GetHeight(), 1, py_size);
    gx.SetZero();
  }
  if (gy.size()==0 || !im.SamePixelAndChannelCount(*gy[0])){
    if (!gy.IsEmpty()) gy.Clear();    
    Image<CalculationType> gr(im.GetWidth(),im.GetHeight()); //dummy image
      gy.Init(gr,py_size);
    //gy.Init(im.GetWidth(), im.GetHeight(), 1, py_size);
    gy.SetZero();
  }

  //  cout << "pim after init:\n";
  //   pim.Dump();

#ifdef TIMING
  timer.Stop();
  cerr << "initializing pyramid images took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif

#ifdef BIAS_DEBUG
  ostringstream os;
  if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
    os.str("");
    os << "input-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
       <<".mip";
    Image<StorageType> tmp_im=im;
    //if (ImageIO::Save(os.str(), tmp_im)!=0){
    if (ImageIO::Save(os.str(), tmp_im)!=0){
      BIASERR("error writing "<<os.str());
    }
  }
#endif

  //lowpass_->AddDebugLevel(D_CONV_KERNEL);
  //lowpass_->AddDebugLevel(D_FILTERBASE_CALLSTACK);
  //lowpass_->AddDebugLevel(D_WRITE_IMAGES);
  if((fres=lowpass.Filter(im, *(pim[0])))!=0){
    BIASERR("filtering failed "<<fres);
  }
  //  lowpass_->PrintKernel();

#ifdef BIAS_DEBUG
  if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
    os.str("");
    os << "pyim-lowpass-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
       << "-pyind"<<setw(1)<<setfill('0')<<0<<".mip";
    //if (ImageIO::Save(os.str(), *pim[0])!=0){
    if (ImageIO::Save(os.str(), *pim[0])!=0){
      BIASERR("error writing "<<os.str());
    }
  }
#endif

  //lowpass_->RemoveDebugLevel(D_CONV_KERNEL);
#ifdef TIMING
  timer.Stop();
  cerr << "lowpass  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif

  pim.Downsample();

#ifdef TIMING
  timer.Stop();
  cerr << "downsampling  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif

  const int pysize=pim.size(); // comes from parameter
  for (int p=0; p<pysize; p++){
    if((gres=grad.Filter(*pim[p], *gx[p], *gy[p]))!=0){
      BIASERR("gradient failed "<<gres);
      break;
    }
#ifdef BIAS_DEBUG
    if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
      os.str("");
      os << "pyim-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      //if (ImageIO::Save(os.str(), *pim[p])!=0){
      if (ImageIO::Save(os.str(), *pim[p])!=0){
        BIASERR("error writing "<<os.str());
      }
      os.str("");
      os << "pygx-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      //if (ImageIO::Save(os.str(), *gx[p])!=0){
      if (ImageIO::Save(os.str(), *gx[p])!=0){
        BIASERR("error writing "<<os.str());
      }
      os.str("");
      os << "pygy-im"<<setw(3)<<setfill('0')<<Tracker_GlobalImageCounter
         << "-pyind"<<setw(1)<<setfill('0')<<p<<".mip";
      //if (ImageIO::Save(os.str(), *gy[p])!=0){
      if (ImageIO::Save(os.str(), *gy[p])!=0){
        BIASERR("error writing "<<os.str());
      }
    }
#endif
  }
#ifdef BIAS_DEBUG
  Tracker_GlobalImageCounter++;
#endif
#ifdef TIMING
  timer.Stop();
  cerr << "gradients  took "<<timer.GetRealTime()<<" us\n";
  timer.Reset();
  timer.Start();
#endif
  return (fres==0 && gres==0)?0:-1;
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
PreparePyramide(const Image<StorageType>& im,
                PyramidImageInterface<CalculationType>& pim)
{
#ifdef BIAS_DEBUG
  BIASASSERT(im.GetChannelCount()==1);
  BIASASSERT(im.GetColorModel()==ImageBase::CM_Grey);
#endif

  const unsigned py_size = pim.size();
  if (pim.size()==0 || !im.SamePixelAndChannelCount(*pim[0])){
    if (!pim.IsEmpty()) pim.Clear();
    pim.Init(im.GetWidth(), im.GetHeight(), 1, py_size);
    pim.SetZero();
  }


  if (im.GetStorageType()==pim[0]->GetStorageType()) {
    *pim[0] = im;
  } else {
    ImageConvert::ConvertST(im, *pim[0], pim[0]->GetStorageType());
  }

  pim.Downsample();

  return 0;
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
PreparePyramide(const Image<StorageType>& im,
                PyramidImageInterface<CalculationType>& pim,
                PyramidImageInterface<CalculationType>& gx,
                PyramidImageInterface<CalculationType>& gy)
{
  GradientSobel3x3<CalculationType, CalculationType> grad;
  Binomial<StorageType, CalculationType> lowpass;
  return PreparePyramide(im, pim, gx, gy, grad, lowpass);
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImageInterface<CalculationType>& pim)
{
  return Track(pim, Points_[PyIndex_]);
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImageInterface<CalculationType>& pim,
      const vector<HomgPoint2D>& prediction,
      const vector<Matrix2x2<KLT_TYPE> >& AffinePred)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int num=0;
  const int pysize=(int)pim.size();
  const double factor=pim.GetRescaleFactor();
  const double offset=pim.GetPositionOffset();
  const Matrix2x2<double> K(factor, 0.0, 0.0, factor);
  //tb_->AddDebugLevel(D_TRACKERB_GKLT);
  //tb_->AddDebugLevel(D_TRACKERB_INIT);
  //tb_->AddDebugLevel(D_TRACKERB_REJECT);

  BIASASSERT(PyIndex_>=0 && PyIndex_<pysize);

  if (LastImage_){
#ifdef BIAS_DEBUG
    if ((&pim)==LastImage_){
      BIASERR("do not remove image between successive calls of Track\n"
              <<&pim<<"!="<<LastImage_);
      //      BIASERR("do not remove image between successive calls of Track\n"
      //              <<(int)(&pim)<<"!="<<(int)(LastImage_));
    }
    BIASASSERT((&pim)!=LastImage_);   
    // AddDebugLevel(D_TRACKER_IMAGES);
    if (DebugLevelIsSet(D_TRACKER_IMAGES)){
      pim.WriteImages("act-py"); LastImage_->WriteImages("last-py");
    }
#endif

    const int pyindex=PyIndex_;
    const int nump=Points_[pyindex].size();

    // fill PredictedPoints_ and Points_
    PredictedPoints_[pyindex]=prediction;

    const double ifac=1.0/factor;
    for (int p=pyindex+1; p<pysize; p++){
      for (int i=0; i<nump; i++){
        Points_[p][i][0]=(Points_[p-1][i][0]-offset)*ifac;
        Points_[p][i][1]=(Points_[p-1][i][1]-offset)*ifac;
        PredictedPoints_[p][i][0]=(PredictedPoints_[p-1][i][0]-offset)*ifac;
        PredictedPoints_[p][i][1]=(PredictedPoints_[p-1][i][1]-offset)*ifac;
      }
    }

    const bool Affine = (dynamic_cast<TrackerBaseAffine<CalculationType>*>(tb_) != NULL) || (dynamic_cast<TrackerBaseAffine2<CalculationType>*>(tb_) != NULL) ;
    if (Affine) {
      if ((int)AffinePred.size()==nump) {
        // copy from vector
        AffinePrediction_ = AffinePred;
      } else {
        // have no prediction, start with identity
        AffinePrediction_.resize(nump);
        for (int i=0; i<nump; i++) {
          AffinePrediction_[i].SetIdentity();
        }
      }
    }

#ifdef TIMING
    timer.Stop();
    cerr << "initializing "<<nump<<" points took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif

    // now do the actual tracking, starting on smallest pyramid level
    for (int p=pysize-1;p>=pyindex; p--){
      BIASCDOUT(D_TRACKER_TRACK, "---- pyramid level "<<p<<" ----\n");

      tb_->SetHalfWinSize(HalfWinSize_[p]);
      tb_->SetMaxIterations(MaxIter_[p]);
      tb_->SetMaxError(MaxError_[p]);
      tb_->SetMaxResiduumMAD(MaxResiduumMAD_[p]);

      // check if size of weight matrices match pyramid size
      if ((int)WeightMatrices_.size() == pysize)
      {
        tb_->SetWeightMatrix(WeightMatrices_[p]);
      }

      
      if (DebugLevelIsSet(D_TRACKER_IMAGES)){
        pim.WriteImages("act1-py"); LastImage_->WriteImages("last1-py");
      }
      tb_->Init(*(*LastImage_)[p], *pim[p], LowpassFM_,
                GradXFM_, GradYFM_);
#ifdef TIMING
      timer.Stop();
      cerr << "initializing tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

      tb_->Track(Points_[p], PredictedPoints_[p], p2tracked_,
                 DisplInLastIter_[p], NumIter_[p], ResiduiMAD_[p],
                 ResiduiMSD_[p], res_[p],Cov_[p], AffinePrediction_,
                 &AffineResults_[p]);

      if (Affine) {
        // update prediction with last estimate
        AffinePrediction_ = AffineResults_[p];
      }

#ifdef TIMING
      timer.Stop();
      cerr << "tracking tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

      // check for result
      for (int i=0; i<nump; i++){
        BIASCDOUT(D_TRACKER_TRACK, setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
                  <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
                  <<", "<<PredictedPoints_[p][i][1]<<") ");
        if (res_[p][i]>=0){
          Points_[p][i]=p2tracked_[i];
          BIASCDOUT(D_TRACKER_TRACK, " -> ("<<p2tracked_[i][0]<<", "
                    <<p2tracked_[i][1]<<") in "<<NumIter_[p][i]
                    <<" iterations with err "<<DisplInLastIter_[p][i]
                    <<" code "<<res_[p][i]
                    <<"   residuumMAD "<<ResiduiMAD_[p][i]
                    <<"   residuumMSD "<<ResiduiMSD_[p][i]<<endl);
          if (p==pyindex) {
            num++;
          } else {
            PredictedPoints_[p-1][i][0]=Points_[p][i][0]*factor+offset;
            PredictedPoints_[p-1][i][1]=Points_[p][i][1]*factor+offset;
          }
        } else { // if (res_[p][i]>=0){
          BIASCDOUT((D_TRACKER_ERROR | D_TRACKER_TRACK), " error tracking: "
                    <<Points_[p][i]<<" on py level "<<p<<" result code "
                    <<res_[p][i]<<endl);
          // cerr <<" error tracking: "<<res_[p][i]<<endl;
          // cerr << setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
          //      <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
          //      <<", "<<PredictedPoints_[p][i][1]<<") "<<endl;

          if ( UseCoarserPointIfLost_ &&
               (p!=(pysize-1) && res_[p+1][i]>=0 && p!=pyindex) ) {
             PredictedPoints_[p-1][i][0]=Points_[p+1][i][0]*factor+offset;
             PredictedPoints_[p-1][i][1]=Points_[p+1][i][1]*factor+offset;

             if (res_[p+1][i]!=0){
               res_[p][i]=res_[p+1][i];
               NumIter_[p][i]=NumIter_[p+1][i];
               DisplInLastIter_[p][i]=DisplInLastIter_[p+1][i]*factor;
               ResiduiMAD_[p][i]=ResiduiMAD_[p+1][i];
               ResiduiMSD_[p][i]=ResiduiMSD_[p+1][i];
               // if a random variable is transformed by a linear function
               // with matrix F, the covariance transformed as J*cov*J^T,
               // where J is the jacobian of F. Here the Jacobian is a
               // matrix where the only entries on the main diagonal are the
               // factors
               Cov_[p][i]=(factor*factor)*Cov_[p+1][i];
             }
          } else {
            // manifest the failure in homogenous coordinate
            for (int k=pyindex; k<pysize; k++)
              Points_[k][i][2]=0.0;
          }
        } // if (res_[p][i]>=0){
      }

#ifdef TIMING
      timer.Stop();
      cerr << "checking results of tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

#ifdef BIAS_DEBUG
      if (DebugLevelIsSet(D_TRACKER_NUMITER)){
        //check number of iterations
        double itersum=0.0;
        for (unsigned h=0; h<NumIter_[p].size(); h++)
          itersum+=NumIter_[p][h];
        //cout << p << " average number of iterations "
        //     <<itersum/(double)NumIter_[p].size()<<endl;
        totalitersum+=itersum;
        BIASCDOUT(D_TRACKER_NUMITER, "totalitersum "<<totalitersum<<endl);
      }
#endif
    }

    //DumpResults(); cerr << endl << endl;

#ifdef TIMING
    timer.Stop();
    cerr << "propagating result codes through pyramide took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif


#ifdef BIAS_DEBUG
    if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
      Image<CalculationType> wim;
      Image<CalculationType> gim;
      pim.GetSingleImage(wim);
      //ImageIO::Save("py2.mip", wim);
      ImageIO::Save("py2.mip", wim);
      LastImage_->GetSingleImage(wim);
      //ImageIO::Save("py1.mip", wim);
      ImageIO::Save("py1.mip", wim);
    }
#endif


  } else {
    BIASCDOUT(D_TRACKER_TRACK, "initialization call\n");
  }

  ReplaceLastImages(&pim, NULL, NULL);

#ifdef TIMING
  timer.Stop();
  cerr << "updateing iternals  took "<<timer.GetRealTime()<<" us\n";
#endif

#ifdef BIAS_DEBUG
  //WriteTrackLog_(num);
#endif

  return num;
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImageInterface<CalculationType>& pim,
      PyramidImageInterface<CalculationType>& gradx,
      PyramidImageInterface<CalculationType>& grady)
{
  return Track(pim, gradx, grady, Points_[PyIndex_]);
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
Track(PyramidImageInterface<CalculationType>& pim,
      PyramidImageInterface<CalculationType>& gradx,
      PyramidImageInterface<CalculationType>& grady,
      const vector<HomgPoint2D>& prediction,
      const vector<Matrix2x2<KLT_TYPE> >& AffinePred)
{
#ifdef TIMING
  TimeMeasure timer;
  timer.Start();
#endif
  int num=0;
  const int pysize=(int)pim.size();
  const double factor=pim.GetRescaleFactor();
  const double offset=pim.GetPositionOffset();
  const Matrix2x2<double> K(factor, 0.0, 0.0, factor);
  //tb_->AddDebugLevel(D_TRACKERB_GKLT);
  //tb_->AddDebugLevel(D_TRACKERB_INIT);
  //tb_->AddDebugLevel(D_TRACKERB_REJECT);

  BIASASSERT(PyIndex_>=0 && PyIndex_<pysize);

  if (LastImage_ && LastGradX_ && LastGradY_){
#ifdef BIAS_DEBUG
    if ((&pim)==LastImage_ || (&gradx)==LastGradX_ || (&grady)==LastGradY_ ){
      /*
      BIASERR("do not remove image between successive calls of Track\n"
              <<(int)(&pim)<<"!="<<(int)(LastImage_)<<","<<(int)(&gradx)<<"!="
              <<(int)(LastGradX_)<<","<<(int)(&grady)<<"!="<<(int)(LastGradY_));
      */
      BIASERR("do not remove image between successive calls of Track\n"
              <<&pim<<"!="<<LastImage_<<","<<&gradx<<"!="
              <<LastGradX_<<","<<&grady<<"!="<<LastGradY_);
    }
    BIASASSERT((&pim)!=LastImage_ && (&gradx)!=LastGradX_ &&
               (&grady)!=LastGradY_ );
    if (DebugLevelIsSet(D_TRACKER_IMAGES)){
      pim.WriteImages("act-py"); LastImage_->WriteImages("last-py");
      gradx.WriteImages("act-gx"); LastGradX_->WriteImages("last-gx");
      grady.WriteImages("act-gy"); LastGradY_->WriteImages("last-gy");
    }
#endif

    //cout << Points_.size() << endl;
    //cout << PyIndex_ << endl;

    const int pyindex=PyIndex_;
    const int nump=Points_[pyindex].size();

    // fill PredictedPoints_ and Points_
    PredictedPoints_[pyindex]=prediction;

    const double ifac=1.0/factor;
    for (int p=pyindex+1; p<pysize; p++){
      for (int i=0; i<nump; i++){
        Points_[p][i][0]=(Points_[p-1][i][0]-offset)*ifac;
        Points_[p][i][1]=(Points_[p-1][i][1]-offset)*ifac;
        PredictedPoints_[p][i][0]=(PredictedPoints_[p-1][i][0]-offset)*ifac;
        PredictedPoints_[p][i][1]=(PredictedPoints_[p-1][i][1]-offset)*ifac;
      }
    }

    const bool Affine = (dynamic_cast<TrackerBaseAffine<CalculationType>*>(tb_) != NULL)  ||(dynamic_cast<TrackerBaseAffine2<CalculationType>*>(tb_) != NULL);
    if (Affine) {
      if ((int)AffinePred.size()==nump) {
        // copy from vector
        AffinePrediction_ = AffinePred;
      } else {
        // have no prediction, start with identity
        AffinePrediction_.resize(nump);
        for (int i=0; i<nump; i++) {
          AffinePrediction_[i].SetIdentity();
        }
      }
    }

#ifdef TIMING
    timer.Stop();
    cerr << "initializing "<<nump<<" points took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif

    // now do the actual tracking, starting on smalles pyramid level
    for (int p=pysize-1;p>=pyindex; p--){
      BIASCDOUT(D_TRACKER_TRACK, "---- pyramid level "<<p<<" ----\n");

      tb_->SetHalfWinSize(HalfWinSize_[p]);
      tb_->SetMaxIterations(MaxIter_[p]);
      tb_->SetMaxError(MaxError_[p]);
      tb_->SetMaxResiduumMAD(MaxResiduumMAD_[p]);

      // check if size of weight matrices match pyramid size
      if ((int)WeightMatrices_.size() == pysize)
      {
        tb_->SetWeightMatrix(WeightMatrices_[p]);
      }

      tb_->Init(*(*LastImage_)[p], *pim[p], *(*LastGradX_)[p],
                *(*LastGradY_)[p], *gradx[p], *grady[p]);
#ifdef TIMING
      timer.Stop();
      cerr << "initializing tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

      tb_->Track(Points_[p], PredictedPoints_[p], p2tracked_,
                 DisplInLastIter_[p], NumIter_[p], ResiduiMAD_[p],
                 ResiduiMSD_[p], res_[p],Cov_[p], AffinePrediction_,
                 &AffineResults_[p]);

      if (Affine) {
        // update prediction with last estimate
        AffinePrediction_ = AffineResults_[p];
      }

#ifdef TIMING
      timer.Stop();
      cerr << "tracking tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

      // check for result
      for (int i=0; i<nump; i++){
        BIASCDOUT(D_TRACKER_TRACK, setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
                  <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
                  <<", "<<PredictedPoints_[p][i][1]<<") ");
        if (res_[p][i]>=0){
          Points_[p][i]=p2tracked_[i];
          BIASCDOUT(D_TRACKER_TRACK, " -> ("<<p2tracked_[i][0]<<", "
                    <<p2tracked_[i][1]<<") in "<<NumIter_[p][i]
                    <<" iterations with err "<<DisplInLastIter_[p][i]
                    <<" code "<<res_[p][i]
                    <<"   residuumMAD "<<ResiduiMAD_[p][i]
                    <<"   residuumMSD "<<ResiduiMSD_[p][i]
                    <<"   cov         "<<Cov_[p][i]   <<endl);
          if (p==pyindex) {
            num++;
          } else {
            PredictedPoints_[p-1][i][0]=Points_[p][i][0]*factor+offset;
            PredictedPoints_[p-1][i][1]=Points_[p][i][1]*factor+offset;
          }
        } else { // if (res_[p][i]>=0){
          BIASCDOUT((D_TRACKER_ERROR | D_TRACKER_TRACK), " error tracking: "
                    <<Points_[p][i]<<" on py level "<<p<<" result code "
                    <<res_[p][i]<<endl);
          // cerr <<" error tracking: "<<res_[p][i]<<endl;
          // cerr << setw(3)<<i<<"  ("<<Points_[p][i][0]<<", "
          //      <<Points_[p][i][1]<<")  <-->  ("<<PredictedPoints_[p][i][0]
          //      <<", "<<PredictedPoints_[p][i][1]<<") "<<endl;
          if (!UseCoarserPointIfLost_) {
            // manifest the failure in homogenous coordinate
            // do manifest it in all pyramid levels smaller than actual
            // the bigger pyramid levels are taken care of by code below
            for (int k=p; k<pysize; k++)
              Points_[k][i][2]=0.0;
          } else {
            // try to propagate results from last successfull track
            if (p!=(pysize-1) && res_[p+1][i]>=0){
              if (p!=pyindex) {
                PredictedPoints_[p-1][i][0]=Points_[p+1][i][0]*factor+offset;
                PredictedPoints_[p-1][i][1]=Points_[p+1][i][1]*factor+offset;
              }
              if (res_[p+1][i]!=0){
                res_[p][i]=res_[p+1][i];
                NumIter_[p][i]=NumIter_[p+1][i];
                DisplInLastIter_[p][i]=DisplInLastIter_[p+1][i]*factor;
                ResiduiMAD_[p][i]=ResiduiMAD_[p+1][i];
                ResiduiMSD_[p][i]=ResiduiMSD_[p+1][i];
                // if a random variable is transformed by a linear function with
                // matrix F, the covariance transformed as J*cov*J^T, where J is
                // the jacobian of F. Here the Jacobian is a matrix where the
                // only entries on the main diagonal are the factors
                Cov_[p][i]=(factor*factor)*Cov_[p+1][i];
              }
            }
          }
        } // if (res_[p][i]>=0){
      }

#ifdef TIMING
      timer.Stop();
      cerr << "checking results of tb took "<<timer.GetRealTime()<<" us\n";
      timer.Reset();
      timer.Start();
#endif

#ifdef BIAS_DEBUG
      if (DebugLevelIsSet(D_TRACKER_NUMITER)){
        //check number of iterations
        double itersum=0.0;
        for (unsigned h=0; h<NumIter_[p].size(); h++)
          itersum+=NumIter_[p][h];
        //cout << p << " average number of iterations "
        //     <<itersum/(double)NumIter_[p].size()<<endl;
        totalitersum+=itersum;
        BIASCDOUT(D_TRACKER_NUMITER, "totalitersum "<<totalitersum<<endl);
      }
#endif
    }

    //DumpResults(); cerr << endl << endl;

#ifdef TIMING
    timer.Stop();
    cerr << "propagating result codes through pyramide took "<<timer.GetRealTime()<<" us\n";
    timer.Reset();
    timer.Start();
#endif


#ifdef BIAS_DEBUG
    if (DebugLevelIsSet(D_TRACKER_WRITE_IM)){
      Image<CalculationType> wim;
      Image<CalculationType> gim;
      pim.GetSingleImage(wim);
      //ImageIO::Save("py2.mip", wim);
      ImageIO::Save("py2.mip", wim);
      LastImage_->GetSingleImage(wim);
      //ImageIO::Save("py1.mip", wim);
      ImageIO::Save("py1.mip", wim);
      gradx.GetSingleImage(gim);
      //ImageIO::Save("gradx2.mip", gim);
      ImageIO::Save("gradx2.mip", gim);
      LastGradX_->GetSingleImage(gim);
      //ImageIO::Save("gradx1.mip", gim);
      ImageIO::Save("gradx1.mip", gim);
      grady.GetSingleImage(gim);
      //ImageIO::Save("grady2.mip", gim);
      ImageIO::Save("grady2.mip", gim);
      LastGradY_->GetSingleImage(gim);
      //ImageIO::Save("grady1.mip", gim);
      ImageIO::Save("grady1.mip", gim);
    }
#endif


  } else {
    BIASCDOUT(D_TRACKER_TRACK, "initialization call\n");
  }

  ReplaceLastImages(&pim, &gradx, &grady);

#ifdef TIMING
  timer.Stop();
  cerr << "updateing iternals  took "<<timer.GetRealTime()<<" us\n";
#endif

#ifdef BIAS_DEBUG
  //WriteTrackLog_(num);
#endif

  return num;
}


template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
WriteTrackLog_(int num)
{
  static bool first = true;
  static int img_num = 0;
  const int min_err_code = -7;
  static ofstream os("tracker.log");
  if (first){
    // open the file
    if (!os.good()){
      BIASERR("error opening \"tracker.log\"");
      BIASABORT;
    }
    // write header
    os << "# reasons for lost tracks\n"
       << "#" << setw(4)<<"no"
       << setw(7) << "num"
       << setw(7) << "lost"
       << setw(7) << "ignrd";
    for (int r = 0; r >= min_err_code; r--){
      os << setw(7) << r;
    }
    os << endl;
    first = false;
  }

  const int pyindex = PyIndex_;
  if (!Points_.empty()){
    const int nump = Points_[pyindex].size();
    // count the different failure reasons
    // and compute mean and variance of covariances
    vector<int> reason(-min_err_code + 1, 0);
    int reas;
    for (int i=0; i<nump; i++){
      reas = res_[pyindex][i];
      if (reas < 0){
        if (reas<min_err_code) {
          BIASABORT;
        }
        reason[-reas]++;
      } else {
        reason[0]++;
      }
    }


    os << setw(5) << img_num
       << setw(7) << nump
       << setw(7) << nump - num - reason[6]
       << setw(7) << reason[6];
    for (int r = 0; r <= -min_err_code; r++){
      os << setw(7) << reason[r];
    }
    os << endl << flush;

    img_num++;
  }
}


template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
WriteTrackFile(deque<vector<HomgPoint2D> >& p, deque<vector<int> >& results,
               const string& name)
{
  const int width=20;
  int i, k;
  ofstream os(name.c_str());
  if (!os){
    BIASERR("error opening "<<name);
    return -1;
  }

  os << "Feel free to place comments here.\n\n";

  os << "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n";
  os << "!!! Warning:  This is a KLT data file.  Do not modify below this line !!!\n\n";

  os << "------------------------------\n";
  os << "KLT Feature Table\n";
  os << "------------------------------\n\n";

  os << "nFrames = "<<p.size()<<", nFeatures = "<<p[0].size()<<"\n\n";

  os << "feature |          frame\n";
  os << "        |";
  for (i = 0 ; i < (int)p.size() ; i++)
    os << setw(width)<<i;
  os << "\n";
  os << "--------+";
  for (i = 0 ; i < (int)p.size() ; i++)
    for (k=0; k<width; k++) os << "-";
  os << "\n";
  if (!os){
    BIASERR("error writing "<<name);
    return -1;
  }
  os.close();


  FILE *fp=fopen(name.c_str(), "ab");
  for (k=0; k<(int)p[0].size(); k++){ // k = feature
    fprintf(fp, "%7d | ", k);
    for (i=0; i<(int)p.size(); i++){ // i = image
      //      cout << p[i][k] << endl;
      //cout << residui[i][k] << endl;
      if (!p[i][k].IsAtInfinity()){
        fprintf(fp, "(%7.3f,%7.3f)=%5d ", p[i][k][0], p[i][k][1],
               results[i][k]);
      } else {
        fprintf(fp, "(%7.3f,%7.3f)=%5d ", -1.0, -1.0, results[i][k]);
      }
    }
    fprintf(fp, "\n");
  }

  fclose(fp);
  return 0;
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
PadPoints_(const int start)
{
  const int pysize = (int)LastImage_->size();
  const int pyindex = PyIndex_;
  const int num = static_cast<int>(Points_[pyindex].size());
  for (int i=start; i<num; i++){
    for (int p=pyindex; p<pysize; p++){
      Points_[p][i][2] = 0.0;
    }
  }
}

template <class StorageType, class CalculationType>
int Tracker<StorageType, CalculationType>::
GetTrackerType()
{
  // check Tracker Type
  if (dynamic_cast<TrackerBaseSimple<CalculationType> *>(tb_))
    return TT_Simple;
  if (dynamic_cast<TrackerBaseHomography<CalculationType> *>(tb_))
    return TT_Homography;
  if (dynamic_cast<TrackerBaseAffine<CalculationType> *>(tb_))
  {
    if ((dynamic_cast<TrackerBaseAffine<CalculationType> *>(tb_))->IsSimilarityTransformOnly())
      return TT_Similar;
    else
      return TT_Affine;
  }
  if (dynamic_cast<TrackerBaseAffine2<CalculationType> *>(tb_))
  {
    if ((dynamic_cast<TrackerBaseAffine2<CalculationType> *>(tb_))->IsSimilarityTransformOnly())
      return TT_Similar;
    else
      return TT_Affine2;
  }
  if (dynamic_cast<TrackerBaseWeighted<CalculationType> *>(tb_))
    return TT_SimpleWeighted;
  if (dynamic_cast<TrackerBaseWeighted1D<CalculationType> *>(tb_))
    return TT_Weighted1D;

  return TT_Invalid;
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
SetTrackerType(int tracker_type)
{
  if (tb_) {
    delete tb_;
    tb_ = NULL;
  }
  switch(tracker_type){
  case TT_Simple:
    tb_ = new TrackerBaseSimple<CalculationType>();
    break;
  case TT_Similar:
    tb_ = new TrackerBaseAffine<CalculationType>(true);
    break;
  case TT_Affine:
    tb_ = new TrackerBaseAffine<CalculationType>(false);
    break;
  case TT_Homography:
    tb_ = new TrackerBaseHomography<CalculationType>();
    break;
  case TT_Affine2:
    tb_ = new TrackerBaseAffine2<CalculationType>(false);
    break;
  case TT_SimpleWeighted:
    tb_ = new TrackerBaseWeighted<CalculationType>();
    break;
  case TT_Weighted1D:
    tb_ = new TrackerBaseWeighted1D<CalculationType>();
    break;
  default:
    BIASERR("Unknown tracker type");
    BIASABORT;
    break;
  }
}

template <class StorageType, class CalculationType>
void Tracker<StorageType, CalculationType>::
Vector2FilterMask_(Vector<double>& input, FilterMask& mask, bool separable)
{
  if (separable){
    Vector<float> hkernel, vkernel;
    if (input.Size()<2u){
      BIASERR("invalid vector size "<<input.Size());
      BIASABORT;
    }
    const int hsize = (int)rint(input[0]), vsize = (int)rint(input[1]);
    if ((int)input.Size() != hsize + vsize + 2){
      BIASERR("invalid vector entries");
      BIASABORT;
    }
    hkernel.newsize(hsize);
    for (int i=0; i<hsize; i++){
      hkernel[i] = (float)(CalculationType)input[i+2];
    }
    vkernel.newsize(vsize);
    for (int i=0; i<vsize; i++){
      vkernel[i] = (float)(CalculationType)input[i+2+hsize];
    }
    mask.Init(hkernel, vkernel);
  } else {
    Matrix<float> kernel;
    if (input.Size()<2u){
      BIASERR("invalid vector size "<<input.Size());
      BIASABORT;
    }
    const int rows = (int)rint(input[0]), cols = (int)rint(input[1]);
    const int num = cols * rows;
    if ((int)input.Size() != num + 2){
      BIASERR("invalid vector entries");
      BIASABORT;
    }
    kernel.newsize(rows, cols);
    for (int i=0; i<num; i++){
      kernel.GetData()[i] = (float)(CalculationType)input[i+2];
    }
    mask.Init(kernel);
  }
}

//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////
namespace BIAS{
template class Tracker<unsigned char, float>;
template class Tracker<float, float>;

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class Tracker<unsigned char, int>;
#endif
#ifdef BUILD_IMAGE_CHAR
//template class Tracker<unsigned char, char>;
#endif
#ifdef BUILD_IMAGE_SHORT
#endif
#ifdef BUILD_IMAGE_USHORT
template class Tracker<unsigned short, float>;
#ifdef BUILD_IMAGE_INT
template class Tracker<unsigned short, int>;
#endif
#endif
#ifdef BUILD_IMAGE_UINT
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif
}
