#ifndef __LevenbergMarquardtHelper_hh__
#define __LevenbergMarquardtHelper_hh__

#include <bias_config.h>

namespace BIAS
{
  /**
     @class LevenbergMarquardtHelper
     @brief Simple helper class for Levenberg-Marquardt optimization.

     This class keep track of the best solution computed by a Levenberg-
     Marquardt optimization.
  */
  class BIASMathAlgo_EXPORT LevenbergMarquardtHelper
  {
  public:

    /** @brief Create empty instance.
        @note Call Init() to provide the number of variables and residuals! */
    LevenbergMarquardtHelper();

    /** @brief Create new instance for given number of variables and
               objective function residuals. */
    LevenbergMarquardtHelper(int numVars, int numResiduals);

    /** @brief Releases the internally stored best variables. */
    ~LevenbergMarquardtHelper();

    /** @brief Relocates internal array of best variables and resets the
               best solution. */
    void Init(int numVars, int numResiduals);

    /** @brief Add variables and results, and computes the sum of squares.
        @return Returns if the new solution is better than the one stored. */
    bool Update(const double *x, const double *fvec);

    /** @brief Returns pointer to internal array of best variables.
         @note Returned pointer must NOT be released! */
    inline double* GetBestVars() const { return vars_; }

  protected:

    double *vars_;
    double sumOfSquares_;
    int numVars_, numResiduals_;

  };

}

#endif // __LevenbergMarquardtHelper_hh__
