/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __PCA_hh__
#define __PCA_hh__

#include <bias_config.h>

#include <Base/Debug/Debug.hh>
#include <vector>
#include <Base/Math/Vector.hh>
#include <Base/Math/Matrix.hh>


#define PCAType float

namespace BIAS { 

  /** @class PCA
      @brief principal component analysis on a set of vectors
      with PCA it is possible to find the most important dimensions
      of a set of vectors. the dimension can be reduced while the
      vectors stay separable
      @author haertel 01/06
  */
  class BIASMathAlgo_EXPORT PCA {
  
    public:
    PCA():reductionSize_(0) {}

    /** @brief Computes a reduction-matrix
    	
    	the first dimension of the matrix is taken out of the first vector
        @param vec the input: a vector of data-vectors
        @param matrix the matrix returned
        @param reductionSize the second dimension of the matrix
    */
    void ComputeReductionMatrix(const std::vector<BIAS::Vector<PCAType> >&vec, 
                                BIAS::Matrix<PCAType> &matrix, 
                                int reductionSize=-1);
    

    /** @brief Computes a reduction-matrix
        this method uses all float-vectors for analyzing
    	
        @param vec the input: a vector of vectors
        @param matrix the matrix returned
        @param reductionSize the second dimension of the matrix
    */

    void ComputeReductionMatrix(const std::vector<std::vector<
                                BIAS::Vector<PCAType> > > & vec, 
                                BIAS::Matrix<PCAType> &matrix, 
                                int reductionSize=-1);


    /** @brief uses scatter matrix cov to compute reduction matrix
        @param normalize set to true when you want only the most important
        directions, set to false if you want to compare distances in reduced
        space without mahalanobis distance */
    void ComputeReductionMatrix(BIAS::Matrix<PCAType> &cov,
                                BIAS::Matrix<PCAType> &matrix, 
                                bool normalize = false,
                                int reductionSize=-1);



    ///computes mean of a set of vectors
    void ComputeMean(const std::vector<BIAS::Vector<PCAType> > &vec, 
                     BIAS::Vector<PCAType> &mean);

    /// compute unnomalized covariance
    void ComputeScatter(const std::vector<BIAS::Vector<PCAType> > &vec,
                        const BIAS::Vector<PCAType> &mean,
                        BIAS::Matrix<PCAType> &cov);
    
    ///get mean of a vector
    void GetMean(BIAS::Vector<PCAType> &mean);
    
    ///get eigenvalues of data (call after ComputeReductionMatrix)
    void GetVariances(BIAS::Vector<double> &S){
      S=S_;
    };
    
    protected:
    
    void SetReductionSize(int size);
    
    int reductionSize_;
    BIAS::Vector<PCAType> mean_;
    BIAS::Vector<double> S_;
    
  };
  
}

#endif
