/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __VECTOR_STATISTICS_HH__
#define __VECTOR_STATISTICS_HH__

// std
#include <vector>
#include <math.h>
#include <algorithm>

// BIAS
#include <Base/Debug/Error.hh>

namespace BIAS
{
  /**
   * @class VectorStatistics
   * @brief Provides methods to compute statistical values from vector<double>.
   *
   * @ingroup g_mathalgo
   * @author rwulff
   * @date   07/2013
   */
  class VectorStatistics
  {
  public:

    /**
     * @brief Computes the sum of the values and returns it.
     */
    static inline
    void ComputeSum(const std::vector<double>& values, double& sum)
    {
      sum = 0.0;

      for (std::size_t cnt = 0; cnt < values.size(); ++cnt)
      {
        sum += values[cnt];
      }
    }

    /**
     * @brief Computes the sum and the mean of the values and returns them.
     */
    static inline
    void ComputeSumAndMean(const std::vector<double>& values,
                           double& sum, double& mean)
    {
      BIASASSERT(!values.empty());

      ComputeSum(values, sum);
      mean = sum / (double)values.size();
    }

    /**
     * @brief Convenience method. Computes the mean of the values and returns it.
     */
    static inline
    void ComputeMean(const std::vector<double>& values,
                     double& mean)
    {
      double sum; // ignored
      ComputeSumAndMean(values, sum, mean);
    }

    /**
     * @brief Computes the sum, the mean and the variance of the values and
     *        returns them.
     */
    static inline
    void ComputeSumMeanAndVariance(const std::vector<double>& values,
                                   double& sum, double& mean, double& variance)
    {
      ComputeSumAndMean(values, sum, mean);

      variance = 0.0;

      for (std::size_t cnt = 0; cnt < values.size(); ++cnt) {
        const double diff = values[cnt] - mean;
        variance += diff * diff;
      }

      variance /= (double)values.size();
    }

    /**
     * @brief Convenience method. Computes the variance of the values and
     *        returns it.
     */
    static inline
    void ComputeVariance(const std::vector<double>& values,
                         double& variance)
    {
      double mean, sum; // ignored
      ComputeSumMeanAndVariance(values, sum, mean, variance);
    }

    /**
     * @brief Computes the sum, the mean, the variance and the standard
     *        deviation of the values and returns them.
     */
    static inline
    void ComputeSumMeanVarianceAndStdDev(const std::vector<double>& values,
                                         double& sum, double& mean,
                                         double& variance, double& stdDev)
    {
      ComputeSumMeanAndVariance(values, sum, mean, variance);

      stdDev = sqrt(variance);
    }

    /**
     * @brief Convenience method. Computes the standard deviation of the values
     *        and returns it.
     */
    static inline
    void ComputeStdDev(const std::vector<double>& values,
                       double& stdDev)
    {
      double sum, mean, variance; // ignored
      ComputeSumMeanVarianceAndStdDev(values, sum, mean, variance, stdDev);
    }

    /**
     * @brief Computes median, min and max of the given values and returns them.
     */
    static inline
    void ComputeMedianMinAndMax(const std::vector<double>& values,
                                double& median, double& min, double& max)
    {
      BIASASSERT(!values.empty());

      std::vector<double> sortedValues(values);
      std::sort(sortedValues.begin(), sortedValues.end());

      median = sortedValues[sortedValues.size() / 2];
      min    = sortedValues[0];
      max    = sortedValues[sortedValues.size() - 1];
    }

  private:

    /**
     * @brief Hidden constructor.
     */
    VectorStatistics();

  };
}

#endif // __VECTOR_STATISTICS_HH__
