__constant sampler_t smp = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_NEAREST;

__kernel void RecursiveGaussianForwardY(__read_only image2d_t src, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int xcoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;
  
  float4 xp = read_imagef(src, smp, (int2)(xcoo,0));
  float4 yb = xp * coefp; 
  float4 yp = yb;

  for (int ycoo = 0; ycoo < h; ycoo++) {
    float4 xc = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
	write_imagef(dst, (int2)(xcoo*XM,ycoo), yc);
    xp = xc;
    yb = yp;
    yp = yc;
  }
}

__kernel void RecursiveGaussianBackwardY(__read_only image2d_t src, __read_only image2d_t tmp, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int xcoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;

  float4 xn = read_imagef(src, smp, (int2)(xcoo,h-1));
  float4 xa = xn;
  float4 yn = xn * (float)coefn;
  float4 ya = yn;

  for (int ycoo = h - 1; ycoo >= 0; ycoo--) {
    float4 xc = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    float4 tv = read_imagef(tmp, smp, (int2)(xcoo,ycoo));
	write_imagef(dst, (int2)(xcoo*XM,ycoo), tv+yc);
  }
}

__kernel void RecursiveGaussianForwardX(__read_only image2d_t src, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int ycoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;

  float4 xp = read_imagef(src, smp, (int2)(0,ycoo));
  float4 yb = xp * coefp; 
  float4 yp = yb;

  float4 xc;
  float4 yc;
  for (int xcoo = 0; xcoo < w; xcoo++) {
    xc = read_imagef(src, smp, (int2)(xcoo,ycoo));
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
	write_imagef(dst, (int2)(xcoo*XM,ycoo), yc);
    xp = xc;
    yb = yp;
    yp = yc;
  }
}

__kernel void RecursiveGaussianBackwardX(__read_only image2d_t src, __read_only image2d_t tmp, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int ycoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;

  float4 xn = read_imagef(src, smp, (int2)(w-1,ycoo));
  float4 xa = xn;
  float4 yn = xn * (float)coefn;
  float4 ya = yn;

  float4 xc;
  float4 yc;
  float4 tv;
  for (int xcoo = w - 1; xcoo >= 0; xcoo--) {
    xc = read_imagef(src, smp, (int2)(xcoo,ycoo));
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    tv = read_imagef(tmp, smp, (int2)(xcoo,ycoo));
	write_imagef(dst, (int2)(xcoo*XM,ycoo), (float4)(tv+yc));
  }
}


__kernel void RowPackGreyToRGBA(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0);
  int cooY = get_global_id(1)*4;
  
  float4 color;
  color.x = read_imagef(src, smp, (int2)(cooX, cooY+0)).x; 
  color.y = read_imagef(src, smp, (int2)(cooX, cooY+1)).x; 
  color.z = read_imagef(src, smp, (int2)(cooX, cooY+2)).x; 
  color.w = read_imagef(src, smp, (int2)(cooX, cooY+3)).x;
  
  cooY /= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX, cooY), color);
}


__kernel void ColPackGreyToRGBA(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0)*4;
  int cooY = get_global_id(1);
  
  float4 color;
  color.x = read_imagef(src, smp, (int2)(cooX+0, cooY)).x; 
  color.y = read_imagef(src, smp, (int2)(cooX+1, cooY)).x; 
  color.z = read_imagef(src, smp, (int2)(cooX+2, cooY)).x; 
  color.w = read_imagef(src, smp, (int2)(cooX+3, cooY)).x;
  
  cooX /= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX, cooY), color);
}

__kernel void RowUnPackRGBAToGrey(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0);
  int cooY = get_global_id(1);
  
  float4 color = read_imagef(src, smp, (int2)(cooX, cooY));
  
  cooY *= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX, cooY+0), (float4)(color.x));
  write_imagef(dst, (int2)(cooX, cooY+1), (float4)(color.y));
  write_imagef(dst, (int2)(cooX, cooY+2), (float4)(color.z));
  write_imagef(dst, (int2)(cooX, cooY+3), (float4)(color.w));
}

__kernel void ColUnPackRGBAToGrey(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0);
  int cooY = get_global_id(1);
  
  float4 color = read_imagef(src, smp, (int2)(cooX, cooY));
  
  cooX *= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX+0*XM, cooY), (float4)(color.x));
  write_imagef(dst, (int2)(cooX+1*XM, cooY), (float4)(color.y));
  write_imagef(dst, (int2)(cooX+2*XM, cooY), (float4)(color.z));
  write_imagef(dst, (int2)(cooX+3*XM, cooY), (float4)(color.w));
}

__kernel void ColPackRGBAToRowPackRGBA(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0);
  int cooY = get_global_id(1);
  
  float4 color[4];
  
  color[0] = read_imagef(src, smp, (int2)(cooX, cooY*4+0));
  color[1] = read_imagef(src, smp, (int2)(cooX, cooY*4+1));
  color[2] = read_imagef(src, smp, (int2)(cooX, cooY*4+2));
  color[3] = read_imagef(src, smp, (int2)(cooX, cooY*4+3));
  
  cooX *= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX+0*XM, cooY), (float4)(color[0].x, color[1].x, color[2].x, color[3].x));
  write_imagef(dst, (int2)(cooX+1*XM, cooY), (float4)(color[0].y, color[1].y, color[2].y, color[3].y));
  write_imagef(dst, (int2)(cooX+2*XM, cooY), (float4)(color[0].z, color[1].z, color[2].z, color[3].z));
  write_imagef(dst, (int2)(cooX+3*XM, cooY), (float4)(color[0].w, color[1].w, color[2].w, color[3].w));
}

__kernel void RowPackRGBAToColPackRGBA(__read_only image2d_t src, __write_only image2d_t dst)
{
  int cooX = get_global_id(0);
  int cooY = get_global_id(1);
  
  float4 color[4];
  
  color[0] = read_imagef(src, smp, (int2)(cooX*4+0, cooY));
  color[1] = read_imagef(src, smp, (int2)(cooX*4+1, cooY));
  color[2] = read_imagef(src, smp, (int2)(cooX*4+2, cooY));
  color[3] = read_imagef(src, smp, (int2)(cooX*4+3, cooY));
  
  cooY *= 4;
  cooX *= XM;
  write_imagef(dst, (int2)(cooX, cooY+0), (float4)(color[0].x, color[1].x, color[2].x, color[3].x));
  write_imagef(dst, (int2)(cooX, cooY+1), (float4)(color[0].y, color[1].y, color[2].y, color[3].y));
  write_imagef(dst, (int2)(cooX, cooY+2), (float4)(color[0].z, color[1].z, color[2].z, color[3].z));
  write_imagef(dst, (int2)(cooX, cooY+3), (float4)(color[0].w, color[1].w, color[2].w, color[3].w));
}

__kernel void RecursiveGaussianForwardYPacked(__read_only image2d_t src, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int xcoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;
  
  float4 color = read_imagef(src, smp, (int2)(xcoo,0));
  float xp = color.x;
  float yb = xp * coefp; 
  float yp = yb;

  for (int ycoo = 0; ycoo < h/4; ycoo++) {
    float4 c = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 result;
    float xc = c.x;
    float yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.x = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.y;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.y = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.z;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
	result.z = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.w;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.w = yc;
	write_imagef(dst, (int2)(xcoo*XM,ycoo), result);
    xp = xc;
    yb = yp;
    yp = yc;
  }
}

__kernel void RecursiveGaussianBackwardYPacked(__read_only image2d_t src, __read_only image2d_t tmp, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int xcoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;

  float4 color = read_imagef(src, smp, (int2)(xcoo,(h/4)-1));
  float xn = color.w;
  float xa = xn;
  float yn = xn * (float)coefn;
  float ya = yn;

  for (int ycoo = (h/4) - 1; ycoo >= 0; ycoo--) {
    float4 c = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 tc = read_imagef(tmp, smp, (int2)(xcoo,ycoo));
    float4 result;
    float xc = c.w;
    float yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    float tv = tc.w;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.w = tv+yc;
    xc = c.z;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.z;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.z = tv+yc;
    xc = c.y;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.y;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.y = tv+yc;
    xc = c.x;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.x;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.x = tv+yc;
    write_imagef(dst, (int2)(xcoo*XM,ycoo), result);
  }
}

__kernel void RecursiveGaussianForwardXPacked(__read_only image2d_t src, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int ycoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;
  
  float4 color = read_imagef(src, smp, (int2)(0,ycoo));
  float xp = color.x;
  float yb = xp * coefp; 
  float yp = yb;

  for (int xcoo = 0; xcoo < w/4; xcoo++) {
    float4 c = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 result;
    float xc = c.x;
    float yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.x = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.y;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.y = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.z;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
	result.z = yc;
    xp = xc;
    yb = yp;
    yp = yc;
    xc = c.w;
    yc = (xc * a0) + (xp * a1) - (yp * b1) - (yb * b2);
    result.w = yc;
	write_imagef(dst, (int2)(xcoo*XM,ycoo), result);
    xp = xc;
    yb = yp;
    yp = yc;
  }
}

__kernel void RecursiveGaussianBackwardXPacked(__read_only image2d_t src, __read_only image2d_t tmp, __write_only image2d_t dst, 
                                    float a0, float a1, 
                                    float a2, float a3, 
                                    float b1, float b2, 
                                    float coefp, float coefn)
{
  unsigned int ycoo = get_global_id(0);
  int2 imagedim = get_image_dim(dst);
  int w = imagedim.x;
  int h = imagedim.y;

  float4 color = read_imagef(src, smp, (int2)((w/4)-1,ycoo));
  float xn = color.w;
  float xa = xn;
  float yn = xn * (float)coefn;
  float ya = yn;

  for (int xcoo = (w/4) - 1; xcoo >= 0; xcoo--) {
    float4 c = read_imagef(src, smp, (int2)(xcoo,ycoo));
    float4 tc = read_imagef(tmp, smp, (int2)(xcoo,ycoo));
    float4 result;
    float xc = c.w;
    float yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    float tv = tc.w;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.w = tv+yc;
    xc = c.z;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.z;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.z = tv+yc;
    xc = c.y;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.y;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.y = tv+yc;
    xc = c.x;
    yc = (xn * a2) + (xa * a3) - (yn * b1) - (ya * b2);
    tv = tc.x;
    xa = xn;
    xn = xc;
    ya = yn;
    yn = yc;
    result.x = tv+yc;
    write_imagef(dst, (int2)(xcoo*XM,ycoo), result);
  }
}

