#define BLOCK_SIZE 16

__constant sampler_t smp = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_NEAREST;

__kernel void Warp(__read_only image2d_t im1,
					__read_only image2d_t im2,
					__read_only image2d_t uvw,
                    __write_only image2d_t warped
                       )
{
   const sampler_t fsmp = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP_TO_EDGE | CLK_FILTER_LINEAR;
         
   int i = get_global_id(0);
   int j = get_global_id(1);

   int2 coords = (int2)(i,j);
   float2 bilcoo = (float2)(i,j)+(float2)(read_imagef(uvw, smp, coords).x + 0.5, read_imagef(uvw, smp, coords).y + 0.5);
   float4 warpedPixel = read_imagef(im2,fsmp, bilcoo);
   warpedPixel.z = warpedPixel.z - read_imagef(im1, smp, coords).x;
   
   write_imagef(warped, (int2)(i*XM,j), warpedPixel);
}

__kernel void UvwGrad(__read_only image2d_t uvwold,
				      __write_only image2d_t uvwX,
				      __write_only image2d_t uvwY)
{
    __local float4 P[BLOCK_SIZE+1][BLOCK_SIZE+1];
    
    //Identification of this workgroup
   int i = get_group_id(0);
   int j = get_group_id(1);

   //Identification of work-item
   int idX = get_local_id(0);
   int idY = get_local_id(1); 

   int ii = i*BLOCK_SIZE + idX; // == get_global_id(0);
   int jj = j*BLOCK_SIZE + idY; // == get_global_id(1);
   
   int2 coords = (int2)(ii, jj);
   P[idX][idY] = read_imagef(uvwold, smp, coords);
   if (idX+1 == BLOCK_SIZE) {
     P[idX+1][idY] = read_imagef(uvwold, smp, (int2)(ii+1, jj));
   }
   if (idY+1 == BLOCK_SIZE) {
     P[idX][idY+1] = read_imagef(uvwold, smp, (int2)(ii, jj+1));
   }
   if (idX+1 == BLOCK_SIZE && idY+1 == BLOCK_SIZE) {
     P[idX+1][idY+1] = read_imagef(uvwold, smp, (int2)(ii+1, jj+1));
   }
   barrier(CLK_LOCAL_MEM_FENCE);

   float4 valX = P[idX+1][idY] - P[idX][idY];   
   float4 valY = P[idX][idY+1] - P[idX][idY];   
   
   coords = (int2)(ii*XM, jj);
   write_imagef(uvwX, coords, valX);
   write_imagef(uvwY, coords, valY);
}

__kernel void UpdateP(
 					__read_only image2d_t pold1,
					__read_only image2d_t pold2,
					__read_only image2d_t uvwX,
					__read_only image2d_t uvwY,
                    __write_only image2d_t p1,
                    __write_only image2d_t p2,
                     float sigma
                       )
{
   int i = get_global_id(0);
   int j = get_global_id(1);

   int2 coords = (int2)(i,j);
   float4 xval = sigma * read_imagef(uvwX, smp, coords);
   float4 yval = sigma * read_imagef(uvwY, smp, coords);
   
   float4 pval1 = read_imagef(pold1, smp, coords);
   float4 pval2 = read_imagef(pold2, smp, coords);
   pval1 += (float4)( xval.x, yval.x, xval.y, yval.y);
   pval2 += (float4)( xval.z, yval.z, 0.0, 0.0 );

   float nval1 = max (1.0f, fast_length(pval1) );   
   float nval2 = max (1.0f, fast_length(pval2) );   

   write_imagef(p1, (int2)(i*XM,j), pval1 / nval1 );
   write_imagef(p2, (int2)(i*XM,j), pval2 / nval2 );
}

__kernel void UpdateDiv(__read_only image2d_t p1,
					__read_only image2d_t p2,
                    __write_only image2d_t divimg)
{
    __local float4 P[BLOCK_SIZE+1][BLOCK_SIZE+1];
    __local float4 Q[BLOCK_SIZE+1][BLOCK_SIZE+1];
    
    //Identification of this workgroup
   int i = get_group_id(0);
   int j = get_group_id(1);

   //Identification of work-item
   int idX = get_local_id(0);
   int idY = get_local_id(1); 

   int ii = i*BLOCK_SIZE + idX; // == get_global_id(0);
   int jj = j*BLOCK_SIZE + idY; // == get_global_id(1);
   
   int2 coords = (int2)(ii, jj);
   P[idX+1][idY+1] = read_imagef(p1, smp, coords);
   if (idX == 0) {
     P[idX][idY+1] = read_imagef(p1, smp, (int2)(ii-1, jj));
   }
   if (idY == 0) {
     P[idX+1][idY] = read_imagef(p1, smp, (int2)(ii, jj-1));
   }
   Q[idX+1][idY+1] = read_imagef(p2, smp, coords);
   if (idX == 0) {
     Q[idX][idY+1] = read_imagef(p2, smp, (int2)(ii-1, jj));
   }
   if (idY == 0) {
     Q[idX+1][idY] = read_imagef(p2, smp, (int2)(ii, jj-1));
   }
   barrier(CLK_LOCAL_MEM_FENCE);

   float4 result;
   result.x = (P[idX+1][idY+1].x - P[idX][idY+1].x) + (P[idX+1][idY+1].y - P[idX+1][idY].y);
   result.y = (P[idX+1][idY+1].z - P[idX][idY+1].z) + (P[idX+1][idY+1].w - P[idX+1][idY].w);
   result.z = (Q[idX+1][idY+1].x - Q[idX][idY+1].x) + (Q[idX+1][idY+1].y - Q[idX+1][idY].y);
   result.w = 0.0;
   write_imagef(divimg, (int2)(ii*XM,jj), result );
   
}

__kernel void UpdateUVW(__read_only image2d_t uvwold,
					__read_only image2d_t div,
					__read_only image2d_t uvw0,
					__read_only image2d_t Ixyg,
					__read_only image2d_t warped,
					__write_only image2d_t uvw,
                    float tau, float lambda, float gamma
                       )
{
   int i = get_global_id(0);
   int j = get_global_id(1);

   int2 coords = (int2)(i,j);
   float4 oldval = read_imagef(uvwold, smp, coords);
   float4 divval = read_imagef(div, smp, coords);
   float4 uvw0val = read_imagef(uvw0, smp, coords);
   float4 Ixygval = read_imagef(Ixyg, smp, coords);
   float4 warpedval = read_imagef(warped, smp, coords);
   
   float4 tauval = oldval + tau * divval;
   float4 result;
   float rho = warpedval.z + (tauval.x-uvw0val.x) * Ixygval.x + (tauval.y-uvw0val.y) * Ixygval.y + gamma * tauval.z;
   float compval = tau*lambda*Ixygval.z;

   if (rho < - compval) {
     result.x = tauval.x + tau*lambda*Ixygval.x;
     result.y = tauval.y + tau*lambda*Ixygval.y;
     result.z = tauval.z + tau*lambda*gamma;
   } else if (rho > compval) {
     result.x = tauval.x - tau*lambda*Ixygval.x;
     result.y = tauval.y - tau*lambda*Ixygval.y;
     result.z = tauval.z - tau*lambda*gamma;
   } else if (fabs(rho) <= compval) {
     result.x = tauval.x - rho*Ixygval.x / Ixygval.z;
     result.y = tauval.y - rho*Ixygval.y / Ixygval.z;
     result.z = tauval.z - rho*gamma / Ixygval.z;
   }

   write_imagef(uvw, (int2)(i*XM,j), result );
}

__kernel void UpdateUVWold(__read_only image2d_t uvwoldsrc,
					__read_only image2d_t uvw,
					__write_only image2d_t uvwolddst
                       )
{
   int i = get_global_id(0);
   int j = get_global_id(1);

   int2 coords = (int2)(i,j);
   float4 oldval = read_imagef(uvwoldsrc, smp, coords);
   float4 uvwval = read_imagef(uvw, smp, coords);

   float4 result = 2.0*uvwval - oldval;

   write_imagef(uvwolddst, (int2)(i*XM,j), result );
}

__constant float weightedmask[5] = 
    { 1.0f / 12.0f, -8.0f / 12.0f, 0.0, 8.0f / 12.0f, -1.0f / 12.0f }; 


__kernel void WeightedGradX(__read_only image2d_t src,
                       __write_only image2d_t dst)
{
   //int w = 5;
   int wBy2 = 2;

    __local float4 P[BLOCK_SIZE+4][BLOCK_SIZE];
    
    //Identification of this workgroup
   int i = get_group_id(0);
   int j = get_group_id(1);

   //Identification of work-item
   int idX = get_local_id(0);
   int idY = get_local_id(1); 

   int ii = i*BLOCK_SIZE + idX; // == get_global_id(0);
   int jj = j*BLOCK_SIZE + idY; // == get_global_id(1);
   
   int2 coords = (int2)(ii, jj);

   //Reads pixels
   P[idX+wBy2][idY] = read_imagef(src, smp, coords);

   //Needs to read extra elements for the filter in the borders
   if (idX < wBy2)
   {
      coords.x = ii - wBy2;
      P[idX][idY] = read_imagef(src, smp, coords);
      coords.x = ii + BLOCK_SIZE;
      P[idX+BLOCK_SIZE+wBy2][idY] = read_imagef(src, smp, coords);
   }

   barrier(CLK_LOCAL_MEM_FENCE);
   
   //Computes convolution
   float4 convPix = 
   		P[idX+0][idY] * weightedmask[0]
      + P[idX+1][idY] * weightedmask[1]
      + P[idX+2][idY] * weightedmask[2]
      + P[idX+3][idY] * weightedmask[3]
      + P[idX+4][idY] * weightedmask[4];

   coords = (int2)(XM*(ii), jj);
   write_imagef(dst, coords, convPix);
}

__kernel void WeightedGradY(__read_only image2d_t src,
                       __read_only image2d_t GradX,
                       __write_only image2d_t dst,
                       float gamma)
{
   //int w = 5;
   int wBy2 = 2; //w divided by 2

    __local float4 P[BLOCK_SIZE+4][BLOCK_SIZE];
    
    //Identification of this workgroup
   int i = get_group_id(0);
   int j = get_group_id(1);

   //Identification of work-item
   int idX = get_local_id(0);
   int idY = get_local_id(1); 

   int ii = i*BLOCK_SIZE + idX; // == get_global_id(0);
   int jj = j*BLOCK_SIZE + idY; // == get_global_id(1);
   
   int2 coords = (int2)(ii, jj);

   //Reads pixels
   P[idY+wBy2][idX] = read_imagef(src, smp, coords);

   //Needs to read extra elements for the filter in the borders
   if (idY < wBy2)
   {
      coords.y = jj - wBy2;
      P[idY][idX] = read_imagef(src, smp, coords);
      coords.y = jj + BLOCK_SIZE;
      P[idY+BLOCK_SIZE+wBy2][idX] = read_imagef(src, smp, coords);
   }
   barrier(CLK_LOCAL_MEM_FENCE);
   
   //Computes convolution
   float4 convPix = 
        P[idY+0][idX] * weightedmask[0]
      + P[idY+1][idX] * weightedmask[1]
      + P[idY+2][idX] * weightedmask[2]
      + P[idY+3][idX] * weightedmask[3]
      + P[idY+4][idX] * weightedmask[4];
   
   convPix.x = read_imagef(GradX, smp, (int2)(ii, jj)).x;
   
   convPix.z = convPix.x*convPix.x+convPix.y*convPix.y+gamma*gamma;
   
   coords = (int2)(XM*(ii), jj);
   write_imagef(dst, coords, convPix);
}

__kernel void CreateColorCoded(__read_only image2d_t src,
					__write_only image2d_t dst
                       )
{
   int i = get_global_id(0);
   int j = get_global_id(1);

   int2 coords = (int2)(i,j);
   float4 uvw = read_imagef(src, smp, coords);
   uvw.z=0;
   uvw.w=0;
   float4 result;
   result.x = atan2pi(uvw.y, uvw.x)*180.0 + 180.0;
   result.y = min(1.0f, length(uvw)/10);
   result.z = min(1.0f, length(uvw)/10);
   result.w = 255;
   
   int d;
   float r,g,b;
   	float p1, p2, p3, f;
	float xh;
	xh = result.x / 60.0;                  /* convert hue to be in 0,6       */
	d = floor(xh);    /* i = greatest integer <= h    */
	f = xh - d;                      /* f = fractional part of h     */
	p1 = result.z * (1 - result.y);
	p2 = result.z * (1 - (result.y * f));
	p3 = result.z * (1 - (result.y * (1 - f)));
	switch ((int) d)
	{
	case 0:
		r = result.z; g = p3;  b = p1; break;
	case 1:
		r = p2; g = result.z;  b = p1; break;
	case 2:
		r = p1; g = result.z;  b = p3; break;
	case 3:
		r = p1; g = p2; b = result.z;  break;
	case 4:
		r = p3; g = p1; b = result.z;  break;
	case 5:
		r = result.z; g = p1; b = p2; break;
    default:
    r=0;g=0;b=0;
	}
	
	result = (float4)(r*255.0,g*255.0,b*255.0,result.w);
    
   write_imagef(dst, (int2)(i*XM,j), result );
}

