/*
 * clfRecursiveGauss.cpp
 *
 *  Created on: Oct 28, 2011
 *      Author: fkellner
 */

#include <OpenCLFramework/Algorithm/clfRecursiveGauss.hh>
#include <Base/Common/BIASpragma.hh>
#include <Base/Image/ImageIO.hh>

using namespace BIAS;
using namespace BIAS;

clfRecursiveGauss::clfRecursiveGauss(float sigma, clfContext *ctx, bool sharedGL, unsigned int device) :
  clfAlgorithm(ctx,sharedGL,device) {
  intermediateImage_ = NULL;
  oneDirFilteredImage_ = NULL;
  packedGreyImage_ = NULL;

  // add code to our program
  programCL_->AddSource("/OpenCLFramework/Algorithm/cl/recursivegauss.cl");
  programCL_->Build();

  kernelNames_.clear();
  kernelNames_.push_back("RecursiveGaussianForwardX");
  kernelNames_.push_back("RecursiveGaussianForwardY");
  kernelNames_.push_back("RecursiveGaussianForwardXPacked");
  kernelNames_.push_back("RecursiveGaussianForwardYPacked");
  kernelNames_.push_back("RecursiveGaussianBackwardX");
  kernelNames_.push_back("RecursiveGaussianBackwardY");
  kernelNames_.push_back("RecursiveGaussianBackwardXPacked");
  kernelNames_.push_back("RecursiveGaussianBackwardYPacked");

  programCL_->AddKernels(kernelNames_);
  SetSigma(sigma);

  programCL_->AddKernel("RowPackGreyToRGBA");
  programCL_->AddKernel("ColPackGreyToRGBA");
  programCL_->AddKernel("RowUnPackRGBAToGrey");
  programCL_->AddKernel("ColUnPackRGBAToGrey");
  programCL_->AddKernel("ColPackRGBAToRowPackRGBA");
  programCL_->AddKernel("RowPackRGBAToColPackRGBA");
}

clfRecursiveGauss::~clfRecursiveGauss() {
  Reset();
}

void clfRecursiveGauss::Reset() {
  delete intermediateImage_;
  intermediateImage_ = NULL;
  delete oneDirFilteredImage_;
  oneDirFilteredImage_ = NULL;
  delete packedGreyImage_;
  packedGreyImage_ = NULL;
}


int clfRecursiveGauss::Filter(clfImage2D *src, clfImage2D *&dst) {

  unsigned int w,h;
  src->GetImageDim(w,h);

  if (w % 16 == 0 && h % 16 == 0) {
    FilterX(src, oneDirFilteredImage_);
    FilterY(oneDirFilteredImage_, dst);
  } else {
    BIASERR("recursive gauss needs images divisible by 16");
    BIASABORT;
  }
  return 0;
}

int clfRecursiveGauss::FilterX(clfImage2D *src, clfImage2D *&dst) {
  programCL_->KernelSetArgument("RecursiveGaussianForwardX", 0, *src);
  programCL_->KernelSetArgument("RecursiveGaussianBackwardX", 0, *src);

  unsigned int w,h;
  src->GetImageDim(w,h);

  if (dst == NULL) {
    dst = context_->CreateImage2D();
    dst->AllocateFromTemplate(*src);
  }
  programCL_->KernelSetArgument("RecursiveGaussianBackwardX", 2, *dst);

  if (intermediateImage_ == NULL) {
    intermediateImage_ = context_->CreateImage2D();
    intermediateImage_->AllocateFromTemplate(*src);
  }
  programCL_->KernelSetArgument("RecursiveGaussianForwardX", 1, *intermediateImage_);
  programCL_->KernelSetArgument("RecursiveGaussianBackwardX", 1, *intermediateImage_);

  context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardX", h, 16);
  context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardX", h, 16);
  return 0;
}

int clfRecursiveGauss::FilterY(clfImage2D *src, clfImage2D *&dst) {

  programCL_->KernelSetArgument("RecursiveGaussianForwardY", 0, *src);
  programCL_->KernelSetArgument("RecursiveGaussianBackwardY", 0, *src);

  unsigned int w,h;
  src->GetImageDim(w,h);

  if (dst == NULL) {
    dst = context_->CreateImage2D();
    dst->AllocateFromTemplate(*src);
  }
  programCL_->KernelSetArgument("RecursiveGaussianBackwardY", 2, *dst);

  if (intermediateImage_ == NULL) {
    intermediateImage_ = context_->CreateImage2D();
    intermediateImage_->AllocateFromTemplate(*src);
  }
  programCL_->KernelSetArgument("RecursiveGaussianForwardY", 1, *intermediateImage_);
  programCL_->KernelSetArgument("RecursiveGaussianBackwardY", 1, *intermediateImage_);

  context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardY", w, 16);
  context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardY", w, 16);
  return 0;
}



int clfRecursiveGauss::FilterGrey(clfImage2D *src, clfImage2D *&dst) {

  { // Allocate temporary images
    if (packedGreyImage_ == NULL) {
      packedGreyImage_ = context_->CreateImage2D();
      packedGreyImage_->AllocateFromTemplate(*src);
    }
    if (intermediateImage_ == NULL) {
      intermediateImage_ = context_->CreateImage2D();
      intermediateImage_->AllocateFromTemplate(*src);
    }
    if (oneDirFilteredImage_ == NULL) {
      oneDirFilteredImage_ = context_->CreateImage2D();
      oneDirFilteredImage_->AllocateFromTemplate(*src);
    }
    if (dst == NULL) {
      dst = context_->CreateImage2D();
      dst->AllocateFromTemplate(*src);
    } else {
      if (dst->IsReadOnly()) {
        THROW_CL_EXCEPTION(cl::Error(-100, "Need to use dst as tempory read/write"));
      }
    }
  }
  unsigned int w,h;
  { // make sure everything is divisible by 16*4
    src->GetImageDim(w,h);
    if (h % 64 != 0) {
      h = h+(64-(h%64));
    }
    if (w % 64 != 0) {
      w = w+(64-(w%64));
    }
  }
  { // pack grey image from pixels R000R000R000 to RGBARGBARGBA
    // rgbargbargbargba     rgba
    // R000R000R000R000     RRRR
    // R000R000R000R000  -> RRRR
    // R000R000R000R000     RRRR
    programCL_->KernelSetArgument("ColPackGreyToRGBA", 0, *src);
    programCL_->KernelSetArgument("ColPackGreyToRGBA", 1, *packedGreyImage_);
    context_->RunOn2DRange(*programCL_, "ColPackGreyToRGBA", w/4, h, 16, 16);
  }

  { // filter packed image, needing only w/4 read/write operations per row
    programCL_->KernelSetArgument("RecursiveGaussianForwardXPacked", 0, *packedGreyImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 0, *packedGreyImage_);

    programCL_->KernelSetArgument("RecursiveGaussianForwardXPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 2, *dst);

    context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardXPacked", h, 16);
    context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardXPacked", h, 16);
  }

  { // "repack" image such that the grey image is now stored (RGBA)^T
    // rgba     rgba
    // RGBA     RRRR
    // RGBA  -> GGGG
    // RGBA     BBBB
    // RGBA     AAAA
    programCL_->KernelSetArgument("ColPackRGBAToRowPackRGBA", 0, *dst);
    programCL_->KernelSetArgument("ColPackRGBAToRowPackRGBA", 1, *oneDirFilteredImage_);
    context_->RunOn2DRange(*programCL_, "ColPackRGBAToRowPackRGBA", w/4, h/4, 16, 16);
  }

  { // filter packed image, needing only h/4 read/write operations per column
    programCL_->KernelSetArgument("RecursiveGaussianForwardYPacked", 0, *oneDirFilteredImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 0, *oneDirFilteredImage_);

    programCL_->KernelSetArgument("RecursiveGaussianForwardYPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 1, *intermediateImage_);

    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 2, *packedGreyImage_);

    context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardYPacked", w, 16);
    context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardYPacked", w, 16);
  }

  { // unpack to grey image again
    programCL_->KernelSetArgument("RowUnPackRGBAToGrey", 0, *packedGreyImage_);
    programCL_->KernelSetArgument("RowUnPackRGBAToGrey", 1, *dst);
    context_->RunOn2DRange(*programCL_, "RowUnPackRGBAToGrey", w, h/4, 16, 16);
  }
  return 0;
}

int clfRecursiveGauss::FilterGreyColPacked(clfImage2D *src, clfImage2D *&dst) {
//  static clfImage2D *temp1 = context_->CreateImage2D();
//  static clfImage2D *temp2 = context_->CreateImage2D();
//  temp1->Allocate(src->StorageType(), src->ColorModel(), src->Width()*4, src->Height(), false, false);
//  temp2->Allocate(src->StorageType(), src->ColorModel(), src->Width()*4, src->Height(), false, false);
//  programCL_->KernelSetArgument("ColUnPackRGBAToGrey", 0, *src);
//  programCL_->KernelSetArgument("ColUnPackRGBAToGrey", 1, *temp1);
//  context_->RunOn2DRange(*programCL_, "ColUnPackRGBAToGrey", src->Width(), src->Height(), 16, 16);
//  context_->Finish();
//  FilterGrey(temp1, temp2);
//  programCL_->KernelSetArgument("ColPackGreyToRGBA", 0, *temp2);
//  programCL_->KernelSetArgument("ColPackGreyToRGBA", 1, *dst);
//  context_->RunOn2DRange(*programCL_, "ColPackGreyToRGBA", src->Width(), src->Height(), 16, 16);
//  return 0;
//
  unsigned int w,h, worg, horg;
  src->GetImageDim(worg,horg);
  src->GetImageDim(w,h);
  w *= 4;

  if (h % 64 != 0) {
    h = h+(64-(h%64));
  }
  if (w % 64 != 0) {
    w = w+(64-(w%64));
  }

  { // Allocate temporary images
    if (packedGreyImage_ == NULL) {
      packedGreyImage_ = context_->CreateImage2D();
      packedGreyImage_->Allocate(src->StorageType(), src->ColorModel(), w, h, 0, false, false);
    }
    if (intermediateImage_ == NULL) {
      intermediateImage_ = context_->CreateImage2D();
      intermediateImage_->AllocateFromTemplate(*packedGreyImage_);
    }
    if (oneDirFilteredImage_ == NULL) {
      oneDirFilteredImage_ = context_->CreateImage2D();
      oneDirFilteredImage_->AllocateFromTemplate(*packedGreyImage_);
    }
    if (dst == NULL) {
      dst = context_->CreateImage2D();
      dst->AllocateFromTemplate(*src);
    } else {
      if (dst->IsReadOnly()) {
        THROW_CL_EXCEPTION(cl::Error(-100, "Need to use dst as temporary read/write"));
      }
    }
  }

  { // filter packed image, needing only w/4 read/write operations per row
    programCL_->KernelSetArgument("RecursiveGaussianForwardXPacked", 0, *src);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 0, *src);

    programCL_->KernelSetArgument("RecursiveGaussianForwardXPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardXPacked", 2, *packedGreyImage_);

    context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardXPacked", h, 16);
    context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardXPacked", h, 16);
    context_->Finish();
  }

  { // "repack" image such that the grey image is now stored (RGBA)^T
    // rgba     rgba
    // RGBA     RRRR
    // RGBA  -> GGGG
    // RGBA     BBBB
    // RGBA     AAAA
    programCL_->KernelSetArgument("ColPackRGBAToRowPackRGBA", 0, *packedGreyImage_);
    programCL_->KernelSetArgument("ColPackRGBAToRowPackRGBA", 1, *oneDirFilteredImage_);
    context_->RunOn2DRange(*programCL_, "ColPackRGBAToRowPackRGBA", w/4, h/4, 16, 16);
    context_->Finish();
  }

  { // filter packed image, needing only h/4 read/write operations per column
    programCL_->KernelSetArgument("RecursiveGaussianForwardYPacked", 0, *oneDirFilteredImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 0, *oneDirFilteredImage_);

    programCL_->KernelSetArgument("RecursiveGaussianForwardYPacked", 1, *intermediateImage_);
    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 1, *intermediateImage_);

    programCL_->KernelSetArgument("RecursiveGaussianBackwardYPacked", 2, *packedGreyImage_);

    context_->RunOn1DRange(*programCL_, "RecursiveGaussianForwardYPacked", w, 16);
    context_->RunOn1DRange(*programCL_, "RecursiveGaussianBackwardYPacked", w, 16);
  }
  { // unpack to grey image again
    programCL_->KernelSetArgument("RowPackRGBAToColPackRGBA", 0, *packedGreyImage_);
    programCL_->KernelSetArgument("RowPackRGBAToColPackRGBA", 1, *intermediateImage_);
    context_->RunOn2DRange(*programCL_, "RowPackRGBAToColPackRGBA", w/4, h/4, 16, 16);
  }
  // copy to dst
  intermediateImage_->CopyToImage(*dst, 0,0, 0,0, worg, horg);
  return 0;
}


void clfRecursiveGauss::SetSigma(float sigma, int order) {
//      float &a0=params_[0], float &a1, float &a2, float &a3=params_[3],
//      float &b1=params_[4], float &b1, float &coefp=params_[6], float &coefn=params_[7]);
  // pre-compute filter coefficients
  float alpha = 1.695f / sigma;
  float ema = exp(-alpha);
  float ema2 = exp(-2.0f * alpha);
  params_.resize(8);
  params_[0] = 0.0f;
  params_[1] = 0.0f;
  params_[2] = 0.0f;
  params_[3] = 0.0f;
  params_[4] = -2.0f * ema;
  params_[5] = ema2;
  params_[6] = 0.0f;
  params_[7] = 0.0f;
  switch (order) {
    case 0: {
      const float k = (1.0f - ema)*(1.0f - ema)/(1.0f + (2.0f * alpha * ema) - ema2);
      params_[0] = k;
      params_[1] = k * (alpha - 1.0f) * ema;
      params_[2] = k * (alpha + 1.0f) * ema;
      params_[3] = -k * ema2;
    }
    break;
    case 1: {
      params_[0] = (1.0f - ema) * (1.0f - ema);
      params_[1] = 0.0f;
      params_[2] = -params_[0];
      params_[3] = 0.0f;
    }
    break;
    case 2: {
      const float ea = exp(-alpha);
      const float k = -(ema2 - 1.0f)/(2.0f * alpha * ema);
      float kn = -2.0f * (-1.0f + (3.0f * ea) - (3.0f * ea * ea) + (ea * ea * ea));
      kn /= (((3.0f * ea) + 1.0f + (3.0f * ea * ea) + (ea * ea * ea)));
      params_[0] = kn;
      params_[1] = -kn * (1.0f + (k * alpha)) * ema;
      params_[2] = kn * (1.0f - (k * alpha)) * ema;
      params_[3] = -kn * ema2;
    }
    break;
    default:
    break;
  }
  params_[6] = (params_[0] + params_[1])/(1.0f + params_[4] + params_[5]);
  params_[7] = (params_[2] + params_[3])/(1.0f + params_[4] + params_[5]);

  for (unsigned int i=0;i<kernelNames_.size();i++) {
    for (unsigned int j=0;j<params_.size();j++) {
      if (i<4)
        programCL_->KernelSetArgument(kernelNames_[i], j+2, params_[j]);
      else
        programCL_->KernelSetArgument(kernelNames_[i], j+3, params_[j]);
    }
  }
}

