/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <OpenCLFramework/Algorithm/clfTVL1Flow.hh>

//#define WRITE_DBG_IMAGES
#undef WRITE_DBG_IMAGES

using namespace std;
using namespace BIAS;
using namespace BIAS;

clfTVL1Flow::clfTVL1Flow(clfContext *ctx) {
  
  {  // default parameter set
    sigma_ = 1.0f/sqrt(8.0f); // was 1.0f/sqrt(8.0f);
    tau_ = 1.0f/sqrt(8.0f); // 1.0f/sqrt(8.0f);
    lambda_ = 20;               // was 50
    gamma_ = 0.02f;             //was 0.02f
    warps_ = 1;                //was 1
    maxIterations_ = 20;       // was 10
  }

  { // initialize all the cl stuff
    context_ = ctx;
    maxComputeUnits_ = context_->GetDeviceInfo().maxComputeUnits;
    
    colorConvCL_ = new clfColorConversion(context_);
    downsampleCL_ = new clfSimpleFilter(context_, "DownsampleBy2");
    upsampleCL_ = new clfSimpleFilter(context_, "UpsampleBy2");
    
    clearFilter_ = new clfFilter<float, float>(context_);
    
    programCL_ = context_->CreateProgram();
    
    // add code to our program
    programCL_->AddSource("/OpenCLFramework/Algorithm/cl/tvl1flow.cl");
    programCL_->Build();
    programCL_->AddKernel("Warp");
    programCL_->AddKernel("UvwGrad");
    programCL_->AddKernel("UpdateP");
    programCL_->AddKernel("UpdateDiv");
    programCL_->AddKernel("UpdateUVW");
    programCL_->AddKernel("UpdateUVWold");
    programCL_->AddKernel("WeightedGradX");
    programCL_->AddKernel("WeightedGradY");
    programCL_->AddKernel("CreateColorCoded");
  }
  { // initialize base images and variables
    numIm_ = 0;
    image1_ = context_->CreateImage2D();
    image2_ = context_->CreateImage2D();
    debug_ = context_->CreateImage2D();
    levels_ = 0;
  }
}

clfTVL1Flow::~clfTVL1Flow() {
  // free all the memory
  for (unsigned int i=0;i<pyramid1_.size();i++) {
    delete pyramid1_[i];
    delete pyramid2_[i];
    delete uvw_[i];
    delete uvwX_[i];
    delete uvwY_[i];
    delete uvw0_[i];
    delete uvwOld_[i];
    delete image2warpedt_[i];
    delete imageDiv_[i];
    delete P1_[i];
    delete P2_[i];
    delete imageCopy_[i];
    delete imageCopy2_[i];
    delete Ixyg_[i];
  }
  delete programCL_;
  delete colorConvCL_;
  delete downsampleCL_;
  delete upsampleCL_;
  delete clearFilter_;

  delete image1_;
  delete image2_;
  delete debug_;
}

void clfTVL1Flow::SetNumberOfIterations(unsigned int iternumber)
{
  maxIterations_ = iternumber;
}

void clfTVL1Flow::SetSmoothness(const float& smoothness)
{
  lambda_ = smoothness;
}

void clfTVL1Flow::GetUVW(BIAS::Image<float> &image) {
  uvw_[0]->CopyToBiasImage(image);
}

void clfTVL1Flow::GetColorCoded(BIAS::Image<float> &image) {
  // run the color coding kernel on biggest uvw image
  programCL_->KernelSetArgument("CreateColorCoded", 0, *uvw_[0]);
  programCL_->KernelSetArgument("CreateColorCoded", 1, *imageCopy_[0]);
  unsigned int w,h;
  imageCopy_[0]->GetImageDim(w,h);
  context_->RunOn2DRange(*programCL_, "CreateColorCoded", w,h,16,16);
  imageCopy_[0]->CopyToBiasImage(image);
}


void clfTVL1Flow::AddImage(BIAS::Image<unsigned char> image) {
  if (numIm_ == 0) {
    // on first image added, reserve all the memory needed
    image1_->AllocateFromBiasImage(image, true, false, true);
    image2_->AllocateFromBiasImage(image, true, false, true);
    int curWidth = image.GetWidth();
    int curHeight = image.GetHeight();
    while (curHeight % 16 == 0) {
      pyramid1_.push_back(newImage_(curWidth, curHeight));
      pyramid2_.push_back(newImage_(curWidth, curHeight));
      uvw_.push_back(newImage_(curWidth, curHeight,true));
      uvwX_.push_back(newImage_(curWidth, curHeight));
      uvwY_.push_back(newImage_(curWidth, curHeight));
      uvw0_.push_back(newImage_(curWidth, curHeight));
      uvwOld_.push_back(newImage_(curWidth, curHeight));
      image2warpedt_.push_back(newImage_(curWidth, curHeight));
      imageDiv_.push_back(newImage_(curWidth, curHeight));
      P1_.push_back(newImage_(curWidth, curHeight,true));
      P2_.push_back(newImage_(curWidth, curHeight,true));
      Ixyg_.push_back(newImage_(curWidth, curHeight));
      imageCopy_.push_back(newImage_(curWidth, curHeight));
      imageCopy2_.push_back(newImage_(curWidth, curHeight));
      curWidth /= 2;
      curHeight /= 2;
    }
    numIm_ = 1;
  } else if (numIm_ == 1) {
    // on second image, store the new image data
    image2_->WriteToImage( image.GetImageData() );
    // convert both images to grey, float, normalized by 1/255
    if (image.GetColorModel() == ImageBase::CM_RGB) {
      colorConvCL_->UCharRGBToFloatGreyNormalized( image1_, pyramid1_[0]);
      colorConvCL_->UCharRGBToFloatGreyNormalized( image2_, pyramid2_[0]);
    } else {
      colorConvCL_->UCharGreyToFloatGrey( image1_, pyramid1_[0], true);
      colorConvCL_->UCharGreyToFloatGrey( image2_, pyramid2_[0], true);
    }
    // create pyramids
    CreatePyramid_(pyramid1_);
    CreatePyramid_(pyramid2_);
    levels_ = (unsigned int)pyramid1_.size();
    numIm_ = 2;
  } else {
    // on every following image
    // switch pointers
    clfImage2D *tmp = image1_;
    image1_ = image2_;
    image2_ = tmp;
    for (unsigned int i=0;i<levels_;i++) {
      tmp = pyramid1_[i];
      pyramid1_[i] = pyramid2_[i];
      pyramid2_[i] = tmp;
    }
    // upload new image and convert
    image2_->WriteToImage( image.GetImageData() );
    if (image.GetColorModel() == ImageBase::CM_RGB) {
      colorConvCL_->UCharRGBToFloatGreyNormalized( image2_, pyramid2_[0]);
    } else {
      colorConvCL_->UCharGreyToFloatGrey( image1_, pyramid1_[0], true);
    }
    // create new pyramid
    CreatePyramid_(pyramid2_);
    // clear necessary lowest level images
    clearFilter_->Clear(0.0f, uvw_[levels_-1]);
    clearFilter_->Clear(0.0f, P1_[levels_-1]);
    clearFilter_->Clear(0.0f, P2_[levels_-1]);
    numIm_++;
  }
}

clfImage2D* clfTVL1Flow::newImage_(unsigned int w, unsigned int h, bool clear) {
  clfImage2D* pim = context_->CreateImage2D();
  pim->Allocate(ImageBase::ST_float, ImageBase::CM_RGBA, w, h, 0, false, false);
  if (clear) {
    clearFilter_->Clear(0.0f, pim);
  }
  return pim;
}


void clfTVL1Flow::CreatePyramid_(std::vector<clfImage2D *> &pyramid) {
  for (unsigned int i=1;i<pyramid.size();i++) {
    downsampleCL_->FilterScale( pyramid[i-1], pyramid[i], 0.5f);
  }
}

void clfTVL1Flow::Upsample_(clfImage2D *src, clfImage2D *dst, float scale) {
  upsampleCL_->GetProgram()->KernelSetArgument("UpsampleBy2", 2, scale);
  upsampleCL_->Filter(src,dst);
}

void clfTVL1Flow::Compute() {
  if (numIm_ < 2) return;
  // from bottom to top in pyramid
  for (int level = (int)pyramid1_.size() -1;level>=0;level--) {
    if (level != (int) pyramid1_.size() -1) {
      // upsample results from previous level to current level
      Upsample_(uvw_[level+1], uvw_[level], 2.0f);
      Upsample_(P1_[level+1], P1_[level]);
      Upsample_(P2_[level+1], P2_[level]);
    }
    // compute flow for this level
    ComputeFlow_(level);
    
#ifdef WRITE_DBG_IMAGES
    Image<float> debug;
    pyramid1_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-im0.mip", debug);
    pyramid2_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-im1.mip", debug);
    uvw_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-uvw.mip", debug);
    uvwX_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-uvwX.mip", debug);
    uvwY_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-uvwY.mip", debug);
    image2warpedt_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-image2warpedt_.mip", debug);
    Ixyg_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-Ixyg_.mip", debug);
    P1_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-p1_.mip", debug);
    P2_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-p2_.mip", debug);
    imageDiv_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-div_.mip", debug);
    Ixyg_[level]->CopyToBiasImage(debug);
    ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-grad.mip", debug);
#endif // WRITE_DBG_IMAGES
  }
}

void clfTVL1Flow::Warp_(unsigned int level) {
  clfImage2D *Ixyg = Ixyg_[level];
  clfImage2D *image1 = pyramid1_[level];
  clfImage2D *image2 = pyramid2_[level];
  clfImage2D *uvw = uvw_[level];
  clfImage2D *warped = image2warpedt_[level];
  clfImage2D *tmp = imageCopy_[level];

  unsigned int w,h;
  image1->GetImageDim(w,h);

  // create gradient image in x dir
  programCL_->KernelSetArgument("WeightedGradX", 0, *image1);
  programCL_->KernelSetArgument("WeightedGradX", 1, *tmp);
  context_->RunOn2DRange(*programCL_, "WeightedGradX", w,h, 16,16);

  // create gradient image in y dir
  // the kernel will also compute squared grad
  programCL_->KernelSetArgument("WeightedGradY", 0, *image1);
  programCL_->KernelSetArgument("WeightedGradY", 1, *tmp);
  programCL_->KernelSetArgument("WeightedGradY", 2, *Ixyg);
  programCL_->KernelSetArgument("WeightedGradY", 3, gamma_);
  context_->RunOn2DRange(*programCL_, "WeightedGradY", w,h, 16,16);
  
  // warp images, kernel will create difference image (It in channel 3 of warped image
  programCL_->KernelSetArgument("Warp", 0, *image1); // read
  programCL_->KernelSetArgument("Warp", 1, *image2); // read
  programCL_->KernelSetArgument("Warp", 2, *uvw); // read
  programCL_->KernelSetArgument("Warp", 3, *warped); // write
  context_->RunOn2DRange(*programCL_, "Warp", w,h,16,16);

}

void clfTVL1Flow::ComputeFlow_(unsigned int level) {
  // copy pointers for more readability
  clfImage2D *uvw = uvw_[level];
  clfImage2D *uvwold = uvwOld_[level];
  clfImage2D *uvw0 = uvw0_[level];
  clfImage2D *Ixyg = Ixyg_[level];
  clfImage2D *uvwX = uvwX_[level];
  clfImage2D *uvwY = uvwY_[level];
  clfImage2D *p1 = P1_[level];
  clfImage2D *p2 = P2_[level];
  clfImage2D *div = imageDiv_[level];
  clfImage2D *tmp = imageCopy_[level];
  clfImage2D *tmp2 = imageCopy2_[level];
  clfImage2D *warped = image2warpedt_[level];

  // set all the parameters
  programCL_->KernelSetArgument("UvwGrad", 0, *uvwold); // read
  programCL_->KernelSetArgument("UvwGrad", 1, *uvwX); // write
  programCL_->KernelSetArgument("UvwGrad", 2, *uvwY); // write
  
  programCL_->KernelSetArgument("UpdateP", 0, *tmp); //read
  programCL_->KernelSetArgument("UpdateP", 1, *tmp2); //read
  programCL_->KernelSetArgument("UpdateP", 2, *uvwX); //read
  programCL_->KernelSetArgument("UpdateP", 3, *uvwY); //read
  programCL_->KernelSetArgument("UpdateP", 4, *p1); //write
  programCL_->KernelSetArgument("UpdateP", 5, *p2); //write
  programCL_->KernelSetArgument("UpdateP", 6, sigma_);
  
  programCL_->KernelSetArgument("UpdateDiv", 0, *p1); //read
  programCL_->KernelSetArgument("UpdateDiv", 1, *p2); //read
  programCL_->KernelSetArgument("UpdateDiv", 2, *div);//write

  programCL_->KernelSetArgument("UpdateUVW", 0, *uvwold); // read
  programCL_->KernelSetArgument("UpdateUVW", 1, *div); //read
  programCL_->KernelSetArgument("UpdateUVW", 2, *uvw0); //read
  programCL_->KernelSetArgument("UpdateUVW", 3, *Ixyg); //read
  programCL_->KernelSetArgument("UpdateUVW", 4, *warped); //read

  programCL_->KernelSetArgument("UpdateUVW", 5, *uvw); // write
  programCL_->KernelSetArgument("UpdateUVW", 6, tau_);
  programCL_->KernelSetArgument("UpdateUVW", 7, lambda_);
  programCL_->KernelSetArgument("UpdateUVW", 8, gamma_);

  programCL_->KernelSetArgument("UpdateUVWold", 0, *tmp);
  programCL_->KernelSetArgument("UpdateUVWold", 1, *uvw);
  programCL_->KernelSetArgument("UpdateUVWold", 2, *uvwold);
  
  // get pyramid level size
  unsigned int w,h;
  uvw->GetImageDim(w,h);

  // backup uvw
  uvw->CopyToImage(*uvwold);

  // default value is to do one warp only
  for (unsigned int j=0;j<warps_;j++) {
    // keep state before iterating
    uvw->CopyToImage(*uvw0);
    // warp image2 according to previous result
    Warp_(level);

    for (unsigned int k=0;k<maxIterations_;k++) {
      // gradients of direction
      context_->RunOn2DRange(*programCL_, "UvwGrad", w,h,16,16);

      // update statistics
      p1->CopyToImage( *tmp );
      p2->CopyToImage( *tmp2 );
      context_->RunOn2DRange(*programCL_, "UpdateP", w,h,16,16);
      // update div
      context_->RunOn2DRange(*programCL_, "UpdateDiv", w,h,16,16);
      
      uvw->CopyToImage(*uvwold);
      // generate new uvw
      context_->RunOn2DRange(*programCL_, "UpdateUVW", w,h,16,16);
      // update for next iteration
      uvwold->CopyToImage(*tmp);
      context_->RunOn2DRange(*programCL_, "UpdateUVWold", w,h, 16,16);
      
#ifdef WRITE_DBG_IMAGES
      if (level==4) {
        Image<float> debug;
        uvwold->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-uvwold.mip", debug);
        uvw->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-uvw.mip", debug);
        p1->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-p1.mip", debug);
        p2->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-p2.mip", debug);
        div->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-div.mip", debug);
        Ixyg->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-Ixyg.mip", debug);
        uvwX->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-uvwX.mip", debug);
        uvwY->CopyToBiasImage(debug);
        ImageIO::Save("cbg-"+FileHandling::LeadingZeroString(level,2)+"-"+FileHandling::LeadingZeroString(k,2)+"-uvwY.mip", debug);
      }
#endif // WRITE_DBG_IMAGES
    }     
  }
}

