/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Utils/Param.hh>
#include <Utils/IOUtils.hh>
#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>
#include <OpenCLFramework/clfContext.hh>
#include <iostream>
#include <OpenGLFramework/RenderingContext/glfPBuffer.hh>
#include <OpenGLFramework/RenderingContext/glfRenderingContext.hh>

using namespace BIAS;
using namespace std;

/**
 * @brief mean filtering an image using opencl
 * 
 * test program for clfImage2D objects. The kernel mean3x3 filters an image
 * 
 */
int main(int argc, char* argv[]) {
  
  Param par;
  string *inname = par.AddParamString("image", "image", "", 'i');
  string *outname = par.AddParamString("out", "result", "climage.mip", 'o');
  
  if (IOUtils::ParseCommandLineEvalHelp(par, argc, argv) == 0) {
    par.Usage();
  }
  
  Image<unsigned char> in,rgba;
  Image<float> inf;
  ImageIO::Load(*inname, in);
//  in.Init(1024,1024,3);
  unsigned int w = in.GetWidth();
  unsigned int h = in.GetHeight();
  rgba.Init(w,h,4);
  unsigned int c = rgba.GetChannelCount();
  ImageConvert::Convert(in, rgba, ImageBase::CM_RGBA);
  ImageConvert::ConvertST(rgba, inf, ImageBase::ST_float);
  Image<float> out;
  cout << "WxHxC: " << w << " " << h << " " << c << endl;
  out.Init(w,h,c);
  out.SetColorModel(ImageBase::CM_RGBA);

  try {

    // create context on first GPU found
    clfContext context(false);
    // print some info
    context.PrintPlatformsInfo();
    context.PrintDeviceInfo();
    
    // create a program in this context
    clfProgram *program = context.CreateProgram();
    
    // add code to our program
    program->AddSource("/OpenCLFramework/Examples/mean.cl");
    // compile program
    program->Build();
    // we want to use the mean_filter kernel
    program->AddKernel("mean_filter");
    
    // allocate opencl image buffers in the context
    clfImage2D *inBuffer = context.CreateImage2D();
    clfImage2D *outBuffer = context.CreateImage2D();
    
    // the first image is created from input
    inBuffer->AllocateFromBiasImage(
        // source image
        inf,
        // read-only (true)
        true,
        // write-only (false)
        false,
        // copy from host
        true);
    
    // the second image is allocated on device, but using out params width, height, etc.
    outBuffer->AllocateFromBiasTemplate(
        // take parameters from
        out,
        // read-only (false)
        false,
        // write-only (true)
        true);
    
    // set the arguments of compute kernel
    program->KernelSetArgument("mean_filter", 0, *inBuffer);
    program->KernelSetArgument("mean_filter", 1, *outBuffer);

    // run 2D kernel program
    context.RunOn2DRange(*program, "mean_filter", w,h, 1,1);

    // now we can either read directly to the output image
    // outBuffer->ReadFromImage( out.GetImageData() );

    // or let opencl map the image on some host address and copy it
    float *data = (float*)outBuffer->MapImage();
    out.CopyIn_NoInit(data);
    // free mapped area
    outBuffer->UnMap(data);
    
  } catch (clfException &err) {
    // the detailed error string
    cout << err.GetDetailedString() << endl;
  }
  // save result
  ImageIO::Save(*outname, out);
  return 0;
}
