/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <OpenCLFramework/Filter/clfFilter.hh>
#include <Base/Common/BIASpragma.hh>
using namespace std;
using namespace BIAS;

namespace BIAS {

  template <class InputStorageType, class OutputStorageType>
  clfFilter<InputStorageType, OutputStorageType>::
  clfFilter(clfContext *ctx) {
    context_ = ctx;
    filterMask_ = NULL;
    winSize_ = 0;
    tmpImage_ = context_->CreateImage2D();
    filterMask_ = context_->CreateBuffer();
    program_ = context_->CreateProgram();
    program_->AddSource("/OpenCLFramework/Filter/cl/filteroperation.cl");
    program_->Build();
    programName_ = "Convolve";
    program_->AddKernel(programName_);
    program_->AddKernel("Clear");
  }

  template <class InputStorageType, class OutputStorageType>
  clfFilter<InputStorageType, OutputStorageType>::
  ~clfFilter() {
    delete tmpImage_;
    delete filterMask_;
    delete program_;
  }
  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  SetFilter_(unsigned int size, const float *mask) {
    winSize_ = size;
    try {
      filterMask_->Allocate( winSize_*winSize_ * sizeof(float), true, false );
      filterMask_->WriteToBuffer( (float*)mask );
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return 0;
  }
  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  SetFilter1D_(unsigned int size, const float *mask) {
    winSize_ = size;
    try {
      filterMask_->Allocate( winSize_ * sizeof(float), true, false );
      filterMask_->WriteToBuffer( (float*)mask );
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return 0;
  }
  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  Build(unsigned int size) {
    winSize_ = size;
    // defaulting to mean filter
    float *mask = new float[winSize_*winSize_];
    for (unsigned int i=0;i<winSize_*winSize_;i++) 
      mask[i] = 1.0f / float(winSize_*winSize_);
    SetFilter_(winSize_, mask);
    delete mask;
    return 0;
  }
  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  Filter(clfImage2D *src, clfImage2D *dst) {
    int ret = 0;
    try {
      ret = Filter_(programName_, src, dst);
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return ret;
  }
  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  Clear(float color, clfImage2D *image) {
    try {
      program_->KernelSetArgument( "Clear", 0, color );
      program_->KernelSetArgument( "Clear", 1, *image );
      unsigned int w,h;
      image->GetImageDim(w,h);
      unsigned int localX = 16;
      unsigned int localY = 16;
      if (w%16) localX = 1;
      if (h%16) localY = 1;
      context_->RunOn2DRange( *program_, "Clear", w,h, localX,localY);
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return 0;
  }

  
  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  FilterSep_(std::string nameX,std::string nameY, clfImage2D *src, clfImage2D *dst) {
    int ret = 0;
    try {
      tmpImage_->AllocateFromTemplate( *src, false, false);
      ret = Filter_(nameX, src, tmpImage_);
      ret += Filter_(nameY, tmpImage_, dst);
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return ret;
  }

  template <class InputStorageType, class OutputStorageType>
  int clfFilter<InputStorageType, OutputStorageType>::
  Filter_(std::string name, clfImage2D *src, clfImage2D *dst, int blockX, int blockY) {
    try {
      program_->KernelSetArgument( "Clear", 0, 0 );
      program_->KernelSetArgument( "Clear", 1, *dst );
      unsigned int w,h;
      dst->GetImageDim(w,h);
      context_->RunOn2DRange( *program_, "Clear", w,h, 16,16);
      program_->KernelSetArgument( name, 0, *src );
      program_->KernelSetArgument( name, 1, *filterMask_ );
      program_->KernelSetArgument( name, 2, (int)winSize_ );
      program_->KernelSetArgument( name, 3, *dst );
      context_->RunOn2DRange( *program_, name, w, h, 16, 16);
    } catch (clfException &e) {
      BIASERR(e.GetDetailedString());
      return 1;
    }
    return 0;
  }

  template class BIASOpenCLFramework_EXPORT clfFilter<float, float>;
  template class BIASOpenCLFramework_EXPORT clfFilter<unsigned char, unsigned char>;

} /* namespace DECS */
