/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CLFBUFFER_HH_
#define CLFBUFFER_HH_

#include <OpenCLFramework/clfOpenCL.hh>
#include <OpenCLFramework/clfMemory.hh>
#include <OpenGLFramework/Base/glfVertexBuffer.hh>

namespace BIAS {
  
  /**
   * 
   * @brief OpenCL Buffer wrapper
   * 
   * An clfBuffer is used for Buffers both host and GL objects.
   * All buffers are created from a valid clfContext, hence the constructor is protected.
   * 
   * @author fkellner 06/11
   */
  class BIASOpenCLFramework_EXPORT clfBuffer : public clfMemory {
    public:

      virtual ~clfBuffer();
      
      /**
       * @brief Allocation of a memory buffer
       * A memory buffer can be created on device or host, it can be initialized from a host ptr. if the buffer is not readonly and not writeonly
       * it is declared read/write. 
       * @param bufsize size of buffer in byte
       * @param readonly if this buffer is read only
       * @param writeonly if this buffer is write only
       * @param hostptr ptr to host memory which is used for initialization (if NULL, do not initialize)
       * @param copy copy host ptr data to device or use host memory
       */
      void Allocate(unsigned int bufsize, bool readonly=false, bool writeonly=false, void *hostptr = NULL, bool copy = false);
      void Allocate(unsigned int rows, unsigned int cols, unsigned int channels, unsigned int itemsize, bool readonly=false, bool writeonly=false, void *hostptr = NULL, bool copy = false);
      void SetCached(bool cache);
      /**
       * @brief Allocation of a memory buffer from a GL vertexBuffer (works only on shared context!)
       */
      void AllocateFromVertexBuffer(BIAS::glfVertexBuffer &vbo, bool readonly=false, bool writeonly=false);
      
      /**
       * @brief write from host memory to buffer object
       * @param data host memory pointer
       * @param offset offset in bytes
       * @param size number of bytes to write
       */
      void WriteToBuffer(const void *data, unsigned int offset = 0, unsigned int size = 0);

      /**
       * @brief read from buffer object to host memory
       * @param data host memory pointer
       * @param offset offset in bytes
       * @param size number of bytes to read
       */
      void ReadFromBuffer(void *data, unsigned int offset = 0, unsigned int size = 0);

      /**
       * @brief copy from one buffer to another
       * @param outputbuffer buffer object (to)
       * @param size number of bytes to copy
       */
      void CopyBuffer(clfBuffer &outputbuffer, unsigned int srcoffset = 0, unsigned int dstoffset = 0, unsigned int size = 0);
      
      void* MapBuffer(bool write = false, unsigned int offset = 0, unsigned int size = 0);
      
      unsigned int GetRows() const;
      unsigned int GetCols() const;
      unsigned int GetChannels() const;
      unsigned int GetItemSize() const;
      unsigned int GetStride() const;

    protected:
      clfBuffer(cl::Context *context, cl::CommandQueue *queue_);
      cl::Buffer& buffer();
      
    private:
      friend class clfContext;
      friend class clfProgram;
      
      unsigned int rows_, cols_, channels_, itemsize_, stride_;
      bool cache_;
  };
}

#endif /* CLFBUFFER_HH_ */
