/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <OpenCLFramework/clfException.hh>
#include <Base/Common/BIASpragma.hh>

#include <iosfwd>
#include <Base/Debug/Error.hh>

namespace BIAS {

  clfException::clfException(cl::Error clerror, std::string file, int line) throw() {
    fileName_ = file;
    lineNumber_ = line;
    errorCode_ = clerror.err();
    message_ = clerror.what();
    
    switch(errorCode_) {
      case CL_SUCCESS:
        reason_ = "success.";
        break;
      case CL_DEVICE_NOT_FOUND:
        reason_ = "CL_DEVICE_NOT_FOUND.";
        break;
      case CL_DEVICE_NOT_AVAILABLE:
        reason_ = "CL_DEVICE_NOT_AVAILABLE.";
        break;
      case CL_COMPILER_NOT_AVAILABLE:
        reason_ = "CL_COMPILER_NOT_AVAILABLE.";
        break;
      case CL_MEM_OBJECT_ALLOCATION_FAILURE:
        reason_ = "CL_MEM_OBJECT_ALLOCATION_FAILURE.";
        break;
      case CL_OUT_OF_RESOURCES:
        reason_ = "CL_OUT_OF_RESOURCES.";
        break;
      case CL_OUT_OF_HOST_MEMORY:
        reason_ = "CL_OUT_OF_HOST_MEMORY.";
        break;
      case CL_PROFILING_INFO_NOT_AVAILABLE:
        reason_ = "CL_PROFILING_INFO_NOT_AVAILABLE.";
        break;
      case CL_MEM_COPY_OVERLAP:
        reason_ = "CL_MEM_COPY_OVERLAP.";
        break;
      case CL_IMAGE_FORMAT_MISMATCH:
        reason_ = "CL_IMAGE_FORMAT_MISMATCH.";
        break;
      case CL_IMAGE_FORMAT_NOT_SUPPORTED:
        reason_ = "CL_IMAGE_FORMAT_NOT_SUPPORTED.";
        break;
      case CL_BUILD_PROGRAM_FAILURE:
        reason_ = "CL_BUILD_PROGRAM_FAILURE.";
        break;
      case CL_MAP_FAILURE:
        reason_ = "CL_MAP_FAILURE.";
        break;
      case CL_MISALIGNED_SUB_BUFFER_OFFSET:
        reason_ = "CL_MISALIGNED_SUB_BUFFER_OFFSET.";
        break;
      case CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST:
        reason_ = "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST.";
        break;
      case CL_INVALID_VALUE:
        reason_ = "CL_INVALID_VALUE.";
        break;
      case CL_INVALID_DEVICE_TYPE:
        reason_ = "CL_INVALID_DEVICE_TYPE.";
        break;
      case CL_INVALID_PLATFORM:
        reason_ = "CL_INVALID_PLATFORM.";
        break;
      case CL_INVALID_DEVICE:
        reason_ = "CL_INVALID_DEVICE.";
        break;
      case CL_INVALID_CONTEXT:
        reason_ = "CL_INVALID_CONTEXT.";
        break;
      case CL_INVALID_QUEUE_PROPERTIES:
        reason_ = "CL_INVALID_QUEUE_PROPERTIES.";
        break;
      case CL_INVALID_COMMAND_QUEUE:
        reason_ = "CL_INVALID_COMMAND_QUEUE.";
        break;
      case CL_INVALID_HOST_PTR:
        reason_ = "CL_INVALID_HOST_PTR.";
        break;
      case CL_INVALID_MEM_OBJECT:
        reason_ = "CL_INVALID_MEM_OBJECT.";
        break;
      case CL_INVALID_IMAGE_FORMAT_DESCRIPTOR:
        reason_ = "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR.";
        break;
      case CL_INVALID_IMAGE_SIZE:
        reason_ = "CL_INVALID_IMAGE_SIZE.";
        break;
      case CL_INVALID_SAMPLER:
        reason_ = "CL_INVALID_SAMPLER.";
        break;
      case CL_INVALID_BINARY:
        reason_ = "CL_INVALID_BINARY.";
        break;
      case CL_INVALID_BUILD_OPTIONS:
        reason_ = "CL_INVALID_BUILD_OPTIONS.";
        break;
      case CL_INVALID_PROGRAM:
        reason_ = "CL_INVALID_PROGRAM.";
        break;
      case CL_INVALID_PROGRAM_EXECUTABLE:
        reason_ = "CL_INVALID_PROGRAM_EXECUTABLE.";
        break;
      case CL_INVALID_KERNEL_NAME:
        reason_ = "CL_INVALID_KERNEL_NAME.";
        break;
      case CL_INVALID_KERNEL_DEFINITION:
        reason_ = "CL_INVALID_KERNEL_DEFINITION.";
        break;
      case CL_INVALID_KERNEL:
        reason_ = "CL_INVALID_KERNEL.";
        break;
      case CL_INVALID_ARG_INDEX:
        reason_ = "CL_INVALID_ARG_INDEX.";
        break;
      case CL_INVALID_ARG_VALUE:
        reason_ = "CL_INVALID_ARG_VALUE.";
        break;
      case CL_INVALID_ARG_SIZE:
        reason_ = "CL_INVALID_ARG_SIZE.";
        break;
      case CL_INVALID_KERNEL_ARGS:
        reason_ = "CL_INVALID_KERNEL_ARGS.";
        break;
      case CL_INVALID_WORK_DIMENSION:
        reason_ = "CL_INVALID_WORK_DIMENSION.";
        break;
      case CL_INVALID_WORK_GROUP_SIZE:
        reason_ = "CL_INVALID_WORK_GROUP_SIZE.";
        break;
      case CL_INVALID_WORK_ITEM_SIZE:
        reason_ = "CL_INVALID_WORK_ITEM_SIZE.";
        break;
      case CL_INVALID_GLOBAL_OFFSET:
        reason_ = "CL_INVALID_GLOBAL_OFFSET.";
        break;
      case CL_INVALID_EVENT_WAIT_LIST:
        reason_ = "CL_INVALID_EVENT_WAIT_LIST.";
        break;
      case CL_INVALID_EVENT:
        reason_ = "CL_INVALID_EVENT.";
        break;
      case CL_INVALID_OPERATION:
        reason_ = "CL_INVALID_OPERATION.";
        break;
      case CL_INVALID_GL_OBJECT:
        reason_ = "CL_INVALID_GL_OBJECT.";
        break;
      case CL_INVALID_BUFFER_SIZE:
        reason_ = "CL_INVALID_BUFFER_SIZE.";
        break;
      case CL_INVALID_MIP_LEVEL:
        reason_ = "CL_INVALID_DECS_LEVEL.";
        break;
      case CL_INVALID_GLOBAL_WORK_SIZE:
        reason_ = "CL_INVALID_GLOBAL_WORK_SIZE.";
        break;
      case CL_INVALID_PROPERTY:
        reason_ = "CL_INVALID_PROPERTY.";
        break;
      case -100:
        reason_ = "OpenCLFramework internal error";
        break;
      default:
        reason_ = "reason unknown.";
        break;
    }
    
    std::stringstream ss;
    ss << "OpenCL Exception:\n" << fileName_ << " : " << lineNumber_ << "\n"
    << message_ << " (" << errorCode_ << ")\n" << "Reason: " << reason_;
    detailed_ = ss.str();    
    //if (errorCode_ != -39)
    BIASERR(detailed_);
    BIASABORT;
  }

  const std::string& clfException::GetFileName() const throw() { return fileName_; }
  int clfException::GetLineNumber() const throw() { return lineNumber_; }
  int clfException::GetErrorCode() const throw() { return errorCode_; }
  const std::string& clfException::GetMessageString() const throw() { return message_; }
  const std::string& clfException::GetReasonString() const throw() { return reason_; }

  const std::string& clfException::GetDetailedString() const throw() { return detailed_; }

  const char* clfException::what() const throw() {
    return detailed_.c_str();
  }
}
