/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CLFIMAGE3D_HH_
#define CLFIMAGE3D_HH_

#include <bias_config.h>
#include <Base/Common/BIASpragmaStart.hh>
#include <OpenCLFramework/clfOpenCL.hh>
#include <OpenCLFramework/clfMemory.hh>
#include <OpenGLFramework/Base/glfTexture3D.hh>

namespace BIAS {

  /**
   *
   * @brief OpenCL Image3D wrapper
   *
   * An clfImage3D is used for both cl::Images, host and GL objects.
   * All buffers are created from a valid clfContext, hence the constructor is protected.
   *
   * @author fkellner 06/11
   */
  class BIASOpenCLFramework_EXPORT clfImage3D : public clfMemory {
    public:

      virtual ~clfImage3D();

      /**
       * @brief Allocation of a memory buffer as 3D image, either call directly or use wrapper for BIAS::ImageBase
       */
      void Allocate(
          BIAS::ImageBase::EStorageType st, BIAS::ImageBase::EColorModel cm,
          unsigned int width, unsigned int height, unsigned int levels, unsigned int stride=0,
          bool readonly=false, bool writeonly=false, const void *hostptr = NULL, bool copy = false);

      /**
       * @brief Allocation of a memory buffer as 3D image (2d+level), biasimage is used as template for one level
       */
      void AllocateFromBiasTemplate(const BIAS::ImageBase &image, unsigned int levels, bool readonly=false, bool writeonly=false);

      void AllocateFromTemplate(const clfImage3D &src, bool readonly=false, bool writeonly=false);

      /**
       * @brief Allocation of a memory buffer from a GL Texture3D (works only on shared context!)
       */
      void AllocateFromTexture3D(BIAS::glfTexture3D &tex, bool readonly=false, bool writeonly=false);

      inline unsigned int Width() { return width_; }
      inline unsigned int Height() { return height_; }
      inline unsigned int Levels() { return depth_; }
      inline unsigned int Stride() { return stride_; }
      inline unsigned int SlideStride() { return slideStride_; }

      /// @brief write from host memory to image
      void WriteToImage(const void *data,
          unsigned int originX=0, unsigned int originY=0, unsigned int originZ=0,
          unsigned int regionX=0, unsigned int regionY=0, unsigned int regionZ=0);

      /// @brief read from image to host memory
      void ReadFromImage(void *data,
          unsigned int originX=0, unsigned int originY=0, unsigned int originZ=0,
          unsigned int regionX=0, unsigned int regionY=0, unsigned int regionZ=0);

      void CopyToImage(clfImage3D &outputimage,
          unsigned int srcoriginX=0, unsigned int srcoriginY=0, unsigned int srcoriginZ=0,
          unsigned int dstoriginX=0, unsigned int dstoriginY=0, unsigned int dstoriginZ=0,
          unsigned int regionX=0, unsigned int regionY=0, unsigned int regionZ=0);

      void* MapImage(bool write = false,
          unsigned int originX=0, unsigned int originY=0, unsigned int originZ=0,
          unsigned int regionX=0, unsigned int regionY=0, unsigned int regionZ=0
          );

      void CopyToBiasImage(BIAS::ImageBase &image, unsigned int level,
          unsigned int originX=0, unsigned int originY=0,
          unsigned int regionX=0, unsigned int regionY=0
          );

      std::vector<std::string> GetSupportedImageFormats(bool readonly=false, bool writeonly=false);

      void GetImageDim(unsigned int &width, unsigned int &height, unsigned int &levels) {
        width = width_;
        height = height_;
        levels = depth_;
      }

    protected:
      clfImage3D(cl::Context *context, cl::CommandQueue *queue);
      cl::Image3D& image();

    private:
      friend class clfContext;
      friend class clfProgram;

      cl::ImageFormat DetermineImageFormat_(BIAS::ImageBase::EStorageType st, BIAS::ImageBase::EColorModel cm);
      bool GetFormatSupported_(cl::ImageFormat test, bool readonly, bool writeonly);
      void MakeDim_(cl::size_t<3> &dest, int x, int y, int z);

      unsigned int width_, height_, depth_, stride_, slideStride_;

      BIAS::ImageBase::EStorageType st_;
      BIAS::ImageBase::EColorModel cm_;

      static std::vector<cl::ImageFormat> imageFormatsReadable_;
      static std::vector<cl::ImageFormat> imageFormatsWritable_;
      static std::vector<cl::ImageFormat> imageFormatsReadWrite_;

  };
}
#include <Base/Common/BIASpragmaEnd.hh>
#endif /* CLFIMAGE3D_HH_ */
