/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef OPENCLPROGRAM_HH_
#define OPENCLPROGRAM_HH_

#include <bias_config.h>
#include <Base/Common/BIASpragmaStart.hh>
#include <OpenCLFramework/clfOpenCL.hh>
#include <OpenCLFramework/clfBuffer.hh>
#include <OpenCLFramework/clfImage2D.hh>
#include <OpenCLFramework/clfImage3D.hh>
#include <map>
#include <OpenCLFramework/clfKernelSources.hh>

namespace BIAS {
  
  typedef std::map<std::string, cl::Kernel> cl_kernel_map;
  
  /**
   * 
   * @brief OpenCL Program wrapper
   * 
   * A program can have multiple kernels and can be set up from different source
   * files.
   * 
   * Programs can only be created from a valid OpenCLContext Object, hence the
   * constructor is protected.
   * 
   * @author fkellner 06/11
   */
  class BIASOpenCLFramework_EXPORT clfProgram {
    public:
      
      virtual ~clfProgram();
      
      /** 
       * @brief adds source code from a file
       * @throw clfException
       * @return source code for anatol
       */
      std::string AddSource(std::string filename);

      /** 
       * @brief adds source code from a string
       * @throw clfException
       */
      void AddSourceFromString(std::string sourceCode);
      
      /** 
       * @brief builds the sources added by AddSource and AddSourceFromString
       * @throw clfException
       */
      void Build(int deviceNr = 0, std::string options = "");
      
      std::string GetBuildLog(int deviceNr = 0);

      /** 
       * @brief adds a kernel to the program. program needs to be built first!
       * @throw clfException
       */
      void AddKernel(std::string kernelname);

      /** 
       * @brief adds a vector of kernels to the program. program needs to be built first!
       * @throw clfException
       */
      void AddKernels(std::vector<std::string> kernelnames);
      
      /** 
       * @brief set kernel argument
       * @param kernelname name of kernel function
       * @param argnumber number of kernel argument
       * @param buffer the argument
       * @throw clfException
       */
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, clfBuffer &buffer);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, clfImage2D &buffer);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, clfImage3D &buffer);

      // todo implement more if needed!
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_float2 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_float4 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_float8 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_float16 arg);

      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_int2 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_int4 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_int8 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_int16 arg);

      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_uint2 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_uint4 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_uint8 arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, cl_uint16 arg);

      void KernelSetArgument(std::string kernelname, unsigned int argnumber, int arg);
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, float arg);     
      void KernelSetArgument(std::string kernelname, unsigned int argnumber, float* arg);

      void KernelSetLocalArgument(std::string kernelname, unsigned int argnumber, int size);

      static void AddSourceToDB(const std::string fname, const char *source, unsigned int &blocks) {
        kernelsourcesdb_.AddSource(fname, source, blocks);
      }
      static void AddAdditionalSourceDir(const std::string dir) {
        additionalsourcedirs_.push_back(dir);
      }

    protected:
      clfProgram(cl::Context *context, std::vector<cl::Device> *devices);
      cl::Kernel& operator()(std::string kernelname);
      
    private:
      friend class clfContext;

      cl::Context *context_;
      std::vector<cl::Device> *devices_;
      
      cl::Program::Sources sources_;
      cl::Program program_;
      cl_kernel_map kernels_;

      bool addWriteMultiplier_;

      static clfKernelSources kernelsourcesdb_;
      static std::vector<std::string> additionalsourcedirs_;
  };
}
#include <Base/Common/BIASpragmaEnd.hh>
#endif /* clfPROGRAM_HH_ */
