/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfElementBuffer_hh__
#define __glfElementBuffer_hh__

#include "glfCommon.hh"
#include <vector>

namespace BIAS {

  /**
   * \brief An element buffer contains vertex indices that form primitives.
   * With element buffers, a single vertex can be used by multiple
   * primitives, although it is only stored once in the vertex buffer.
   * \ingroup g_openglframework	
   * \author jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfElementBuffer {
  public:
    glfElementBuffer();
    ~glfElementBuffer();

    /**
     * Creates the element buffer for the given number of indices.
     * \param numIndices Number of indices to fit into the buffer.
     * \param type Data type of indices. May be GL_UNSIGNED_BYTE,
     *             GL_UNSIGNED_SHORT, or GL_UNSIGNED_INT.
     * \param mode Primitive type. May be GL_POINTS, GL_LINE_STRIP,
                   GL_LINE_LOOP, GL_LINES, GL_TRIANGLE_STRIP, GL_TRIANGLE_FAN,
                   GL_TRIANGLES, GL_QUAD_STRIP, GL_QUADS, or GL_POLYGON.
     * \param indices Pointer to vertex indices.
     */
    void Create(int numIndices, GLenum type, GLenum mode, const void* indices);

    /**
     * Secure convenience wrapper for glfElementBuffer::Create method with
     * type = GL_UNSIGNED_BYTE.
     */
    void Create(const std::vector<unsigned char>& indices, GLenum mode) {
      Create((int)indices.size(), GL_UNSIGNED_BYTE, mode, &indices[0]);
    }

    /**
     * Secure convenience wrapper for glfElementBuffer::Create method with
     * type = GL_UNSIGNED_SHORT.
     */
    void Create(const std::vector<unsigned short>& indices, GLenum mode) {
      Create((int)indices.size(), GL_UNSIGNED_SHORT, mode, &indices[0]);
    }

    /**
     * Secure convenience wrapper for glfElementBuffer::Create method with
     * type = GL_UNSIGNED_INT;
     */
    void Create(const std::vector<unsigned int>& indices, GLenum mode) {
      Create((int)indices.size(), GL_UNSIGNED_INT, mode, &indices[0]);
    }

    /**
     * Maps the element buffer to system memory, so it can be modified after creation.
     * \attention Access is write-only!
     */
    void* Map();

    /**
     * Unmaps the element buffer. Must be called after glfElementBuffer::Map and before
     * the element buffer is used for rendering again.
     */
    void Unmap();

    /**
     * Draw elements from the buffer.
     */
    void Draw(int first = 0, int count = -1) const;

  private:
    GLenum mode_;
    GLenum type_;
    int typeSize_;
    int numIndices_;
    bool useBuffer_;
    GLuint bufferID_;
    bool mapped_;
    std::vector<GLubyte> array_;

    /** An intuitive assignment operator should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this operator cannot be used.
     */
    glfElementBuffer& operator=(const glfElementBuffer& b);
    
    /** An intuitive copy constructor should generate a copy, as long as nobody is willing to implement
     * the generation of a copy in gl controlled memory this constructor cannot be used.
     */
    glfElementBuffer(const glfElementBuffer& b);    

    
  };

} // namespace BIAS

#endif // __glfElementBuffer_hh__
