/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef IMAGEWARPER_HH
#define IMAGEWARPER_HH

#include <BIAS_DeclSpec.hh>
#include <Image/Camera.hh>
#include <OpenGLFramework/Base/glfFramebufferObject.hh>
#include <OpenGLFramework/RenderingContext/glfPBuffer.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <OpenGLFramework/SpecializedBatches/PMDWarp.hh>

namespace BIAS {

/**
 * \class ImageWarper
 * \brief this class warpes depth images from one persective to the perspective of
 * another camera. In principle it wraps the classes PMDWarp, glfPBuffer,
 * glfFramebufferObject and others to warp using an own context and own textures
 * \author ischiller
 * \date 10/2010
 */
template<class StorageType>
class BIASOpenGLFramework_EXPORT ImageWarper : public BIAS::Camera<StorageType> {

public:
  ImageWarper(bool ignoreRadialDistortion = false);
  ~ImageWarper();

  /** \brief Init the warping with projection parameters   */
  int Init(BIAS::ProjectionParametersPerspective& source,
            BIAS::ProjectionParametersPerspective& target,
            float minZ=500, float maxZ=7500.0);
  /** \brief Process one warping from source to target parameters
   * \param depthImage[in]: the depth image
  *  \param warpedDepthImage[out]: the resulting warped depth image
  *  \param getdepth[in]: say whether to get depth (true)
  *                       or zdepth from 0 to 1 (false). default = false
  */
  int Process(BIAS::Image<StorageType>& depthImage,
              BIAS::Image<StorageType>& warpedDepthImage,
              float minZ=500, float maxZ=7500.0, bool getdepth=false);

  /** \brief Process one warping from source to target parameters including color
   * texture information which has to have the same dimension as the depth images.
   * \param depthImage[in]: the depth image
   * \param colorImage[in]: the color image
   * \param warpedDepthImage[out]: the resulting warped depth image
   * \param warpedColorImage[out]: the resulting warped color image
   *  \param getdepth[in]: say whether to get depth (true)
   *                       or zdepth from 0 to 1 (false). default = false*/
  int ProcessWithColor(BIAS::Image<StorageType>& depthImage,
  		BIAS::Image<unsigned char>& colorImage,
  		BIAS::Image<StorageType>& warpedDepthImage,
  		BIAS::Image<unsigned char>& warpedColorImage,
  		float minZ, float maxZ, bool getdepth=false);
  
  /** \brief ppp extrinsicis have to be in global coordinates of pppd of Init call!   **/
  void UseProjectiveTexturing(const BIAS::ProjectionParametersPerspective& ppp){
  	depthMeshing_.UseProjectiveTexturing(ppp);
  };
  
  int SetTargetCamera(BIAS::ProjectionParametersPerspective& target,
                      float minZ=500, float maxZ=7500.0);
  
  void IgnoreRadialDistortion(bool ignoreRadialDistortion = true){
  	ignoreRadialDist_=ignoreRadialDistortion;
  };
  
  bool IsInitialized(){return  inited_;};
  
private:
  
	void SetDefaultTextureParameters_(BIAS::glfTexture2D& t);
  void InitFBOTexture_(BIAS::glfTexture2D& tex,unsigned width,
                       unsigned height,GLenum intFormat);

private:
  bool                         inited_,ignoreRadialDist_;
  unsigned                     width_,height_;
  BIAS::glfFramebufferObject   depthWarpFBO_;

  BIAS::glfTexture2D           resultTexture;
  BIAS::glfTexture2D           depthTexture;

  BIAS::PMDWarp                depthMeshing_;
  BIAS::glfPBuffer             pBuffer_;
  BIAS::ProjectionParametersPerspective source_,target_;

};
}
#endif
