/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __glfRenderingContext_hh__
#define __glfRenderingContext_hh__

#include <OpenGLFramework/Base/glfCommon.hh>
#include <Base/Image/Image.hh>

namespace BIAS {

  /**
   * Configuration for a rendering context.
   */
  struct BIASOpenGLFramework_EXPORT glfRenderingContextConfig
  {
    glfRenderingContextConfig();
    void Dump(std::ostream& out = std::cout);

    int   width;
    int   height;
    bool  doubleBuffer;
    int   redSize;
    int   greenSize;
    int   blueSize;
    int   alphaSize;
    int   depthSize;
    int   stencilSize;
  };

  /**
   * Interface for OpenGL rendering contexts.
   * \author 05/2007 jkollmann
   */
  class BIASOpenGLFramework_EXPORT glfRenderingContext {
  public:

    virtual ~glfRenderingContext() {}

    /**
     * Initializes the rendering context with the given configuration.
     * All implementations should call <code>MakeCurrent</code> and
     * <code>InitGlew</code> in this method. A context may be initialized
     * multiple times.
     * \param config Requested configuration
     */
    virtual void Init(const glfRenderingContextConfig& config) = 0;

    /**
     * Uninitializes the rendering context. This method must be called
     * before the application quits and after all Open GL resources
     * (textures etc.) have been deleted.
     */
    virtual void Destroy() = 0;

    /**
     * Makes this context the current target for OpenGL calls. Must be
     * called when switching between multiple rendering contexts.
     */
    virtual void MakeCurrent() = 0;

    /**
     * Removes the current state of this context. Must be
     * called before MakeCurrent is called from another thread.
     */
    virtual void DoneCurrent(){
     std::cerr << "WARNING: DoneCurrent() is not implemented. This function should be overloaded." << std::endl;
    };

    /**
     * Gets the currently used config of the rendering context. The rendering
     * context must have been successfully initialized before this method may
     * be called.
     * \param config The config structure to be filled
     */
    virtual void GetConfig(glfRenderingContextConfig& config) = 0;

    /**
     * Frame buffer data that can be retrieved by <code>Grab</code>.
     */
    enum GrabSource {
      GRAB_COLOR,   /**< Grab RGB image. */
      GRAB_DEPTH,   /**< Grab depth buffer. */
      GRAB_STENCIL  /**< Grab stencil buffer. */
    };

    /**
     * Copies the contents of the framebuffer to an image.
     * This method calls <code>MakeCurrent</code> on this rendering context.
     */
    void Grab(BIAS::Image<unsigned char>& image, GrabSource source = GRAB_COLOR,
              int x = 0, int y = 0, int width = 0, int height = 0);

  protected:

    /**
     * Initializes GLEW if enabled in the BIAS configuration. If not,
     * this method does nothing. Should be called during <code>Init</code> by
     * every <code>glfRenderingContext</code> implementation.
     */
    void InitGlew();
  };
  
} // namespace BIAS

#endif // __glfRenderingContext_hh__
