/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C)  2008, 2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <OpenGLFramework/Utils/Texture2DPool.hh>

using namespace BIAS;
using namespace std;

Texture2DPool::~Texture2DPool()
{
  ClearAll();
}

BIAS::glfTexture2D*
Texture2DPool::operator[](const std::string& name)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      return it->second;
    }
  else
    {
      BIASWARN("pool does not contain texture named: " << name);
      return NULL;

    }
}

/** Allocates an empty texture **/
BIAS::glfTexture2D*
Texture2DPool::Create(const std::string& name, int width, int height,
    GLenum internalFormat, GLenum minFilter, GLenum magFilter, GLenum wrapS,
    GLenum wrapT, int mipmap)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      BIASERR("already used pool name : " << name
          << ". use clear before reusing the name!");
      ClearAll();
      BIASABORT;
    }
  glfTexture2D* newTexture(NULL);
  texturePool_[name] = newTexture = new glfTexture2D;
  newTexture->Create();
  newTexture->Set(minFilter, magFilter, wrapS, wrapT);
  newTexture->Allocate(width, height, internalFormat, mipmap);
  return newTexture;
}

/** Uploads BIAS image to texture space **/
BIAS::glfTexture2D*
Texture2DPool::Create(const std::string& name, const BIAS::ImageBase& img,
    GLenum minFilter, GLenum magFilter, GLenum wrapS, GLenum wrapT,
    GLenum internalFormat, int mipmap)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      BIASERR("already used pool name : " << name
          << ". use clear before reusing the name!");
      ClearAll();
      BIASABORT;
    }
  glfTexture2D* newTexture(NULL);
  texturePool_[name] = newTexture = new glfTexture2D;
  newTexture->Create();
  newTexture->Set(minFilter, magFilter, wrapS, wrapT);
  newTexture->UploadImage(img, internalFormat, mipmap);
  return newTexture;

}

/** Allocates an empty texture **/
glfTexture2D*
Texture2DPool::Create(const string& name, int width, int height,
    ImageBase::EStorageType storageType, ImageBase::EColorModel colorModel,
    GLenum minFilter, GLenum magFilter, GLenum wrapS, GLenum wrapT, int mipmap)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      BIASERR("already used pool name : " << name
          << ". use clear before reusing the name!");
      ClearAll();
      BIASABORT;
    }
  glfTexture2D* newTexture(NULL);
  texturePool_[name] = newTexture = new glfTexture2D;
  newTexture->Create();
  newTexture->Set(minFilter, magFilter, wrapS, wrapT);
  newTexture->Allocate(width, height, storageType, colorModel, mipmap);
  return newTexture;
}

glfTexture2D*
Texture2DPool::UploadImage(const string& name, const BIAS::ImageBase& img,
    GLenum minFilter, GLenum magFilter, GLenum wrapS, GLenum wrapT,
    GLenum internalFormat, int mipmap)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      glfTexture2D* textureP = it->second;
      textureP->UploadImage(img, internalFormat, mipmap);
      return textureP;
    }
  else
    {
      glfTexture2D* textureP = new glfTexture2D();
      textureP->Create();
      textureP->Set(minFilter, magFilter, wrapS, wrapT, GL_TEXTURE0);
      textureP->UploadImage(img, internalFormat, mipmap);
      texturePool_[name] = textureP;
      return textureP;
    }
}

void
Texture2DPool::ClearAll()
{
  map<string, glfTexture2D*>::iterator it = texturePool_.begin();
  for (; it != texturePool_.end(); it++)
    {
      delete it->second;
      it->second = NULL;
    }
  texturePool_.clear();
}

void
Texture2DPool::Clear(const std::string& name)
{
  map<string, glfTexture2D*>::iterator it = texturePool_.find(name);
  if (it != texturePool_.end())
    {
      delete it->second;
      texturePool_.erase(it);
    }
}

