/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __DPPanTiltControl_hh__
#define __DPPanTiltControl_hh__

#include <bias_config.h>

#include <stdio.h>
#ifndef WIN32
#  include <unistd.h>
#endif
#include <iostream>
#include <vector>

// Include headers for Directed Perception PTU
extern "C"
{
#include "include/ptu.h"
#ifndef WIN32
#  include "include/linuxser.h"
#else
#  include "include/w32seria.h"
#endif
}

// @note Call QueryResolution to get actual pan/tilt resolution
//       because it depends on the PTU model and speed settings!

#define DPPTU_DEF_RES  0.051428
#define DPPTU_MAXSPEED 2901

#define DPPTU_MINPAN -3000
#define DPPTU_MAXPAN  3000
#define DPPTU_MINTILT -500
#define DPPTU_MAXTILT  300

#define SPEEDMODE      0
#define ABSOLUTEMODE   1

namespace BIAS
{
  /**
     @class DPPanTiltControl
     @brief Basic controller interface for Directed Perception Pan Tilt Unit.
     @author MIP
  */ 
  class BIASPanTilt_EXPORT DPPanTiltControl
  {
  public:

    enum DPPanTiltStepping {
      AUTO = 0,
      HALF,
      QUARTER,
      EIGHTH,
      FULL
    };

    /** @brief Standard constructor */
    DPPanTiltControl();

    /** @brief Standard destructor */
    ~DPPanTiltControl();
    
    /** @brief Set the mode in which the camera is to operate
        @param[in] nMode Defines the mode, 0 is speed control mode,
                         1 is absolute positioning mode
        @return Returns 0 on success, != 0 for errors */
    int SetMode(int nMode);

    /** @brief Reset the position of the pan tilt unit to 0.
        @return Returns true on success, false for errors */
    bool ResetPanTiltUnit();
    
    /** @brief Stop the currently moving pan tilt unit. */
    void StopPanTiltUnit();

    /** @brief Await completion of currently executed command. */
    void AwaitCompletion();

    /** @brief Open the given serial port and initialize the pan tilt unit.
        @param[in] nPort The number of the serial port (or USB port)
        @param[in] bUSB  Use USB port instead of serial port
        @return Returns true on success, false for errors */
    bool InitPanTiltUnit(int nPort, bool bUSB = false);

    /** @brief Open the given port and initialize the pan tilt unit.
        @param[in] portName The name of the serial port or USB port to open,
                   e.g. /dev/ttyUSB0
        @return Returns true on success, false for errors */
    bool InitPanTiltUnit(const char *portName);
    
    /** @brief closes the comport stream connected to PTU */
    void ClosePanTiltUnit();
    
    /** @brief Move camera nValue steps upwards */
    void MoveUp(int nValue);

    /** @brief Move camera nValue steps downwards */
    void MoveDown(int nValue);

    /** @brief Move camera nValue steps left */
    void MoveLeft(int nValue);

    /** @brief Move camera nValue steps right */
    void MoveRight(int nValue);

    /** @brief Start a thread moving the PTU in a circle */
    void StartCircularMovement();

    /** @brief Set speed for movement */
    char SetSpeed(int nPanSpeed,int nTiltSpeed);

    /** @brief Set acceleration for movement */
    char SetAcceleration(int nPanAcc, int nTiltAcc);

    /** @brief Get the resolution of the pan axis
        @return Returns the pan resolution in radians per step */
    double GetPanResolution() { return panResolution_; }

    /** @brief Get the resolution of the tilt axis
        @return Returns the tilt resolution in radians per step */
    double GetTiltResolution() { return tiltResolution_; }
   
    /** @return Returns the current speed */
    void GetSpeed(long &panSpeed, long &tiltSpeed);

    /** @return Returns the current acceleration */
    void GetAcceleration(long &panAcc, long &tiltAcc);

    /** @return Returns the current position of the PTU in steps */
    void GetCurrentPosition(long &lPanPos, long &lTiltPos);

    /** @return Returns the current position of the PTU in degrees */
    void GetCurrentPositionDegrees(double &dPanDeg, double &dTiltDeg);

    /** @return Returns the current position of the PTU in radians */
    void GetCurrentPositionRadians(double &dPanRad, double &dTiltRad);

    /** @brief Set the current position of the PTU in steps */
    void SetPosition(int nPanPos, int nTiltPos, bool bWaitComplete = true);

    /** @brief Send a command string terminated by a space or carriage
               return character to the serial (or USB) port. */
    void SendRawCommand(std::string cmd);

    /** @brief Read a line from the serial (or USB) port.
        @return Returns the length of the read string, or 0 when nothing
                has been read */
    int ReadRawLine(std::string &cmd);
	
    /** @brief Set the stepping of the motor. */
    int SetMotorStepping(int panStepping, int tiltStepping);

    /** @brief Get resolution of the motor.
        @param[out] panResolution  Return the pan resolution
        @param[out] tiltResolution Returns the tilt resolution */
    void QueryResolution(double &panResolution, double &tiltResolution,
                         bool updateInternal = false);

    /** @brief Set the orientation of the PTU (default or upside-down).
        @param[in] use Use true for default orientation and false for mounting
                       upside down (e.g. when mounted under helicopter)
        @note Implemented for helicopter project by Ingo Schiller */
    inline void UseDefaultOrientation(bool use) { bDefaultOrientation_ = use; }

  protected:

    portstream_fd COMstream_;
    short int maxSpeed_;
    double panResolution_, tiltResolution_;
    bool bDefaultOrientation_;

  };

}

#endif // __DPPanTiltControl_hh__
