/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef WIN32

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <termios.h>

#else
#include <windows.h>
#endif

#include <iostream>
#include <sstream>
#include <stdio.h>
#include "ViscaControl.hh"
#include <Base/Common/W32Compat.hh>

using namespace std;
using namespace BIAS;

ViscaControl::
ViscaControl()
{
  connected_ = false;
  zoomed_ = false;
  used_ = false;
  maxZoom_ = 31424;
  //maxZoom_ = 1023;

  // Set default pan and tilt speed for units without speed support.
  // WARNING: Sending any other speed values to those units will
  // result in an engine lock!
  panSpeed_ = 0x18;
  tiltSpeed_ = 0x1C;

  // Set default active unit
  activeUnit_ = 0x81;
}

ViscaControl::
~ViscaControl()
{}

bool ViscaControl::
InitPanTiltUnit(std::string port)
{
  bool ret = false;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_init(&camMutex,NULL);
#endif
  //init pointers to the send and receive buffers
  sendPtr=sendBuffer;

	//Windows: "\\\\.\\COM3"
	cout<<"trying to open "<<port<<" for serial communication...";
  ret = serialIO_.OpenPort(port);
  if(!ret){ BIASERR("Could not open serial port!") return false; }
  serialIO_.SetOptions(9600,SPPM_NoParity, false,false,false);
  connected_ = true;
 
  // init the lens and start ccd pixel correction
  CamInitialize_();
  return true;
}

bool ViscaControl::ClosePanTiltUnit()
{
  // turn the camera off
  bool ret = serialIO_.ClosePort();
  connected_=false;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_destroy(&camMutex);
#endif
  return ret;
}


int ViscaControl::
SendCommand_(unsigned char* buf, int length,
						 bool manageAnswer)
{

	if(!connected_) return -1;
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_lock(&camMutex);
#endif
  serialIO_.WriteBytes(buf,length);
  if(manageAnswer)ManageAnswer();
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_unlock(&camMutex);
#endif
  return 0;
}


void ViscaControl::ContinuousRead(){
  while(true) {
    biasmsleep(500);
    ManageAnswer();
  }
}

void ViscaControl::ManageAnswer()
{
  int length = 40;
  //clear receiveBuffer to delete memory
  unsigned char* receiveBuffer= new unsigned char[length];

  for (int i=0;i<length;i++){ receiveBuffer[i]=0; }

  cout<<"Reading answer as reply..";
  int ret = serialIO_.ReadBytes(receiveBuffer,length,true);
  if(ret < 0){BIASERR("Could not read answer")}
  else cout<<"done, "<<ret<<" Bytes!"<<endl;

  cout<<"MSG:";
  for (int i=0;i<ret;i++){
    char temp[8];
    sprintf(temp,"0x%02X, ", receiveBuffer[i]);
    cout<<temp<<" ";
  }
  cout<<endl;

  if (receiveBuffer[0]==0x90 && receiveBuffer[1]==0x41 && receiveBuffer[2]==0xFF) {
    cout << "ACK!" << endl;
  }
  if (receiveBuffer[1]==0x61)
  {
    if(receiveBuffer[2]==0x01){ BIASERR("Message length error.")}
    else if(receiveBuffer[2]==0x02){ BIASERR("Syntax error.")}
    else if(receiveBuffer[2]==0x03){ BIASERR("Command buffer full.")}
    else if(receiveBuffer[2]==0x04){ BIASERR("Command cancelled.")}
    else if(receiveBuffer[2]==0x05){ BIASERR("No socket (to be cancelled).")}
    else if(receiveBuffer[2]==0x41){ BIASERR("Command not executable.")}
  }
  if(receiveBuffer[3]==0x90 && receiveBuffer[4]==0x51 && receiveBuffer[5]==0xFF) {
    cout << "COMMAND COMPLETED!" << endl;
  }
  else{
  	cout<<"Uninterpreted reply:"<<endl;
  }

  //Pt Position / ZOOM Inquiry Management:
  //if ptpos was requested store it
  //if zoompos was requested store it
  if(receiveBuffer[0]==0x90 && receiveBuffer[1]==0x50){
    if(!zoomed_){
      pan_=0;
      pan_|=(short)receiveBuffer[2]<<12;
      pan_|=(short)receiveBuffer[3]<<8;
      pan_|=(short)receiveBuffer[4]<<4;
      pan_|=(short)receiveBuffer[5];
      cout << "pan_ " << pan_ << endl;

      tilt_=0;
      tilt_|=(short)receiveBuffer[6]<<12;
      tilt_|=(short)receiveBuffer[7]<<8;
      tilt_|=(short)receiveBuffer[8]<<4;
      tilt_|=(short)receiveBuffer[9];
      cout << "tilt_ " << tilt_ << endl;
      used_ = false;
    }
    else {
      //zoom pos was requested
      zoom_=0;
      zoom_|=(int)receiveBuffer[2]<<12;
      zoom_|=(int)receiveBuffer[3]<<8;
      zoom_|=(int)receiveBuffer[4]<<4;
      zoom_|=(int)receiveBuffer[5];
      cout << "zoom int " << zoom_ << endl;

      zoomed_ = false;
      used_ = false;
    }
  }
  delete[] receiveBuffer;
}


void ViscaControl::MoveRelative(short relPan, short relTilt)
{
  sendBuffer[0] =  activeUnit_;
  sendBuffer[1] =  0x01;
  sendBuffer[2] =  0x06;
  sendBuffer[3] =  0x03;
  sendBuffer[4] =  panSpeed_;
  sendBuffer[5] =  tiltSpeed_;
  sendBuffer[6] =  (unsigned char)((relPan >> 12) & 0x0F);
  sendBuffer[7] =  (unsigned char)((relPan >> 8) & 0x0F);
  sendBuffer[8] =  (unsigned char)((relPan >> 4) & 0x0F);
  sendBuffer[9] =  (unsigned char)(relPan & 0x0F);
  sendBuffer[10]=  (unsigned char)((relTilt >> 12) & 0x0F);
  sendBuffer[11]=  (unsigned char)((relTilt >> 8) & 0x0F);
  sendBuffer[12]= (unsigned char)((relTilt >> 4) & 0x0F);
  sendBuffer[13]=  (unsigned char)(relTilt & 0x0F);
  sendBuffer[14]= 0xFF;

  this->SendCommand_(sendBuffer,15);
}

void ViscaControl::MoveAbsolute(short panPos, short tiltPos)
{
  sendBuffer[0] = activeUnit_;
  sendBuffer[1] = 0x01;
  sendBuffer[2] = 0x06;
  sendBuffer[3] = 0x02;
  sendBuffer[4] = panSpeed_;
  sendBuffer[5] = tiltSpeed_;
  sendBuffer[6] = (unsigned char)((panPos >> 12)& 0x0F);
  sendBuffer[7] = (unsigned char)((panPos>> 8)& 0x0F);
  sendBuffer[8] = (unsigned char)((panPos>> 4) & 0x0F);
  sendBuffer[9] = (unsigned char)(panPos & 0x000F);
  sendBuffer[10]= (unsigned char)((tiltPos >> 12)& 0x0F);
  sendBuffer[11]= (unsigned char)((tiltPos >> 8)& 0x0F);
  sendBuffer[12]= (unsigned char)((tiltPos >> 4)& 0x0F);
  sendBuffer[13]= (unsigned char)(tiltPos & 0x0F);
  sendBuffer[14]= 0xFF;

  this->SendCommand_(sendBuffer,15);
}

void ViscaControl::MovePhysical(short panSteps, short tiltSteps)
{
  sendBuffer[0] = activeUnit_;
  sendBuffer[1] = 0x01;
  sendBuffer[2] = 0x06;
  sendBuffer[3] = 0x08;
  sendBuffer[4] = panSpeed_;
  sendBuffer[5] = tiltSpeed_;
  sendBuffer[6] = (unsigned char)((panSteps >> 12)& 0x0F);
  sendBuffer[7] = (unsigned char)((panSteps >> 8)& 0x0F);
  sendBuffer[8] = (unsigned char)((panSteps >> 4)& 0x0F);
  sendBuffer[9] = (unsigned char)(panSteps & 0x000F);
  sendBuffer[10]= (unsigned char)((tiltSteps >> 12)& 0x0F);
  sendBuffer[11]= (unsigned char)((tiltSteps >> 8)& 0x0F);
  sendBuffer[12]= (unsigned char)((tiltSteps >> 4)& 0x0F);
  sendBuffer[13]= (unsigned char)(tiltSteps & 0x0F);
  sendBuffer[14]= 0xFF;

  this->SendCommand_(sendBuffer,15);

}

void ViscaControl::LightOn() {

	sendBuffer[0] = activeUnit_;
	sendBuffer[1] = 0x01;
	sendBuffer[2] = 0x04;
	sendBuffer[3] = 0x02;
	sendBuffer[4] = 0x03;
	sendBuffer[5] = 0x02;
	sendBuffer[6] = 0x01;
	sendBuffer[7] = 0xFF;

	this->SendCommand_(sendBuffer, 8);
}


void ViscaControl::LightOff() {
	
	sendBuffer[0] = activeUnit_;
	sendBuffer[1] = 0x01;
	sendBuffer[2] = 0x04;
	sendBuffer[3] = 0x02;
	sendBuffer[4] = 0x03;
	sendBuffer[5] = 0x03;
	sendBuffer[6] = 0x01;
	sendBuffer[7] = 0xFF;

	this->SendCommand_(sendBuffer, 8);
}


void ViscaControl::Wash() {

  sendBuffer[0] = activeUnit_;
  sendBuffer[1] = 0x01;
  sendBuffer[2] = 0x04;
  sendBuffer[3] = 0x02;
  sendBuffer[4] = 0x02;
  sendBuffer[5] = 0x01; // Timer / On / Off
  sendBuffer[6] = 0x01; // Set timer (seconds)
  sendBuffer[7] = 0xFF;

	this->SendCommand_(sendBuffer, 8);
}


void ViscaControl::Wipe() {
	
  sendBuffer[0] = activeUnit_;
  sendBuffer[1] = 0x01;
  sendBuffer[2] = 0x05;
  sendBuffer[3] = 0x06;
  sendBuffer[4] = 0x01;
  sendBuffer[5] = 0x01; // Intervals / On / Off  
  sendBuffer[6] = 0x01; // Set number of intervals
  sendBuffer[7] = 0xFF;

	this->SendCommand_(sendBuffer, 8);
}



void ViscaControl::CamPower(bool mode)
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x00;

  if(mode) { sendBuffer[4]=0x02; } // POWER ON
  else     { sendBuffer[4]=0x03; } // POWER OFF

  sendBuffer[5]= 0xFF;

  this->SendCommand_(sendBuffer,6);
}

int ViscaControl::CamInitialize_()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x19;
  sendBuffer[4]= 0x01; // Lens Initialisation Start
  sendBuffer[5]= 0xFF;

  this->SendCommand_(sendBuffer,6);

  sendBuffer[4]=0x02; // Correction of CCD pixel blemishes

  return SendCommand_(sendBuffer,6);
}

void ViscaControl::PanTiltPosInq()
{
  sendBuffer[0]= activeUnit_; //PANTILTPHYPOSING
  sendBuffer[1]= 0x09;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x12; //0x15 fuer PTPhy
  sendBuffer[4]= 0xFF;  
  this->SendCommand_(sendBuffer,5);
}

void ViscaControl::MoveHome()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x04;
  sendBuffer[4]= 0xFF;
  this->SendCommand_(sendBuffer,5);
}


void ViscaControl::AdressSet()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x30;
  sendBuffer[2]= 0x01;
  sendBuffer[3]= 0xFF;

  this->SendCommand_(sendBuffer,4);
}

void ViscaControl::MoveLeft()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
  sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x01;
  sendBuffer[7]= 0x03;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveRight()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x02;
  sendBuffer[7]= 0x03;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}
void ViscaControl::MoveUp()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x03;
  sendBuffer[7]= 0x01;
  sendBuffer[8]= 0xFF;
  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveUpLeft()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x01;
  sendBuffer[7]= 0x01;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveUpRight()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x02;
  sendBuffer[7]= 0x01;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveDown()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x03;
  sendBuffer[7]= 0x02;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveDownLeft()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x01;
  sendBuffer[7]= 0x02;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveDownRight()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x02;
  sendBuffer[7]= 0x02;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}

void ViscaControl::MoveStop()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x01;
	sendBuffer[4]= panSpeed_;
  sendBuffer[5]= tiltSpeed_;
  sendBuffer[6]= 0x03;
  sendBuffer[7]= 0x03;
  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}


void ViscaControl::ZoomTele()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x07;
  sendBuffer[4]= 0x02;
  sendBuffer[5]= 0xFF;

  this->SendCommand_(sendBuffer,6);

  ZoomInq();
}

void ViscaControl::ZoomWide()
{
  sendBuffer[0]= activeUnit_;
	sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x07;
  sendBuffer[4]= 0x03;
  sendBuffer[5]= 0xFF;

  this->SendCommand_(sendBuffer,6);
}

void ViscaControl::ZoomStop()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x07;
  sendBuffer[4]= 0x00;
  sendBuffer[5]= 0xFF;

  zoomed_ = true;

  this->SendCommand_(sendBuffer,6);
}

void ViscaControl::ResetPanTilt()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x06;
  sendBuffer[3]= 0x05;
  sendBuffer[4]= 0xFF;

  this->SendCommand_(sendBuffer,5);
}

void ViscaControl::ResetCamCustom()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x3F;
  sendBuffer[4]= 0x00;
  sendBuffer[5]= 0x7F;
  sendBuffer[6]= 0xFF;

  this->SendCommand_(sendBuffer,7);
}

void ViscaControl::ZoomInq()
{
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x09;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x47;
  sendBuffer[4]= 0xFF;

  zoomed_=true;

  this->SendCommand_(sendBuffer,5);
}


void ViscaControl::ZoomDirect(short zoom)
{
  if(zoom>maxZoom_){
    zoom=(short)maxZoom_;
  }
  if(zoom<0){
    zoom=0;
  }

  cout<<"Zoom:"<<zoom<<endl;
  sendBuffer[0]= activeUnit_;
  sendBuffer[1]= 0x01;
  sendBuffer[2]= 0x04;
  sendBuffer[3]= 0x47;

  sendBuffer[4]= (unsigned char)((zoom >> 12)& 0x0F);
  sendBuffer[5]= (unsigned char)((zoom >> 8)& 0x0F);
  sendBuffer[6]= (unsigned char)((zoom >> 4)& 0x0F);
  sendBuffer[7]= (unsigned char)(zoom & 0x0F);

  sendBuffer[8]= 0xFF;

  this->SendCommand_(sendBuffer,9);
}



bool ViscaControl::SetCCDScanningMode(VISCA_CONSTANT mode){
	//register 72, mode 0=interlaced, 1=progresive
	sendBuffer[0]= activeUnit_;
	sendBuffer[1]= 0x01;
	sendBuffer[2]= 0x04;
	sendBuffer[3]= 0x24;
	sendBuffer[4]= 0x72;
	sendBuffer[5]= 0x00;

	if(mode == VISCA_INTERLACED){
			sendBuffer[6]= 0x00;
	}
	else if(mode == VISCA_PROGRESSIVE){
		sendBuffer[6]= 0x01;
	}

	sendBuffer[7]= 0xFF;
	return this->SendCommand_(sendBuffer,8) != 0;
}

ViscaControl::VISCA_CONSTANT ViscaControl::GetCCDScanningMode(){
	sendBuffer[0]= activeUnit_;
	sendBuffer[1]= 0x09;
	sendBuffer[2]= 0x04;
	sendBuffer[3]= 0x24;
	sendBuffer[4]= 0x72;
	sendBuffer[5]= 0xFF;
	this->SendCommand_(sendBuffer,6,false);
	ViscaControl::VISCA_CONSTANT result;
	int length = 20;
	unsigned char* receiveBuffer= new unsigned char[length];
	for (int i=0;i<length;i++) receiveBuffer[i]=0;

	bool ret = serialIO_.ReadBytes(receiveBuffer,length,true) != 0;
	if(!ret)BIASERR("Could not read answer");

	if (receiveBuffer[2]==0x00 && receiveBuffer[3]==0x00) {
		result= VISCA_INTERLACED;
	}
	else if(receiveBuffer[2]==0x00 && receiveBuffer[3]==0x01) {
		result= VISCA_PROGRESSIVE;
	}
	else{
		BIASERR("Unknown CCD Scanning Mode!");
		result= VISCA_ERROR;
	}
  delete[] receiveBuffer;
  return result;
}





