/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __ViscaControl_hh__
#define __ViscaControl_hh__

#include <bias_config.h>
#ifdef BIAS_HAVE_PTHREADS
	#include <pthread.h>
#endif
#include <string>
#ifdef WIN32
#include <windows.h>
#include <Base/Common/W32Compat.hh>
#endif
// BIAS
#include <Utils/SerialPortIO.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Debug/Error.hh>

namespace BIAS {

/**
 * \class ViscaControl
 * \brief for use with Sony PanTilt Cameras, also compatible with
 * 	WISKA PanTilt Cameras. This is used to control a camera settings.
 *  It uses the class SerialPortIO to send commands and receive answers.
 *  Also included are some functions of the Camera (MoveUp...), they can
 *  be extended as needed (for a full list of functions see the reference
 *  manuals for the parts).
 *  The commands always start with 8X where X is the camera number, so 1 is the first camera!!
 *  and X \element of {1,...,7}. A broadcast is done with 88.
 *  Replies start with 9X and 0!! (unequal to request command!)
 * \author  ischiller, fnommensen, mfranke
 */
class BIASPanTilt_EXPORT ViscaControl: public BIAS::Debug
{

public:
	enum VISCA_CONSTANT{
		VISCA_ERROR=-1,
		VISCA_INTERLACED=0,
		VISCA_PROGRESSIVE
	};

  ViscaControl();
  ~ViscaControl();

  bool InitPanTiltUnit(std::string port); //PowerOn+Cam calibration etc

  // Moves the PanTilt unit in various directions
  // until MoveStop() is called. Use with caution!
  void MoveUp();
  void MoveUpLeft();
  void MoveUpRight();
  void MoveDown();
  void MoveDownLeft();
  void MoveDownRight();
  void MoveLeft();
  void MoveRight();
  void MoveStop();
  
  void LightOn();
  void LightOff();

  void Wash();
  void Wipe();

  void ZoomTele();
  void ZoomWide();
  void ZoomStop();

  void MoveRelative(short relPan, short relTilt);
  void MoveAbsolute(short panPos, short tiltPos);
  void MovePhysical(short panSteps, short tiltSteps);

  void CamPower(bool mode);
  bool ClosePanTiltUnit();
  void PanTiltPosInq();
  void MoveHome();

	// Change pan and tilt speed for units with speed support.
	// WARNING: Sending other values than the defaults to units
	// without speed support will result in an engine lock!
	void SetPanSpeed(unsigned char panSpeed) {
		panSpeed_ = panSpeed;
	}
	void SetTiltSpeed(unsigned char tiltSpeed) {
		tiltSpeed_ = tiltSpeed;
	}

	// Change active unit (default: 0x81)
	void SetActiveUnit(unsigned char activeUnit) {
		activeUnit_ = activeUnit;
	}

  /** Answers start with X0, where X is the camera number +8, e.g. 90 is camera 1 */
  void ManageAnswer();

  void AdressSet();
  void WaitForAck();
  void ResetPanTilt();
  void ResetCamCustom();
  void ZoomInq();
  void ZoomDirect(short zoom);

  bool IsConnected(){return connected_;}
  bool SetCCDScanningMode(VISCA_CONSTANT mode);

  ViscaControl::VISCA_CONSTANT GetCCDScanningMode();
  int GetMaxZoom(){return maxZoom_;}
  int GetPan(){return pan_;}
  int GetTilt(){return tilt_;}
  int GetZoom(){return zoom_;}
  void ContinuousRead();

  // values for pan, tilt and zoom
  short int pan_,tilt_,zoom_;
  // have the values been used yet?
  bool used_;

protected:
  /** adresses of cameras start with 1, eg, 0x81 is first cam
      the reply starts with 0 on 90, e.g. 0.90 is first cam */
  int SendCommand_(unsigned char* buf,int length,
			 	 	 	 	 	 	 bool manageAnswer=true);
  int CamInitialize_();

protected:
#ifdef BIAS_HAVE_PTHREADS
  pthread_mutex_t camMutex; 
#endif
  bool zoomed_;
  bool connected_;
  int maxZoom_;
  unsigned char sendBuffer[16];
  unsigned char* sendPtr;

	unsigned char panSpeed_;
	unsigned char tiltSpeed_;
	unsigned char activeUnit_;

	BIAS::SerialPortIO serialIO_;
};

}//end namespace
#endif
