/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Utils/Param.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>
#include <sstream>

using namespace std;
using namespace BIAS;

/**
   @file
   @ingroup g_tools
   @brief Convert input image format to output image format preserving timestamps see biasConvertMDtimestamps.cpp
   @author MIP
*/
int convertImage(string &input, string &output) {

  //tmp input image
  ImageBase img;

  // load input image, cancel programm if image read fails
  cout << "Loading " << input << endl << flush;
  if (ImageIO::Load(input,img) < 0) {
    cerr << "Error loading image!" << endl << flush;
    return -1;
  }

  //convert image to BIAS::Camera format by loading, parsing meta data using
  //old format, update with new format and writing back
  switch (img.GetStorageType()) {
  case ImageBase::ST_invalid :
    cerr << "Invalid storage type for image: " << input << endl << flush;
    return -1;
    break;
  case ImageBase::ST_unsignedchar : 
    {
      Camera<unsigned char> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#ifdef BUILD_IMAGE_CHAR
  case ImageBase::ST_char : 
    {
      Camera<char> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_CHAR
#ifdef BUILD_IMAGE_USHORT
  case ImageBase::ST_unsignedshortint :
    {
      Camera<unsigned short int> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_USHORT
#ifdef BUILD_IMAGE_SHORT
  case ImageBase::ST_shortint :
    {
      Camera<short int> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_UINT
#ifdef BUILD_IMAGE_UINT
  case ImageBase::ST_unsignedint :
    {
      Camera<unsigned int> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_UINT
#ifdef BUILD_IMAGE_INT
  case ImageBase::ST_int :
    {
      Camera<int> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_INT
  case ImageBase::ST_float :
    {
      Camera<float> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#ifdef BUILD_IMAGE_DOUBLE
  case ImageBase::ST_double :
    {
      Camera<double> cam(img);
      cam.ParseMetaData(false);
      cam.UpdateMetaData();
      if (ImageIO::Save(output,cam) < 0) {
        cerr << "Error writing file " << output << endl;
        return -1;
      }
    }
    break;
#endif //BUILD_IMAGE_DOUBLE
  default:
    cerr << "unsupported storage type" << endl;
    break;
  }
  cout << "Processed image: " << input << endl;
  cout << "       wrote to: " << output << endl;

  return 0;
}

int main(int argc, char* args[]) {

  //prepare and parse parameters
  Param parameters;
  parameters.AddParamString("inputlist",
                            "Give filename for input filelist (txt-format with images seperated by whitespaces)",
                            "",'I', GRP_ALL);
  parameters.AddParamString("inputfile", "Give filename for input file","",'i', GRP_ALL);
  parameters.AddParamString("outputlist",
                            "Give filename for output filelist (txt-format with images seperated by whitespaces)",
                            "",'O', GRP_ALL);
  parameters.AddParamString("outputfile", "Give filename for output file","",'o', GRP_ALL);
  parameters.ParseCommandLine (argc, args);
  bool bUseList = false;
  bool bUseFile = false;
  string *in_filelist = NULL;
  in_filelist = parameters.GetParamString("inputlist");
  string *in_file = NULL;
  in_file = parameters.GetParamString("inputfile");
  string *out_filelist = NULL;
  out_filelist = parameters.GetParamString("outputlist");
  string *out_file = NULL;
  out_file = parameters.GetParamString("outputfile");
  if ((*in_filelist != "")&&(*out_filelist != "")) {
    bUseList = true;
  }
  if ((*in_file != "")&&(*out_file != "")) {
    bUseFile = true;
  }

  //no filename or filelist given
  if (!bUseFile && !bUseList) {
    cerr << "Give either in-/output file or in-/output filelist (--help for options)!" << endl;
    exit(-1);
  }
  
  //convert single img
  if (bUseFile) {
    string inFile(*in_file);
    string outFile(*out_file);
    if (convertImage(inFile, outFile) < 0) {
      exit(-1);
    }
  }

  //convert img list
  if (bUseList) {
    string inFile("");
    string outFile("");
    ifstream inlist(in_filelist->c_str());
    ifstream outlist(out_filelist->c_str());
    while((inlist >> inFile) && (outlist >> outFile)) {
      if (convertImage(inFile, outFile) < 0)
        exit(-1);
      inFile.clear();
      outFile.clear();
    }
  }

  cout << "finished processing" << endl << flush;    
}
