#include <Base/Common/BIASpragma.hh>
#include <bias_config.h>

#include <stdio.h>
#include <map>
#include <string>

#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Geometry/Projection.hh>
#include <Utils/TriangleMesh.hh>
#include <Utils/ThreeDOut.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Math/Vector3.hh>
#include <Filter/Gauss.hh>

#ifdef BIAS_HAVE_OPENSCENEGRAPH
	#include <GLviewer/ThreeDOutOpenSceneGraph.hh>
#endif

using namespace BIAS;
using namespace std;

//forward declarations
void usage(int argc, char* argv[]);
void GenerateCheckerboard(int size,int xCorners, int yCorners);
void GenerateIntensityTestCheckerboard(int size,int xCorners, int yCorners,float maxDev);


/**
    @file biasGenerateCheckerboardModel
    @ingroup g_tools
    @brief generates a checkerboard vrml file
    @author ischiller
*/
int main (int argc, char *argv[])
{
  if(argc<4){
    usage(argc, argv);
    return -1;
  }

  bool intensity = false;
  float maxDev = 0.0;

  int cols = atoi(argv[1]);
  int rows = atoi(argv[2]);
  int size = atoi(argv[3]);

  if(argc > 4) intensity = bool(atoi(argv[4]) != 0);
  if(argc > 5) maxDev = atof(argv[5]);


  if(intensity){
     cout<<"Intensity with MaxDev:"<<maxDev<<endl;
    GenerateIntensityTestCheckerboard(size,cols, rows, maxDev);
  }
  else{
    cout<<"Standard:"<<endl;
    GenerateCheckerboard(size,cols, rows);
  }
}
///////////////////////////////////////////////////////////////////////
void usage(int argc, char* argv[])
{
  cout<<"This program will create a checkerboard model and write it to a file checkerboard.wrl. "<<
      "If OpenSceneGraph support is activated in BIAS it will also write corresponding files with "<<
      "suffixes .ive, .3ds .osg\n";
  cout<<"ATTENTION: The ThreeDOut.AddImage() method has a bug therefore the textures in models .ive, .3ds and "<<
      ".osg models are wrong. (ischiller 29/06/2011).\n\n";
  cout<<"Usage:"<<argv[0]<<" cols rows size [intensityVariable] [maxDeviation]\n"<<endl;
}
///////////////////////////////////////////////////////////////////////
void GenerateCheckerboard(int size,int xCorners, int yCorners){

  BIAS::Vector3<unsigned int> UL,UR,LL,LR;
  BIAS::Image<unsigned char> black,white;
  BIAS::Image<unsigned char> image;
  unsigned int factor = 20;
  BIAS::ThreeDOut threedout;
  Vector3<double> ul(-size,-size,0);
  Vector3<double> ur((xCorners)*size,-size,0);
  Vector3<double> ll(-size,(yCorners)*size,0);
  Vector3<double> lr((xCorners)*size,(yCorners)*size,0);


  black.Init(factor,factor,1); 
  black.FillImageWithConstValue((unsigned char)0);  
  white.Init(factor,factor,1); 
  white.FillImageWithConstValue((unsigned char)255);

  // The checkerboard image  
  if(!image.IsEmpty())  image.Release();
  unsigned newWidth=(unsigned)((xCorners+1)*factor);
  unsigned newHeight=(unsigned)((yCorners+1)*factor);
 
  image.Init(newWidth,newHeight,1);
  
  bool topLeftBlack = true;
  bool isBlack = topLeftBlack;  

  for(int y=0;y < yCorners+1; y++){ //rows
    for(int x= 0;x< xCorners+1 ; x++){   //cols 
   
      if((y % 2) == 0 && x == 0)
        isBlack = topLeftBlack;      
      else if(x == 0)
        isBlack = !topLeftBlack;
                
      UL[0]=x*factor;
      UL[1]=y*factor;
      UL[2]=0;
      LR[0]=(x+1)*factor;
      LR[1]=(y+1)*factor;
      LR[2]=0;

      image.SetROICorners(UL[0],UL[1], LR[0],LR[1]);

      if(isBlack)             
        image.Paste2ROI(black);
      else
        image.Paste2ROI(white);
      
      isBlack = !isBlack;
    }
  }


  BIAS::ImageIO::Save("checkerboard.jpg",image);

  unsigned char *pData = image.GetImageData();
  unsigned xSize = (unsigned)rint(image.GetWidth());
  unsigned ySize = (unsigned)rint(image.GetHeight());
  threedout.AddImage(ul, ur,ll,lr,xSize ,ySize,pData,1,"cb",false);

  threedout.VRMLOut("checkerboard.wrl");
#ifdef BIAS_HAVE_OPENSCENEGRAPH
  ThreeDOutOpenSceneGraph t3dOSG;
  t3dOSG.AddImage(ul, ur,ll,lr,xSize ,ySize,pData,1,"cb",false);
  t3dOSG.OpenSceneGraphOut("checkerboard.osg");
  t3dOSG.OpenSceneGraphOut("checkerboard.ive");
  t3dOSG.OpenSceneGraphOut("checkerboard.3ds");
#endif
}

///////////////////////////////////////////////////////////////////////
void GenerateIntensityTestCheckerboard(int size,int xCorners, int yCorners,float maxDev)
{
  BIAS::Vector3<unsigned int> UL,UR,LL,LR;
  BIAS::Image<unsigned char> black,grey80,grey60,grey40,grey20,white;
  BIAS::Image<unsigned char> image;
  BIAS::Image<float> imageD,blackD,grey80D,grey60D,grey40D,grey20D,whiteD;

  unsigned int factor = 20;
  BIAS::ThreeDOut threedout;
  
  Vector3<double> ul(-size,-size,0);
  Vector3<double> ur((xCorners)*size,-size,0);
  Vector3<double> ll(-size,(yCorners)*size,0);
  Vector3<double> lr((xCorners)*size,(yCorners)*size,0);

  black.Init(factor,factor,1); 
  black.FillImageWithConstValue((unsigned char)0);
  grey80.Init(factor,factor,1); 
  grey80.FillImageWithConstValue((unsigned char)(255.0*0.20));
  grey60.Init(factor,factor,1); 
  grey60.FillImageWithConstValue((unsigned char)(255.0*0.40));
  grey40.Init(factor,factor,1); 
  grey40.FillImageWithConstValue((unsigned char)(255.0*0.60));
  grey20.Init(factor,factor,1); 
  grey20.FillImageWithConstValue((unsigned char)(255.0*0.80));
  white.Init(factor,factor,1); 
  white.FillImageWithConstValue((unsigned char)255);

  //fill depth images, black comes to front, white is at back plane
  if(maxDev<=0) maxDev =0.0;
  blackD.Init(factor,factor,1); 
  blackD.FillImageWithConstValue(maxDev);
  grey80D.Init(factor,factor,1); 
  grey80D.FillImageWithConstValue(maxDev*0.80);
  grey60D.Init(factor,factor,1); 
  grey60D.FillImageWithConstValue(maxDev*0.60);
  grey40D.Init(factor,factor,1); 
  grey40D.FillImageWithConstValue(maxDev*0.40);
  grey20D.Init(factor,factor,1); 
  grey20D.FillImageWithConstValue(maxDev*0.20);
  whiteD.Init(factor,factor,1); 
  whiteD.FillImageWithConstValue(1.0);
  
  // The checkerboard image
  unsigned newWidth=(unsigned)((xCorners+1)*factor);
  unsigned newHeight=(unsigned)((yCorners+1)*factor);
 
  image.Init(newWidth,newHeight,1);
  imageD.Init(newWidth,newHeight,1);
  
  bool topLeftBlack = true;
  bool isBlack = topLeftBlack;  
  unsigned blackCounter=0;
  for(int y=0;y < yCorners+1; y++){ //rows
    for(int x= 0;x< xCorners+1 ; x++){   //cols 
   
      if((y % 2) == 0 && x == 0){
        isBlack = topLeftBlack;
      }
      else if(x == 0)
        isBlack = !topLeftBlack;
          
      UL[0]=x*factor;
      UL[1]=y*factor;
      UL[2]=0;
      LR[0]=(x+1)*factor;
      LR[1]=(y+1)*factor;
      LR[2]=0;

      image.SetROICorners(UL[0],UL[1], LR[0],LR[1]);
      imageD.SetROICorners(UL[0],UL[1], LR[0],LR[1]);

      if(isBlack){
        if( y!=0 && y!=yCorners && x != 0 && x!=xCorners){        
          switch(blackCounter){
          case 0:
            image.Paste2ROI(black);
            imageD.Paste2ROI(blackD);
            break;
          case 1:
            image.Paste2ROI(grey80);
            imageD.Paste2ROI(grey80D);
            break;
          case 2:
            image.Paste2ROI(grey60);
            imageD.Paste2ROI(grey60D);
            break;
          case 3:
            image.Paste2ROI(grey40);
            imageD.Paste2ROI(grey40D);
            break;
          case 4:
            image.Paste2ROI(grey20);
            imageD.Paste2ROI(grey20D);
            break;
          default:
            BIASERR("blackCounter out of bounds");
          }        
          blackCounter++;
          if(blackCounter == 5)
            blackCounter =0;
        }
        else{
          image.Paste2ROI(black);
          imageD.Paste2ROI(blackD);
        }
      }
      else{
        image.Paste2ROI(white);
        imageD.Paste2ROI(whiteD);
      }
      isBlack = !isBlack;
    }
  }
  BIAS::ImageIO::Save("checkerboard_distorted.mip",image);

  /*
  BIAS::Gauss<float,float> gauss;
  gauss.Filter7x7Grey(imageD,imageD);  
  BIAS::ImageIO::Save("checkerboardGaussFiltered.mip",imageD);
  
  TriangleMesh mesh(0.0,360.0,1);
  unsigned w = 800;
  unsigned h = 600;
  float f = 470.0;
  KMatrix K;K.Set(f,f,w/2.0,h/2.0);
  Projection P; 
  P.CreatePerspective(Pose(),K,w,h);*/

  unsigned char *pData = image.GetImageData();
 
  unsigned xSize = (unsigned)rint(image.GetWidth());
  unsigned ySize = (unsigned)rint(image.GetHeight());

  threedout.AddImage(ul, ur,ll,lr,xSize ,ySize,pData,1,"cb",false);
  threedout.VRMLOut("checkerboard_distorted.wrl");

#ifdef BIAS_HAVE_OPENSCENEGRAPH
  ThreeDOutOpenSceneGraph t3dOSG;
  t3dOSG.AddImage(ul, ur,ll,lr,xSize ,ySize,pData,1,"cb",false);
  t3dOSG.OpenSceneGraphOut("checkerboard_distorted.3ds");
  t3dOSG.OpenSceneGraphOut("checkerboard_distorted.ive");
  t3dOSG.OpenSceneGraphOut("checkerboard_distorted.osg");
#endif
}

