#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Geometry/Projection.hh>
#include <Geometry/ProjectionParametersPerspective.hh>
#include <Image/Camera.hh>
#include <Image/ProjectionMapping.hh>
#include <Utils/Param.hh>

using namespace std;
using namespace BIAS;

/** @file biasUndistort
    @brief Undistorts an image using backward mapping. Projection image
           metadata is updating when reading/writing MIP image files.
    @note Supports only perspective cameras and unsigned char images yet!
    @ingroup g_tools
    @author esquivel 02/2013
 */

int main(int argc, char *argv[])
{  
  cout << "\n---- biasUndistort ----\n";

  // Read parameters
  BIAS::Param params;
  bool *help = params.AddParamBool("help", "Show information about parameters.", false, 'h');
  std::string *srcFile = params.AddParamString("input", "Filename of input image (needed)", "", 'i');
  std::string *outFile = params.AddParamString("output", "Filename of output image (needed)", "", 'o');
  std::string *projFile = params.AddParamString("proj", "Filename of projection (optional)", "", 'p');
  params.ParseCommandLine(argc, argv);

  if (*help || srcFile->empty() || outFile->empty()) {
    params.Usage();
    exit(0);
  }

  // Load source image
  BIAS::Camera<unsigned char> srcImage;
  if (BIAS::ImageIO::Load(*srcFile, srcImage) != 0) {
    cout << "[ERROR] Failed to load image from file " << *srcFile << "!\n\n";
    exit(-1);
  }
  srcImage.ParseMetaData();

  // Load projection or read from metadata
  BIAS::Projection srcProj;
  if (!projFile->empty()) {
    if (srcProj.Load(*projFile) != 0) {
      cout << "[ERROR] Failed to load projection from file " << *projFile << "!\n\n";
      exit(-1);
    } else {
      cout << "- Loaded projection from file " << *projFile << "\n";
    }
  } else if (srcImage.IsProjValid()) {
    srcProj = srcImage.GetProj();
    cout << "- Read projection from image metadata\n";
  } else {
    cout << "[ERROR] Failed to read projection from image metadata!\n\n";
    exit(-1);
  }
  BIAS::ProjectionParametersPerspective* srcPPP =
    (BIAS::ProjectionParametersPerspective*)srcProj.GetParameters();
  if (srcProj.Size() != 1 || srcPPP == NULL) {
    cout << "[ERROR] Invalid projection found! Supports only perspecive cameras!\n\n";
    exit(-1);
  }

  // Determine if projection contains distortion
  if (!srcPPP->IsDistorted()) {
    cout << "[ERROR] Image projection contains no distortion!\n\n";
    exit(-1);
  }

  // Create output image
  unsigned int srcWidth, srcHeight;
  //const unsigned int srcChannels = srcImage.GetChannelCount();
  srcPPP->GetImageSize(srcWidth, srcHeight);
  if (srcWidth != srcImage.GetWidth() || srcHeight != srcImage.GetHeight()) {
    cout << "[ERROR] Image and projection have different dimensions!\n\n";
    exit(-1);
  }

  // Create output projection
  BIAS::Projection outProj = srcProj;
  BIAS::ProjectionParametersPerspective* outPPP =
    (BIAS::ProjectionParametersPerspective*)outProj.GetParameters();
  outPPP->SetUndistortion(0, 0, 0, 0);
  outPPP->SetDistortionType(DISTYPE_NONE);

  // Create output image
  BIAS::Camera<unsigned char> outImage = srcImage;
  outImage.SetProj(outProj);
  outImage.UpdateMetaData();
  outImage.Clear(0);

  // Perform backward mapping to remove distortion
  cout << "- Creating undistorted image via backward mapping\n";
  BIAS::ProjectionMapping<unsigned char, unsigned char> mapping;
  mapping.SetSourceCam(srcProj);
  mapping.SetSinkCam(outProj);
  if (mapping.Map(srcImage, outImage) != 0) {
    cout << "[ERROR] Failed to undistort image via backward mapping!\n\n";
    exit(-1);
  }

  // Write output image to file
  if (BIAS::ImageIO::Save(*outFile, outImage) != 0)
    cout << "[ERROR] Failed to save image to file " << *outFile << "!\n\n";
  else
    cout << "- Saved undistorted image to file " << *outFile << "\n\n";

  return 0;
}
