/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Base/Image/Image.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>
#include <Utils/IOUtils.hh>
#include <Utils/Param.hh>
#include <Base/Common/FileHandling.hh>

#include <Image/ConvertHDR.hh>

using namespace std;
using namespace BIAS;

/**
 @file
 @ingroup g_tools
 @brief convert a (hdr) bayer image to unsigned char rgb, see biasbayer2rgb.cpp
 @author bartczak 12/2008
 */

int main(int argc, char* argv[])
{
  // Parse parameters
  Param *p = new Param();
  p->AddParamString("image", "Image in Bayer pattern", "", 'i');
  p->AddParamString("imageList", "List of images in Bayer pattern", "", 'l');
  p->AddParamString("bayer", "Bayer pattern (e.g. RGGB)", "", 'b');
  p->AddParamInt("colorDepth", "Specify color depth (e.g. use 255 for 8 bit image)", 255);
  if (!IOUtils::ParseCommandLineEvalHelp(*p, argc, argv)) {
    return -1;
  }
  
  // Read image filenames
  string imageFileName = *(p->GetParamString("image"));
  string imageListName = *(p->GetParamString("imageList"));
  string bayer = *(p->GetParamString("bayer"));
  int colorDepth = *(p->GetParamInt("colorDepth"));
  vector<string> imageNames;
  if (!imageListName.empty()) {
    if (Param::ParseListFile(imageListName, imageNames) != 0) {
      BIASERR("Failed to load image list file " << imageListName << "!");
      return -1;
    } else {
      cout << "Loaded image list file " << imageListName << endl;
    }
  } else if (!imageFileName.empty()) {
    imageNames.push_back(imageFileName);
  } else {
    p->Usage();
    return 0;
  }

  // Parse Bayer pattern
  BIAS::ImageBase::EColorModel pattern = BIAS::ImageBase::CM_invalid;
  if (strcmp(bayer.c_str(), "BGGR") == 0) {
    cout << "Converting Bayer pattern BGGR to RGB" << endl;
    pattern = ImageBase::CM_Bayer_BGGR;
  } else if (strcmp(bayer.c_str(), "GBRG") == 0) {
    cout << "Converting Bayer pattern GBRG to RGB" << endl;
    pattern = ImageBase::CM_Bayer_GBRG;
  } else if (strcmp(bayer.c_str(), "GRBG") == 0) {
    cout << "Converting Bayer pattern GRBG to RGB" << endl;
    pattern = ImageBase::CM_Bayer_GRBG;
  } else if (strcmp(bayer.c_str(), "RGGB") == 0) {
    cout << "Converting Bayer pattern RGGB to RGB" << endl;
    pattern = ImageBase::CM_Bayer_RGGB;
  } else {
    cout << "Unknown Bayer pattern " << bayer << " given!" << endl;
    return -1;
  }
  
  // Convert images
  for (unsigned int i = 0; i < imageNames.size(); i++)
  {
    string imageName = imageNames[i];
    string newImageName = "rgb_" + imageName;
    ImageBase tmp;
    if (ImageIO::Load(imageName, tmp) != 0) {
      BIASERR("Failed to load image " << imageName << "!");
      return -1;
    }

    // convert HDR image
    if (colorDepth > 256)
    {
      // convert HDR image to float image
      Image<float> hdrFloat;
      ImageConvert::ConvertST(tmp, hdrFloat, ImageBase::ST_float);
      ConvertHDR<float> hdrconverter;
      Image<unsigned char> tmpUC;
      // convert float image to unsigned char image
      hdrconverter.ToUnsignedCharGamma(hdrFloat, tmpUC, 1.0, colorDepth);
      tmp = tmpUC;
    }
    
    // convert Bayer image to RGB image
    if (tmp.GetStorageType() == ImageBase::ST_unsignedchar)
    {
      tmp.SetColorModel(pattern);

      Camera<unsigned char> bayerImage(tmp);
      Camera<unsigned char> rgbImage;
      bayerImage.ParseMetaData();
      ImageConvert::ToRGB(bayerImage, rgbImage);
      cout << "Converted " << imageName << " : " << bayerImage.GetColorModel()
           << " ---> " << newImageName << " : " << rgbImage.GetColorModel() << endl;
      rgbImage.SetMetaData(*bayerImage.GetMetaData());
      if (ImageIO::Save(newImageName, rgbImage) != 0) {
        BIASERR("Failed to save RGB image to file " << newImageName << "!");
        continue;
      }
    } else {
       BIASERR("Bayer image " << imageName << " has invalid storage type "
               "(unsigned char needed)!");
        continue;
    }
  }

  return 0;
}
