/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Utils/Param.hh>
#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>

#include <Base/ImageUtils/ImageCalc.hh>

using namespace std;
using namespace BIAS;

// operations for two input images
#define OP_ADD               0
#define OP_SUBTRACT          1
#define OP_MULTIPLY          2
#define OP_DIVIDE            3
#define OP_AND               4
#define OP_NAND              5
#define OP_OR                6
#define OP_NOR               7
#define OP_XOR               8
// operations for one image and a scalar
#define OP_ADD_SCALAR        9
#define OP_SUBTRACT_SCALAR  10
#define OP_MULTIPLY_SCALAR  11
#define OP_DIVIDE_SCALAR    12
// operations for one image
#define OP_LOG              13
#define OP_LOG10            14
// the total number of operations
#define NUM_OPS             15

/**
 * @brief Loads the image given by filename and returns it as image
 *
 * @param filename
 *        the path of the file to load
 * @param image
 *        the returned image
 * @param floatOutput
 *        if set to true, the loaded image is converted to float
 *
 * @return true on success or false otherwise
 *
 * @author rwulff
 * @date   03/2010
 * @file biasimagecalc
 * @internal
 */
bool LoadImage(const string &filename, ImageBase &image,
               const bool *floatOutput) {
  if (filename.empty()) {
    cerr << "ERROR: please specify an input image. try '-h' for help"
         << endl;
    return false;
  }
  else { // image specified -> try to load it
    // convert to float, if floatOutput is true
    if (*floatOutput) {
      ImageBase tempIm;

      if (ImageIO::Load(filename, tempIm)!= 0) {
        cerr << "ERROR: image (" << filename << ") could not be loaded"
             << endl;
        return false;
      }

      ImageConvert::ConvertST(tempIm, image, ImageBase::ST_float);
    }
    else { // floatOuput is false
      if (ImageIO::Load(filename, image)!= 0) {
        cerr << "ERROR: image (" << filename << ") could not be loaded"
             << endl;
        return false;
      }
    }
  }

  return true;
}

/**
 * @brief Helper function. Performs the calculations on two images.
 *
 * @param imagesInPath
 *        filenames of the two input images
 * @param output
 *        filename of the output image
 * @param operation
 *        the operation to perform between the input images
 * @param floatOutput
 *        if set to true, the operation will be performed on float images
 *
 * @return true on success or false otherwise
 *
 * @author rwulff
 * @date   03/2010
 * @file biasimagecalc
 * @internal
 */
bool CalculateOnTwoImages(string *imagesInPath[2], const string *output,
                          const int *operation,
                          const bool *floatOutput) {
  ImageBase imagesIn[2];
  ImageBase imageOut;

  // load images
  if (!LoadImage(*imagesInPath[0], imagesIn[0], floatOutput)) {
    return false;
  }

  if (!LoadImage(*imagesInPath[1], imagesIn[1], floatOutput)) {
    return false;
  }

  // assert same size
  BIASASSERT(   imagesIn[0].GetWidth()        == imagesIn[1].GetWidth()
             && imagesIn[0].GetHeight()       == imagesIn[1].GetHeight()
             && imagesIn[0].GetChannelCount() == imagesIn[1].GetChannelCount())

  // ensure same storage type
  if (imagesIn[0].GetStorageType() != imagesIn[1].GetStorageType()) {
    cout << "ERROR: input images must have same storage type" << endl;
    return false;
  }

  // initialise output image with correct size, channel count and storage type
  imageOut.Init(imagesIn[0].GetWidth(), imagesIn[0].GetHeight(),
                imagesIn[0].GetChannelCount(), imagesIn[0].GetStorageType());

  // perform operation
  switch (*operation) {
    case OP_ADD:                                                          // ADD
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Add((const Image<unsigned char> &)imagesIn[0],
                                        (const Image<unsigned char> &)imagesIn[1],
                                        (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Add((const Image<float> &)imagesIn[0],
                                (const Image<float> &)imagesIn[1],
                                (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_SUBTRACT:                                                // SUBTRACT
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Subtract((const Image<unsigned char> &)imagesIn[0],
                                             (const Image<unsigned char> &)imagesIn[1],
                                             (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Subtract((const Image<float> &)imagesIn[0],
                                     (const Image<float> &)imagesIn[1],
                                     (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_MULTIPLY:                                                // MULTIPLY
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Multiply((const Image<unsigned char> &)imagesIn[0],
                                             (const Image<unsigned char> &)imagesIn[1],
                                             (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Multiply((const Image<float> &)imagesIn[0],
                                     (const Image<float> &)imagesIn[1],
                                     (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_DIVIDE:                                                    // DIVIDE
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Divide((const Image<unsigned char> &)imagesIn[0],
                                           (const Image<unsigned char> &)imagesIn[1],
                                           (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Divide((const Image<float> &)imagesIn[0],
                                   (const Image<float> &)imagesIn[1],
                                   (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_AND:                                                          // AND
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::And((const Image<unsigned char> &)imagesIn[0],
                                        (const Image<unsigned char> &)imagesIn[1],
                                        (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::And((const Image<float> &)imagesIn[0],
                                (const Image<float> &)imagesIn[1],
                                (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_NAND:                                                        // NAND
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Nand((const Image<unsigned char> &)imagesIn[0],
                                         (const Image<unsigned char> &)imagesIn[1],
                                         (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Nand((const Image<float> &)imagesIn[0],
                                 (const Image<float> &)imagesIn[1],
                                 (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_OR:                                                            // OR
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Or((const Image<unsigned char> &)imagesIn[0],
                                       (const Image<unsigned char> &)imagesIn[1],
                                       (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Or((const Image<float> &)imagesIn[0],
                               (const Image<float> &)imagesIn[1],
                               (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_NOR:                                                          // NOR
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Nor((const Image<unsigned char> &)imagesIn[0],
                                        (const Image<unsigned char> &)imagesIn[1],
                                        (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Nor((const Image<float> &)imagesIn[0],
                                (const Image<float> &)imagesIn[1],
                                (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_XOR:                                                          // XOR
      switch (imagesIn[0].GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Xor((const Image<unsigned char> &)imagesIn[0],
                                        (const Image<unsigned char> &)imagesIn[1],
                                        (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Xor((const Image<float> &)imagesIn[0],
                                (const Image<float> &)imagesIn[1],
                                (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
  }

  // save output image
  ImageIO::Save(*output, imageOut);

  return true;
}


/**
 * @brief Helper function. Performs the calculations on the images and the
 *        scalar.
 *
 * @param imageInPath
 *        filename of the input images
 * @param scalar
 *        the scalar value
 * @param output
 *        filename of the output image
 * @param operation
 *        the operation to perform on the input image and the scalar
 * @param floatOutput
 *        if set to true, the operation will be performed on float images
 *
 * @return true on success or false otherwise
 *
 * @author rwulff
 * @date   03/2010
 * @file biasimagecalc
 * @internal
 */
bool CalculateOnOneImageAndScalar(const string *imageInPath,
                                  const double *scalar,
                                  const string *output,
                                  const int *operation,
                                  const bool *floatOutput) {
  ImageBase imageIn;
  ImageBase imageOut;

  // load image
  if (!LoadImage(*imageInPath, imageIn, floatOutput)) {
    return false;
  }

  // initialise output image with correct size, channel count and storage type
  imageOut.Init(imageIn.GetWidth(), imageIn.GetHeight(),
                imageIn.GetChannelCount(), imageIn.GetStorageType());

  // perform operation
  switch (*operation) {
    case OP_ADD_SCALAR:                                            // ADD_SCALAR
      switch (imageIn.GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::AddScalar((const Image<unsigned char> &)imageIn,
                                              (unsigned char)*scalar,
                                              (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::AddScalar((const Image<float> &)imageIn,
                                      (float)*scalar,
                                      (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_SUBTRACT_SCALAR:                                  // SUBTRACT_SCALAR
      switch (imageIn.GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::SubtractScalar((const Image<unsigned char> &)imageIn,
                                                   (unsigned char)*scalar,
                                                   (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::SubtractScalar((const Image<float> &)imageIn,
                                           (float)*scalar,
                                           (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_MULTIPLY_SCALAR:                                  // MULTIPLY_SCALAR
      switch (imageIn.GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::MultiplyScalar((const Image<unsigned char> &)imageIn,
                                                   (unsigned char)*scalar,
                                                   (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::MultiplyScalar((const Image<float> &)imageIn,
                                           (float)*scalar,
                                           (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_DIVIDE_SCALAR:                                      // DIVIDE_SCALAR
      switch (imageIn.GetStorageType()) {
        // the scalar must not be 0
        if (*scalar == 0) {
          cout << "ERROR: the scalar must not be 0" << endl;
          return false;
        }

        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::DivideScalar((const Image<unsigned char> &)imageIn,
                                                 (unsigned char)*scalar,
                                                 (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::DivideScalar((const Image<float> &)imageIn,
                                         (float)*scalar,
                                         (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
  }

  // save output image
  ImageIO::Save(*output, imageOut);

  return true;
}


/**
 * @brief Helper function. Performs the calculations on the image.
 *
 * @param imageInPath
 *        filename of the input image
 * @param output
 *        filename of the output image
 * @param operation
 *        the operation to perform on the input image
 * @param floatOutput
 *        if set to true, the operation will be performed on float images
 *
 * @return true on success or false otherwise
 *
 * @author rwulff
 * @date   03/2010
 * @file biasimagecalc
 * @internal
 */
bool CalculateOnOneImage(const string *imageInPath,
                         const string *output,
                         const int *operation,
                         const bool *floatOutput) {
  ImageBase imageIn;
  ImageBase imageOut;

  // load image
  if (!LoadImage(*imageInPath, imageIn, floatOutput)) {
    return false;
  }

  // initialise output image with correct size, channel count and storage type
  imageOut.Init(imageIn.GetWidth(), imageIn.GetHeight(),
                imageIn.GetChannelCount(), imageIn.GetStorageType());

  // perform operation
  switch (*operation) {
    case OP_LOG:                                                          // LOG
      switch (imageIn.GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Log((const Image<unsigned char> &)imageIn,
                                        (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Log((const Image<float> &)imageIn,
                                (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
    case OP_LOG10:                                                      // LOG10
      switch (imageIn.GetStorageType()) {
        case ImageBase::ST_unsignedchar:
          ImageCalc<unsigned char>::Log10((const Image<unsigned char> &)imageIn,
                                          (Image<unsigned char> &)imageOut);
          break;
        case ImageBase::ST_float:
          ImageCalc<float>::Log10((const Image<float> &)imageIn,
                                  (Image<float> &)imageOut);
          break;
        default:
          cout << "ERROR: operation not implemented for this storage type" << endl;
          return false;
      } // end of storage type switch
      break;
  }

  // save output image
  ImageIO::Save(*output, imageOut);

  return true;
}


/**
 * @brief Main function. Evaluates the command line parameters and calls one
 *        of the helper functions.
 *
 * @author rwulff
 * @date   03/2010
 * @file biasimagecalc
 */
int main(int argc, char *argv[]) {
  // init command line parameters
  Param param;

  bool *help = param.AddParamBool("help",
                                  "print help information and exit",
                                  false,
                                  'h',
                                  1);

  string *imagesInPath[2];
  imagesInPath[0] = param.AddParamString("input1",
                                         "path to the first output image (mandatory for all operations)",
                                         "",
                                         'i',
                                         1);
  imagesInPath[1] = param.AddParamString("input2",
                                         "path to the second output image (must have same size and storage type as first input image)",
                                         "",
                                         'j',
                                         1);
  double *scalar = param.AddParamDouble("scalar",
                                        "the scalar value (will be cast to the storage type of the input image)",
                                        0.0,
                                        -DBL_MAX,
                                        DBL_MAX,
                                        's',
                                        1);
  string *output = param.AddParamString("output",
                                        "path to the output image (default: out.mip)"
                                        " will be initialised with same storage type and size as first input image"
                                        " (except if param '--float' is set)",
                                        "out.mip",
                                        'o',
                                        1);

  bool *floatOutput = param.AddParamBool("float",
                                         "convert input images to float and perform the operations on these float images. output image will be of type float as well.",
                                         false,
                                         'f',
                                         1);

  // init string representations of operations
  vector<string> operations(NUM_OPS);

  // operations for two input images
  operations[OP_ADD]             = "add";
  operations[OP_SUBTRACT]        = "subtract";
  operations[OP_MULTIPLY]        = "multiply";
  operations[OP_DIVIDE]          = "divide";
  operations[OP_AND]             = "and";
  operations[OP_NAND]            = "nand";
  operations[OP_OR]              = "or";
  operations[OP_NOR]             = "nor";
  operations[OP_XOR]             = "xor";
  // operations for one image and a scalar
  operations[OP_ADD_SCALAR]      = "addScalar";
  operations[OP_SUBTRACT_SCALAR] = "subtractScalar";
  operations[OP_MULTIPLY_SCALAR] = "multiplyScalar";
  operations[OP_DIVIDE_SCALAR]   = "divideScalar";
  // operations for one image
  operations[OP_LOG]             = "log";
  operations[OP_LOG10]           = "log10";

  int *operation =
    param.AddParamEnum("operation",
                       "the arithmetic operation to perform.\n"
                       "\toperations that require TWO IMAGES:\n"
                       "\t- add:            computes the pixelwise sum\n"
                       "\t- subtract:       computes the pixelwise difference\n"
                       "\t- multiply:       computes the pixelwise product\n"
                       "\t- divide:         computes the pixelwise quotient\n"
                       "\t- and:            computes the pixelwise AND\n"
                       "\t- nand:           computes the pixelwise NAND\n"
                       "\t- or:             computes the pixelwise OR\n"
                       "\t- nor:            computes the pixelwise NOR\n"
                       "\t- xor:            computes the pixelwise XOR\n"
                       "\toperations that require ONE IMAGE AND A SCALAR:\n"
                       "\t- addScalar:      adds the scalar to each pixel\n"
                       "\t- subtractScalar: subtracts the scalar from each pixel\n"
                       "\t- multiplyScalar: multiplies the scalar with each pixel\n"
                       "\t- divideScalar:   divides each pixel by the scalar\n"
                       "\toperations that require ONE IMAGE:\n"
                       "\t- log:            computes the natural logarithm (base e) for each pixel\n"
                       "\t- log10:          computes the common logarithm (base 10) for each pixel",
                       operations,
                       -1,
                       NULL,
                       'p',
                       1);

  param.SetGroupName(1, "parameters");

  param.ParseCommandLine(argc, argv);

  // print help?
  if (*help) {
    cout << endl
         << "This tool performs pixelwise calculations on images."
         << endl;
    param.Usage();
    cout << "Examples:" << endl
         << "\tbiasimagecalc -i im1.jpg -j im2.jpg -p add      -o sum.jpg" << endl
         << "\tbiasimagecalc -i im.mip  -s 2       -p multiply -o product.mip" << endl
         << "\tbiasimagecalc -i im.mip             -p log      -o logarithm.mip" << endl
         << endl;
    return -1;
  }

  // perform calculations
  switch (*operation) {
    case OP_ADD:            // operations that need two input images
    case OP_SUBTRACT:
    case OP_MULTIPLY:
    case OP_DIVIDE:
    case OP_AND:
    case OP_NAND:
    case OP_OR:
    case OP_NOR:
    case OP_XOR:
      CalculateOnTwoImages(imagesInPath, output, operation, floatOutput);
      break;
    case OP_ADD_SCALAR:     // operations that need one input image and a scalar
    case OP_SUBTRACT_SCALAR:
    case OP_MULTIPLY_SCALAR:
    case OP_DIVIDE_SCALAR:
      CalculateOnOneImageAndScalar(imagesInPath[0], scalar,
                                   output,
                                   operation, floatOutput);
      break;
    case OP_LOG:            // operations that need only one image
    case OP_LOG10:
      CalculateOnOneImage(imagesInPath[0], output, operation, floatOutput);
      break;
    default:                // default case
      cout << "ERROR: operation not specified or unknown. try '-h' for help" << endl;
      return -1;
  }

  return 0;
}
