/*
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003-2009    (see file CONTACT for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <Base/Image/Image.hh>
#include <Base/Image/ImageConvert.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/ImagePackage.hh>
#include <Image/Camera.hh>
#include <Utils/IOUtils.hh>
#include <Utils/Param.hh>
#if !defined(WIN32) && defined(BIAS_HAVE_DC1394)
#include <Image/ConvertBayerPattern.hh>
#endif

using namespace std;
using namespace BIAS;

/**
 @file
 @ingroup g_tools
 @brief Unpacks image packages written by ShowCamWx, see biasimagepackage.cpp
 @author fkellner
 */
int main(int argc, char* argv[])
{

  Param param;
  bool *pack = param.AddParamBool("pack", "pack ImageList to package", false,
      'p');
  bool *unpack = param.AddParamBool("unpack",
      "unpack package, use output to override filenames from package", false,
      'u');
  string *infile = param.AddParamString("input", "file to process", "", 'i');
  string
      *outfile =
          param.AddParamString(
              "output",
              "basename of output file(s) (a number and file extension will be added by the program)",
              "", 'o');

  string *proj = param.AddParamString("camera",
      "optional camera projection xml file to be set in extracted images", "",
      'c');
  string *lut = param.AddParamString("lut",
      "apply lookuptable mapping with lut file", "", 'l');
  bool *toRgb = param.AddParamBool("rgb", "optional convert to rgb", false);
  string
      *bayer =
          param.AddParamString(
              "bayer",
              "if convert to rgb is active, assume this bayer pattern if color model is CM_Grey in the input image",
              "", 'b');
  string *ext = param.AddParamString("ext",
      "extension of images to save (in unpack mode)", "mip", 'e');

  if (!IOUtils::ParseCommandLineEvalHelp(param, argc, argv))
    {
      return 0;
    }

  bool bSetProj = false;
  Projection p;
  if (proj->length() > 0)
    {
      p.Load(*proj);
      bSetProj = true;
    }

  bool bLutMap = false;
  InterpolationMethod method = MapTrilinear;
  unsigned int w = 0, h = 0;
  ProjectionMapping<unsigned char, unsigned char> bm;
  if (lut->length() > 0)
    {
      if (!bSetProj)
        {
            BIASERR("you have to add a sink proj for mapping to work");
            BIASABORT;
        }
      p.GetParameters()->GetImageSize(w, h);
      bm.SetLookupTable(*lut, method);
      bLutMap = true;
    }

  ImageBase im;
  Camera<unsigned char> tmp;
#if !defined(WIN32) && defined(BIAS_HAVE_DC1394)
  ConvertBayerPattern<unsigned char> demosaic;
#endif

  if (*pack)
    {
      vector<string> imageNames;
      if (Param::ParseListFile(*infile, imageNames) != 0)
        {
            BIASERR("error parsing image names!");
          return -1;
        }
      ImagePackage ip;
      ip.Create(*outfile);

      for (unsigned int i = 0; i < imageNames.size(); i++)
        {

          cout << "adding image: " << imageNames[i] << endl;
          if (!im.IsEmpty()) {
            im.Release(true);
          }
          ImageIO::Load(imageNames[i], im);
          if (*toRgb)
            {
              MetaData meta = *im.GetMetaData();
              if (im.GetStorageType() != ImageBase::ST_unsignedchar) {
                BIASERR("to rgb only for unsigned char images.");
                BIASABORT;
              }
              Image<unsigned char> ucharsrc( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              Image<unsigned char> uchardst( im.GetWidth(), im.GetHeight(), 3);
              ucharsrc.CopyIn_NoInit( (unsigned char*)im.GetImageData());
#if !defined(WIN32) && defined(BIAS_HAVE_DC1394)
              if (ucharsrc.GetChannelCount() == 1)
                demosaic.Convert(ucharsrc, uchardst);
              else
                ImageConvert::ToRGB(ucharsrc, uchardst);
#else
              ImageConvert::ToRGB(ucharsrc, uchardst);
#endif
              uchardst.SetMetaData( meta );
              im = uchardst;
            }
          if (bLutMap)
            {
              if (im.GetStorageType() != ImageBase::ST_unsignedchar) {
                BIASERR("lut map only for unsigned char images.");
                BIASABORT;
              }
              Image<unsigned char> ucharsrc( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              Image<unsigned char> uchardst( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              ucharsrc.CopyIn_NoInit( (unsigned char*)im.GetImageData());
              uchardst.FillImageWithConstValue((unsigned char)0);
              bm.MapWithLookupTable(ucharsrc, uchardst, method);
              uchardst.SetMetaData( *im.GetMetaData() );
              im = uchardst;
            }
          if (bSetProj)
            {
              MetaData meta = *im.GetMetaData();
              tmp.SetMetaData( meta );
              tmp.SetProj(p);
              tmp.UpdateMetaData();
              im.SetMetaData( *tmp.GetMetaData() );
            }

          ip.AddImage(im, imageNames[i]);

        }

      ip.Close();

    }
  else if (*unpack)
    {
      if (infile->find(".pack") == string::npos)
        {
          cout << "input file should end with .pack" << endl;
          return 0;
        }
      string fname = *infile;
      unsigned pos = fname.find(".pack");
      pos -=4; //substract 4 for 0001 etc
      fname = fname.substr(0,pos);
      ImagePackage ip;
      ip.Open(fname);
      cout<<"*** Opened package ***: "<<fname<<endl;

      //ip.WriteAllImages(*outfile);
      string filename;
      int n = 0;
      while (ip.GetNextImage(im, filename) == 0)
        {

          if (*toRgb && bayer->length() != 0 && im.GetColorModel()
              == ImageBase::CM_Grey)
            {
              BIAS::ImageBase::EColorModel pattern =
                  BIAS::ImageBase::CM_invalid;
              if (strcmp(bayer->c_str(), "BGGR") == 0)
                {
                  cout << "converting bayer BGGR to RGB" << endl;
                  pattern = ImageBase::CM_Bayer_BGGR;
                }
              else if (strcmp(bayer->c_str(), "GBRG") == 0)
                {
                  cout << "converting bayer GBRG to RGB" << endl;
                  pattern = ImageBase::CM_Bayer_GBRG;
                }
              else if (strcmp(bayer->c_str(), "GRBG") == 0)
                {
                  cout << "converting bayer GRBG to RGB" << endl;
                  pattern = ImageBase::CM_Bayer_GRBG;
                }
              else if (strcmp(bayer->c_str(), "RGGB") == 0)
                {
                  cout << "converting bayer RGGB to RGB" << endl;
                  pattern = ImageBase::CM_Bayer_RGGB;
                }
              else
                {
                  cout << *bayer << "? i dont know this pattern... abort!"
                      << endl;
                    BIASABORT;
                }

              im.SetColorModel(pattern);
            }

          if (*toRgb)
            {
              MetaData meta = *im.GetMetaData();
              if (im.GetStorageType() != ImageBase::ST_unsignedchar) {
                BIASERR("to rgb only for unsigned char images.");
                BIASABORT;
              }
              Image<unsigned char> ucharsrc( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              Image<unsigned char> uchardst( im.GetWidth(), im.GetHeight(), 3);
              ucharsrc.CopyIn_NoInit( (unsigned char*)im.GetImageData());
#if !defined(WIN32) && defined(BIAS_HAVE_DC1394)
              if (ucharsrc.GetChannelCount() == 1)
                demosaic.Convert(ucharsrc, uchardst);
              else
                ImageConvert::ToRGB(ucharsrc, uchardst);
#else
              ImageConvert::ToRGB(ucharsrc, uchardst);
#endif
              uchardst.SetMetaData( meta );
              im = uchardst;
            }
          if (bLutMap)
            {
              if (im.GetStorageType() != ImageBase::ST_unsignedchar) {
                BIASERR("lut map only for unsigned char images.");
                BIASABORT;
              }
              Image<unsigned char> ucharsrc( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              Image<unsigned char> uchardst( im.GetWidth(), im.GetHeight(), im.GetChannelCount());
              ucharsrc.CopyIn_NoInit( (unsigned char*)im.GetImageData());
              uchardst.FillImageWithConstValue((unsigned char)0);
              bm.MapWithLookupTable(ucharsrc, uchardst, method);
              uchardst.SetMetaData( *im.GetMetaData() );
              im = uchardst;
            }
          if (bSetProj)
            {
              MetaData meta = *im.GetMetaData();
              tmp.SetMetaData( meta );
              tmp.ParseMetaData();
              tmp.SetProj(p);
              tmp.UpdateMetaData();
              im.SetMetaData( *tmp.GetMetaData());
            }
          if (outfile->length() > 0)
            {
              filename = *outfile + FileHandling::LeadingZeroString(++n, 5)
                  + "." + (*ext);
            }
          else if (ext->compare("mip") != 0)
            {
              filename = filename.substr(0, filename.rfind("."));
              filename += "." + (*ext);
            }
          cout << "saving: " << filename << endl;
          ImageIO::Save(filename, im);
        }

    }
  return 0;

}
