/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Base/Image/ImageBase.hh>
#include <Base/Image/ImageIO.hh>
#include <stdio.h>


using namespace BIAS;
using namespace std;


/**
   @file
   @ingroup g_tools
   @brief load image given by the command line, assign
          a uuid and write back to disk, see  biassetuid.cpp
   @author koeser
*/
void usage()
{
  cout << "Usage: biassetuid [options] files"<<endl;
  cout <<" Options: "<< endl
       <<" -f         : force setting of a new uid"<<endl
       << endl;
  cout <<"This program opens the image files given by the command line"
       <<", checks for a valid id and assigns a new one if there is no."<<endl
       <<"Returns: 0 on success, 1 if errors occured."<<endl<<endl;
}

int DetermineFiletype(const string &fn, ImageIO::TFileFormat &thetype)
{
  string::size_type EndOfPrefix;
  string suffix;

  EndOfPrefix = fn.rfind('.');
  if (EndOfPrefix == string::npos) return -1;
  suffix = fn.substr(EndOfPrefix+1,fn.length());
  if (suffix == "pnm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "ppm") thetype =  ImageIO::FF_ppm;
  else if (suffix == "pgm") thetype =  ImageIO::FF_pgm;
  else if (suffix == "mip") thetype =  ImageIO::FF_mip;
  else return -1;

  return 0;
}

int main (int argc, char *argv[])
{
  bool ForceNewUUID = false;
  int i = 1;
  if (argc>1) if (!strcmp("-f",argv[1])) {
    std::cout << "Existing UIDs are replaced."<<endl;
    ForceNewUUID = true;
    i = 2;
  }
  if (i>=argc) {
    usage();
    return 0; // for Dart testing
  }

  ImageIO::TFileFormat FileType;
  bool ErrorFlag = false;
  int IOResult = 0;
  BIAS::ImageBase CurImage;

  for (;i<argc;i++) {
    if (DetermineFiletype(argv[i], FileType)!=0) {
      BIASERR("Skipping current file because of unknown/unsuported format: "
              <<argv[i]<<endl<<" UIDs can only be written to mip,pgm,ppm,pnm");
      ErrorFlag = true;
      continue;
    }

    IOResult = ImageIO::Load(argv[i], CurImage);
    if (IOResult!=0) {
      ErrorFlag = true;
      BIASERR("Error reading "<< argv[i]);
      continue;
    }
    // flag indicating the previous/new id status of the file
    int idstatus = 0;
    if (CurImage.GetUID().IsValid()) {
      idstatus = 1;
      if (ForceNewUUID) idstatus = 2;
    }
    /*
    if (FileType==ImageIO::FF_mip) {
      IOResult = ImageIO::Save(argv[i], CurImage, false, ForceNewUUID );
    } else {
      IOResult = ImageIO::Save(argv[i], CurImage, FileType, false,
                                      0, ForceNewUUID);
    } */
    IOResult = ImageIO::Save(argv[i], 
                             CurImage,
                             FileType,
                             false,
                             0,
                             ForceNewUUID);

    if (IOResult!=0) {
      ErrorFlag = true;
      BIASERR("Error writing "<< argv[i]);
      continue;
    }
    cout << "File " << argv[i] << " has UID "<<CurImage.GetUID();
    switch (idstatus) {
    case 1: cout<<" (kept existing)"; break;
    case 2: cout<<" (replaced old)"; break;
    default:cout<<" (new)";
    }
    cout << endl;
  }
  if (ErrorFlag) cout << "There were errors." <<endl;
  if (ErrorFlag) return 1;
  return 0;

}












