/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __CheckUtils_hh__
#define __CheckUtils_hh__

#include <Base/Math/Vector2.hh>
#include <Base/Geometry/HomgPoint2D.hh>
#include <cmath>

namespace BIAS {

  class CheckUtils {
  public:
    
    template<class T>
    static bool InImageDisc(const unsigned int& width, 
                            const unsigned int& height,
                            const Vector2<T>& pt) {
      const T w = static_cast<T>(width-1);
      const T h = static_cast<T>(height-1);
      return (pt[0] >= 0 && pt[1] >= 0 &&  
              pt[0] <= w && pt[1] <= h);
    }

    static bool InImageDisc(const unsigned int& width, 
                            const unsigned int& height,
                            const HomgPoint2D& pt) {
      const double w = static_cast<double>(width-1);
      const double h = static_cast<double>(height-1);
      return (pt[0] >= 0 && pt[1] >= 0 &&  
              pt[0] <= w && pt[1] <= h && 
              pt[2]==1.0);
    }

    template<class T>
    static bool InImageDisc(const unsigned int& width, 
                            const unsigned int& height,
                            const T x,
                            const T y) {
      const T w = static_cast<T>(width-1);
      const T h = static_cast<T>(height-1);
      return (x >= 0 &&  y >= 0 &&  
              x <= w && y <= h);
    }

    
    /** Returns integer coordinate range for the window around pt
     *  which is in the image, pt also has to lie within the image
     * \param hwSize is the window's half win size.
     * \param xMin minimal VALID x coordintate in the window
     * \param xMax maximal VALID x coordintate in the window
     * \param yMin minimal VALID y coordintate in the window
     * \param yMax maximal VALID y coordintate in the window
     * \returns false if no valid window region was detected or if pt is not in the image.
     **/
    static bool GetIntegerWindow(const unsigned int& width, 
                                 const unsigned int& height,
                                 const HomgPoint2D& pt,
                                 const unsigned int hwSize,
                                 unsigned int& xMin, 
                                 unsigned int& xMax, 
                                 unsigned int& yMin, 
                                 unsigned int& yMax) {

      //center point has to be in image
      if(!InImageDisc(width, height, pt)) return false;


      const double w = static_cast<double>(width-1);
      const double h = static_cast<double>(height-1);
      float hw  = static_cast<float>(hwSize);

      float xMinTmp = float(std::max(rint(pt[0])-hw, 0.0));
      float yMinTmp = float(std::max(rint(pt[1])-hw, 0.0));
      float xMaxTmp = float(std::min(rint(pt[0])+hw, w));
      float yMaxTmp = float(std::min(rint(pt[1])+hw, h));
      
      if(xMaxTmp < xMinTmp || yMaxTmp < yMinTmp) return false;
     
      xMin = static_cast<unsigned int>(xMinTmp);
      yMin = static_cast<unsigned int>(yMinTmp);
      xMax = static_cast<unsigned int>(xMaxTmp);
      yMax = static_cast<unsigned int>(yMaxTmp);

      return true;
      
    }


  };

}//end of namespace


#endif
