/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @brief Example for ColourGradient
 *
 * Loads an image, converts it to grayscale (if necessary) and recolours it
 * using ColourGradient.
 *
 * @author Robert Wulff
 * @date 10/2010
 */

// std
#include <iostream>
#include <vector>

// BIAS
#include <Utils/ColourGradient.hh>
#include <Base/Image/Image.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageConvert.hh>

using namespace std;
using namespace BIAS;

// flag determines if black pixels are to be leaved untouched
const bool SKIP_BLACK_PIXELS = true;

/**
 * @brief main
 */
int main(int argc, char* argv[])
{
  // ensure filenames are given as parameters
  if (argc != 3) {
    cout << endl
         << "Loads an image, recolours it and saves it." << endl << endl
         << "USAGE:  "
         << argv[0] << " <input_image> <output_image>" << endl << endl;
    exit(-1);
  }

  // load image
  Image<unsigned char> imageIn;
  if (ImageIO::Load(argv[1], imageIn) != 0) {
    cout << "loading image failed" << endl;
    exit(-1);
  }

  // ensure gray scale
  if (imageIn.GetChannelCount() != 1) {
    cout << "converting image to gray scale" << endl;
    ImageConvert::IP_ToGrey(imageIn);
  }

  // either create custom gradient...
//  vector<unsigned char> red(3), green(3), blue(3);
//  red[0]   = 255; red[1]   =   0; red[2]   =   0;
//  green[0] =   0; green[1] = 255; green[2] =   0;
//  blue[0]  =   0; blue[1]  =   0; blue[2]  = 255;

//  ColourGradient<unsigned char> gradient(red, blue);
//  gradient.AddColour(0.5,  green);

  // ... or a preset
  ColourGradient<unsigned char> gradient(ColourGradient<unsigned char>::Jet);

  // recolour the image with the gradient
  Image<unsigned char> imageOut(imageIn.GetWidth(),
                                imageIn.GetHeight(),
                                3);
  unsigned char* idaIn  = imageIn.GetImageData();
  unsigned char* idaOut = imageOut.GetImageData();
  vector<unsigned char> col(3);

  for (unsigned int y = 0; y < imageIn.GetHeight(); ++y) {
    for (unsigned int x = 0; x < imageIn.GetWidth(); ++x) {
      const unsigned int offsetIn = y * imageIn.GetWidth() + x;
      const unsigned int offsetOut = y * imageIn.GetWidth() * imageOut.GetChannelCount()
                                     + x * imageOut.GetChannelCount();

      const float curColFloat = float(idaIn[offsetIn]) / 255.0f;
      if (!SKIP_BLACK_PIXELS || curColFloat != 0.0f) {
        gradient.GetColour(curColFloat, col);
        idaOut[offsetOut]     = col[0];
        idaOut[offsetOut + 1] = col[1];
        idaOut[offsetOut + 2] = col[2];
      }
    }
  }

  // save the image
  ImageIO::Save(argv[2],
                imageOut);

  return 0;
}
