/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

// std
#include <iostream>

// BIAS
#include <Utils/TriangleMesh.hh>

using namespace std;
using namespace BIAS;

/**
 * @brief Creates a TriangleMesh programmatically.
 *
 * The TriangleMesh consists of two non-connected subsegments and will look
 * like this:
 * 5-----6-----7     8-----9     <- point indices
 * |    /|    /|     |    /|
 * | 0 / | 2 / |     | 4 / |     <- triangle indices
 * |  /  |  /  |     |  /  |
 * | / 1 | / 3 |     | / 5 |     <- triangle indices
 * |/    |/    |     |/    |
 * 0-----1-----2     3-----4     <- point indices
 *
 * The composition is relatively simple. When processed by the algorithm no
 * subsegmentes need to be fused.
 *
 * @author rwulff
 * @date   02/2011
 */
void GenerateMesh1(TriangleMesh& mesh)
{
  vector<Vector3<double> > vertices;
  vertices.push_back(Vector3<double>(0, 0, 0)); // point 0
  vertices.push_back(Vector3<double>(1, 0, 0)); // point 1
  vertices.push_back(Vector3<double>(2, 0, 0)); // point 2
  vertices.push_back(Vector3<double>(3, 0, 0)); // point 3
  vertices.push_back(Vector3<double>(4, 0, 0)); // point 4
  vertices.push_back(Vector3<double>(0, 1, 0)); // point 5
  vertices.push_back(Vector3<double>(1, 1, 0)); // point 6
  vertices.push_back(Vector3<double>(2, 1, 0)); // point 7
  vertices.push_back(Vector3<double>(3, 1, 0)); // point 8
  vertices.push_back(Vector3<double>(4, 1, 0)); // point 9

  vector<Vector3<double> > normals;
  for (size_t vertexCnt = 0; vertexCnt < vertices.size(); ++vertexCnt) {
    normals.push_back(Vector3<double>(0, 0, 1));
  }

  vector<Vector4<unsigned char> > colors;
  for (size_t vertexCnt = 0; vertexCnt < vertices.size(); ++vertexCnt) {
    colors.push_back(Vector4<unsigned char>(0, 0, 255, 255));
  }

  vector<Vector<int> > triangleIndices;
  Vector<int> tmpTriangle(3);
  tmpTriangle[0] = 0; // triangle 0
  tmpTriangle[1] = 6;
  tmpTriangle[2] = 5;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] = 0; // triangle 1
  tmpTriangle[1] = 1;
  tmpTriangle[2] = 6;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] = 1; // triangle 2
  tmpTriangle[1] = 7;
  tmpTriangle[2] = 6;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] = 1; // triangle 3
  tmpTriangle[1] = 2;
  tmpTriangle[2] = 7;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] = 3; // triangle 4
  tmpTriangle[1] = 9;
  tmpTriangle[2] = 8;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] = 3; // triangle 5
  tmpTriangle[1] = 4;
  tmpTriangle[2] = 9;
  triangleIndices.push_back(tmpTriangle);

  mesh.SetMesh(vertices, normals, colors, triangleIndices);
}

/**
 * @brief Creates a TriangleMesh programmatically.
 *
 * The TriangleMesh consists of two non-connected subsegments and will look
 * like this:
 * 0-----4-----7-----3    10-----11    <- point indices
 * |    /|    /|    /|     |    /|
 * | 0 / | 2 / | 4 / |     | 6 / |     <- triangle indices
 * |  /  |  /  |  /  |     |  /  |
 * | / 1 | / 3 | / 5 |     | / 7 |     <- triangle indices
 * |/    |/    |/    |     |/    |
 * 2-----6-----5-----1     8-----9     <- point indices
 *
 * The composition is a little more complex and requires fusion of segments
 * during processing.
 *
 * @author rwulff
 * @date   02/2011
 */
void GenerateMesh2(TriangleMesh& mesh)
{
  vector<Vector3<double> > vertices;
  vertices.push_back(Vector3<double>(0, 1, 0)); // point  0
  vertices.push_back(Vector3<double>(3, 0, 0)); // point  1
  vertices.push_back(Vector3<double>(0, 0, 0)); // point  2
  vertices.push_back(Vector3<double>(3, 1, 0)); // point  3
  vertices.push_back(Vector3<double>(1, 1, 0)); // point  4
  vertices.push_back(Vector3<double>(2, 0, 0)); // point  5
  vertices.push_back(Vector3<double>(1, 0, 0)); // point  6
  vertices.push_back(Vector3<double>(2, 1, 0)); // point  7
  vertices.push_back(Vector3<double>(4, 0, 0)); // point  8
  vertices.push_back(Vector3<double>(5, 0, 0)); // point  9
  vertices.push_back(Vector3<double>(4, 1, 0)); // point 10
  vertices.push_back(Vector3<double>(5, 1, 0)); // point 11


  vector<Vector3<double> > normals;
  for (size_t vertexCnt = 0; vertexCnt < vertices.size(); ++vertexCnt) {
    normals.push_back(Vector3<double>(0, 0, 1));
  }

  vector<Vector4<unsigned char> > colors;
  for (size_t vertexCnt = 0; vertexCnt < vertices.size(); ++vertexCnt) {
    colors.push_back(Vector4<unsigned char>(0, 0, 255, 255));
  }

  vector<Vector<int> > triangleIndices;
  Vector<int> tmpTriangle(3);
  tmpTriangle[0] =  2; // triangle 0
  tmpTriangle[1] =  4;
  tmpTriangle[2] =  0;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  2; // triangle 1
  tmpTriangle[1] =  6;
  tmpTriangle[2] =  4;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  6; // triangle 2
  tmpTriangle[1] =  7;
  tmpTriangle[2] =  4;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  6; // triangle 3
  tmpTriangle[1] =  5;
  tmpTriangle[2] =  7;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  5; // triangle 4
  tmpTriangle[1] =  3;
  tmpTriangle[2] =  7;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  5; // triangle 5
  tmpTriangle[1] =  1;
  tmpTriangle[2] =  3;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  8; // triangle 6
  tmpTriangle[1] = 11;
  tmpTriangle[2] = 10;
  triangleIndices.push_back(tmpTriangle);
  tmpTriangle[0] =  8; // triangle 7
  tmpTriangle[1] =  9;
  tmpTriangle[2] = 11;
  triangleIndices.push_back(tmpTriangle);

  mesh.SetMesh(vertices, normals, colors, triangleIndices);
}


/**
 * @brief Creates a TriangleMesh and searches for the biggest connected segment.
 *
 * @author rwulff
 * @date   02/2011
 */
int main(int argc, char* argv[])
{
  // create the mesh
  cout << "creating triangle mesh..." << endl;

  TriangleMesh origMesh;
//  GenerateMesh1(origMesh);
  GenerateMesh2(origMesh);

  // get biggest connected segment
  cout << "getting biggest segment..." << endl;

  TriangleMesh biggestSegment;
  origMesh.GetBiggestConnectedSegment(biggestSegment);

  // output
  cout << "orig mesh has " << origMesh.GetTriangleIndicesRef().size()
       << " triangles" << endl;
  cout << "biggest segment has " << biggestSegment.GetTriangleIndicesRef().size()
       << " triangles" << endl;

  return 0;
}
