/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/W32Compat.hh>
#include <Base/ImageUtils/ImageDraw.hh>
#include "GraphPlotter.hh"

using namespace BIAS;
using namespace std;


// constructor
GraphPlotter::
GraphPlotter()
{
  width_ = 1024;
  height_ = 768;
  xOffset_ = 0;
  yOffset_ = 0;
  xTics_ = yTics_ = 10;
  xMin_=yMin_=DBL_MAX;
  xMax_=yMax_=DBL_MIN;
  xLabel_="x";
  yLabel_="y";
  bDrawGrid_ = false;
  bAutoDetectBordersX_=true;
  bAutoDetectBordersY_=true;
  bPrepared_ = false;
  bFixXAxisToBottom_=true;
  bFixYAxisToLeft_=true;
  dLineThickness_ = 1;
  dTextThickness_ = 1;
}

// destructor
GraphPlotter::
~GraphPlotter()
{
  if(!image_.IsEmpty())
    image_.Release();
}


BIAS::Image<unsigned char>* GraphPlotter::
GetImage(){
  return &image_;
}

BIAS::Image<unsigned char> GraphPlotter::
GetCopyOfImage(){
  return image_;
}


void GraphPlotter::
SetXRange(double minX, double maxX){
  if(minX > maxX){
    xMin_ = maxX;
    xMax_ = minX;
  }else{
    xMin_ = minX;
    xMax_ = maxX;
  }
  bAutoDetectBordersX_=false;
}

void GraphPlotter::
SetYRange(double minY, double maxY){
  if(minY > maxY){
    yMin_ = maxY;
    yMax_ = minY;
  }else{
    yMin_ = minY;
    yMax_ = maxY;
  }
  bAutoDetectBordersY_=false;
}

void GraphPlotter::
SetXTics(unsigned xtics){
  xTics_ = xtics;
}

void GraphPlotter::
SetYTics(unsigned ytics){
  yTics_ = ytics;
}

void GraphPlotter::
SetXLabel(std::string label){
  xLabel_=label;
}

void GraphPlotter::
SetYLabel(std::string label){
  yLabel_=label;
}

void GraphPlotter::
SetImageSize(unsigned width, unsigned height){
  width_ = width;
  height_ = height;
}

void GraphPlotter::
SetDrawGrid(bool drawGrid){
  bDrawGrid_=drawGrid;
}


void GraphPlotter::
SetAxesToYLeftAndXBottom(bool fixAxes){
  bFixXAxisToBottom_=fixAxes;
  bFixYAxisToLeft_=fixAxes;
}

void GraphPlotter::
SetLineThickness(unsigned thickness){
  dLineThickness_ = thickness;
}

void GraphPlotter::
SetTextThickness(unsigned thickness){
  dTextThickness_ = thickness;
}

void GraphPlotter::
Reset(){
  width_ = 1024;
  height_ = 768;
  xOffset_ = 0;
  yOffset_ = 0;
  xTics_ = yTics_ = 10;
  xMin_=yMin_=0.0;
  xMax_=yMax_=1000.0;
  xLabel_="x";
  yLabel_="y";
  bDrawGrid_ = false;
  bAutoDetectBordersX_=true;
  bAutoDetectBordersY_=true;
  bPrepared_ = false;
  bFixXAxisToBottom_=true;
  bFixYAxisToLeft_=true;
  dLineThickness_ = 1;
  dTextThickness_ = 1;
}

void GraphPlotter::
Prepare_(){

  BIAS::ColourRGB<unsigned char> black(0,0,0);
  BIAS::ColourRGB<unsigned char> white(255,255,255);
  BIAS::ColourRGB<unsigned char> red(255,0,0);

  xCoeff_ = width_/(xMax_-xMin_);
  yCoeff_ = height_/(yMax_-yMin_);

  //  cout<<"xCoeff_:"<<xCoeff_<<" xMax:"<<xMax_<<endl;
  //   cout<<"yCoeff_:"<<yCoeff_<<" yMax:"<<yMax_<<endl;

  //TODO calc offsets correctly
  xOffset_ = int(-rint(xMin_*xCoeff_))+10;
  yOffset_ = int(rint(yMin_*yCoeff_))+height_-20;

  //  cout<<"xOffset:"<<xOffset_<<endl;
  //   cout<<"yOffset:"<<yOffset_<<endl;

  if(!image_.IsEmpty()) image_.Release();

  image_.Init(width_,height_,3);
  unsigned char color[] ={255,255,255};
  image_.FillImageWithConstValue(color);

  int xAxisInc = abs(int(rint((xMax_-xMin_)/double(xTics_))));
  int yAxisInc = abs(int(rint((yMax_-yMin_)/double(yTics_))));

  //check if axis is in image, if not set to 10 pixels from left
  unsigned int xTicsPos = 0;
  if(yOffset_>=10 && yOffset_ < (int)height_-20)
    xTicsPos = yOffset_;
  else
    xTicsPos = height_-20;

  if(bFixXAxisToBottom_)
    xTicsPos = height_-20;

  //draw 0-x line
  BIAS::ImageDraw<unsigned char>::Line (image_,
                                        0,//x
                                        yOffset_,//y
                                        width_, //x end
                                        yOffset_, //y end
                                        black,dLineThickness_);

  //x-axis line
  BIAS::ImageDraw<unsigned char>::Line (image_,
                                        0,//x
                                        xTicsPos,//y
                                        width_, //x end
                                        xTicsPos, //y end
                                        black,dLineThickness_);

  // x direction text and tics 0->xMax
  int i=0;
  if(xMin_>=0) i = int(xMin_);
  else i=0;
  for(; i<=xMax_;i+=xAxisInc){
    stringstream message; message<<i;
    BIAS::ImageDraw<unsigned char>::
      Text (image_,message.str(),int(rint(i*xCoeff_))+xOffset_+5,xTicsPos+15,
            black,CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);
    BIAS::ImageDraw<unsigned char>::
      Line(image_,
           int(rint(i*xCoeff_))+xOffset_,
           xTicsPos,
           int(rint(i*xCoeff_))+xOffset_,
           xTicsPos+15,black,dLineThickness_);

    if(bDrawGrid_){
      BIAS::ImageDraw<unsigned char>::
        Line(image_,
             int(rint(i*xCoeff_))+xOffset_,
             0,
             int(rint(i*xCoeff_))+xOffset_,
             height_,black);
    }
  }

  if(xMin_ < 0){
    //x direction text and tics 0->xMin
    for(i=0;i>=xMin_;i-=xAxisInc){
      stringstream message; message<<i;
      BIAS::ImageDraw<unsigned char>::
        Text (image_,message.str(),int(rint(i*xCoeff_))+xOffset_+5,xTicsPos+15,
              black,CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);
      BIAS::ImageDraw<unsigned char>::
        Line(image_,
             int(rint(i*xCoeff_))+xOffset_,
             xTicsPos,
             int(rint(i*xCoeff_))+xOffset_,
             xTicsPos+15,black,dLineThickness_);

      if(bDrawGrid_){
        BIAS::ImageDraw<unsigned char>::
          Line(image_,
               int(rint(i*xCoeff_))+xOffset_,
               0,
               int(rint(i*xCoeff_))+xOffset_,
               height_,black);
      }
    }
  }
  //x axis text
  BIAS::ImageDraw<unsigned char>::
    Text(image_,xLabel_,width_-250,xTicsPos-15, black,
         CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);


  //check if axis is in image, if not set to 10 pixels from left
  unsigned int yTicsPos = 0;
  if(xOffset_>=10 && xOffset_ < (int)width_-10)
    yTicsPos = xOffset_;
  else
    yTicsPos = 10;

  if(bFixYAxisToLeft_)
    yTicsPos = 10;
  //y-axis line
  BIAS::ImageDraw<unsigned char>::Line(image_,
                                       yTicsPos,0,yTicsPos,
                                       height_,black,dLineThickness_);
  if(yMin_>=0) i = int(yMin_);
  else i=0;
  //y direction text and tics
  for(; i<=yMax_;i+=yAxisInc){
    stringstream message;  message<<i;
    BIAS::ImageDraw<unsigned char>::
      Text (image_,message.str(),
            yTicsPos+5,
            yOffset_-int(rint(i*yCoeff_))-5,black,CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);
    BIAS::ImageDraw<unsigned char>::
      Line(image_,
           yTicsPos,
           yOffset_-int(rint(i*yCoeff_)),
           yTicsPos+15,
           yOffset_-int(rint(i*yCoeff_)), black,dLineThickness_);

    if(bDrawGrid_){
      BIAS::ImageDraw<unsigned char>::
        Line(image_,
             0,
             yOffset_-int(rint(i*yCoeff_)),
             width_,
             yOffset_-int(rint(i*yCoeff_)), black);
    }
  }

  if(yMin_ < 0){
    for(i=0; i>=yMin_;i-=yAxisInc){
      stringstream message;  message<<i;
      BIAS::ImageDraw<unsigned char>::
        Text (image_,message.str(),
              yTicsPos+5,
              yOffset_-int(rint(i*yCoeff_))-5,black,CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);
      BIAS::ImageDraw<unsigned char>::
        Line(image_,
             yTicsPos,
             yOffset_-int(rint(i*yCoeff_)),
             yTicsPos+15,
             yOffset_-int(rint(i*yCoeff_)), black,dLineThickness_);

      if(bDrawGrid_){
        BIAS::ImageDraw<unsigned char>::
          Line(image_,
               0,
               yOffset_-int(rint(i*yCoeff_)),
               width_,
               yOffset_-int(rint(i*yCoeff_)), black);
      }
    }
  }


  //y  axis text
  BIAS::ImageDraw<unsigned char>::
    Text(image_,yLabel_,yTicsPos+50,20,black,CV_FONT_HERSHEY_SIMPLEX,0.5,0.5,0,dTextThickness_);

  //signal that the graph is inited
  bPrepared_=true;
}

void GraphPlotter::
Draw(std::vector<double>& xValues,
     std::vector<double>& yValues,
     BIAS::ColourRGB<unsigned char> color,
     BIAS::GraphPlotter_DrawStyle ds){

  if(!bPrepared_){
    if(bAutoDetectBordersX_){
      for(unsigned int i=0;i<xValues.size();i++){
        if(xValues[i] < xMin_) xMin_ = xValues[i];
        if(xValues[i] > xMax_) xMax_ = xValues[i];
      }
      xMin_ -= 10;
      xMax_ += 10;
    }
    if(bAutoDetectBordersY_){
      for(unsigned int i=0;i<yValues.size();i++){
        if(yValues[i] < yMin_) yMin_ = yValues[i];
        if(yValues[i] > yMax_) yMax_ = yValues[i];
      }
      yMin_ -= 10;
      yMax_ += 10;
    }
    Prepare_();
  }
  switch(ds){
  case DS_POINT_1:
  case DS_POINT_2:
  case DS_POINT_3:
  case DS_POINT_4:
  case DS_POINT_5:
    unsigned char tmpColor[3];
    tmpColor[0] = color[0];
    tmpColor[1] = color[1];
    tmpColor[2] = color[2];
    for(unsigned i=0;i<xValues.size();i++){
      BIAS::ImageDraw<unsigned char>::
        CircleCenterFilled(image_,
                           int(rint(xValues[i]*xCoeff_))+xOffset_,
                           yOffset_-int(rint(yValues[i]*yCoeff_)),
                           ds,tmpColor);

    }
    break;
  case DS_LINE:
    double oldx = 0.0;
    double oldy = 0.0;

    for(unsigned i=0;i<xValues.size();i++){
      if(oldx != 0.0){
        BIAS::ImageDraw<unsigned char>::
          Line(image_,
               int(rint(oldx*xCoeff_))+xOffset_,//startx
               yOffset_-int(rint(oldy*yCoeff_)),
               int(rint(xValues[i]*xCoeff_))+xOffset_,//stopx
               yOffset_-int(rint(yValues[i]*yCoeff_)),color,dLineThickness_);

        //  cout<<int(rint(oldx*xCoeff_))+xOffset_<<","<<
        //           yOffset_-int(rint(oldy*yCoeff_))<<" \t-> "<<
        //           int(rint(xValues[i]*xCoeff_))+xOffset_<<","<<//stopx
        //           yOffset_-int(rint(yValues[i]*yCoeff_))<<endl;
      }
      oldx = xValues[i];
      oldy = yValues[i];
    }

    //  cout<<"\n\n\n"<<endl;
    break;
  }
}

void GraphPlotter::
Draw(std::vector<double>& xValues,
     std::vector<double>& yValues,
     std::vector<double>& yStdDeviation,
     BIAS::ColourRGB<unsigned char> color,
     BIAS::ColourRGB<unsigned char> colorSTD,
     BIAS::GraphPlotter_DrawStyle ds, 
     std::vector<string> *labels) {

  if(!bPrepared_){
    if(bAutoDetectBordersX_){
      for(unsigned int i=0;i<xValues.size();i++){
        if(xValues[i] < xMin_) xMin_ = xValues[i];
        if(xValues[i] > xMax_) xMax_ = xValues[i];
      }
      xMin_ -= 10;
      xMax_ += 10;
    }
    if(bAutoDetectBordersY_){
      for(unsigned int i=0;i<yStdDeviation.size();i++){
        if(-(yStdDeviation[i]+yValues[i]) < yMin_) yMin_ = -(yStdDeviation[i]+yValues[i]);
        if(yStdDeviation[i]+yValues[i] > yMax_) yMax_ = yStdDeviation[i]+yValues[i];
      }
      yMin_ -= 20;
      yMax_ += 20;
    }
    Prepare_();
  }

  //draw standard deviations
  for(unsigned i=0;i<yStdDeviation.size();i++){
    //vertical line
    BIAS::ImageDraw<unsigned char>::
      Line(image_,
           int(rint(xValues[i]*xCoeff_))+xOffset_,//startx
           yOffset_-int(rint(yValues[i]*yCoeff_))-
           int(rint(yStdDeviation[i]*yCoeff_)),//startY
           int(rint(xValues[i]*xCoeff_))+xOffset_,//stopx
           yOffset_-int(rint(yValues[i]*yCoeff_))+
           int(rint(yStdDeviation[i]*yCoeff_)),//stopy
           colorSTD,dLineThickness_);
    //upper horizontal line
    BIAS::ImageDraw<unsigned char>::
      Line(image_,
           int(rint(xValues[i]*xCoeff_))-5+xOffset_,//startx
           yOffset_-int(rint(yValues[i]*yCoeff_))-
           int(rint(yStdDeviation[i]*yCoeff_)),//startY
           int(rint(xValues[i]*xCoeff_))+5+xOffset_,//stopx
           yOffset_-int(rint(yValues[i]*yCoeff_))-
           int(rint(yStdDeviation[i]*yCoeff_)),//stopy
           colorSTD,dLineThickness_);

    //lower horizontal line
    BIAS::ImageDraw<unsigned char>::
      Line(image_,
           int(rint(xValues[i]*xCoeff_))-5+xOffset_,//startx
           yOffset_-int(rint(yValues[i]*yCoeff_))+
           int(rint(yStdDeviation[i]*yCoeff_)),//startY
           int(rint(xValues[i]*xCoeff_))+5+xOffset_,//stopx
           yOffset_-int(rint(yValues[i]*yCoeff_))+
           int(rint(yStdDeviation[i]*yCoeff_)),//stopy
           colorSTD,dLineThickness_);
    if (labels != NULL) {
      if (labels->size() <= i) {
        BIASWARN("label names not complete");
      }
      BIAS::ImageDraw<unsigned char>::Text(image_, (*labels)[i], int(rint(xValues[i]*xCoeff_))+5+xOffset_,yOffset_-int(rint(yValues[i]*yCoeff_))+int(rint(yStdDeviation[i]*yCoeff_)),colorSTD);
    }
  }
  Draw(xValues,yValues,color,ds);
}

void GraphPlotter::
DrawLabels(){
	
	for(unsigned i=0;i<labels_.size();i++){
	   BIAS::ImageDraw<unsigned char>::Text(image_, labels_[i].first, width_-350,height_-50-(i+1)*20,labels_[i].second);	  
	}
		
}
