#ifndef INTEGRALIMAGE_HH
#define INTEGRALIMAGE_HH

#include <Base/Image/Image.hh>

namespace BIAS {

        /**
         * \class InategralImage
         * \brief Class to create a summed area table (aka integral image). It provides fast access to area sums, average area values
         *  as well difference of means (DOM).
         * \author jordt
         * \date March 2013
         */
template<class StorageType>
  class BIASUtils_EXPORT IntegralImage{

  public:

    /** Create an uninitialized integral image. */
    IntegralImage();

    /** Create an integral image instance based on the given image.*/
    IntegralImage(const Image<StorageType> &im);

    ~IntegralImage();

    /** Create or update the integral image based on the given image*/
    void GenerateFrom(const ImageBase &im);

    /** Get the raw data pointer to the integral image. */
    double* GetIntImageData();

    /** Calculate the difference of means (DOM) at a given position with a
     * given scale (halfWinSize). This can be used as a fast alternative to
     * the difference of gaussian (DOG) based scale space.
     * @param x x position in the DOM space
     * @param y y position in the DOM space
     * @param halfWinSize The half windows size of the inner area (which is
     *        subtracked by an area with 2*halfWinSize)
     */
    double GetDOM(int x, int y, int halfLevelSize);

  protected:

    int width_;
    int height_;
    int channels_;
    double* intImg_;

  public:

    /**
     * Evaluate the integral image. Returns the sum of values in the rectangle with
     * the upper left corner (xA,yA) and lower right corner (xB,yB).
     * @param xA left border
     * @param yA upper border
     * @param xB right border
     * @param yB lower border
     */
    inline double GetAreaSum(const int &xA,const int &yA, const int &xB,const int &yB){
      BIASASSERT(xA >= 0 && xA < width_)
      BIASASSERT(xB >= 0 && xB < width_)
      BIASASSERT(yA >= 0 && yA < height_)
      BIASASSERT(yB >= 0 && yB < height_)
      return (intImg_[yB*width_+xB] - intImg_[yB*width_+xA] - intImg_[yA*width_+xB] + intImg_[yA*width_+xA]);
    }

    /**
     * Evaluate the integral image. Returns the average value in the rectangle with
     * the upper left corner (xA,yA) and lower right corner (xB,yB).
     * @param xA left border
     * @param yA upper border
     * @param xB right border
     * @param yB lower border
     */
    inline double GetAreaAvg(const int &xA,const int &yA, const int &xB,const int &yB){
      BIASASSERT(xA >= 0 && xA < width_)
      BIASASSERT(xB >= 0 && xB < width_)
      BIASASSERT(yA >= 0 && yA < height_)
      BIASASSERT(yB >= 0 && yB < height_)
      return (intImg_[yB*width_+xB] - intImg_[yB*width_+xA] - intImg_[yA*width_+xB] + intImg_[yA*width_+xA])
          /((double)(xB-xA)*(yB-yA));
    }

    /**
     * Evaluate the integral image. Returns the rounded average value in the rectangle with
     * the upper left corner (xA,yA) and lower right corner (xB,yB) cast to StorageType.
     * @param xA left border
     * @param yA upper border
     * @param xB right border
     * @param yB lower border
     */
    inline StorageType GetAreaAvgRd(const int &xA,const int &yA, const int &xB,const int &yB){
      BIASASSERT(xA >= 0 && xA < width_)
      BIASASSERT(xB >= 0 && xB < width_)
      BIASASSERT(yA >= 0 && yA < height_)
      BIASASSERT(yB >= 0 && yB < height_)
      return (StorageType)rint((intImg_[yB*width_+xB] - intImg_[yB*width_+xA] - intImg_[yA*width_+xB] + intImg_[yA*width_+xA])
          /((double)(xB-xA)*(yB-yA)));
    }


  }; //IntegralImage

} //namespace


#endif // INTEGRALIMAGE_HH
