/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _TRIANGLEMESH_HH_
#define _TRIANGLEMESH_HH_


#include <math.h>
#include <vector>
#include <set>
#include <Base/Image/Image.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/Projection.hh>

namespace BIAS {

  /**
   * Struct for easier handling of edges
   */
  typedef struct Edge_s
  {
    int IDv1;
    int IDv2;
    Vector3<double> newV;
    double err;

    bool operator()(Edge_s a,Edge_s b) const{
      return a.err < b.err;
    }

    bool operator==(const Edge_s& E){
      if(E.IDv1 == IDv1 && E.IDv2 == IDv2)
        return true;
      else
        return false;
    }

    Edge_s& operator =(const Edge_s& E){

      IDv1 = E.IDv1;
      IDv2 = E.IDv2;
      newV = E.newV;
      err = E.err;

      return *this;
    }

  }Edge;

  /** \class TriangleMesh
   *  \ingroup g_utils
   *  \brief Create and represent a 3D triangle mesh. The parameter minCornerAngle
   *  defines a lower bound on the allowed angle in each triangle. If a triangle
   *  contains an angle smaller than this threshold, it is dropped.
   *  The parameter maxAngle defines an upper bound on the angle between the
   *  viewing ray and the normal of the triangle. If the angle if bigger than
   *  the threshold, it is dropped.
   *  The parameter depthScale can be used to scale the model prior to triangle
   *  generation.
   * \author bartczak
   */
  class BIASUtils_EXPORT TriangleMesh
  {
  public:

    TriangleMesh(const double& minCornerAngle = M_PI * 5.0 / 180.0,
                 const double& maxAngle = M_PI * 60.0 / 180.0,
                 const double depthScale = 1.0);

    void SetParams(const double& minCornerAngle = M_PI * 5.0 / 180.0,
                   const double& maxAngle = M_PI * 60.0 / 180.0,
                   const double depthScale = 1.0);

    inline
    void SetMesh(BIAS::Image<unsigned char> texture,
                 std::vector<BIAS::Vector3<double> > meshVertices,
                 std::vector<BIAS::Vector3<double> > vertexNormals,
                 std::vector<BIAS::Vector2<float> > textureCoords,
                 std::vector<BIAS::Vector<int> > triangleIndices){
      texture_ = texture;
      meshVertices_ = meshVertices;
      vertexNormals_ = vertexNormals;
      textureCoords_ = textureCoords;
      triangleIndices_ = triangleIndices;
    }

    /**
     *
     * @param texture
     *        set to empty image to ignore and use texture that was set before
     */
    inline
    void SetMesh(std::vector<BIAS::Vector3<double> > meshVertices,
                 std::vector<BIAS::Vector3<double> > vertexNormals,
                 std::vector<BIAS::Vector2<float> > textureCoords,
                 std::vector<BIAS::Vector<int> > triangleIndices,
                 BIAS::Image<unsigned char> texture = Image<unsigned char>()){
      meshVertices_ = meshVertices;
      vertexNormals_ = vertexNormals;
      textureCoords_ = textureCoords;
      triangleIndices_ = triangleIndices;

      if (texture.GetWidth() > 0 && texture.GetHeight() > 0) {
        texture_ = texture;
      }
    }

    inline
    void SetMesh(std::vector<BIAS::Vector3<double> > meshVertices,
                 std::vector<BIAS::Vector3<double> > vertexNormals,
                 std::vector<BIAS::Vector4<unsigned char> > vertexColors,
                 std::vector<BIAS::Vector<int> > triangleIndices){
      meshVertices_ = meshVertices;
      vertexNormals_ = vertexNormals;
      vertexColors_ = vertexColors;
      triangleIndices_ = triangleIndices;
    }

    /** Calculate a triangle mesh from dense depth map without PMatrix.
     * That's a 3D representation of an Image. All entries in depth map are
     * simply combined to little triangles, if neighboring pixels are set.
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          const BIAS::Image<unsigned char>& Texture);

    /** Calculate a triangle mesh from dense depth map.
     * No triangles are discarded, hence no smoothing.
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::PMatrix& P,
                          const BIAS::Image<unsigned char>& Texture);

    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                              BIAS::Projection& P,
                              const BIAS::Image<unsigned char>& Texture,
                              BIAS::Projection& TexP,
                              bool reduce,
                              double reduceValue);

    /** @brief Calculate a triangle mesh only using the given cutout
    recalculates depthmap, image and p and calls generic interface
    @author koeser  */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::PMatrix& P,
                          const BIAS::Image<unsigned char>& Texture,
                          const ROI& cutout);

    /** Calculate a triangle mesh from dense depth map.
     * No triangles are discarded, hence no smoothing.  */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& P,
                          const BIAS::Image<unsigned char>& Texture,
                          BIAS::Projection& TexP,
                          const unsigned int distBetweenPoints=1);

    /** old interface without a projection for the texture  */
    inline
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& P,
                          const BIAS::Image<unsigned char>& Texture,
                          const unsigned int distBetweenPoints=1)
    {
      if( ( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
             (DenseDepthMap.GetHeight()!=Texture.GetHeight()) )
             && ( !(Texture.IsEmpty()) )
        ) {
        BIASWARNONCE("Texturesize "<<Texture.GetWidth()<<"x"<<Texture.GetHeight()
            <<"!= DepthMapsize "<<DenseDepthMap.GetWidth()<<"x"
            <<DenseDepthMap.GetHeight()<<std::endl);
        }
      return(GenerateDenseMesh(DenseDepthMap,P,Texture,P,
             distBetweenPoints));
    }

    /** calculates trianglemesh but only for given part of depth map (can be used for tile mode)!
     *  Generated texture coordinates reference are relative to the
     *  full depth map size!
     *  If distBetweenPoints > 1, the intermediate point are simply ignored, when combining point
     *  to triangles   */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& DepthP,
                          const BIAS::Image<unsigned char>& Texture,
                          BIAS::Projection& TexP,
                          const unsigned int x0, const unsigned int y0,
                          const unsigned int width, const unsigned int height,
                          const unsigned int distBetweenPoints=1);

    /** old interface without a projection for the texture  */
    inline
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& DepthP,
                          const BIAS::Image<unsigned char>& Texture,
                          const unsigned int x0, const unsigned int y0,
                          const unsigned int width, const unsigned int height,
                          const unsigned int distBetweenPoints=1)
    {
      if( ( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
             (DenseDepthMap.GetHeight()!=Texture.GetHeight()) )
             && ( !(Texture.IsEmpty()) )
        ) {
        BIASWARNONCE("Texturesize "<<Texture.GetWidth()<<"x"<<Texture.GetHeight()
            <<"!= DepthMapsize "<<DenseDepthMap.GetWidth()<<"x"
            <<DenseDepthMap.GetHeight()<<std::endl);
        }
      return(GenerateDenseMesh(DenseDepthMap,DepthP,Texture,DepthP,x0,y0,
                               width,height,distBetweenPoints));
    }

    /** backproject image onto a plane in space
        so far only implemented for perspective cameras */
    int GenerateImagePlane(const BIAS::Projection& P,
                           const BIAS::Image<unsigned char>& Texture,
                           const double& w = 1.0);

    /** generates the sensor plane / cylinder / sphere in 3D space
        @author koeser
        @param texture orignal image, must be rgb !!!  */
    void GenerateTexturedCamera(const ProjectionParametersBase* PPB,
                                Image<unsigned char>& rgbtexture,
                                const double& scale = 1.0,
                                const double& opacity = 1.0,
                                const double& resolution = 1.0);

    inline
    void GetTexture(Image<unsigned char>& img) const {
      img = texture_;
    }

    inline
    const BIAS::Image<unsigned char> & GetTextureRef() const {
      return texture_;
    }

    inline
    void GetVertices(std::vector<BIAS::Vector3<double> >& vertices) const {
      vertices = meshVertices_;
    }

    inline
    const std::vector<BIAS::Vector3<double> >& GetVerticesRef() const
    {return meshVertices_;}

    inline
    void GetTexCoords(std::vector<BIAS::Vector2<float> >& texCoo) const {
      texCoo = textureCoords_;
    }

    inline
    void GetVertexColors(std::vector<BIAS::Vector4<unsigned char> >& colors) const {
    	colors= vertexColors_;
    }

    inline
    const std::vector<BIAS::Vector2<float> >& GetTexCoordsRef() const
    {return textureCoords_;}

    inline
    void GetTriangleIndices(std::vector<BIAS::Vector<int> >& triangleIndices) const {
      triangleIndices = triangleIndices_;
    }

    inline
    const std::vector<BIAS::Vector<int> >& GetTriangleIndicesRef()  const
    {return triangleIndices_;}

    inline
    void GetNormals(std::vector<BIAS::Vector3<double> >& normals) const
    {
      normals = vertexNormals_;
    }

    inline
    const std::vector<BIAS::Vector3<double> >& GetNormalsRef() const
    {
      return vertexNormals_;
    }

    inline
    int GenerateSimplifiedMesh(const Image<float>& DenseDepthMap,
                               const Projection& P,
                               const Image<unsigned char>& Texture,
                               const unsigned int x0, const unsigned int y0,
                               const unsigned int width,
                               const unsigned int height,
                               double minCosAngle = 0.8,
                               double maxDiagRatio = 3.0)
    {
      if( ( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
            (DenseDepthMap.GetHeight()!=Texture.GetHeight()) )
             && ( !(Texture.IsEmpty()) )
        ) {
        BIASWARN("Texturesize != DepthMapsize \n");
        }
      return(GenerateSimplifiedMesh(DenseDepthMap,P,Texture,P,x0,y0,width,
             height,minCosAngle,maxDiagRatio));
    }

    int GenerateSimplifiedMesh(const Image<float>& DenseDepthMap,
                               const Projection& P,
                               const Image<unsigned char>& Texture,
                               const Projection& TexP,
                               const unsigned int x0, const unsigned int y0,
                               const unsigned int width,
                               const unsigned int height,
                               double minCosAngle = 0.8,
                               double maxDiagRatio = 3.0);

    /** @brief Generate planar quad in 3D space with given texture coordinates.
               The quad will be represented by two triangles.
        @note If texture has zero width, an untextured quad is created.
    */
    int GenerateTexturedQuad(const Image<unsigned char>& Texture,
                             const Vector3<double>& UL, const Vector3<double>& UR,
                             const Vector3<double>& LL, const Vector3<double>& LR,
                             const Vector2<double>& ul, const Vector2<double>& ur,
                             const Vector2<double>& ll, const Vector2<double>& lr,
                             const Vector3<double>& normal = Vector3<double>(0,0,0));

    /** @brief Generate planar quad in 3D space with texture projection
               using projection matrix P. Will be represented by two triangles.
        @note If texture has zero width, an untextured quad is created.
              If matrix P is zero, image corners of texture will be mapped to
              quad corners.
    */
    int GenerateTexturedQuad(PMatrix P, const Image<unsigned char>& Texture,
                             const Vector3<double>& UL,
                             const Vector3<double>& UR,
                             const Vector3<double>& LL,
                             const Vector3<double>& LR,
                             const Vector3<double>& normal = Vector3<double>(0,0,0));

    /** @brief Generate a strip of planar quads in 3D space using the given
               texture coordinates. Each quad will be represented by two
               triangles.
        @attention The ordering of the 3d points in vector X and corresponding
               2d texture coordinates in vector x is upper/lower from left
               to right, i.e.:
               UL0, LL0, UR0 = UL1, LR0 = LL1, UR1 = UL2, LR1 = LL2, ...
    */
    int GenerateTexturedQuadStrip(const Image<unsigned char>& Texture,
                                  const std::vector< Vector3<double> > &X,
                                  const std::vector< Vector2<double> > &x);

    /** @brief Generate a strip of planar quads in 3D space. Projection of the
               texture onto the quad is done by the given projection matrix P.
               Each quad will be represented by two triangles.
        @attention The ordering of the 3d points in vector X is upper/lower
               from left to right, i.e.:
               UL0, LL0, UR0 = UL1, LR0 = LL1, UR1 = UL2, LR1 = LL2, ...
    */
    int GenerateTexturedQuadStrip(const Projection& P,
                                  const Image<unsigned char>& Texture,
                                  const std::vector< Vector3<double> > &X);

    /** @brief Replace the texture of this mesh. */
    int ChangeTexture(const BIAS::Image<unsigned char>& texture);

    /**
     * @brief Returns the biggest connected segment of this triangle mesh.
     *
     * @note The returned mesh will only contain the 3D points that are actually
     *       referenced by at least one triangle. Therefore the indices of the
     *       triangles may differ from the original mesh.
     *
     * @param[out] biggestSegment
     *             the biggest subsegment of this triangle mesh
     *
     * @author rwulff
     * @date   02/2012
     */
    void GetBiggestConnectedSegment(TriangleMesh& biggestSegment) const;

    /**
     * @brief Returns the minimal and maximal point of the axis-aligned
     *        bounding box of this triangle mesh.
     *
     * @note The result is not cached and recomputed each time this method is
     *       called.
     *
     * @param[out] minPoint
     *             point with minimal values in each dimension
     * @param[out] maxPoint
     *             point with maximal values in each dimension
     */
    void GetBoundingBox(Vector3<double>& minPoint,
                        Vector3<double>& maxPoint) const;

    /**
     * @brief Computes the normals for the vertices by averaging the normals of
     *        adjacent triangles.
     *
     * @param leftHanded
     *        determines direction of normals. true = compute in left-handed
     *        fashion, false = compute in right-handed fashion
     * @param normalForUnusedVertices
     *        normal to assign to vertices that do not occur in any triangle
     */
    void GenerateVertexNormals(bool leftHanded = true,
                               Vector3<double> normalForUnusedVertices
                               = Vector3<double>(-1.0, 0.0, 0.0));

  protected:

    bool MakeTriangles_(const Projection& P,
                        const Projection& TexP,
                        const BIAS::Image<unsigned char>& Texture,
                        std::vector<Image<float> > & depthmaps,
                        float x,float y, float offset,
                        float minoffset,
                        double minCosAngle, double maxDiagRatio,
                        bool simulateOnly);

    bool SimplifyMeshSurface_(const BIAS::Image<unsigned char>& Texture,BIAS::Projection& TexP, double reduceValue);
    double ThreePointsToPlane_(Vector3<double> v1, Vector3<double> v2, Vector3<double> v3, Vector4<double>& erg);
    void GetNeighbours_(int v1, int v2, std::vector<int>& dead_triangles, std::vector<int>& moved_triangles,int &moved_pivot);
    void TestContraction(int v1, int v2, int edge, bool init=false);
    
    BIAS::Image<unsigned char> texture_;
    std::vector<BIAS::Vector3<double> > meshVertices_;
    std::vector<BIAS::Vector3<double> > vertexNormals_;
    std::vector<BIAS::Vector4<unsigned char> > vertexColors_;
    std::vector<BIAS::Vector2<float> > textureCoords_;
    std::vector<BIAS::Vector<int> > triangleIndices_;

    /// a set with all neighbours of a vertex v
    std::vector<std::set<int> > neighboursOfVertex_;
    /// a set with triangles that contain the vertex v
    std::vector<std::set<int> > vertexIsInTriangles_;
    /// a set with pairs that contain the vertex v
    std::vector<std::set<int> > vertexIsInThisPairs_;
    // a set that contains the edges in the order of their quadric error value
    std::multiset<BIAS::Edge,BIAS::Edge> sorted;

    std::vector<BIAS::Edge> validPairs_;

    /// minimum angle of any corner to accept a triangle
    double minCornerAngle_;
    /// maximum angle between triangle normal and viewing ray
    double maxAngle_;

    /// apply this scale to depth values.
    double depthScale_;

    int currentIndex_;

  };  //end class


}//end namespace


#endif
