/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * FlirThermalCameraSerialControl.cpp
 *
 *  Created on: Jan 12, 2011
 *      Author: ischiller
 */

#include "FlirThermalCameraSerialControl.hh"
#include <iostream>
#include <Utils/Checksum.hh>
#include <Base/Debug/Error.hh>
#ifdef WIN32
	#include <windows.h>
#endif
#include <Base/Common/W32Compat.hh>

using namespace BIAS;
using namespace std;

FlirThermalCameraSerialControl::
FlirThermalCameraSerialControl()
{
	connected_ = false;
	serialNumber_ = "";
}

FlirThermalCameraSerialControl::
~FlirThermalCameraSerialControl()
{
	if(connected_) DisconnectCamera();
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
ConnectCamera(std::string port)
{
	FLIR_STATUS_MESSAGE ret = CAM_OK;
	cout<<"trying to open "<<port<<" for serial communication...";
  bool res = serialIO_.OpenPort(port);
  if(!res){
  	cout<<"failed!"<<endl;
  	return CAM_NOT_READY;
  }
	cout<<"success!"<<endl;
  serialIO_.SetOptions(57600,SPPM_NoParity,false,true,false,1);
  serialIO_.SetTimeOuts(5,1,10,1,10);
  connected_ = true;
  return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
DisconnectCamera()
{
	if(!connected_) return CAM_NOT_READY;
	serialNumber_ ="";
  bool ret = serialIO_.ClosePort();
  if(ret){
  	connected_ = false;
  	return CAM_OK;
  }
  return CAM_NOT_READY;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
GetSerialNumber(std::string &serial){
	FLIR_STATUS_MESSAGE ret = CAM_OK;
	unsigned char buffer[10];
	for(int i=0;i<10;i++) buffer[i]=0;
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x04; //Function serial number
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x00; //byte count LSB

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB
	Checksum::CalcCRC_CCITT16(crc2,buffer,8);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	buffer[8]=(unsigned char)msb; //CRC2 MSB
	buffer[9]=(unsigned char)lsb; //CRC2 LSB
	ret = SendCommand_(buffer,10);
	if(ret !=CAM_OK){ cout<<__LINE__<<StatusCodeToString(ret)<<endl; return ret;}
	biasusleep(500);
	unsigned char buffer2[18];
	ret = GetAnswer_(buffer2,18,false);
	if(ret !=CAM_OK){ cout<<__LINE__<<StatusCodeToString(ret)<<endl; return ret;}

	cout<<"GetSerial Answer:"<<endl;
	for(unsigned i=0;i<18;i++)
		cout<<showbase<<hex<<(int)buffer2[i]<<" ";
	cout<<noshowbase<<dec<<endl;

	//5 and 6 is byte count
	unsigned int c1 = (unsigned)buffer2[4];
	unsigned int c2 = (unsigned)buffer2[5];
  cout<<"Byte counts:"<<c1<<","<<c2<<endl;
	//c2 must be 8 byte
	if(c1 !=0 || c2 != 8) {
		return CAM_BYTE_COUNT_ERROR;
	}
	serialNumber_ = "";

  //from bytes 8 to 8+c2 there is the data, 
  //4 bytes for camera serial number (0-1 high word, 2-3 low word)
  //and 4 bytes for sensor serial number (0-1 high word, 2-3 low word) 
  char a = buffer2[8],  b = buffer2[9], c = buffer2[10], d =buffer2[11];
  char tmp[4]={a,b,c,d};
  a = buffer2[12];  b = buffer2[13]; c = buffer2[14]; d =buffer2[15];
  char tmp2[4]={a,b,c,d};

  int camSerNum= tmp[0]; 
  camSerNum= camSerNum<<8;
  camSerNum+= tmp[1]; 
  camSerNum= camSerNum<<8;
  camSerNum+= tmp[2];
  camSerNum= camSerNum<<8;
  camSerNum+= tmp[3]; 
 
  int senSerNum= tmp2[0]; 
  senSerNum= senSerNum<<8;
  senSerNum+= tmp2[1]; 
  senSerNum= senSerNum<<8;
  senSerNum+= tmp2[2];
  senSerNum= senSerNum<<8;
  senSerNum+= tmp2[3]; 

  stringstream txt; txt << camSerNum <<":"<<senSerNum;
	serialNumber_ = txt.str();
  serial=serialNumber_;
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetDefaults(){
	unsigned char buffer[10];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x01; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x00; //byte count LSB

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
	//cout<<"Checksum1: "<<showbase<<hex<<crc1<<endl;
	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	//cout<<"msb: "<<showbase<<hex<<(int)msb<<endl;
	//cout<<"lsb: "<<showbase<<hex<<(int)lsb<<endl;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB
	//no data

	Checksum::CalcCRC_CCITT16(crc2,buffer,8);
	//cout<<"Checksum2: "<<hex<<crc2<<endl;
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	//cout<<"msb: "<<showbase<<hex<<msb<<endl;
	//cout<<"lsb: "<<showbase<<hex<<lsb<<endl;

	buffer[8]=(unsigned char)msb; //CRC2 MSB
	buffer[9]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,10);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,10);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetColoring(unsigned color){
	cout<<"Setting Video color:"<<color<<endl;
	if(color > 12) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x10; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
//	cout<<showbase<<endl;
//	cout<<"Checksum1: "<< hex <<crc1<<endl;
	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
//	cout<<"msb: "<<showbase<<hex<<(int)msb<<endl;
//	cout<<"lsb: "<<showbase<<hex<<(int)lsb<<endl;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(color);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	//cout<<"Checksum2: "<<hex <<crc2<<endl;
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	//cout<<"msb: "<<showbase<<hex<<msb<<endl;
	//cout<<"lsb: "<<showbase<<hex<<lsb<<endl;

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetFFCMode(unsigned mode){
  cout<<noshowbase<<dec<<endl;
	cout<<"Setting FFC Mode:"<<mode<<endl;
	if(mode != 0 && mode != 1 && mode != 2) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x0B; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
	cout<<showbase<<endl;
	cout<<"Checksum1: "<< hex <<crc1<<endl;
	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	cout<<"msb: "<<showbase<<hex<<(int)msb<<endl;
	cout<<"lsb: "<<showbase<<hex<<(int)lsb<<endl;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(mode);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	cout<<"Checksum2: "<<hex <<crc2<<endl;
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	cout<<"msb: "<<showbase<<hex<<msb<<endl;
	cout<<"lsb: "<<showbase<<hex<<lsb<<endl;

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetFFCInterval(unsigned interval, unsigned intervalLG)
{ cout<<noshowbase<<dec<<endl;
  cout<<"Setting FFC Interval:"<<interval<<","<<intervalLG<<endl;
	unsigned char buffer[14];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x0D; //Function, FFC_PERIOD
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x04; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
	cout<<showbase<<endl;

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//4 bytes data
	buffer[8]=(unsigned char)(interval>>8);
	buffer[9]=(unsigned char)(interval);
  buffer[10]=(unsigned char)(intervalLG>>8);
	buffer[11]=(unsigned char)(intervalLG);


	Checksum::CalcCRC_CCITT16(crc2,buffer,12);
	
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	
	buffer[12]=(unsigned char)msb; //CRC2 MSB
	buffer[13]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,14);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,14);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetFFCTempChange(unsigned tmpChange,unsigned tmpChangeLG)
{
  cout<<noshowbase<<dec<<endl;
  cout<<"Setting FFC TmpChange:"<<tmpChange<<","<<tmpChangeLG<<endl;
	unsigned char buffer[14];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x0E; //Function, FFC_PERIOD
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x04; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);
	cout<<showbase<<endl;

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//4 bytes data
	buffer[8]=(unsigned char)(tmpChange>>8);
	buffer[9]=(unsigned char)(tmpChange);
  buffer[10]=(unsigned char)(tmpChangeLG>>8);
	buffer[11]=(unsigned char)(tmpChangeLG);


	Checksum::CalcCRC_CCITT16(crc2,buffer,12);
	
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	
	buffer[12]=(unsigned char)msb; //CRC2 MSB
	buffer[13]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,14);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,14);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetGainMode(unsigned mode){
  cout<<"Setting Gain Mode:"<<mode<<endl;
	if(mode > 3) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x0A; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(mode);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetExternalSync(unsigned mode){
  cout<<"Setting External Sync Mode:"<<mode<<endl;
	if(mode > 2) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x21; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(mode);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetVideoMode(unsigned level){
	cout<<"Setting Video Mode:"<<level<<endl;
	if(level != 0 && level != 1 && level != 4 && level != 8)
		return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x0F; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(level);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}

/** \brief Set the algorithm for Automatic Gain Control*/
FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetAGCAlgorithm(unsigned algo){
  cout<<"Set AGC algorithm:"<<algo<<endl;
  if(algo > 5) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x13; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(algo);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB


	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetVideoOrientation(unsigned orientation){
	cout<<"Setting Video Orientation:"<<orientation<<endl;
	if(orientation > 3) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x11; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(orientation);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB


	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetVideoStandard(unsigned std){
	if(std > 1) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x72; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(std);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetSpotMeterMode(unsigned spotmeter){
	if(spotmeter > 2) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x1F; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(spotmeter);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB
	cout<<"Setting spot meter mode:"<<spotmeter<<endl;

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetSpotMeterDisplay(unsigned display){
	if(display > 3) return CAM_RANGE_ERROR;

	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x2B; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x02; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//2 bytes data
	buffer[8]=0x00;
	buffer[9]=(unsigned char)(display);

	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);

	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	cout<<"Setting spot display:"<<display<<endl;
	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK){ cout<<StatusCodeToString(ret)<<endl; return ret;}
	ret = GetAnswer_(buffer,12);

	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
GetSpotMeterCelcius(double& degree){

	unsigned char buffer[10];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x43; //Function
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x00; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//No Data

	Checksum::CalcCRC_CCITT16(crc2,buffer,8);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	buffer[8]=(unsigned char)msb; //CRC2 MSB
	buffer[9]=(unsigned char)lsb; //CRC2 LSB

	cout<<"Get spot value:"<<endl;
	//sent Get Message
	cout<<"Writing:";
	for(unsigned i=0;i<10;i++)
		cout<<showbase<<hex<<(int)buffer[i]<<" ";
	cout<<noshowbase<<dec<<endl;

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,10);
	if(ret !=CAM_OK) return ret;

	unsigned char buffer2[12];
	ret = GetAnswer_(buffer2,12);
	if(ret!= CAM_OK) return ret;

	for(unsigned i=0;i<12;i++)
		cout<<showbase<<hex<<(int)buffer2[i]<<" ";
	cout<<noshowbase<<dec<<endl;

  // in 8 the temperature in C*10 is stored
	degree = buffer2[8];
  degree*=0.1;
	return ret;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
GetSensorTemp(double& degree){
	unsigned char buffer[12];
	unsigned int crc1,crc2;

	buffer[0]=0x6E; //process code
	buffer[1]=0x00; //Status
	buffer[2]=0x00; //Reserved
	buffer[3]=0x20; //Function read sensor
	buffer[4]=0x00; //byte count MSB
	buffer[5]=0x00; //byte count LSB //2 bytes

	Checksum::CalcCRC_CCITT16(crc1,buffer,6);

	unsigned int msb = crc1>>8;
	unsigned int lsb = (unsigned char)crc1;
	buffer[6]=(unsigned char)msb; //CRC1 MSB
	buffer[7]=(unsigned char)lsb; //CRC1 LSB

	//No Data
	buffer[8] =0x00;
	buffer[9] =0x00;
	Checksum::CalcCRC_CCITT16(crc2,buffer,10);
	msb  = crc2>>8;
	lsb = (unsigned char)(crc2);
	buffer[10]=(unsigned char)msb; //CRC2 MSB
	buffer[11]=(unsigned char)lsb; //CRC2 LSB

	cout<<"Get sensor tmp:"<<endl;
	//sent Get Message
	cout<<"Writing:";
	for(unsigned i=0;i<12;i++)
		cout<<showbase<<hex<<(int)buffer[i]<<" ";
	cout<<noshowbase<<dec<<endl;

	FLIR_STATUS_MESSAGE ret = SendCommand_(buffer,12);
	if(ret !=CAM_OK) return ret;

	ret = GetAnswer_(buffer,12);
	if(ret!= CAM_OK) return ret;

	for(unsigned i=0;i<12;i++)
		cout<<showbase<<hex<<(int)buffer[i]<<" ";
	cout<<noshowbase<<dec<<endl;

  // in 8 the temperature in C*10 is stored
	degree = buffer[8];
  degree*=0.1;
	return ret;
}


FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetOptimization(unsigned optimization){
	return CAM_NOT_READY;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetContrast(unsigned contrast){
	return CAM_NOT_READY;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SetBrightness(unsigned brigthness){
	return CAM_NOT_READY;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
SendCommand_(unsigned char* buf, int length)
{
	if(!connected_ || length < 2) return CAM_NOT_READY;
	serialIO_.WriteBytes(buf,length);
	return CAM_OK;
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
GetAnswer_(unsigned char* buf, int length, bool wait)
{
	if(!connected_ || length < 2) return CAM_NOT_READY;
	//clear buf to wipe memory
	for (int i=0;i<length;i++) buf[i]=0;
	serialIO_.ReadBytes(buf,length,wait);
	return CheckCameraReply_(buf);
}

FLIR_STATUS_MESSAGE FlirThermalCameraSerialControl::
CheckCameraReply_(unsigned char* buf){
	if(buf[1]==0x00) return CAM_OK;
	else if(buf[1]==0x01) return CAM_BUSY;
	else if(buf[1]==0x02) return CAM_NOT_READY;
	else if(buf[1]==0x03) return CAM_RANGE_ERROR;
	else if(buf[1]==0x04) return CAM_CHECKSUM_ERROR;
	else if(buf[1]==0x05) return CAM_UNDEFINED_PROCESS_ERROR;
	else if(buf[1]==0x06) return CAM_UNDEFINED_FUNCTION_ERROR;
	else if(buf[1]==0x07) return CAM_TIMEOUT_ERROR;
	else if(buf[1]==0x08) return CAM_BYTE_COUNT_ERROR;
	else if(buf[1]==0x09) return CAM_FEATURE_ERROR;
	else if(buf[1]==0x0a) return CAM_NOT_READY;
	else return CAM_GENERAL_ERROR;
}

std::string FlirThermalCameraSerialControl::
StatusCodeToString(FLIR_STATUS_MESSAGE code){
	if(code == CAM_OK) return "CAM_OK";
	else if(code== CAM_BUSY)return "CAM_BUSY";
	else if(code==CAM_NOT_READY)return "CAM_NOT_READY";
	else if(code==CAM_RANGE_ERROR)return "CAM_RANGE_ERROR";
	else if(code==CAM_CHECKSUM_ERROR)return "CAM_CHECKSUM_ERROR";
	else if(code==CAM_UNDEFINED_PROCESS_ERROR)return "CAM_UNDEFINED_PROCESS_ERROR";
	else if(code==CAM_UNDEFINED_FUNCTION_ERROR)return "CAM_UNDEFINED_FUNCTION_ERROR";
	else if(code==CAM_TIMEOUT_ERROR)return "CAM_TIMEOUT_ERROR";
	else if(code==CAM_BYTE_COUNT_ERROR)return "CAM_BYTE_COUNT_ERROR";
	else if(code==CAM_FEATURE_ERROR)return "CAM_FEATURE_ERROR";
	else if(code==CAM_NOT_READY)return "CAM_NOT_READY";
	else return "CAM_GENERAL_ERROR";
}
