/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIASVIDEOSOURCE_DCAM_HH__
#define __BIASVIDEOSOURCE_DCAM_HH__

#include <bias_config.h>
#include <string>

#if defined(WIN32) && !defined(BIAS_HAVE_1394CAMERACMU)
#  error You need 1394CAMERACMU for BIAS Videosource_DCAM. Please recompile BIAS with USE_1394CAMERACMU.
#endif

#if !defined(WIN32) && !defined(BIAS_HAVE_DC1394)
#  error You need DC1394 for BIAS Videosource_DCAM. Please recompile BIAS with USE_DC1394.
#endif

// consistency check, DCAM is set if one of the above is set
#if !defined(BIAS_HAVE_1394CAMERACMU) && !defined(BIAS_HAVE_DC1394)
#  error You need either DC1394 or 1394CameraCMU  for BIAS Videosource_DCAM. Please recompile BIAS with one of them.
#endif




#include <Base/Common/W32Compat.hh>

#ifndef WIN32
#  include <sys/time.h>
#endif
#ifdef BIAS_HAVE_PTHREADS
#  include <pthread.h>
#endif

#include <Image/Camera.hh>

#ifdef WIN32
#  include <1394Camera.h>
//typedef unsigned long long int u_int64_t;
typedef unsigned long long int uint64_t;
#else
#  include <dc1394/control.h>
#  include <dc1394/utils.h>
#endif

#include <VideoSource/VideoSource_Base.hh>

// only this number of cards are probed
#ifndef MAX_1394_CARDS
#define MAX_1394_CARDS 2
#endif

#define MAX_1394_SOURCES 16
#ifndef MAX_1394_CHANNELS
#define MAX_1394_CHANNELS 256
#endif

#define MAX_DMABUFFERS 10




namespace BIAS {

#ifdef WIN32
  enum dc_videomode;
  typedef dc_videomode dc1394video_mode_t;
#endif

/**
 * \struct CameraParameter_s
 * \brief Stores the camera parameters such as Shutter, Gain, etc.
 * Only used internally in VideoSource_DCAM
 */
typedef struct CameraParameter_s
{
  // these are the real parameters
  float Shutter;
  float Gain;
  unsigned int WhiteBalance;
  bool AutoShutter;
  bool AutoGain;
  bool AutoWhiteBalance;
  unsigned char Sharpen;
  unsigned char GammaMode;
  unsigned char OpticalFilter;

  // these bools describe active changes
  bool ShutterChanged;
  bool GainChanged;
  bool WhiteBalanceChanged;
  bool AutoShutterChanged;
  bool AutoGainChanged;
  bool AutoWhiteBalanceChanged;
  bool SharpenChanged;
  bool GammaModeChanged;
  bool OpticalFilterChanged;

  // one-push  action
  bool WhiteBalanceOnePush;

} CameraParameterSet;

/**
 * \struct CameraInfo_s
 * \brief Stores camera information such as model, vendor, framerates.
 * Only used internally in VideoSource_DCAM
 */
typedef struct CameraInfo_s
{
  int camid;
  std::string vendor;
  std::string model;
  std::vector<std::vector<float> > framerates;
  bool b_capable;
  float bestFramerate;
#ifdef WIN32
  BIAS::dc1394video_mode_t bestMode;
  std::vector<BIAS::dc1394video_mode_t> videoModes;
  std::vector<BIAS::dc1394video_mode_t> f7_modes;
  LARGE_INTEGER guid;
#else
  dc1394video_mode_t bestMode;
  std::vector<dc1394video_mode_t> videoModes;
  std::vector<dc1394video_mode_t> f7_modes;
  uint64_t guid;
#endif
  std::vector<unsigned int> f7_maxWidth;
  std::vector<unsigned int> f7_maxHeight;
  std::vector<unsigned int> f7_unitWidth;
  std::vector<unsigned int> f7_unitHeight;
  std::vector<unsigned int> f7_MinBpP;
  std::vector<unsigned int> f7_MaxBpP;
  std::vector<unsigned int> f7_bestBpP;
}CameraInfo;


/// support function to get the number of cameras and IDs of all cameras on one bus
extern int BIASVideoSource_EXPORT ScanBus(std::stringstream &ofs);
extern int BIASVideoSource_EXPORT ScanBus(std::vector<CameraInfo> &camInfos);
/**
 * \enum CameraModel_e
 * \brief Enumeration for the known IEEE1394 camera models
 * Only used internally in VideoSource_DCAM
 */
typedef enum CameraModel_e
{
	CAMERA_SONY_DFW_X700  = 0,        //IEEE1394
	CAMERA_SONY_DFW_V500,             //IEEE1394
	CAMERA_PTGREY_DRAGONFLY_COLOR ,   //IEEE1394
	CAMERA_PTGREY_DRAGONFLY_HIRES ,   //IEEE1394
	CAMERA_PTGREY_DRAGONFLY_BW ,      //IEEE1394
	CAMERA_PTGREY_DRAGONFLY2_COLOR_HIRES,
	CAMERA_PTGREY_SCORPION_COLOR_1600,//IEEE1394
	CAMERA_PTGREY_FIREFLY_COLOR,      //IEEE1394
	CAMERA_PTGREY_FIREFLY_BW,         //IEEE1394
	CAMERA_PTGREY_GRASSHOPPER_COLOR,  //IEEE1394
	CAMERA_PTGREY_FLEA2_BW,           //IEEE1394b
	CAMERA_PTGREY_BUMBLEBEE_XB3,      //IEEE1394b
	CAMERA_SONY_EVID31,               //analog via frame grabber, PTZ-camera
	CAMERA_SONY_DV,
	CAMERA_UNIBRAIN_FIREI,
	CAMERA_NET_FO323SC,
	CAMERA_KAMERAWERK_DRESDEN_LOGILUX_I5,
	CAMERA_PTGREY_BUMBLEBEE_XB2,      //IEEE1394
	CAMERA_UNKNOWN
} CameraModel;

/**
 * \enum LensModel_e
 * \brief Enumeration for the known lenses for IEEE1394 cameras
 * Only used internally in VideoSource_DCAM
 */
typedef enum LensModel_e{
	LENS_SONY_EVID31_BUILTIN,
	LENS_SONY_DV_BUILTIN,
	LENS_PENTAX_CMOUNT_4_8,
	LENS_PENTAX_CMOUNT_8_0,
	LENS_PENTAX_CMOUNT_4_2,
	LENS_UNKNOWN
} LensModel;


#ifdef WIN32
	/**\enum dc1394feature_t
	 * \brief  only define these values in windows. In linux they are defined
	 * in dc1394/control.h	(NOT in namespace BIAS)	*/
  typedef enum dc1394feature_t{
    DC1394_FEATURE_BRIGHTNESS,
    DC1394_FEATURE_EXPOSURE,
    DC1394_FEATURE_SHARPNESS,
    DC1394_FEATURE_WHITE_BALANCE,
    DC1394_FEATURE_HUE,
    DC1394_FEATURE_SATURATION,
    DC1394_FEATURE_GAMMA,
    DC1394_FEATURE_SHUTTER,
    DC1394_FEATURE_GAIN,
    DC1394_FEATURE_IRIS,
    DC1394_FEATURE_FOCUS,
    DC1394_FEATURE_TEMPERATURE,
    DC1394_FEATURE_TRIGGER,
    DC1394_FEATURE_ZOOM,
    DC1394_FEATURE_PAN,
    DC1394_FEATURE_TILT,
    DC1394_FEATURE_OPTICAL_FILTER,
    DC1394_FEATURE_CAPTURE_SIZE,
    DC1394_FEATURE_CAPTURE_QUALITY
  } ;

  /**\enum dc1394feature_mode_t
   * \brief  only define these values in windows. In linux they are defined
   * in dc1394/control.h	(NOT in namespace BIAS)	*/
  typedef enum dc1394feature_mode_t{
    DC1394_FEATURE_MODE_AUTO,
    DC1394_FEATURE_MODE_MANUAL
  };

  /**\enum dc1394color_coding_t
   * \brief  only define these values in windows. In linux they are defined
   * in dc1394/control.h	(NOT in namespace BIAS)	*/
  typedef enum dc1394color_coding_t{
      DC1394_COLOR_CODING_MONO8= 352,
      DC1394_COLOR_CODING_YUV411,
      DC1394_COLOR_CODING_YUV422,
      DC1394_COLOR_CODING_YUV444,
      DC1394_COLOR_CODING_RGB8,
      DC1394_COLOR_CODING_MONO16,
      DC1394_COLOR_CODING_RGB16,
      DC1394_COLOR_CODING_MONO16S,
      DC1394_COLOR_CODING_RGB16S,
      DC1394_COLOR_CODING_RAW8,
      DC1394_COLOR_CODING_RAW16
  } ;

#define FEATURE_MIN   FEATURE_BRIGHTNESS
#define FEATURE_MAX   FEATURE_CAPTURE_QUALITY
#define NUM_FEATURES (FEATURE_MAX - FEATURE_MIN + 1)


  /**\enum dc_videomode
   * \brief Enumeration of video modes. Only define these values in windows.
   * In linux they are defined in dc1394/control.h	(NOT in namespace BIAS)	*/
  typedef enum dc_videomode{
    DC1394_VIDEO_MODE_160x120_YUV444= 64,
    DC1394_VIDEO_MODE_320x240_YUV422,
    DC1394_VIDEO_MODE_640x480_YUV411,
    DC1394_VIDEO_MODE_640x480_YUV422,
    DC1394_VIDEO_MODE_640x480_RGB8,
    DC1394_VIDEO_MODE_640x480_MONO8,
    DC1394_VIDEO_MODE_640x480_MONO16,
    DC1394_VIDEO_MODE_800x600_YUV422,
    DC1394_VIDEO_MODE_800x600_RGB8,
    DC1394_VIDEO_MODE_800x600_MONO8,
    DC1394_VIDEO_MODE_1024x768_YUV422,
    DC1394_VIDEO_MODE_1024x768_RGB8,
    DC1394_VIDEO_MODE_1024x768_MONO8,
    DC1394_VIDEO_MODE_800x600_MONO16,
    DC1394_VIDEO_MODE_1024x768_MONO16,
    DC1394_VIDEO_MODE_1280x960_YUV422,
    DC1394_VIDEO_MODE_1280x960_RGB8,
    DC1394_VIDEO_MODE_1280x960_MONO8,
    DC1394_VIDEO_MODE_1600x1200_YUV422,
    DC1394_VIDEO_MODE_1600x1200_RGB8,
    DC1394_VIDEO_MODE_1600x1200_MONO8,
    DC1394_VIDEO_MODE_1280x960_MONO16,
    DC1394_VIDEO_MODE_1600x1200_MONO16,
    DC1394_VIDEO_MODE_EXIF,
    DC1394_VIDEO_MODE_FORMAT7_0,
    DC1394_VIDEO_MODE_FORMAT7_1,
    DC1394_VIDEO_MODE_FORMAT7_2,
    DC1394_VIDEO_MODE_FORMAT7_3,
    DC1394_VIDEO_MODE_FORMAT7_4,
    DC1394_VIDEO_MODE_FORMAT7_5,
    DC1394_VIDEO_MODE_FORMAT7_6,
    DC1394_VIDEO_MODE_FORMAT7_7
  };

 //typedef video_mode dc1394video_mode_t;

  #define DC1394_VIDEO_MODE_MIN     DC1394_VIDEO_MODE_160x120_YUV444
  #define DC1394_VIDEO_MODE_MAX       DC1394_VIDEO_MODE_FORMAT7_7
  #define DC1394_VIDEO_MODE_NUM      (DC1394_VIDEO_MODE_MAX - DC1394_VIDEO_MODE_MIN + 1)

  // Special min/max are defined for Format_7
  #define DC1394_VIDEO_MODE_FORMAT7_MIN       DC1394_VIDEO_MODE_FORMAT7_0
  #define DC1394_VIDEO_MODE_FORMAT7_MAX       DC1394_VIDEO_MODE_FORMAT7_7
  #define DC1394_VIDEO_MODE_FORMAT7_NUM      (DC1394_VIDEO_MODE_FORMAT7_MAX - DC1394_VIDEO_MODE_FORMAT7_MIN + 1)

#endif // WIN32

  /**
     \class VideoSource_DCAM
     \ingroup g_videosource
     \author evers/angerer

     \brief This class VideoSource_DCAM implements access to IEEE1394 (Firewire, iLink)
     cameras following the DCam 2.0 standard protocol.
  */
  class BIASVideoSource_EXPORT VideoSource_DCAM : public VideoSource{

  public:
    /**
       @brief Activates the timestamp for PointGrey Scorpion SCOR-20SO
    */
    void ActivateScorpionTimestamp();
    bool ScorpionTimestampActive_;

     /**
        @brief Constructor with IEEE1394 card
        @param Card The IEEE1394 card on which the camera is
    */
    VideoSource_DCAM(int Card = 0);

    /**
        @brief Destructor
    */
    ~VideoSource_DCAM();

    /**
        @brief Opens a dc1394 device
        @return int, <0 on failure, 0 on success
    */
    int OpenDevice();
    /**
        @brief Opens a dc1394 device with id camid
        @param camid The camera id of the camera to open
        @return int, <0 on failure, 0 on success
    */
    int OpenDevice(int camid);

    /**
        @brief this a special function to by-pass all automatic initializations and
        Use only in conjuction with SetSize() and SetLeftTop() !
    */
    int SetFormat7(int mode, int bpp, dc1394color_coding_t colorCoding = DC1394_COLOR_CODING_MONO8);

    /**
     * @brief set's the id of a camera manually, has to be called before OpenDevice()
     * @param manualcamid_ the unique id of the camera
     */
    void SetCam_ID(int manualcamid_);

    /**
      @brief closes a camera device and cleans up

    */
    int CloseDevice();

    // evil hack, fixme, please fixme !!! (I forgot the TODO)
    void SetBinning(bool b) {UseBinning_ = b;};

//    /**
//      @brief Select the port (bus-id) to use, only valid before OpenDevice()
//      @param channel The port (bus-id) to use
//      @return int, <0 on failure, 0 on success
//    */
//    int UsePort(int channel);

    /**
     * @brief fetches one image from the camera and writes it int a Camera structure
     * @param &image the Camera structure to write the image from the camera in
     * @return int, <0 on failure, 0 on success
     */
    int GrabSingle(BIAS::Camera <unsigned char> &image);
#ifdef BUILD_IMAGE_USHORT
    int GrabSingle(BIAS::Camera <unsigned short> &image);
#endif
#ifdef WIN32
	/**
     * @brief make GrabSingle faster, include the converts
     * @param  image  
     * @return int, <0 on failure, 0 on success
     */
    int GrabSingleAndConvert(BIAS::Image <unsigned char> *dest, enum ImageBase::EColorModel targetColorModel);
    /**
     * @brief make GrabSingle more faster for grey, no Bayer conversation
     * @param  image  
     * @return int, <0 on failure, 0 on success
     */
    int GrabSingleAndConvertForGrey(BIAS::Image <unsigned char> *dest);
     
#endif
    /**
     * @brief initializes the iso transmission of a camera. this has to be called
     * before capturing any image
     *
     * @return int, <0 on failure, 0 on success
     */
    int PreGrab();

    /**
     * @brief stops iso transmission of a camera. this has to be called
     * after capturing any image
     *
     * @return int, <0 on failure, 0 on success
     */
    int PostGrab();

    //int ShowCapabilities();
    /**
     * @brief set the shutter to auto or manual mode
     * @param b true if the shutter should be set to auto mode, false if not
     */
    void SetAutoShutter(bool b);

    /**
     * @return returns true if the shutter is set to automode, false if not
     */
    bool GetAutoShutter();

    /**
     * @brief set the gain to auto or manual mode
     * @param b true if the gain should be set to auto mode, false if not
     */
    void SetAutoGain(bool b);


    void SetAutoBrightness(bool b);

    bool GetAutoBrightness();

    /**
     * @return returns true if the gain is set to automode, false if not
     */
    bool GetAutoGain();

    /**
      @brief Set a camera feature to auto or to manual mode
      @param feature the feature to set
      @param mode the mode to set as enum, use DC1394_FEATURE_MODE_AUTO for auto mode
      or DC1394_FEATURE_MODE_MANUAL for manual mode
      @return int, <0 on failure, 0 on success;
    */
    int AutoMode( dc1394feature_t feature, dc1394feature_mode_t mode);

    /**
      @brief Set a camera feature by hand
      @param feature the feature to set
      @param value the value to be set for the feature feature
      @return int, <0 on failure, 0 on success
    */
    int SetFeature(dc1394feature_t feature, unsigned int value);

    /**
      @brief get the value of a specific feature
      @param feature the feature of which the value should be requested
      @return unsigned int, the value of the feature
    */
    unsigned int GetFeature( dc1394feature_t feature);

    /**
      @brief
      @param feature the feature to set one push automode for
      @return int, <0 on failure, 0 on success
    */
    int  OnePushAuto(dc1394feature_t feature);

#ifndef WIN32
    /**
     * @brief is a feature available at OnePush?
     * @return returns true, if the feature is available at OnePush mode
     */
    bool OnePushRunning(dc1394feature_t feature);

    /**
     * @brief selects the colormodel for a specific mode
     * @param videoMode the videomode to find the colormodel for
     * @return int, <0 on failure, 0 on success
     */
    int SetColorFromMode_(dc1394video_mode_t videoMode);
#else
    int SetColorFromMode_();
    dc1394video_mode_t FormatAndModeToEnum(int Format, int Mode);
#endif

    bool HasControlShutter(){return true;};
    bool HasControlGain(){return true;};
    bool HasControlBrightness(){return true;};
    bool HasControlWhiteBalance(){return true;};

    /**
     * @brief set value for shutter
     * @param exptime Exposuretime in milliseconds
     * @return int <0 on failure, 0 on success
     */
    int   SetShutter(float exptime);

    /**
     * @brief get the value of the shutter feature
     * @return float returns the shutter value
     */
    float GetShutter();

    /**
     * @brief set value for gain
     * @param g gain in dB
     * @return int, <0 on failure, 0 on success
     */
    int   SetGain(float g);

    /**
     * @brief get the value of the gain feature
     * @return float returns the gain value
     */
    float GetGain();


    /**
     * @brief set value for brightness
     * @param b brightness
     * @return int, <0 on failure, 0 on success
     */
    int SetBrightness(float b);

    /**
     * @brief get the value of the brightness feature
     * @return float returns the brightness value
     */
    float GetBrightness();

    /**
     * @brief sets the white balance of the camera,
     * rvalue and bvalue are in range -0.5 to 0.5
     * @param rvalue
     * @param bvalue
     * @return int, <0 on failure, 0 on success
     */
    int SetWhiteBalance(float rvalue,float bvalue);


    /**
     * @brief set Gamma Mode
     * @param g the gamma mode as unsigned char
     * @return int, <0 on failure, 0 on success
     */
    int SetGammaMode(int g);

    
    /**
        * @brief set External Trigger mode, overrides Base class function
        * @param trigger, bool to activate or deactivate trigger        
        */
    void SetExternalTrigger(bool trigger = true);
    
    /**
     * @brief get the value of Gamma Mode
     * @return unsigned char, returns the value of Gamma Mode
     */
    int GetGammaMode();

  
    // format7 specific methods
    inline unsigned int GetTop(){return Top_;};
    inline unsigned int GetLeft(){return Left_;};


    /**
     * @brief set the topleft corner for partitial scan
     * @param left distance from left
     * @param top distance from top
     * @return int, <0 on failure, 0 on success
     */
    int SetLeftTop(unsigned int left, unsigned  int top);

    inline int GetTransmissionTime(){return TransmissionTime_;};

    /**
     * @brief set camera to parameters specified in complete set
     * @param camparam the CameraParameterSet with the camera parameters to be set
     * @return int, <0 on failure, 0 on success
     */
    int SetCamera(CameraParameterSet camparam);

    virtual void SetCameraID(int /*id*/){
      BIASERR("Can not set camera id for DCams, they are identified automatically");
    };

    // void WriteFeatureDetails(int feature, bool *SuppFeatDetails_[])

    /**
     * @brief this choses f7 mode even for standard resolution
     * @param forcef7 true if format 7 should be forced, false if not
     */
    void ForceFormat7(bool forcef7) {ForceF7_ =forcef7;}

    /**
     * @brief sets the bytes per packet used for format 7
     * @param b the bytes per packet
     */
    void SetBytesPerPacket(const int b){BytesPerPacket_ = b;}

    void SetModeAndFramerate(dc1394video_mode_t mode, float framerate);

    void SetFirewireB(bool on);

    /**
     * @brief selects the default framerate for a cam
     * @return int, <0 on failure, 0 on success
     */
    int SelectDefault();
    void EnsureLatestFrame(const bool ensure) {
      bEnsureLatestFrame_ = ensure;
    }

    void GetFormat7ImageDimensions(int &maxWidth, int &maxHeight,
                                   int &minBpP, int &maxBpP){
      maxWidth = MaxWidth_;
      maxHeight = MaxHeight_;
      minBpP = MinBpP_;
      maxBpP = MaxBpP_;
    };

    /**
     * @brief returns min and max bytes per packet for a specified resolution in format 7
     * @param camid the id of the cam
     * @param mode the format 7 mode
     * @param width of the resulting picture
     * @param height of the resulting picture
     * @param bppmin here the minimum bytes per packet are stored
     * @parma bppmax here the maximum bytes per packet are stored
     */
    void GetBytesPerPacketForFormat7(int camid, dc1394video_mode_t mode, int width, int height, unsigned int& bppmin, unsigned int& bppmax);
    /**
    Sets the Camera model
    */
    void  SetCameraModel( CameraModel model);

    /** \brief returns a string from a IEEE1394 video mode*/
    static std::string VideoModeToString(dc1394video_mode_t mode);



  protected:

  	
  	 /**
       * @brief if external triggering is true, activate it here in Opendevice  
       * @param trigger true if external trigger should be acticated, false if deactivated
       * @return int, <0 on failure, 0 on success
       */
      int ActivateExternalTrigger_(bool trigger = true);

    /**
     * @brief initializes the camera and tries to set the optimal settings for the camera if no settings are specified
     * @return int, <0 on failure, 0 on success
     */
    int InitCam();
    //int StartIso();
    //int StopIso();

    /**
     * @brief querys the cam for available modes, framerates and features
     * @return int, <0 on failure, 0 on success
     */
    int QueryCam();

    /**
     * @brief tries to find a camera mode from user given width and height
     * @return int,  <0 on failure, 0 on success
     */
    int UserSizeToMode();

    /**
     * @brief for all formats with 16bit per pixel per channel (aka short int)
     * @return int,  <0 on failure, 0 on success
     */
    int UserSizeToMode_16bit_();

    /**
     * @brief for all formats with standard 8bpp per channel (unsigned char image)
     * @return int,  <0 on failure, 0 on success
     */
    int UserSizeToMode_8bit_();

    /**
     * @brief if user requested format 7
     * @return int,  <0 on failure, 0 on success
     */
    int UserSizeToFormat7_();
    //void PrintFormatAndMode_();

    /**
     * @brief Set this value and read back what really happend.
     * @return int, <0 on failure, 0 on success
     */
    int ApplyBytesPerPacket();

    /**
     * @brief Evaluate Min and Max BpP, framesize and fps, then calc BpP
     * @return unsigned int, returns the bytesperpackage calculated from
     * Min and Max BpPm, framesize and fps
     */
    unsigned int CalcBytesPerPacket();

    // write file with camera infos and features
    //void WriteCamInfos_();
    //void WriteFeatureDetails(int feature, std::ofstream *fout);

    int ShowCapabilities();

    // there is only one handle per bus for multiple devices
#ifdef WIN32
    C1394Camera camera;
    CameraModel GetModelForEuid(LARGE_INTEGER euid);
    unsigned long DcamFps_;
    unsigned long DcamMode_;
#else // WIN32

    static int InstancesTotal_;

    /**
     * @brief searches for the correct cammodel by means of the euid of a camera
     * @return CameraModel, returns the CameraModel specified by the euid out of an enum
     */
    CameraModel GetModelForEuid(const  u_int64_t  euid);
    int camid;

    dc1394error_t err;
    dc1394camera_t * camera;
    dc1394camera_list_t *cameraList;
    dc1394_t *dcbus;
    uint32_t numCamsOnBus;
    dc1394video_modes_t CameraModes_;
    dc1394framerates_t CameraFramerates_[DC1394_VIDEO_MODE_NUM];
    dc1394featureset_t DcamFeatures_;
    dc1394video_mode_t DcamMode_;
    dc1394framerate_t DcamFps_;
    dc1394trigger_mode_t Trigger_;
#endif // WIN32

	dc1394color_coding_t f7ColorCoding_;

    int NumCams_;
    int SelectedCam_;
    int DcamSpeed_;

    int DcamFormat_;
    int DcamIsoChannel_;

    bool ForceF7_;
    bool DirectF7_;
    bool UseDMA_;
    bool bEnsureLatestFrame_;
    int DcamDMABuffers_;
    bool TryFirewireB_;
    bool UserMode_;

    // used to set Format 7 Scalabel Image Size
    unsigned int MaxWidth_,  MaxHeight_, Left_,Top_;
    unsigned int UnitWidth_, UnitHeight_;
    bool UserTopLeft_;
    int BytesPerPacket_;
    unsigned int MaxBpP_, MinBpP_;
    Image<unsigned char>   *imageintern_;
    // time it lasts to transmit one image in usec
    int TransmissionTime_;

    unsigned char Shift16_; // uses this to shift bits down to normal
    CameraModel CamModel_;
    LensModel Lens_;
    bool UseBinning_;

#ifdef BIAS_HAVE_PTHREADS
    // guarantee exclusive usage of GrabSingle
    // (needed for example in SoftwareWhiteBalance)
    pthread_mutex_t grabMutex_;
#endif

  };


} // namespace BIAS


#endif // __BIASVIDEOSOURCE_DCAM_HH__
