/*
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003-2009    (see file CONTACT for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __VIDEOSOURCE_FFMPEG_HH__
#define __VIDEOSOURCE_FFMPEG_HH__

#include <VideoSource/VideoSource_Base.hh>
#include <Image/Camera.hh>

// Check for FFmpeg libraries in BIAS
#ifndef BIAS_HAVE_FFMPEG
  #error BIAS was build without FFmpeg, please enable USE_FFMPEG and recompile BIAS.
#endif

// Fix for FFmpeg on 32 bit systems
#ifndef INT64_C
#define INT64_C(c) (c ## LL)
#define UINT64_C(c) (c ## ULL)
#endif

#ifdef __cplusplus
extern "C" {
#endif
#include <libavcodec/avcodec.h>
#include <libavformat/avformat.h>
#include <libswscale/swscale.h>
#ifdef __cplusplus
}
#endif

namespace BIAS {
  /**
   * @class VideoSource_FFmpeg
   * @author Carsten Heine <carstenheine@gmx.de>
   * @ingroup g_videosource
   * @brief Simple video decoding of video files using FFmpeg library.
   */
  class BIASVideoSource_EXPORT VideoSource_FFmpeg : public VideoSource
  {
  public:

    /**
     * @brief Standard constructor.
     */
    VideoSource_FFmpeg();

    /**
     * @brief Clean up.
     */
    virtual ~VideoSource_FFmpeg();

    /**
     * @brief Open video file.
     * @return 0 on success, -1 on error
     */
    virtual int OpenDevice(const char *filename);

    /**
     * @brief Close video file.
     * @return 0 on success, -1 on error
     */
    virtual int CloseDevice();

    /**
     * @brief Get the next frame from the video file. The number of the frame #[num] and
     * presentation time in microseconds #[pts] are stored in the metadata of the
     * image for later use.
     * @return 0 on success, -1 on error
     */
    virtual int GrabSingle(Camera<unsigned char> &image);

    /**
     * @brief Seek to frame in video stream.
     * @return 0 on success, -1 on error
     */
    int SeekFrame(int64_t frame);

    /**
     * @brief Seek to time in milliseconds in video stream.
     * @return 0 on success, -1 on error
     */
    int SeekMs(int ms);

    /**
     * @brief Set output color model. Not all color models are supported by all decoders.
     * @return 0 on success, -1 on error
     */
    virtual int SetColorModel(ImageBase::EColorModel mode);

    /**
     * @brief Get the framerate of the video stream.
     */
    inline float GetFPS() { return FramesPerSecond_; }

    /**
     * @brief  Get the number of frames in the video stream. 0 if unknown.
     */
    inline int64_t GetNumberOfFrames() const { return numFrames_; }

    /**
     * @brief Get the number of frames decoded so far.
     */
    inline int64_t GetCurrentFrameNumber() const { return frameNumber_; }

    /**
     * @brief Get the timebase numerator of the video stream.
     */
    inline int GetTimeBaseNum() { return timeBaseNum_; }

    /**
     * @brief Get the timebase denumerator of the video stream.
     */
    inline int GetTimeBaseDen() { return timeBaseDen_; }

    /**
     * @brief Get the timestamp of the video stream in microseconds in Unix time.
     */
    inline int64_t GetTimestamp() { return timestamp_; }

    /**
     * @brief Get a string representation of the timestamp of the video stream.
     */
    inline char * GetDateString() { return dateStr_; }

    /**
     * @brief Get the duration of the video stream in microseconds.
     */
    inline int64_t GetDuration() { return duration_; }

    /**
     * @brief Get the presentation timestamp of the last decoded frame in microseconds.
     */
    inline int64_t GetPTS() { return pts_; }

    /**
     * @brief Get the last error message.
     */
    inline std::string GetError() { return errMsg_; }

  protected:

    /**
     * Called by constructor to init variables.
     */
    void Init_();

    /**
     * Try to extract date string of format "YYYYMMDD_HHMMSS" from string.
     */
    char * ExtractDateString_(std::string str);

    AVFormatContext  *pFormatCtx_;   ///< Format context for video decoder
    int              streamIndex_;   ///< Index of the video stream
    AVFrame          *pFrame_;       ///< Video stream frame
    AVFrame          *pOutputFrame_; ///< Output frame
    uint8_t          *pBuffer_;      ///< Ouput frame buffer
    enum PixelFormat outputPixFmt_;  ///< Output frame pixel format
    SwsContext       *pSwsCtx_;      ///< Color conversion context
    int64_t          numFrames_;     ///< Number of frames of the stream, 0 if unknown
    int64_t          frameNumber_;   ///< Number of decoded frames
    int              timeBaseNum_;   ///< Stream timebase numerator
    int              timeBaseDen_;   ///< Stream timebase denumerator
    int64_t          timestamp_;     ///< Stream timestamp
    char             *dateStr_;      ///< String representation of timestamp_
    int64_t          duration_;      ///< Stream duration in milliseconds, 0 if unknown
    int64_t          pts_;           ///< Presentation timestamp of last decoded frame in microseconds
    std::string      errMsg_;        ///< The last error message
  };
}

#endif // __VIDEOSOURCE_FFMPEG_HH__
