#ifndef VideoSource_Kinect2Net_HH
#define VideoSource_Kinect2Net_HH

#include <VideoSource/VideoSource_Base.hh>
#include <Image/Camera.hh>
#include <semaphore.h>
#include <NetworkComm/CScommClient.hh>

namespace BIAS
{
  /** @class VideoSource_Kinect2Net
      @brief Provides remote access to a Microsoft Kinect2 camera.
      @author fkellner 04/15
   */
  class BIASVideoSource_EXPORT VideoSource_Kinect2Net : public VideoSource
  {
  public:

    VideoSource_Kinect2Net();

    virtual ~VideoSource_Kinect2Net();

    /// Opens first device found, OpenDevice(0)
    /// @return Returns -1 if no device is available and -2 if an error happened while opening
    virtual int OpenDevice();

    /// Opens device by id. Use GetNumDevices() to determine the number of available Kinect2s
    /// @return Returns -1 if no device is available and -2 if an error happened while opening
    virtual int OpenDevice(std::string host, int port=8090);

    /// Close device, if no other kinect cameras are in use, context is deleted (i.e. close usb bus connection)
    virtual int CloseDevice();

    /// Start grabbing
    virtual int PreGrab();

    /// Stop grabbing
    virtual int PostGrab();

    /// kinect 2 depth image will be 512x424, float in Millimeters
    virtual int InitDepthImage(BIAS::ImageBase &Image);

    /// kinect 2 ir image will be 512x424, single channel unsigned char
    virtual int InitIrImage(BIAS::ImageBase &Image);

    /// grab the ir sum image computed from the phase images.
    virtual int GrabSingleIR(BIAS::Camera <float> &image);

    /// grab the depth image (slow). Use this function only if no high framerates are required, since
    /// the conversion from ir to depth image is done on the cpu and requires a lot computation.
    virtual int GrabSingleDepth(BIAS::Camera <float> &image);

    /// Return the number of available Kinect 2 devices
    virtual int GetNumDevices();

    /// No standard capabilities here
    virtual int GetCapabilities(VideoSourceCapabilities &caps);

    /// No standard capabilities here
    virtual int GetCapabilities(const char *device, VideoSourceCapabilities &caps);

  protected:

    static void* runner(void *args);

  private:

    static const int buffersize_ = 4;

    class imbuffer
    {
    public:
      imbuffer() : curProd(0), curCons(0) {
        sem_init(&sProd,true,buffersize_);
        sem_init(&sCons,true,0);
        for (int i=0;i<buffersize_;i++) buffer[i].Init(512,424,1);
      }
      Image<float> buffer[buffersize_];
      sem_t sProd;
      sem_t sCons;
      int curProd, curCons;
    };

    imbuffer depthbuffer_;
    imbuffer irbuffer_;
    int numDevices_;

    pthread_t grabThread_;

    CScommClient client_;
    std::string host_;
    int port_;
    bool stopThread_;

  };

}

#endif // VideoSource_Kinect2Net_HH
