/* 
 This file is part of the BIAS library (Basic ImageAlgorithmS).

 Copyright (C) 2003, 2004    (see file CONTACTS for details)
 Multimediale Systeme der Informationsverarbeitung
 Institut fuer Informatik
 Christian-Albrechts-Universitaet Kiel


 BIAS is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or
 (at your option) any later version.

 BIAS is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with BIAS; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef VIDEOSOURCE_OPENNI_HH_
#define VIDEOSOURCE_OPENNI_HH_
#ifndef _WIN32
  #if __cplusplus > 199711L
    #ifndef linux
      #define linux 1
    #endif
    #ifndef __linux
      #define __linux
    #endif
  #endif
#endif

#include <XnCppWrapper.h>
#include <VideoSource/VideoSource_Kinect_Base.hh>
#include <Image/Camera.hh>
#include <pthread.h>

namespace BIAS {

  /**
   * Use OpenNI just like the ToF Cameras. There is no parameters to control the camera yet.
   * Functionally might be added later depending on evolution of OpenNI
   * @author fkellner 02/11
   */
  class BIASVideoSource_EXPORT VideoSource_OpenNI : public VideoSource_Kinect_Base {
    public:

      VideoSource_OpenNI();
      virtual ~VideoSource_OpenNI();
      
      /// Opens first device found, OpenDevice(0)
      virtual int OpenDevice();

      /// Opens device by id. On first call inits the context and looks for available cameras (i.e. scan bus)
      virtual int OpenDevice(int device);

      /// Close device, if no other kinect cameras are in use, context is deleted (i.e. close usb bus connection)
      virtual int CloseDevice();

      /// set appropriate operation mode for kinect and start grabbing thread (see callbacks in VideoSource_OpenNI_Callback)
      virtual int PreGrab();
      /// stop grabbing thread 
      virtual int PostGrab();
      
      /// kinect color image will be 640x480, RGB. 
      virtual int InitImage(BIAS::ImageBase &Image);
      /// kinect depth image will be 640x480, float in Millimeters 
      virtual int InitDepthImage(BIAS::ImageBase &Image);

      /// grab single color image
      virtual int GrabSingle(BIAS::Camera <unsigned char> &image);
      /// depth according to formula  by Stephane Magnenat, see http://openkinect.org/wiki/Imaging_Information, todo: needs validation!
      /// Note that the depth image is shifted by 3 pixels in order to fit the IR image.
      virtual int GrabSingleDepth(BIAS::Camera <float> &image);
      
      virtual int SetVideoModeColor();
      virtual int SetVideoModeIR();
      
      virtual int SetDepthCaptureMode(CaptureMode mode);

      virtual int SetSwitchVideoModeAfterGrab();
      
      virtual int SetTilt(double degrees);
      virtual int SetLed(int mode);
      virtual int GetAccelerometer(double &tilt, double &x, double &y, double &z);

      virtual int GetNumDevices() {
        return numDevices_;
      }
      
      /// get if generating color (=true)
      virtual bool GetColorGeneratorMode() {
        return colorMode_;
      }

      virtual inline xn::Context* GetOpenNIContext() { return &context_; }

      virtual int GetCapabilities(BIAS::VideoSourceCapabilities &caps);
      virtual int GetCapabilities(const char *device, BIAS::VideoSourceCapabilities &caps);

      virtual int CreateDepthLookupTable(double px, double py, double fx, double fy);
      
      virtual int StartDepthStream();
      
      virtual int StopDepthStream();
      
      virtual int ToggleDepthStream();
      
    private:
      
      struct ONIDeviceNode {
        ONIDeviceNode(const xn::NodeInfo& device, const xn::NodeInfo& image, const xn::NodeInfo& ir, const xn::NodeInfo& depth)
        : deviceInfo(device), imageInfo(image), irInfo(ir), depthInfo(depth)
        {
        }
        xn::NodeInfo deviceInfo;
        xn::NodeInfo imageInfo;
        xn::NodeInfo irInfo;
        xn::NodeInfo depthInfo;
      };
      
      xn::DepthGenerator depthGenerator_;
      xn::UserGenerator userGenerator_;
      xn::ImageGenerator colorGenerator_;
      xn::IRGenerator irGenerator_;

      xn::DepthMetaData depthMD_;
      xn::ImageMetaData colorMD_;
      xn::IRMetaData irMD_;

      XnStatus rc_;
      xn::EnumerationErrors errors_;
      
      CaptureMode capMode_;
      bool colorMode_;
      
      Image<float> *depthLookupTable_;
      static int numInstances_;
      static int numDevices_;
      
      
      static xn::Context context_;
      static std::vector<ONIDeviceNode> devices_;
      
      static xn::NodeInfoList device_nodes_;
      static xn::NodeInfoList depth_nodes_;
      static xn::NodeInfoList image_nodes_;
      static xn::NodeInfoList ir_nodes_;

      void CreateLookupTable_(double px = 328.82739, double py = 252.6967098, double fx = 585.9303337, double fy = 585.548233470985);
      void DeleteLookupTable_();
      
      static bool CheckError_(XnStatus rc, std::string what);
      
      static void ScanBus();
      
      bool contextInited_;
      
      int deviceId_;
      
      bool switchModeEveryFrame_;
      
      static pthread_mutex_t *exclusive_mutex;
  };

}

#endif /* VIDEOSOURCE_OPENNI_HH_ */
